package main

import (
	"context"
	"crypto/tls"
	"flag"
	"fmt"
	"net/http"
	"os"
	"os/signal"
	"syscall"
	"time"

	"go.uber.org/zap"
	"go.uber.org/zap/zapcore"

	"git.grosinger.net/tgrosinger/tonyslist/pkg/cron"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/postmark"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/routes"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/services"
)

// configPath is the filepath of the config yaml file to load.
var configPath string

func main() {
	zapConfig := zap.NewProductionConfig()
	zapConfig.EncoderConfig.EncodeTime = zapcore.RFC3339TimeEncoder
	logger := zap.Must(zapConfig.Build())
	defer logger.Sync()

	loadConfigPath(logger)

	// Start a new container
	c := services.NewContainer(configPath, logger)
	defer func() {
		if err := c.Shutdown(); err != nil {
			logger.Fatal("Unable to create new container",
				zap.Error(err))
		}
	}()

	// Initialize and configure required postmark resources
	err := postmark.Initialize(c.Config, c.Logger)
	if err != nil {
		logger.Fatal("Unable to configure required Postmark resources",
			zap.Error(err))
	}

	// Build the router
	routes.BuildRouter(c)

	// Register cron jobs
	cron.RegisterDefaultCronJobs(c.Cron)

	// Start the server
	go func() {
		srv := http.Server{
			Addr:         fmt.Sprintf("%s:%d", c.Config.HTTP.Hostname, c.Config.HTTP.Port),
			Handler:      c.Web,
			ReadTimeout:  c.Config.HTTP.ReadTimeout,
			WriteTimeout: c.Config.HTTP.WriteTimeout,
			IdleTimeout:  c.Config.HTTP.IdleTimeout,
		}

		if c.Config.HTTP.TLS.Enabled {
			certs, err := tls.LoadX509KeyPair(c.Config.HTTP.TLS.Certificate, c.Config.HTTP.TLS.Key)
			if err != nil {
				logger.Fatal("Cannot load TLS certificate", zap.Error(err))
			}

			srv.TLSConfig = &tls.Config{
				Certificates: []tls.Certificate{certs},
			}
		}

		if err := c.Web.StartServer(&srv); err != http.ErrServerClosed {
			logger.Fatal("Shutting down the server", zap.Error(err))
		}
	}()

	// Wait for interrupt signal to gracefully shutdown the server with a timeout of 10 seconds.
	quit := make(chan os.Signal, 1)
	signal.Notify(quit, os.Interrupt)
	signal.Notify(quit, syscall.SIGTERM)
	<-quit
	ctx, cancel := context.WithTimeout(context.Background(), 10*time.Second)
	defer cancel()
	if err := c.Web.Shutdown(ctx); err != nil {
		logger.Error("Shutting down the application", zap.Error(err))
	}
}

// loadConfigPath attempts to find the path to the config file by reading the
// runtime flags. Will panic if the config file is not set or does not exist.
func loadConfigPath(logger *zap.Logger) {
	flag.StringVar(&configPath, "config", "", "Required")
	flag.Parse()

	if configPath == "" {
		logger.Fatal("Must provide a path to the config file")
	}
	_, err := os.Stat(configPath)
	if os.IsNotExist(err) {
		logger.Fatal("Provided config file does not exist")
	} else if err != nil {
		logger.Fatal("Unable to check config file", zap.Error(err))
	}
}
