-- users definition

CREATE TABLE IF NOT EXISTS users (
    id integer NOT NULL PRIMARY KEY AUTOINCREMENT,
    name text NOT NULL,
    email text NOT NULL,
    password text NOT NULL,
    verified bool NOT NULL DEFAULT false,
    created_at datetime NOT NULL
);

CREATE UNIQUE INDEX IF NOT EXISTS `users_email_key` ON `users` (`email`);

-- password_tokens definition

CREATE TABLE IF NOT EXISTS password_tokens (
    id integer NOT NULL PRIMARY KEY AUTOINCREMENT,
    hash text NOT NULL,
    created_at datetime NOT NULL,
    user_id integer NOT NULL,

    CONSTRAINT `password_tokens_user_id`
        FOREIGN KEY (`user_id`)
        REFERENCES `users` (`id`)
        ON DELETE NO ACTION
);

-- groups definition

CREATE TABLE IF NOT EXISTS groups (
    id integer NOT NULL PRIMARY KEY AUTOINCREMENT,
    name text NOT NULL
);

CREATE UNIQUE INDEX IF NOT EXISTS `groups_name` ON `groups` (`name`);

-- user_groups definition

CREATE TABLE IF NOT EXISTS user_groups (
    user_id integer NOT NULL,
    group_id integer NOT NULL,

    PRIMARY KEY (`user_id`, `group_id`),

    CONSTRAINT `user_groups_user_id`
        FOREIGN KEY (`user_id`)
        REFERENCES `users` (`id`)
        ON DELETE CASCADE,

    CONSTRAINT `user_groups_group_id`
        FOREIGN KEY (`group_id`)
        REFERENCES `groups` (`id`)
        ON DELETE CASCADE
);

-- posts_base definition

CREATE TABLE IF NOT EXISTS posts_base (
    id integer NOT NULL PRIMARY KEY AUTOINCREMENT,
    post_type text NOT NULL,
    title text NOT NULL,
    description text NOT NULL,
    phone_call text,
    phone_text text,
    include_email bool NOT NULL DEFAULT true,
    featured_photo integer,
    creator integer NOT NULL,
    created_at datetime NOT NULL,
    published_at datetime,
    listed bool NOT NULL DEFAULT true,
    moderator_delisted bool NOT NULL DEFAULT false,

    CONSTRAINT `posts_featured_photo`
        FOREIGN KEY (`featured_photo`)
        REFERENCES `photos` (`id`)
        ON DELETE SET NULL,

    CONSTRAINT `posts_creator`
        FOREIGN KEY (`creator`)
        REFERENCES `users` (`id`)
        ON DELETE CASCADE
);

CREATE UNIQUE INDEX IF NOT EXISTS `posts_base_featured_photo_key` ON `posts_base` (`featured_photo`);

-- classifieds definition

CREATE TABLE IF NOT EXISTS classifieds (
    id integer NOT NULL PRIMARY KEY,
    classified_type text NOT NULL,
    price real,

    CONSTRAINT `classifieds_posts_base`
        FOREIGN KEY (`id`)
        REFERENCES `posts_base` (`id`)
        ON DELETE CASCADE
);

-- housing_for_rent definition

CREATE TABLE IF NOT EXISTS housing_for_rent (
    id integer NOT NULL PRIMARY KEY,
    rent real,
    bedrooms integer,
    bathrooms real,
    parking_spots integer,
    pets bool NOT NULL DEFAULT false,
    shared bool NOT NULL DEFAULT false,
    street text,
    zipcode text,
    map_filename text,

    CONSTRAINT `housing_for_rent_posts_base`
        FOREIGN KEY (`id`)
        REFERENCES `posts_base` (`id`)
        ON DELETE CASCADE
);

-- housing_requests definition

CREATE TABLE IF NOT EXISTS housing_requests (
    id integer NOT NULL PRIMARY KEY,
    bedrooms integer,
    bathrooms real,
    parking_spots integer,
    pets bool NOT NULL DEFAULT false,
    shared bool NOT NULL DEFAULT false,

    CONSTRAINT `housing_requests_posts_base`
        FOREIGN KEY (`id`)
        REFERENCES `posts_base` (`id`)
        ON DELETE CASCADE
);

-- photos definition

CREATE TABLE IF NOT EXISTS photos (
    id integer NOT NULL PRIMARY KEY AUTOINCREMENT,
    filename text NOT NULL,
    fullsize_filename text,
    thumbnail_filename text,
    webp_fullsize_filename text,
    webp_thumbnail_filename text,
    created_at datetime NOT NULL,
    post_id integer NOT NULL,

    CONSTRAINT `photos_posts_base_id`
        FOREIGN KEY (`post_id`)
        REFERENCES `posts_base` (`id`)
        ON DELETE CASCADE
);

CREATE UNIQUE INDEX IF NOT EXISTS `photos_filename_key` ON `photos` (`filename`);
CREATE UNIQUE INDEX IF NOT EXISTS `photos_fullsize_filename_key` ON `photos` (`fullsize_filename`);
CREATE UNIQUE INDEX IF NOT EXISTS `photos_thumbnail_filename_key` ON `photos` (`thumbnail_filename`);
CREATE UNIQUE INDEX IF NOT EXISTS `photos_webp_fullsize_filename_key` ON `photos` (`webp_fullsize_filename`);
CREATE UNIQUE INDEX IF NOT EXISTS `photos_webp_thumbnail_filename_key` ON `photos` (`webp_thumbnail_filename`);

-- Posts Base - Full Text Search

CREATE VIRTUAL TABLE posts_base_fts USING fts5(title, description, content='posts_base', content_rowid='id');
CREATE TRIGGER posts_base_ai AFTER INSERT ON posts_base BEGIN
  INSERT INTO posts_base_fts(rowid, title, description) VALUES (new.id, new.title, new.description);
END;
CREATE TRIGGER posts_base_ad AFTER DELETE ON posts_base BEGIN
  INSERT INTO posts_base_fts(posts_base_fts, rowid, title, description) VALUES('delete', old.id, old.title, old.description);
END;
CREATE TRIGGER posts_base_au AFTER UPDATE ON posts_base BEGIN
  INSERT INTO posts_base_fts(posts_base_fts, rowid, title, description) VALUES('delete', old.id, old.title, old.description);
  INSERT INTO posts_base_fts(rowid, title, description) VALUES (new.id, new.title, new.description);
END;
