-- name: GetPost :one
SELECT sqlc.embed(pb), COUNT(p.post_id) as count_photos
FROM posts_base pb
LEFT JOIN photos p on p.post_id = pb.id
WHERE
    pb.id = ?
    AND pb.moderator_delisted = false
GROUP BY (p.post_id)
LIMIT 1;

-- name: GetMyPosts :many
SELECT * FROM posts_base
WHERE
    creator = ?
    AND post_type = ?
    AND moderator_delisted = false;

-- name: GetPostIDs :many
SELECT id FROM posts_base;

-- name: GetPublishedPosts :many
-- GetPublishedPosts returns posts which are published, listed, and have not been delisted.
-- classified_type is only valid when post_type == classified.
SELECT pb.id, pb.title, u.email, pb.creator, pb.post_type, pb.published_at, coalesce(c.classified_type, '') as classified_type, coalesce(h.housing_type, '') as housing_type
FROM posts_base pb
JOIN users u ON u.id = pb.creator
LEFT JOIN classifieds c ON c.id = pb.id
LEFT JOIN housing h on h.id = pb.id
WHERE
    moderator_delisted = false
    AND listed = true
    AND published_at is not null;

-- name: GetPostForAnonymousEmail :one
-- GetPostForAnonymousEmail returns the post with a matching anonymous email field.
-- Only active and published posts are returned.
SELECT posts_base.id, posts_base.title, posts_base.post_type, users.email
FROM posts_base
JOIN users on users.id = posts_base.creator
WHERE posts_base.anonymous_email = ?
    AND listed = true
    AND moderator_delisted = false
    AND datetime(published_at) > datetime('now', ?);

-- name: GetActivePostsForUserCount :one
SELECT COUNT(*) FROM posts_base pb
WHERE moderator_delisted = false
    AND listed = true
    AND datetime(published_at) > datetime('now', '-14 days')
    AND creator = ?;

-- name: CreatePost :execlastid
-- CreatePost makes a posts_base object which can then be referenced by other
-- post tables.
INSERT INTO posts_base (
    post_type, title, description, anonymous_email, creator, phone_call, phone_text, created_at, listed
) VALUES (
    ?, ?, ?, ?, ?, ?, ?, datetime('now'), true
);

-- name: RelistPost :exec
-- RelistPost updates the published_at column on the post to the current time so
-- that it no longer is expired and sets listed to true.
UPDATE posts_base
SET
    published_at = datetime('now'),
    listed = true
WHERE id = ?;

-- name: IncrementTimesPosted :exec
-- IncrementTimesPosted increases the counter for the number of times a post
-- has been posted.
UPDATE posts_base
SET
    count_times_posted = count_times_posted + 1
WHERE id = ?;

-- name: IncrementNonOwnerViews :exec
-- IncrementNonOwnerViews increases the counter for the number of times a post
-- has been viewed by someone other than the person that created the post.
UPDATE posts_base
SET
    count_nonowner_views = count_nonowner_views + 1
WHERE id = ?;

-- name: IncrementEmailResponseCount :exec
-- IncrementEmailResponseCount increases the counter for the number of email responses
-- received for a post.
UPDATE posts_base
SET
    count_email_responses = count_email_responses + 1
WHERE id = ?;

-- name: SetPostUnlisted :exec
-- SetPostUnlisted will set that value on the post so it no longer shows up in
-- the general browser queries.
UPDATE posts_base
SET listed = false
WHERE id = ?;

-- name: SetPostPublishedNow :exec
UPDATE posts_base
SET published_at = datetime('now')
WHERE id = ?;

-- name: SetPostPublishedAt :exec
-- SetPostPublishedAt can be used to set the published_at column explicitly.
-- This is most useful for unit tests to expire a post.
UPDATE posts_base
SET published_at = ?
WHERE id = ?;

-- name: SetPostFeaturedPhotoID :exec
-- SetPostFeaturedPhotoID updates the featured photo on the post.
-- This should only be used on classified and for-rent posts.
UPDATE posts_base
SET featured_photo = ?
WHERE id = ?;

-- name: SetPostModeratorDelisted :exec
-- SetPostModeratorDelisted will flag this post as deleted by a moderator,
-- preventing it from being seeing by any users.
UPDATE posts_base
SET moderator_delisted = true
WHERE id = ?;

-- name: UpdatePost :exec
UPDATE posts_base
SET
    title = ?,
    description = ?
WHERE id = ?;

-- name: DeletePost :exec
-- DeletePost completely removes a post, cascading to other tables.
DELETE FROM posts_base
WHERE id = ?;

-- name: DeletePostsExpiredBefore :exec
-- DeletePostsExpiredBefore completely removes posts which are listed, but which
-- have a published_at date before the provided date.
DELETE FROM posts_base
WHERE listed AND date(published_at) < date(?);

-- name: DeletePostsUnlistedBefore :exec
-- DeletePostsUnlistedBefore completely removes posts which are unlisted and
-- which have a published_at date before the provided date.
DELETE FROM posts_base
WHERE NOT listed AND date(published_at) < date(?);
