package auth

import (
	"fmt"
	"time"

	"github.com/golang-jwt/jwt"
	"github.com/labstack/echo-contrib/session"
	"github.com/labstack/echo/v4"

	"git.grosinger.net/tgrosinger/tonyslist/config"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
)

const (
	// sessionName stores the name of the session which contains authentication
	// data.
	sessionName = "ua"

	// sessionKeyUserID stores the key used to store the user ID in the session.
	sessionKeyUserID = "user_id"

	// sessionKeyAuthenticated stores the key used to store the authentication
	// status in the session.
	sessionKeyAuthenticated = "authenticated"
)

// ErrNotAuthenticated is an error returned when a user is not authenticated
var ErrNotAuthenticated = fmt.Errorf("user not authenticated")

// Client is the client that handles authentication requests
type Client struct {
	config *config.Config
	db     *models.Client
}

// New creates a new authentication client
func New(cfg *config.Config, db *models.Client) *Client {
	return &Client{
		config: cfg,
		db:     db,
	}
}

// Login logs in a user of a given ID
func (c *Client) Login(ctx echo.Context, userID int) error {
	sess, err := session.Get(sessionName, ctx)
	if err != nil {
		return err
	}
	sess.Values[sessionKeyUserID] = userID
	sess.Values[sessionKeyAuthenticated] = true
	return sess.Save(ctx.Request(), ctx.Response())
}

// Logout logs the requesting user out
func (c *Client) Logout(ctx echo.Context) error {
	sess, err := session.Get(sessionName, ctx)
	if err != nil {
		return err
	}
	sess.Values[sessionKeyAuthenticated] = false
	return sess.Save(ctx.Request(), ctx.Response())
}

// GetAuthenticatedUserID returns the authenticated user's ID, if the user is
// logged in.
func (c *Client) GetAuthenticatedUserID(ctx echo.Context) (int, error) {
	sess, err := session.Get(sessionName, ctx)
	if err != nil {
		return 0, err
	}

	if sess.Values[sessionKeyAuthenticated] == true {
		return sess.Values[sessionKeyUserID].(int), nil
	}

	return 0, ErrNotAuthenticated
}

// GetAuthenticatedUser returns the authenticated user if the user is logged in
func (c *Client) GetAuthenticatedUser(ctx echo.Context) (sqlc.User, error) {
	if userID, err := c.GetAuthenticatedUserID(ctx); err == nil {
		return c.db.C.GetUserByID(ctx.Request().Context(), userID)
	}

	return sqlc.User{}, ErrNotAuthenticated
}

// GenerateJWTEmailAuthToken generates an email verification token for a
// given email address using JWT which is set to expire based on the duration
// stored in configuration.
func (c *Client) GenerateJWTEmailAuthToken(email string) (string, error) {
	return c.GenerateJWTAuthToken(map[string]interface{}{
		"email": email,
	}, &c.config.App.EmailVerificationTokenExpiration)
}

// GenerateJWTAuthToken generates a verification token over the provided data
// using JWT which is set to expire based on the duration stored in
// configuration.
func (c *Client) GenerateJWTAuthToken(data map[string]interface{},
	expiration *time.Duration,
) (string, error) {
	claims := jwt.MapClaims{}
	if expiration != nil {
		claims["exp"] = time.Now().Add(*expiration).Unix()
	}

	for k, v := range data {
		claims[k] = v
	}

	token := jwt.NewWithClaims(jwt.SigningMethodHS256, claims)
	return token.SignedString([]byte(c.config.App.EncryptionKey))
}

// ValidateJWTAuthToken validates an email verification token and
// returns the associated email address if the token is valid and has not
// expired.
func (c *Client) ValidateJWTAuthToken(token string) (string, error) {
	claims, err := c.ValidateJWTAuthTokenMap(token)
	if err != nil {
		return "", err
	}

	if email, ok := claims["email"].(string); ok {
		return email, nil
	}

	return "", fmt.Errorf("invalid or expired token")
}

// ValidateJWTAuthTokenMap validates an email verification token and
// returns the claims if the token is valid and has not expired.
func (c *Client) ValidateJWTAuthTokenMap(token string) (jwt.MapClaims, error) {
	t, err := jwt.Parse(token, func(t *jwt.Token) (interface{}, error) {
		if _, ok := t.Method.(*jwt.SigningMethodHMAC); !ok {
			return nil, fmt.Errorf("unexpected signing method: %v", t.Header["alg"])
		}

		return []byte(c.config.App.EncryptionKey), nil
	})
	if err != nil {
		return nil, err
	}

	if claims, ok := t.Claims.(jwt.MapClaims); ok && t.Valid {
		return claims, nil
	}

	return nil, fmt.Errorf("invalid or expired token")
}
