package auth

import (
	"os"
	"testing"
	"time"

	"github.com/labstack/echo/v4"
	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
	"go.uber.org/zap"

	"git.grosinger.net/tgrosinger/tonyslist/config"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/tests"
)

var (
	cfg    config.Config
	client *Client
	ctx    echo.Context
	usr    sqlc.User
)

func TestMain(m *testing.M) {
	logger := zap.NewNop()

	var err error
	configPath := tests.LoadConfigPath()
	cfg, err = config.GetConfig(configPath)
	if err != nil {
		panic(err)
	}

	// Create the database
	db, err := models.New(logger, &cfg)
	if err != nil {
		panic(err)
	}

	client = New(&cfg, db)

	// Create a web context
	web := echo.New()
	ctx, _ = tests.NewContext(web, "/")
	tests.InitSession(ctx)

	// Create a test user
	if usr, err = db.User.CreateMockUser(); err != nil {
		panic(err)
	}

	// Run tests
	exitVal := m.Run()

	db.Close()
	web.Close()
	os.Exit(exitVal)
}

func TestAuthClient_Auth(t *testing.T) {
	assertNoAuth := func() {
		_, err := client.GetAuthenticatedUserID(ctx)
		assert.Equal(t, ErrNotAuthenticated, err)
		_, err = client.GetAuthenticatedUser(ctx)
		assert.Equal(t, ErrNotAuthenticated, err)
	}

	assertNoAuth()

	err := client.Login(ctx, usr.ID)
	require.NoError(t, err)

	uid, err := client.GetAuthenticatedUserID(ctx)
	require.NoError(t, err)
	assert.Equal(t, usr.ID, uid)

	u, err := client.GetAuthenticatedUser(ctx)
	require.NoError(t, err)
	assert.Equal(t, u.ID, usr.ID)

	err = client.Logout(ctx)
	require.NoError(t, err)

	assertNoAuth()
}

func TestAuthClient_EmailVerificationToken(t *testing.T) {
	t.Run("valid token", func(t *testing.T) {
		email := "test@localhost.com"
		token, err := client.GenerateJWTEmailAuthToken(email)
		require.NoError(t, err)

		tokenEmail, err := client.ValidateJWTAuthToken(token)
		require.NoError(t, err)
		assert.Equal(t, email, tokenEmail)
	})

	t.Run("invalid token", func(t *testing.T) {
		badToken := "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJlbWFpbCI6InRlc3RAbG9jYWxob3N0LmNvbSIsImV4cCI6MTkxNzg2NDAwMH0.ScJCpfEEzlilKfRs_aVouzwPNKI28M3AIm-hyImQHUQ"
		_, err := client.ValidateJWTAuthToken(badToken)
		assert.Error(t, err)
	})

	t.Run("expired token", func(t *testing.T) {
		cfg.App.EmailVerificationTokenExpiration = -time.Hour
		email := "test@localhost.com"
		token, err := client.GenerateJWTEmailAuthToken(email)
		require.NoError(t, err)

		_, err = client.ValidateJWTAuthToken(token)
		assert.Error(t, err)

		cfg.App.EmailVerificationTokenExpiration = time.Hour * 12
	})
}
