package context

import (
	"context"
	"errors"

	"github.com/labstack/echo/v4"
	"go.uber.org/zap"
	"go.uber.org/zap/zapcore"
)

// Context is a reexport of context.Context from the standard library to help
// avoid clashing package names.
type Context = context.Context

func Background() context.Context {
	return context.Background()
}

var WithValue = context.WithValue

const (
	// AuthenticatedUserKey is the key value used to store the authenticated user in context
	AuthenticatedUserKey = "auth_user"

	// AuthenticatedUserGroupsKey is the key value used to store the list of
	// user groups in the context.
	AuthenticatedUserGroupsKey = "auth_user_groups"

	// UserKey is the key value used to store a user in context
	UserKey = "user"

	// StripeCustomerIDKey is the key value used to store a stripe customer id in context
	StripeCustomerIDKey = "stripe_customer_id"

	// FormKey is the key value used to store a form in context
	FormKey = "form"

	// PostKey is the key value used to store a classified or housing post in the context
	PostBaseKey = "post_base"

	// PostKey is the key value used to store a classified or housing post in the context
	PostKey = "post"

	// DigestMembershipKey is the key value used to store the current digest
	// membership status for the authenticated user in the context. If there is
	// no authenticated user then it will not be set.
	DigestMembershipKey = "digest_membership"

	// EventKey is the key value used to store an event in the context
	EventKey = "event"

	// LoggerKey is the key value used to store a *zap.Logger in the context
	LoggerKey = "logger"
)

// GetLogger retrieves the logger for a specific request from the context.
func GetLogger(c echo.Context) *zap.Logger {
	logger := c.Get(LoggerKey)
	if logger != nil {
		return logger.(*zap.Logger)
	}

	// For some reason we do not have a logger.
	// Make a new one, but log this eror

	zapConfig := zap.NewProductionConfig()
	zapConfig.EncoderConfig.EncodeTime = zapcore.RFC3339TimeEncoder
	newLogger := zap.Must(zapConfig.Build())

	newLogger.Error("Unexpectedly do not have a logger in the request context",
		zap.String("request_id", c.Response().Header().Get(echo.HeaderXRequestID)),
		zap.String("method", c.Request().Method),
		zap.String("route", c.Path()),
		zap.String("path", c.Request().URL.Path))

	return newLogger
}

// IsCanceledError determines if an error is due to a context cancelation
func IsCanceledError(err error) bool {
	return errors.Is(err, context.Canceled)
}
