package controller

import (
	"bytes"
	"fmt"
	"io"
	"net/http"
	"os"
	"path/filepath"

	"github.com/labstack/echo/v4"
	"github.com/yuin/goldmark"
	"github.com/yuin/goldmark/parser"
	"github.com/yuin/goldmark/renderer/html"
	"go.uber.org/zap"

	"git.grosinger.net/tgrosinger/tonyslist/config"
	"git.grosinger.net/tgrosinger/tonyslist/constants"
	"git.grosinger.net/tgrosinger/tonyslist/constants/pages"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/context"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/classified"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/postbase"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/msg"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/services"
)

// Controller provides base functionality and dependencies to routes.
// The proposed pattern is to embed a Controller in each individual route struct and to use
// the router to inject the container so your routes have access to the services within the container
type Controller struct {
	// Container stores a services container which contains dependencies
	Container *services.Container
}

// NewController creates a new Controller
func NewController(c *services.Container) Controller {
	return Controller{
		Container: c,
	}
}

// LoadMarkdown returns the requested markdown file rendered into HTML. The
// markdown parser configured here will pass through any HTML in the source
// file, so it must only be used with trusted content. Attempts
// to use the cache before loading and rendering from file.
func (c *Controller) LoadMarkdown(ctx echo.Context, markdownFilename string) (string, error) {
	rendered, ok := c.Container.Cache.Get().
		Group(constants.CachedMarkdownGroup).
		Key(markdownFilename).
		Fetch()
	if ok {
		context.GetLogger(ctx).Info("loaded markdown content from cache",
			zap.String("filename", markdownFilename))
		return rendered.(string), nil
	}

	filename := filepath.Join(config.TemplateDir(c.Container.Config), "markdown", markdownFilename)
	md, err := os.ReadFile(filename)
	if err != nil {
		return "", c.Fail(err, "failed to load privacy policy markdown")
	}

	converter := goldmark.New(
		goldmark.WithParserOptions(
			parser.WithAutoHeadingID(),
		),
		goldmark.WithRendererOptions(
			html.WithUnsafe(),
		),
	)

	var buf bytes.Buffer
	err = converter.Convert(md, &buf)
	if err != nil {
		return "", c.Fail(err, "failed to parse privacy policy markdown")
	}

	renderedMarkdown := buf.String()

	c.Container.Cache.Set().
		Group(constants.CachedMarkdownGroup).
		Key(markdownFilename).
		Data(renderedMarkdown).
		Save()

	return renderedMarkdown, nil
}

type Renderer interface {
	Render(io.Writer) error
}

// RenderPage renders a Page as an HTTP response
func (c *Controller) RenderHTML(ctx echo.Context, page Page, r Renderer) error {
	return c.renderPage(ctx, page, r, "text/html; charset=UTF-8")
}

func (c *Controller) RenderXML(ctx echo.Context, page Page, r Renderer) error {
	return c.renderPage(ctx, page, r, "application/xml; charset=UTF-8")
}

func (c *Controller) renderPage(ctx echo.Context, page Page, r Renderer, contentType string) error {
	// Set any headers
	ctx.Response().Header().Set("Content-Type", contentType)
	for k, v := range page.Headers {
		ctx.Response().Header().Set(k, v)
	}

	// Set the status code
	// This needs to happen after setting the rest of the headers.
	ctx.Response().Status = page.StatusCode
	ctx.Response().WriteHeader(page.StatusCode)

	// Send the response
	err := r.Render(ctx.Response().Writer)
	if err != nil {
		// TODO: record a metric here
		context.GetLogger(ctx).Error(
			"Failed to write output to client", zap.Error(err))
		return echo.NewHTTPError(http.StatusInternalServerError)
	}
	return nil
}

// Redirect redirects to a given route name with optional route parameters
func (c *Controller) Redirect(ctx echo.Context, route string, routeParams ...interface{}) error {
	url := ctx.Echo().Reverse(route, routeParams...)
	return ctx.Redirect(http.StatusFound, url)
}

// RedirectToPost redirects the request to the post details page for the provided post.
func (c *Controller) RedirectToPost(ctx echo.Context, p *postbase.PostBase) error {
	switch p.PostType {
	case constants.Classified:
		return c.Redirect(ctx, pages.Names.Classified.One, p.ID)
	case constants.Housing:
		return c.Redirect(ctx, pages.Names.Housing.One, p.ID)
	}
	return c.Redirect(ctx, pages.Names.Home)
}

// RedirectToPostBrowser redirects the request to the browse page for the given
// post type stored in the post base. For classified posts, the full post must be
// stored in the context.
func (c *Controller) RedirectToPostBrowser(ctx echo.Context) error {
	p, ok := ctx.Get(context.PostBaseKey).(*postbase.PostBase)
	if !ok {
		msg.Warning(ctx, "The requested post could not be found.")
		c.Redirect(ctx, pages.Names.Home)
	}

	var path string
	switch p.PostType {
	case constants.Classified:
		post, ok := ctx.Get(context.PostKey).(*classified.Post)
		if ok {
			switch post.ClassifiedType {
			case constants.ForSale:
				path = ctx.Echo().Reverse(pages.Names.Classified.ForSale.Browse)
			case constants.ForFree:
				path = ctx.Echo().Reverse(pages.Names.Classified.ForFree.Browse)
			case constants.WantToBuy:
				path = ctx.Echo().Reverse(pages.Names.Classified.WantToBuy.Browse)
			case constants.WantToBorrow:
				path = ctx.Echo().Reverse(pages.Names.Classified.WantToBorrow.Browse)
			}
		} else {
			path = ctx.Echo().Reverse(pages.Names.Classified.ForSale.Browse)
		}
	case constants.Housing:
		path = ctx.Echo().Reverse(pages.Names.Housing.Browse)
	}

	if path == "" {
		msg.Warning(ctx, "The requested page could not be found.")
		c.Redirect(ctx, pages.Names.Home)
	}
	return ctx.Redirect(http.StatusFound, path)
}

// Fail is a helper to fail a request by returning a 500 error and logging the error.
// Neither the error or the log are displayed to the user.
func (c *Controller) Fail(err error, log string) error {
	return echo.NewHTTPError(http.StatusInternalServerError, fmt.Sprintf("%s: %v", log, err))
}

// BadReq is a helper to fail a request by returning a 400 error and logging the error
func (c *Controller) BadReq(log string) error {
	return echo.NewHTTPError(http.StatusBadRequest, log)
}
