package controller

import (
	"strings"

	"github.com/asaskevich/govalidator"
	"github.com/labstack/echo/v4"
)

// FormSubmission represents the state of the submission of a form, not including the form itself
type FormSubmission struct {
	// IsSubmitted indicates if the form has been submitted
	IsSubmitted bool

	// Errors stores a slice of error message strings keyed by form struct field name
	Errors map[string][]string
}

// Process processes a submission for a form
func (f *FormSubmission) Process(ctx echo.Context, form interface{}) error {
	f.Errors = make(map[string][]string)
	f.IsSubmitted = true

	// Validate the form
	if err := ctx.Validate(form); err != nil {
		f.setErrorMessages(err)
	}

	return nil
}

// HasErrors indicates if the submission has any validation errors
func (f FormSubmission) HasErrors() bool {
	if f.Errors == nil {
		return false
	}
	return len(f.Errors) > 0
}

// FieldHasErrors indicates if a given field on the form has any validation errors
func (f FormSubmission) FieldHasErrors(fieldName string) bool {
	return len(f.GetFieldErrors(fieldName)) > 0
}

// SetFieldError sets an error message for a given field name
func (f *FormSubmission) SetFieldError(fieldName string, message string) {
	if f.Errors == nil {
		f.Errors = make(map[string][]string)
	}
	f.Errors[fieldName] = append(f.Errors[fieldName], message)
}

// GetFieldErrors gets the errors for a given field name
func (f FormSubmission) GetFieldErrors(fieldName string) []string {
	if f.Errors == nil {
		return []string{}
	}
	return f.Errors[fieldName]
}

// GetFieldBorderClass returns an HTML class based on the status of the field
func (f FormSubmission) GetFieldBorderClass(fieldName string) string {
	if f.IsSubmitted {
		if f.FieldHasErrors(fieldName) {
			return "border-red-300"
		}
	}
	return "border-gray-300"
}

// IsDone indicates if the submission is considered done which is when it has been submitted
// and there are no errors.
func (f FormSubmission) IsDone() bool {
	return f.IsSubmitted && !f.HasErrors()
}

// setErrorMessages sets errors messages on the submission for all fields that failed validation
func (f *FormSubmission) setErrorMessages(err error) {
	errs := err.(govalidator.Errors)
	f.setGoValidatorErrors(errs)
}

func (f *FormSubmission) setGoValidatorErrors(errs govalidator.Errors) {
	for _, err := range errs {
		// Could be nested govalidator.Errors
		if ve, ok := err.(govalidator.Errors); ok {
			f.setGoValidatorErrors(ve)
			continue
		}

		ve, ok := err.(govalidator.Error)
		if !ok {
			// TODO?
			continue
		}

		var message string

		switch ve.Validator {
		case "required":
			message = "This field is required."
		case "email":
			message = "Enter a valid email address."
		case "numeric":
			message = "Must be a number."
		case "money":
			message = "Must be in the format 12.24."
		case "phone":
			message = "Must be a phone number in for format 123-456-7890."
		default:
			message = strings.TrimPrefix(ve.Error(), ve.Name+": ")
		}

		// Add the error
		f.SetFieldError(ve.Name, message)
	}
}
