package controller

import (
	"net/http"
	"net/url"
	"strings"

	"github.com/casbin/casbin/v2"
	"github.com/labstack/echo/v4"
	echomw "github.com/labstack/echo/v4/middleware"
	"go.uber.org/zap"

	"git.grosinger.net/tgrosinger/tonyslist/config"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/context"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/services"
)

// Page consists of all data that will be used to render a page response for a given controller.
// While it's not required for a controller to render a Page on a route, this is the common data
// object that will be passed to the templates, making it easy for all controllers to share
// functionality both on the back and frontend. The Page can be expanded to include anything else
// your app wants to support.
// Methods on this page also then become available in the templates, which can be more useful than
// the funcmap if your methods require data stored in the page, such as the context.
type Page struct {
	Cfg *config.Config

	// Title stores the title of the page. If empty, the layout may display just the AppName
	Title string

	// Context stores the request context
	Context echo.Context

	// ToURL is a function to convert a route name and optional route parameters to a URL
	ToURL func(name string, params ...interface{}) string

	// Path stores the path of the current request
	Path string

	// URL stores the URL of the current request
	URL string

	// IsAuth stores whether or not the user is authenticated
	IsAuth bool

	// DisplayEmailDigestCTA controls whether to show a button or banner
	// prompting the user to join the email digest.
	DisplayEmailDigestCTA bool

	// AuthUser stores the authenticated user
	AuthUser sqlc.User

	// AuthUserGroups stores the list of groups the authenticated user belongs to
	AuthUserGroups []string

	// StatusCode stores the HTTP status code that will be returned
	StatusCode int

	// Metatags stores metatag values
	Metatags struct {
		// Description stores the description metatag value
		Description string

		// Keywords stores the keywords metatag values
		Keywords []string

		// HasFeed indicates that this page supports retrieval as an RSS feed
		HasFeed bool
	}

	// Pager stores a pager which can be used to page lists of results.
	// Pager is empty by default unless set after the page is created.
	Pager Pager

	// CSRF stores the CSRF token for the given request.
	// This will only be populated if the CSRF middleware is in effect for the given request.
	// If this is populated, all forms must include this value otherwise the requests will be rejected.
	CSRF string

	// Headers stores a list of HTTP headers and values to be set on the response
	Headers map[string]string

	// RequestID stores the ID of the given request.
	// This will only be populated if the request ID middleware is in effect for the given request.
	RequestID string

	// Enforcer is initialized when the page is being rendered and is not
	// available before that point.
	Enforcer *casbin.Enforcer
}

// NewPage creates and initiatizes a new Page for a given request context
func NewPage(ctx echo.Context, container *services.Container) Page {
	p := Page{
		Cfg:        container.Config,
		Context:    ctx,
		Enforcer:   container.Enforcer,
		ToURL:      ctx.Echo().Reverse,
		Path:       ctx.Request().URL.Path,
		URL:        ctx.Request().URL.String(),
		StatusCode: http.StatusOK,
		Headers:    make(map[string]string),
		RequestID:  ctx.Response().Header().Get(echo.HeaderXRequestID),
	}

	if csrf := ctx.Get(echomw.DefaultCSRFConfig.ContextKey); csrf != nil {
		p.CSRF = csrf.(string)
	}

	if u := ctx.Get(context.AuthenticatedUserKey); u != nil {
		p.IsAuth = true
		p.AuthUser = u.(sqlc.User)
	}

	if g := ctx.Get(context.AuthenticatedUserGroupsKey); g != nil {
		p.AuthUserGroups = g.([]string)
	}

	return p
}

// Enforce checks whether the currently authenticated user can perform the
// requested action against the requested reosurce.
func (p Page) Enforce(resource, action string) bool {
	for _, g := range p.AuthUserGroups {
		ok, err := p.Enforcer.Enforce(g, resource, action)
		if err != nil {
			context.GetLogger(p.Context).Warn("failed to check enforcement of action",
				zap.String("subject", g),
				zap.String("resource", resource),
				zap.String("action", action),
				zap.Error(err))
		} else if ok {
			return true
		}
	}
	return false
}

// RSSLink converts the current URL into an RSS link by appending ".rss" to the
// path. Query parameters are preserved.
func (p Page) RSSLink() string {
	u, err := url.Parse(p.URL)
	if err != nil {
		return ""
	}

	// We can modify the URL object here because we're operating on a copy.
	u.Path += ".rss"
	return u.String()
}

// ToUserContentPath converts the provided path (generally a photo) to an
// absolute path for that content as served by the web server.
func (p Page) ToUserContentPath(path string) string {
	filePath := strings.TrimPrefix(path, p.Cfg.Storage.PhotoStorageDir)
	if !strings.HasPrefix(filePath, "/") {
		filePath = "/" + filePath
	}
	return "/" + config.UserContentPrefix + filePath
}
