package conv

import (
	"encoding/xml"
	"fmt"
	"image"
	"image/color"

	qrcode "github.com/skip2/go-qrcode"
	gp "maragu.dev/gomponents"
)

/*

Largely based on svg-qr-code
https://github.com/wamuir/svg-qr-code

Modified to set a ViewBox on the SVG element and to more
easily allow the customization required.

*/

func SVGQRCode(href string, background, foreground color.Color) gp.Node {
	code, err := qrcode.New(href, qrcode.Highest)
	if err != nil {
		fmt.Println(err)
		return nil
	}

	code.DisableBorder = true
	code.Level = qrcode.Low
	code.ForegroundColor = foreground
	code.BackgroundColor = background

	q := QR{code, 16, 0, code.BackgroundColor}
	return gp.Raw(q.String())
}

// QR holds a QR Code (from github.com/skip2/go-qrcode) and SVG settings.
type QR struct {
	Code        *qrcode.QRCode // underlying QR Code
	Blocksize   int            // size of each block in pixels, default = 16 pixels
	Borderwidth int            // size of the border in blocks, default = 4 blocks
	Borderfill  color.Color    // fill color for the border
}

// SVG is the vector representation of a QR code, as a Go struct.
type SVG struct {
	XMLName xml.Name `xml:"svg"`
	NS      string   `xml:"xmlns,attr"`
	Style   string   `xml:"style,attr"`
	Blocks  []Block  `xml:"rect"`
	ViewBox string   `xml:"viewBox,attr"`
}

// Block is a color block in the rendered QR code.
type Block struct {
	X      int    `xml:"x,attr"`
	Y      int    `xml:"y,attr"`
	Width  int    `xml:"width,attr"`
	Height int    `xml:"height,attr"`
	Fill   string `xml:"fill,attr"`
}

// SVG returns the vector representation of a QR code, as a Go struct.
// This could, for instance, be marshaled with encoding/xml.
func (q *QR) SVG() *SVG {
	q.Code.DisableBorder = true
	var i image.Image = q.Code.Image(0)
	var w int = i.Bounds().Max.X

	var svg SVG
	svg.NS = "http://www.w3.org/2000/svg"
	width := (w + 2*q.Borderwidth) * q.Blocksize
	svg.ViewBox = fmt.Sprintf("0 0 %d %[1]d", width)

	background := hex(q.Borderfill)

	blocks := make([]Block, 0, 25)
	blocks = append(blocks, Block{0, 0, width, width, background})

	// Track current block being built
	var currentBlock *Block

	for x := 0; x < w; x++ {
		for y := 0; y < w; y++ {
			color := hex(i.At(x, y))
			if color == background {
				// Reset current block if we hit a background color
				currentBlock = nil
				continue
			}

			// Calculate block coordinates
			blockX := (x + q.Borderwidth) * q.Blocksize
			blockY := (y + q.Borderwidth) * q.Blocksize

			// If no current block or colors don't match, create a new block
			if currentBlock == nil || currentBlock.Fill != color {
				currentBlock = &Block{
					X:      blockX,
					Y:      blockY,
					Width:  q.Blocksize,
					Height: q.Blocksize,
					Fill:   color,
				}
				blocks = append(blocks, *currentBlock)
			} else {
				// Extend the current block horizontally
				blocks[len(blocks)-1].Height += q.Blocksize
			}
		}
		// Reset current block at the end of each row
		currentBlock = nil
	}

	svg.Blocks = blocks
	return &svg
}

// String() returns the SVG as a string and satisfies the fmt.Stringer interface.
func (s *SVG) String() string {
	x, _ := xml.Marshal(s)
	return string(x)
}

// String() returns the QR as an SVG string and satisfies the fmt.Stringer interface.
func (q *QR) String() string {
	return q.SVG().String()
}

func hex(c color.Color) string {
	rgba := color.RGBAModel.Convert(c).(color.RGBA)
	return fmt.Sprintf("#%.2x%.2x%.2x", rgba.R, rgba.G, rgba.B)
}
