package conv

import (
	"database/sql"
	"database/sql/driver"
	"fmt"
	"math/rand"
	"strconv"
	"strings"
	"time"
)

var (
	// validRandomChars is a list of the characters which may be used when
	// generating a random string. It avoids look-alike characters to make it easier
	// to copy. Only uses lowercase characters since emails should not be case sensitive.
	validRandomChars    = "abcdefghjkmnpqrstuvwxyz23456789"
	lenValidRandomChars = int64(len(validRandomChars))
)

func StringToNullString(str string) sql.NullString {
	if str == "" {
		return sql.NullString{Valid: false}
	}
	return sql.NullString{Valid: true, String: str}
}

// RandomNullString returns a random string in a sql.NullString of the requested
// length. If valid is false, no string will be generated and the NullString is
// set to invalid.
func RandomNullString(length int) sql.NullString {
	sb := strings.Builder{}
	sb.Grow(length)
	for i := 0; i < length; i++ {
		sb.WriteByte(validRandomChars[rand.Int63()%lenValidRandomChars])
	}
	return sql.NullString{Valid: true, String: sb.String()}
}

func StringToNullInt(name string, val string) (NullInt, error) {
	if val == "" {
		return NullInt{Valid: false}, nil
	}

	i, err := strconv.Atoi(val)
	if err != nil {
		return NullInt{}, fmt.Errorf("unable to parse %s into numeric value", name)
	}
	return NullInt{Valid: true, Int: i}, nil
}

func StringToNullFloat64(name string, val string) (sql.NullFloat64, error) {
	if val == "" {
		return sql.NullFloat64{Valid: false}, nil
	}

	p, err := strconv.ParseFloat(val, 64)
	if err != nil {
		return sql.NullFloat64{}, fmt.Errorf("unable to parse %s into numeric value", name)
	}
	return sql.NullFloat64{Valid: true, Float64: p}, nil
}

// ParseTimeParts creates a timestamp for the provided time parts.
// Date must be provided an in the format "2006-01-02". Time is optional, but if
// provided it must be in the format "15:04".
func ParseTimeParts(date, t string, loc *time.Location) (time.Time, error) {
	format := "2006-01-02"
	value := date
	if t != "" {
		format = "2006-01-02 15:04"
		value = fmt.Sprintf("%s %s", date, t)
	}

	return time.ParseInLocation(format, value, loc)
}

func IntToNullInt(i int) NullInt {
	return NullInt{Valid: true, Int: i}
}

// NullInt represents an int that may be null.
// NullInt implements the Scanner interface so it can be used as a scan
// destination, similar to NullString.
type NullInt struct {
	Int   int
	Valid bool // Valid is true if Int is not NULL
}

// Scan implements the Scanner interface.
func (n *NullInt) Scan(value any) error {
	internal := &sql.NullInt64{}
	err := internal.Scan(value)
	if err != nil {
		return err
	}

	n.Int = int(internal.Int64)
	n.Valid = internal.Valid
	return nil
}

// Value implements the driver Valuer interface.
func (n NullInt) Value() (driver.Value, error) {
	if !n.Valid {
		return nil, nil
	}
	return int64(n.Int), nil
}
