package conv

import "time"

// StartOfDay returns a new timestamp with the clock portion of the timestamp
// removed, leaving only the date. This is essentally midnight on the date of
// the provided timestamp.
func StartOfDay(a time.Time) time.Time {
	y, m, d := a.Date()
	return time.Date(y, m, d, 0, 0, 0, 0, a.Location())
}

// Returns true iff two times have the same year, month, and day. Ignores the
// time fields.
func DatesEqual(a, b time.Time) bool {
	aYear, aMonth, aDay := a.Date()
	bYear, bMonth, bDay := b.Date()
	return aYear == bYear && aMonth == bMonth && aDay == bDay
}

// DateInRange checks if the provided date is within the date range specified by
// beginning and end. It only considers the day, month, and year. Times are
// ignored.
func DateInRange(d, beginning, end time.Time) bool {
	absBeginning := StartOfDay(beginning)

	absEnd := StartOfDay(end)
	absEnd.AddDate(0, 0, 1)

	return d.Compare(absBeginning) >= 0 && d.Compare(absEnd) <= 0
}

// CombineDateAndTime returns a new time.Time where the date from the first
// provided Time and the clock from the second provided Time are combined.
func CombineDateAndTime(d, t time.Time) time.Time {
	year, month, day := d.Date()
	hour, minute, _ := t.Clock()
	return time.Date(year, month, day, hour, minute, 0, 0, d.Location())
}

// DateRange returns a list of dates which are between the start and end
// inclusive. It's behavior is undefined if the end is before the start.
func DateRange(start, end time.Time) []time.Time {
	dates := make([]time.Time, 0, 31)
	for d := start; d.After(end) == false; d = d.AddDate(0, 0, 1) {
		dates = append(dates, d)
	}
	return dates
}

// FirstDayOfWeek returns the Sunday at the start of the week containing the
// provided date.
func FirstDayOfWeek(d time.Time) time.Time {
	return d.AddDate(0, 0, -1*int(d.Weekday()))
}

// LastDayOfWeek returns the Saturday at the end of the week containing the
// provided date.
func LastDayOfWeek(d time.Time) time.Time {
	return d.AddDate(0, 0, 6-int(d.Weekday()))
}

func FirstOfPreviousMonth(d time.Time) time.Time {
	previousMonth := d.AddDate(0, -1, 0)
	y, m, _ := previousMonth.Date()
	return time.Date(y, m, 1, 0, 0, 0, 0, d.Location())
}

func FirstOfNextMonth(d time.Time) time.Time {
	nextMonth := d.AddDate(0, 1, 0)
	y, m, _ := nextMonth.Date()
	return time.Date(y, m, 1, 0, 0, 0, 0, d.Location())
}

// SundayBeforeFirstOfMonth returns the first Sunday before the first of the
// month containing the provided date.
func SundayBeforeFirstOfMonth(d time.Time) time.Time {
	currentYear, currentMonth, _ := d.Date()
	firstDoM := time.Date(currentYear, currentMonth, 1, 0, 0, 0, 0, d.Location())
	return FirstDayOfWeek(firstDoM)
}

// SaturdayAfterEndOfMonth returns the first Saturday on or after the last day
// of the month containing the provided date.
func SaturdayAfterEndOfMonth(d time.Time) time.Time {
	currentYear, currentMonth, _ := d.Date()
	firstDoM := time.Date(currentYear, currentMonth, 1, 0, 0, 0, 0, d.Location())
	lastDoM := firstDoM.AddDate(0, 1, -1)
	return LastDayOfWeek(lastDoM)
}
