package conv

import (
	"testing"
	"time"

	"github.com/stretchr/testify/assert"
)

func TestDateRange(t *testing.T) {
	t.Parallel()

	testCases := []struct {
		name     string
		start    time.Time
		end      time.Time
		expected []time.Time
	}{
		{
			name:  "Single day",
			start: time.Date(2023, 4, 15, 0, 0, 0, 0, time.Local),
			end:   time.Date(2023, 4, 15, 0, 0, 0, 0, time.Local),
			expected: []time.Time{
				time.Date(2023, 4, 15, 0, 0, 0, 0, time.Local),
			},
		},
		{
			name:  "Two Days",
			start: time.Date(2023, 4, 15, 0, 0, 0, 0, time.Local),
			end:   time.Date(2023, 4, 16, 0, 0, 0, 0, time.Local),
			expected: []time.Time{
				time.Date(2023, 4, 15, 0, 0, 0, 0, time.Local),
				time.Date(2023, 4, 16, 0, 0, 0, 0, time.Local),
			},
		},
		{
			name:  "Eight Days",
			start: time.Date(2023, 4, 15, 0, 0, 0, 0, time.Local),
			end:   time.Date(2023, 4, 22, 0, 0, 0, 0, time.Local),
			expected: []time.Time{
				time.Date(2023, 4, 15, 0, 0, 0, 0, time.Local),
				time.Date(2023, 4, 16, 0, 0, 0, 0, time.Local),
				time.Date(2023, 4, 17, 0, 0, 0, 0, time.Local),
				time.Date(2023, 4, 18, 0, 0, 0, 0, time.Local),
				time.Date(2023, 4, 19, 0, 0, 0, 0, time.Local),
				time.Date(2023, 4, 20, 0, 0, 0, 0, time.Local),
				time.Date(2023, 4, 21, 0, 0, 0, 0, time.Local),
				time.Date(2023, 4, 22, 0, 0, 0, 0, time.Local),
			},
		},
	}

	for _, tc := range testCases {
		t.Run(tc.name, func(t *testing.T) {
			result := DateRange(tc.start, tc.end)
			assert.EqualValues(t, tc.expected, result)
		})
	}
}

func TestDateInRange(t *testing.T) {
	t.Parallel()

	testCases := []struct {
		name      string
		d         time.Time
		beginning time.Time
		end       time.Time
		expected  bool
	}{
		{
			name:      "Sunday of week in month middle",
			d:         time.Date(2023, 4, 16, 0, 0, 0, 0, time.Local),
			beginning: time.Date(2023, 4, 16, 0, 0, 0, 0, time.Local),
			end:       time.Date(2023, 4, 22, 0, 0, 0, 0, time.Local),
			expected:  true,
		},
		{
			name:      "Saturday of week in month middle",
			d:         time.Date(2023, 4, 22, 0, 0, 0, 0, time.Local),
			beginning: time.Date(2023, 4, 16, 0, 0, 0, 0, time.Local),
			end:       time.Date(2023, 4, 22, 0, 0, 0, 0, time.Local),
			expected:  true,
		},
		{
			name:      "Wednesday of week in month middle",
			d:         time.Date(2023, 4, 19, 0, 0, 0, 0, time.Local),
			beginning: time.Date(2023, 4, 16, 0, 0, 0, 0, time.Local),
			end:       time.Date(2023, 4, 22, 0, 0, 0, 0, time.Local),
			expected:  true,
		},
		{
			name:      "Day before window",
			d:         time.Date(2023, 4, 15, 0, 0, 0, 0, time.Local),
			beginning: time.Date(2023, 4, 16, 0, 0, 0, 0, time.Local),
			end:       time.Date(2023, 4, 22, 0, 0, 0, 0, time.Local),
			expected:  false,
		},
		{
			name:      "Day after window",
			d:         time.Date(2023, 4, 23, 0, 0, 0, 0, time.Local),
			beginning: time.Date(2023, 4, 16, 0, 0, 0, 0, time.Local),
			end:       time.Date(2023, 4, 22, 0, 0, 0, 0, time.Local),
			expected:  false,
		},
		{
			name:      "Window crosses year boundary",
			d:         time.Date(2024, 1, 4, 0, 0, 0, 0, time.Local),
			beginning: time.Date(2023, 12, 31, 0, 0, 0, 0, time.Local),
			end:       time.Date(2024, 1, 6, 0, 0, 0, 0, time.Local),
			expected:  true,
		},
	}

	for _, tc := range testCases {
		t.Run(tc.name, func(t *testing.T) {
			result := DateInRange(tc.d, tc.beginning, tc.end)
			assert.EqualValues(t, tc.expected, result)
		})
	}
}

func TestFirstDayOfWeek(t *testing.T) {
	t.Parallel()

	testCases := []struct {
		name     string
		input    time.Time
		expected time.Time
	}{
		{
			name:     "Saturday",
			input:    time.Date(2023, 4, 15, 0, 0, 0, 0, time.Local),
			expected: time.Date(2023, 4, 9, 0, 0, 0, 0, time.Local),
		},
		{
			name:     "Sunday",
			input:    time.Date(2023, 4, 9, 0, 0, 0, 0, time.Local),
			expected: time.Date(2023, 4, 9, 0, 0, 0, 0, time.Local),
		},
		{
			name:     "Wednesday",
			input:    time.Date(2023, 4, 19, 0, 0, 0, 0, time.Local),
			expected: time.Date(2023, 4, 16, 0, 0, 0, 0, time.Local),
		},
	}

	for _, tc := range testCases {
		t.Run(tc.name, func(t *testing.T) {
			result := FirstDayOfWeek(tc.input)
			assert.Equal(t, tc.expected, result)
		})
	}
}

func TestLastDayOfWeek(t *testing.T) {
	t.Parallel()

	testCases := []struct {
		name     string
		input    time.Time
		expected time.Time
	}{
		{
			name:     "Saturday",
			input:    time.Date(2023, 4, 15, 0, 0, 0, 0, time.Local),
			expected: time.Date(2023, 4, 15, 0, 0, 0, 0, time.Local),
		},
		{
			name:     "Sunday",
			input:    time.Date(2023, 4, 9, 0, 0, 0, 0, time.Local),
			expected: time.Date(2023, 4, 15, 0, 0, 0, 0, time.Local),
		},
		{
			name:     "Wednesday",
			input:    time.Date(2023, 4, 19, 0, 0, 0, 0, time.Local),
			expected: time.Date(2023, 4, 22, 0, 0, 0, 0, time.Local),
		},
	}

	for _, tc := range testCases {
		t.Run(tc.name, func(t *testing.T) {
			result := LastDayOfWeek(tc.input)
			assert.Equal(t, tc.expected, result)
		})
	}
}

func TestSundayBeforeFirstOfMonth(t *testing.T) {
	t.Parallel()

	testCases := []struct {
		name     string
		input    time.Time
		expected time.Time
	}{
		{
			name:     "First on Monday from the first",
			input:    time.Date(2023, 5, 1, 0, 0, 0, 0, time.Local),
			expected: time.Date(2023, 4, 30, 0, 0, 0, 0, time.Local),
		},
		{
			name:     "First on Monday from the 10th",
			input:    time.Date(2023, 5, 10, 0, 0, 0, 0, time.Local),
			expected: time.Date(2023, 4, 30, 0, 0, 0, 0, time.Local),
		},
		{
			name:     "First on Monday from the last day",
			input:    time.Date(2023, 5, 31, 0, 0, 0, 0, time.Local),
			expected: time.Date(2023, 4, 30, 0, 0, 0, 0, time.Local),
		},
		{
			name:     "First on Sunday from the first",
			input:    time.Date(2023, 1, 1, 0, 0, 0, 0, time.Local),
			expected: time.Date(2023, 1, 1, 0, 0, 0, 0, time.Local),
		},
		{
			name:     "First on Sunday from the 10th",
			input:    time.Date(2023, 1, 10, 0, 0, 0, 0, time.Local),
			expected: time.Date(2023, 1, 1, 0, 0, 0, 0, time.Local),
		},
		{
			name:     "First on Sunday from the last day",
			input:    time.Date(2023, 1, 31, 0, 0, 0, 0, time.Local),
			expected: time.Date(2023, 1, 1, 0, 0, 0, 0, time.Local),
		},
		{
			name:     "First on Saturday from the first",
			input:    time.Date(2023, 4, 1, 0, 0, 0, 0, time.Local),
			expected: time.Date(2023, 3, 26, 0, 0, 0, 0, time.Local),
		},
		{
			name:     "First on Saturday from the 10th",
			input:    time.Date(2023, 4, 10, 0, 0, 0, 0, time.Local),
			expected: time.Date(2023, 3, 26, 0, 0, 0, 0, time.Local),
		},
	}

	for _, tc := range testCases {
		t.Run(tc.name, func(t *testing.T) {
			result := SundayBeforeFirstOfMonth(tc.input)
			assert.Equal(t, tc.expected, result)
		})
	}
}

func TestSaturdayAfterEndOfMonth(t *testing.T) {
	t.Parallel()

	testCases := []struct {
		name     string
		input    time.Time
		expected time.Time
	}{
		{
			name:     "Last on Friday from the last",
			input:    time.Date(2023, 3, 31, 0, 0, 0, 0, time.Local),
			expected: time.Date(2023, 4, 1, 0, 0, 0, 0, time.Local),
		},
		{
			name:     "Last on Friday from the first",
			input:    time.Date(2023, 3, 1, 0, 0, 0, 0, time.Local),
			expected: time.Date(2023, 4, 1, 0, 0, 0, 0, time.Local),
		},
		{
			name:     "Last on Saturday from the last",
			input:    time.Date(2023, 9, 30, 0, 0, 0, 0, time.Local),
			expected: time.Date(2023, 9, 30, 0, 0, 0, 0, time.Local),
		},
		{
			name:     "Last on Saturday from the first",
			input:    time.Date(2023, 9, 1, 0, 0, 0, 0, time.Local),
			expected: time.Date(2023, 9, 30, 0, 0, 0, 0, time.Local),
		},
	}

	for _, tc := range testCases {
		t.Run(tc.name, func(t *testing.T) {
			result := SaturdayAfterEndOfMonth(tc.input)
			assert.Equal(t, tc.expected, result)
		})
	}
}
