package cron

import (
	"context"
	"fmt"
	"time"

	"go.uber.org/zap"

	"git.grosinger.net/tgrosinger/tonyslist/constants"
	"git.grosinger.net/tgrosinger/tonyslist/constants/pages"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/conv"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/services"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/services/mail"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/services/mail/types"
)

const (
	twiceWeeklyDBKey = "twice-weekly-digest"
)

func sendDigestEmails(c *services.Container) {
	// For now, only send digest emails on Saturday and Wednesday at 7:00 Pacific
	now := time.Now().In(c.Config.App.Loc)

	c.Logger.Debug("Checking if it's 7:00 on a Saturday or Wednesday")

	if now.Hour() != 7 {
		return
	}

	switch now.Weekday() {
	case time.Tuesday, time.Saturday:
		sendTwiceWeeklyDigestEmail(c)
	default:
		return
	}
}

// sendTwiceWeeklyDigestEmail assumes that it is being run at 7:00am local time
// on either Saturday or Wednesday. It will send a link to all listed posts
// which were published since the last time this digest was sent.
func sendTwiceWeeklyDigestEmail(c *services.Container) {
	c.Logger.Info("Sending the twice weekly digest email")

	times, err := getLastRunTime(c.Logger, c.DB, twiceWeeklyDBKey)
	if err != nil {
		c.Logger.Warn("cron failed to retrieve last run time for twice weekly digest",
			zap.Error(err))
		return
	}

	subscribers, err := c.DB.C.GetUserEmailTwiceWeeklyDigestSubscription(context.Background())
	if err != nil {
		c.Logger.Warn("cron failed to retrieve a list of users subscribed to the twice weekly digest",
			zap.Error(err))
		return
	}
	if len(subscribers) == 0 {
		c.Logger.Info("No email addresses subscribed to the twice weekly digest email")
		return
	}

	posts, err := c.DB.C.GetPublishedPosts(context.Background())
	if err != nil {
		c.Logger.Warn("cron failed to retrieve published posts for twice weekly digest",
			zap.Error(err))
		return
	}

	newPosts := make([]sqlc.GetPublishedPostsRow, 0)
	for _, post := range posts {
		if !post.PublishedAt.Valid {
			continue
		}

		// In theory this runs twice a week and posts last a full week, so we
		// don't need to check if the post has expired.

		if post.PublishedAt.Time.After(times.lastRun) {
			newPosts = append(newPosts, post)
		}
	}

	today := conv.StartOfDay(time.Now())
	events, err := c.DB.C.GetUpcomingEventsInRange(context.Background(),
		sqlc.GetUpcomingEventsInRangeParams{
			Date:   today,
			Date_2: today.AddDate(0, 0, 8),
		})
	if err != nil {
		c.Logger.Warn("cron failed to retrieve a list of upcoming events")
	}

	if len(newPosts) == 0 && len(events) == 0 {
		c.Logger.Info("cron skipping twice weekly digest due to no new posts or events")
		return
	}

	toSend := make([]*types.Mail, 0, len(subscribers))

	for _, sub := range subscribers {
		unsubscribeToken, err := c.Auth.GenerateJWTAuthToken(map[string]interface{}{
			"user_id": sub.ID,
			"action":  "digest",
		}, nil) // Does not expire
		if err != nil {
			c.Logger.Warn("cron failed to generate unsubscribe token for twice weekly digest",
				zap.Error(err))
			continue
		}
		unsubscribeHref := fmt.Sprintf("%s%s?token=%s",
			c.Config.App.PublicURL,
			c.Web.Reverse(pages.Names.User.Unsubscribe),
			unsubscribeToken)

		templateData := map[string]interface{}{
			"new_event_url":    c.Config.App.PublicURL + c.Web.Reverse(pages.Names.Event.New),
			"events":           processEvents(c, events),
			"posts":            groupPosts(c, newPosts),
			"profile":          c.Config.App.PublicURL + c.Web.Reverse(pages.Names.User.Account),
			"homepage_href":    c.Config.App.PublicURL,
			"unsubscribe_href": unsubscribeHref,
		}

		m, err := c.Mail.Template(mail.TemplateArgs{
			MessageStream: types.TwiceWeeklyDigest,
			To:            sub.Email,
			From:          c.Config.Mail.DigestsAddress,
			Subject:       c.Config.App.Name + " New Listings Digest",
			HTMLTemplate:  "twice_weekly_digest_html",
			TextTemplate:  "twice_weekly_digest_text",
			TemplateData:  templateData,
			ReplyTo:       c.Config.Mail.SupportAddress,
			Headers: []types.Header{
				{
					Name:  "List-Unsubscribe",
					Value: fmt.Sprintf("<%s>", unsubscribeHref),
				},
			},
		})
		if err != nil {
			c.Logger.Warn("cron failed to template twice weekly digest email", zap.Error(err))
			continue
		}

		toSend = append(toSend, m)
	}

	c.Mail.Send(c.Logger, toSend, nil)
}

type classifieds struct {
	ForSale      map[string]string
	ForFree      map[string]string
	WantToBuy    map[string]string
	WantToBorrow map[string]string
}

type housing struct {
	Requests map[string]string
	ForRent  map[string]string
}

type groupedPosts struct {
	Classifieds classifieds
	Housing     housing
}

func groupPosts(c *services.Container, posts []sqlc.GetPublishedPostsRow) groupedPosts {
	baseURL := c.Config.App.PublicURL

	toReturn := groupedPosts{
		Classifieds: classifieds{
			ForSale:      make(map[string]string),
			ForFree:      make(map[string]string),
			WantToBuy:    make(map[string]string),
			WantToBorrow: make(map[string]string),
		},
		Housing: housing{
			Requests: make(map[string]string),
			ForRent:  make(map[string]string),
		},
	}

	for _, post := range posts {
		switch post.PostType {
		case constants.Classified:
			url := baseURL + c.Web.Reverse(pages.Names.Classified.One, post.ID)
			switch post.ClassifiedType {
			case constants.ForFree:
				toReturn.Classifieds.ForFree[post.Title] = url
			case constants.ForSale:
				toReturn.Classifieds.ForSale[post.Title] = url
			case constants.WantToBuy:
				toReturn.Classifieds.WantToBuy[post.Title] = url
			case constants.WantToBorrow:
				toReturn.Classifieds.WantToBorrow[post.Title] = url
			}
		case constants.Housing:
			url := baseURL + c.Web.Reverse(pages.Names.Housing.One, post.ID)
			switch post.HousingType {
			case constants.ForRent:
				toReturn.Housing.ForRent[post.Title] = url
			case constants.Request:
				toReturn.Housing.Requests[post.Title] = url
			}
		}
	}

	return toReturn
}

type Event struct {
	Title string
	URL   string
}

// processEvents returns a list of Event objects ready for including in the
// newsletter. Unlike the classified posts, events are returned as a list of
// structs to ensure their order is maintained.
func processEvents(c *services.Container, events []sqlc.Event) []Event {
	baseURL := c.Config.App.PublicURL

	toReturn := make([]Event, 0, len(events))
	for _, event := range events {
		toReturn = append(toReturn, Event{
			Title: event.Title,
			URL:   baseURL + c.Web.Reverse(pages.Names.Event.One, event.ID),
		})
	}
	return toReturn
}
