package cron

import (
	"context"
	"database/sql"
	"strings"
	"testing"
	"time"

	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/mock"
	"github.com/stretchr/testify/require"
	"go.uber.org/zap"

	"git.grosinger.net/tgrosinger/tonyslist/constants"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/classified"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
	mailmock "git.grosinger.net/tgrosinger/tonyslist/pkg/services/mail/mock"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/services/mail/types"
)

func TestGetUserEmailTwiceWeeklyDigestSubscription(t *testing.T) {
	// Create a couple more users
	usr2, err := db.User.CreateMockUser()
	require.NoError(t, err)

	usr3, err := db.User.CreateMockUser()
	require.NoError(t, err)

	// Input some settings
	err = db.C.SetUserSetting(context.Background(), sqlc.SetUserSettingParams{
		UserID: usr.ID,
		Key:    string(constants.EmailDigestSubscription),
		Value:  sql.NullString{Valid: true, String: constants.EmailDigestSubscriptionTwiceWeekly},
	})
	require.NoError(t, err)

	err = db.C.SetUserSetting(context.Background(), sqlc.SetUserSettingParams{
		UserID: usr2.ID,
		Key:    string(constants.EmailDigestSubscription),
		Value:  sql.NullString{Valid: true, String: constants.EmailDigestSubscriptionUnsubscribed},
	})
	require.NoError(t, err)

	addresses, err := db.C.GetUserEmailTwiceWeeklyDigestSubscription(context.Background())
	require.NoError(t, err)

	// Assert that the user with no config and the one that is subscribed are returned
	expected := []sqlc.GetUserEmailTwiceWeeklyDigestSubscriptionRow{
		{
			ID:    usr.ID,
			Email: usr.Email,
		},
		{
			ID:    usr3.ID,
			Email: usr3.Email,
		},
	}

	assert.ElementsMatch(t, expected, addresses)
}

func TestSendTwiceWeeklyDigestEmail(t *testing.T) {
	// Not parallel safe because we need the email to only have the post we created.

	// Reset the users in the database so that we only get an email sent for this one user.
	err := c.DB.TruncateUsers()
	require.NoError(t, err)
	usr, err = db.User.CreateMockUser()
	require.NoError(t, err)

	// reset the last run time
	_, err = getLastRunTime(zap.NewNop(), db, twiceWeeklyDBKey)
	require.NoError(t, err)

	// Sleep long enough for the post we're going to be creating to be
	// considered before the lastRunTime.
	time.Sleep(time.Second * 1)

	// create a new post
	postID, err := c.DB.Classified.Create(context.Background(),
		classified.NewPostInput{
			CreatePostParams: sqlc.CreatePostParams{
				PostType:       constants.Classified,
				Title:          "Testing",
				Description:    "Testing",
				Creator:        usr.ID,
				AnonymousEmail: sql.NullString{Valid: true, String: "anonymous"},
			},
			ClassifiedType: constants.ForFree,
		})
	require.NoError(t, err)
	err = c.DB.C.SetPostPublishedNow(context.Background(), postID)
	require.NoError(t, err)

	// setup mock mail
	providerMock := mailmock.New()
	c.Mail.SetProvider(providerMock)
	providerMock.On("Send", mock.Anything).Return([]types.Response{}, nil)

	// send emails
	sendTwiceWeeklyDigestEmail(c)

	// Wait for the worker thread to handle each message that needs to be sent
	time.Sleep(time.Millisecond * 200)

	expectedTextBody := "Hello! Here's your twice-weekly digest of newly posted items on OrcasHub.\n\nUpcoming Events\n\nThere are no events coming up, but you can create one.\n    https://example.com\n\nFor Free:\n- Testing (https://example.com)\n\nOrcasHub homepage:\n\n    https://example.com\n\nUnsubscribe from digest emails:\n\n    https://example.com?token=eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJhY3Rpb24iOiJkaWdlc3QiLCJ1c2VyX2lkIjo0fQ.7bABZD_P8f9DCgHcx30RIf8J8CpxOAaK7Vt4ChKn7Wg\n"
	expectedHTMLBody := "<p>\n    Hello! Here's your twice-weekly digest of newly posted items on OrcasHub.\n</p>\n\n<h2>Upcoming Events</h2>\n\n    <p>There are no events coming up, but you can <a href=\"https://example.com\">create one here</a>.</p>\n\n    <h2>For Free</h2>\n    <ul>\n    \n        <li><a href=\"https://example.com\">Testing</a></li>\n    \n    </ul>\n\n\n<p>\n    <a href=\"https://example.com\">Click here to open the OrcasHub homepage.</a>\n</p>\n\n<p>\n    <a href=\"https://example.com?token=eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJhY3Rpb24iOiJkaWdlc3QiLCJ1c2VyX2lkIjo0fQ.7bABZD_P8f9DCgHcx30RIf8J8CpxOAaK7Vt4ChKn7Wg\">Click here to unsubscribe.</a>\n</p>\n"

	// verify emails sent
	calls := providerMock.Calls
	require.Len(t, calls, 1)
	sentMail := calls[0].Arguments[0].([]*types.Mail)
	assert.Len(t, sentMail, 1)
	assert.Equal(t, c.Config.Mail.FromAddress, sentMail[0].From)
	assert.Equal(t, usr.Email, sentMail[0].To)
	assert.Equal(t, types.TwiceWeeklyDigest, sentMail[0].MessageStream)
	assert.Equal(t, c.Config.App.Name+" New Listings Digest", sentMail[0].Subject)
	assert.Equal(t, expectedTextBody, sentMail[0].TextBody)
	assert.Equal(t, expectedHTMLBody, sentMail[0].HTMLBody)

	headers := sentMail[0].Headers
	assert.Len(t, headers, 1)
	assert.Equal(t, "List-Unsubscribe", headers[0].Name)
	assert.True(t, strings.HasPrefix(headers[0].Value, "<https://example.com?token="))
	assert.True(t, strings.HasSuffix(headers[0].Value, ">"))
}
