package cron

import (
	"context"
	"database/sql"
	"testing"
	"time"

	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"

	"git.grosinger.net/tgrosinger/tonyslist/constants"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/services"
)

func createPost(t *testing.T, title string) int {
	t.Helper()

	id, err := db.C.CreatePost(context.Background(), sqlc.CreatePostParams{
		PostType:    constants.Classified,
		Title:       title,
		Description: title + " description",
		Creator:     usr.ID,
	})
	require.NoError(t, err)

	return int(id)
}

func TestDeleteExpiredPosts(t *testing.T) {
	unlistedNewPost := createPost(t, "post1")
	unlistedOldPost := createPost(t, "post2")
	unpublishedPost := createPost(t, "post3")
	listedOldExpiredPost := createPost(t, "post4")
	listedNewExpiredPost := createPost(t, "post5")
	listedUnexpiredPost := createPost(t, "post6")

	now := time.Now()
	yesterday := now.AddDate(0, 0, -1)
	expiredTime := now.Add(-1 * constants.ClassifiedsDuration)
	oneMonthFiveDays := expiredTime.AddDate(0, -1, -5)
	threeMonthsFiveDays := expiredTime.AddDate(0, -3, -5)

	err := db.C.SetPostUnlisted(context.Background(), unlistedNewPost)
	require.NoError(t, err)
	err = db.C.SetPostPublishedAt(context.Background(), sqlc.SetPostPublishedAtParams{
		ID:          unlistedNewPost,
		PublishedAt: sql.NullTime{Time: now, Valid: true},
	})
	require.NoError(t, err)

	err = db.C.SetPostUnlisted(context.Background(), unlistedOldPost)
	require.NoError(t, err)
	err = db.C.SetPostPublishedAt(context.Background(), sqlc.SetPostPublishedAtParams{
		ID:          unlistedOldPost,
		PublishedAt: sql.NullTime{Time: oneMonthFiveDays, Valid: true},
	})
	require.NoError(t, err)

	err = db.C.SetPostPublishedAt(context.Background(), sqlc.SetPostPublishedAtParams{
		ID:          unpublishedPost,
		PublishedAt: sql.NullTime{Valid: false},
	})
	require.NoError(t, err)

	err = db.C.SetPostPublishedAt(context.Background(), sqlc.SetPostPublishedAtParams{
		ID:          listedOldExpiredPost,
		PublishedAt: sql.NullTime{Time: threeMonthsFiveDays, Valid: true},
	})
	require.NoError(t, err)

	err = db.C.SetPostPublishedAt(context.Background(), sqlc.SetPostPublishedAtParams{
		ID:          listedNewExpiredPost,
		PublishedAt: sql.NullTime{Time: oneMonthFiveDays, Valid: true},
	})
	require.NoError(t, err)

	err = db.C.SetPostPublishedAt(context.Background(), sqlc.SetPostPublishedAtParams{
		ID:          listedUnexpiredPost,
		PublishedAt: sql.NullTime{Time: yesterday, Valid: true},
	})
	require.NoError(t, err)

	c := &services.Container{
		Logger: logger,
		DB:     db,
	}
	deleteExpiredPosts(c)

	ids, err := db.C.GetPostIDs(context.Background())
	require.NoError(t, err)

	assert.Contains(t, ids, unlistedNewPost)
	assert.Contains(t, ids, unpublishedPost)
	assert.Contains(t, ids, listedNewExpiredPost)
	assert.Contains(t, ids, listedUnexpiredPost)

	assert.NotContains(t, ids, unlistedOldPost)
	assert.NotContains(t, ids, listedOldExpiredPost)
}
