package cron

import (
	"context"
	"fmt"

	"go.uber.org/zap"

	"git.grosinger.net/tgrosinger/tonyslist/constants"
	"git.grosinger.net/tgrosinger/tonyslist/constants/pages"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/services"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/services/mail"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/services/mail/types"
)

const (
	expiredPostDBKey = "expired-post-notification"
)

// emailExpiredPostOwners finds posts which have expired since the last
// iteration and sends them an email that their post has expired.
//
// TODO: Ideally this would be filtering the posts in the database so we only
// need to iterate though the ones which have expired since we last checked,
// however sqlite is not handing the times correctly and does not return results
// which should be in the time window.
func emailExpiredPostOwners(c *services.Container) {
	expiredPosts, err := getExpiredPosts(c.DB, c.Logger)
	if err != nil {
		c.Logger.Warn("cron failed to retrieve recently expired posts", zap.Error(err))
		return
	}

	toSend := make([]*types.Mail, 0, len(expiredPosts))

	for _, post := range expiredPosts {
		c.Logger.Info("Notifying post owner about expiration", zap.Int("post_id", post.ID))

		viewHref, err := generateAuthedLinkForPost(c, post.Email, post.Creator, post.ID, post.PostType)
		if err != nil {
			c.Logger.Warn("cron failed to create an authed link for an expired post",
				zap.Error(err))
			continue
		}

		templateData := map[string]string{
			"app_name":      c.Config.App.Name,
			"domain":        c.Config.App.PublicURL,
			"support_email": c.Config.Mail.SupportAddress,
			"title":         post.Title,
			"view_href":     viewHref,
		}

		m, err := c.Mail.Template(mail.TemplateArgs{
			MessageStream: types.AccountManagement,
			To:            post.Email,
			Subject:       fmt.Sprintf("%s Expired Post", c.Config.App.Name),
			HTMLTemplate:  "expired_post_html",
			TextTemplate:  "expired_post_text",
			TemplateData:  templateData,
		})
		if err != nil {
			c.Logger.Warn("cron failed to templae an expired post notification",
				zap.Int("post_id", post.ID),
				zap.Error(err))
		}
		toSend = append(toSend, m)
	}

	c.Mail.Send(c.Logger, toSend, nil)
}

// getExpiredPosts returns a list of posts which have expired in the last hour.
func getExpiredPosts(db *models.Client, logger *zap.Logger) ([]sqlc.GetPublishedPostsRow, error) {
	times, err := getLastRunTime(logger, db, expiredPostDBKey)
	if err != nil {
		return nil, err
	}

	afterTime := times.lastRun.Add(-1 * constants.ClassifiedsDuration)
	beforeTime := times.now.Add(-1 * constants.ClassifiedsDuration)

	logger.Info("Looking for recently expired posts to notify",
		zap.Time("after", afterTime),
		zap.Time("before", beforeTime))

	posts, err := db.C.GetPublishedPosts(context.Background())
	if err != nil {
		return nil, err
	}

	toReturn := make([]sqlc.GetPublishedPostsRow, 0)
	for _, post := range posts {
		if !post.PublishedAt.Valid {
			continue
		}

		publishedAt := post.PublishedAt.Time
		if publishedAt.After(afterTime) && publishedAt.Before(beforeTime) {
			toReturn = append(toReturn, post)
		}
	}

	return toReturn, nil
}

func generateAuthedLinkForPost(c *services.Container, userEmail string, userID, postID int,
	postType constants.PostType,
) (string, error) {
	token, err := c.Auth.GenerateJWTEmailAuthToken(userEmail)
	if err != nil {
		return "", err
	}

	var dest string
	switch postType {
	case constants.Classified:
		dest = c.Web.Reverse(pages.Names.Classified.One, postID)
	case constants.Housing:
		dest = c.Web.Reverse(pages.Names.Housing.One, postID)
	}

	return fmt.Sprintf("%s?return_to=%s",
		c.Web.Reverse(pages.Names.User.LoginToken, userID, token), dest), nil
}
