package cron

import (
	"context"
	"database/sql"
	"testing"
	"time"

	"github.com/stretchr/testify/require"
	"go.uber.org/zap"

	"git.grosinger.net/tgrosinger/tonyslist/constants"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/conv"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/classified"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
)

func TestGetExpiredPosts(t *testing.T) {
	// Create a post with all fields
	id, err := db.Classified.Create(context.Background(), classified.NewPostInput{
		CreatePostParams: sqlc.CreatePostParams{
			PostType:    constants.Classified,
			Title:       "test title",
			Description: "test description",
			PhoneCall:   conv.StringToNullString("123-456-7890"),
			PhoneText:   conv.StringToNullString("098-765-4321"),
			Creator:     usr.ID,
		},
		ClassifiedType: constants.WantToBuy,
		Price:          sql.NullFloat64{Valid: true, Float64: 10.25},
	})
	require.NoError(t, err)

	// Set last run to 1 hour ago
	db.C.SetCronLastRun(context.Background(), sqlc.SetCronLastRunParams{
		Name:    expiredPostDBKey,
		LastRun: time.Now().Add(-1 * time.Hour),
	})

	// Set to have expired 30 minutes ago
	db.C.SetPostPublishedAt(context.Background(), sqlc.SetPostPublishedAtParams{
		PublishedAt: sql.NullTime{
			Valid: true,
			Time:  time.Now().Add((-1 * constants.ClassifiedsDuration) - 30*time.Minute),
		},
		ID: id,
	})

	posts, err := getExpiredPosts(db, zap.NewNop())
	require.NoError(t, err)
	require.Len(t, posts, 1)
	require.Equal(t, id, posts[0].ID)

	// Set last run to 1 hour ago
	db.C.SetCronLastRun(context.Background(), sqlc.SetCronLastRunParams{
		Name:    expiredPostDBKey,
		LastRun: time.Now().Add(-1 * time.Hour),
	})

	// Set to have expired 90 minutes ago
	db.C.SetPostPublishedAt(context.Background(), sqlc.SetPostPublishedAtParams{
		PublishedAt: sql.NullTime{
			Valid: true,
			Time:  time.Now().Add((-1 * constants.ClassifiedsDuration) - 90*time.Minute),
		},
		ID: id,
	})

	posts, err = getExpiredPosts(db, zap.NewNop())
	require.NoError(t, err)
	require.Len(t, posts, 0)

	// Set last run to 1 hour ago
	db.C.SetCronLastRun(context.Background(), sqlc.SetCronLastRunParams{
		Name:    expiredPostDBKey,
		LastRun: time.Now().Add(-1 * time.Hour),
	})

	// Set to expire in 30 minutes
	db.C.SetPostPublishedAt(context.Background(), sqlc.SetPostPublishedAtParams{
		PublishedAt: sql.NullTime{
			Valid: true,
			Time:  time.Now().Add((-1 * constants.ClassifiedsDuration) + 30*time.Minute),
		},
		ID: id,
	})

	posts, err = getExpiredPosts(db, zap.NewNop())
	require.NoError(t, err)
	require.Len(t, posts, 0)
}
