package middleware

import (
	"database/sql"
	"fmt"
	"net/http"

	"github.com/casbin/casbin/v2"
	"github.com/labstack/echo/v4"
	"go.uber.org/zap"

	"git.grosinger.net/tgrosinger/tonyslist/constants/pages"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/auth"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/context"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/msg"
)

// LoadAuthenticatedUser loads the authenticated user, if one, and stores in context
func LoadAuthenticatedUser(authClient *auth.Client, db *models.Client) echo.MiddlewareFunc {
	return func(next echo.HandlerFunc) echo.HandlerFunc {
		return func(c echo.Context) error {
			u, err := authClient.GetAuthenticatedUser(c)
			if err == sql.ErrNoRows {
				context.GetLogger(c).Warn("auth user not found")
				return next(c)
			} else if err == auth.ErrNotAuthenticated {
				return next(c)
			} else if err != nil {
				return echo.NewHTTPError(
					http.StatusInternalServerError,
					fmt.Sprintf("error querying for authenticated user: %v", err))
			}

			c.Set(context.AuthenticatedUserKey, u)
			context.GetLogger(c).Info("auth user loaded in to context",
				zap.Int("user_id", u.ID))

			groups, err := db.C.GetGroupsForUserID(c.Request().Context(), u.ID)
			if err != nil {
				return echo.NewHTTPError(
					http.StatusInternalServerError,
					fmt.Sprintf("error querying for user groups: %v", err))
			}

			c.Set(context.AuthenticatedUserGroupsKey, groups)

			return next(c)
		}
	}
}

// AuthenticateUserViaToken Authenticates a user by a JWT token in the URL,
// then logs them in.
func AuthenticateUserViaToken(authClient *auth.Client, db *models.Client) echo.MiddlewareFunc {
	return func(next echo.HandlerFunc) echo.HandlerFunc {
		return func(c echo.Context) error {
			// If the user is already logged in, then log out.
			err := authClient.Logout(c)
			if err != nil {
				return &echo.HTTPError{
					Code:     http.StatusInternalServerError,
					Message:  " Internal Server Error",
					Internal: fmt.Errorf("unable to log out before authenticating with token %w", err),
				}
			}

			// Extract the user parameter
			if c.Get(context.UserKey) == nil {
				return echo.NewHTTPError(http.StatusInternalServerError)
			}
			u := c.Get(context.UserKey).(sqlc.User)

			token := c.Param("token")
			tokenEmail, err := authClient.ValidateJWTAuthToken(token)
			if err != nil {
				context.GetLogger(c).Info("Failed to validate jwt token", zap.Error(err))
				msg.Warning(c, "The link is either invalid or has expired. Please request a new one.")
				return c.Redirect(http.StatusFound, c.Echo().Reverse(pages.Names.User.Login))
			}

			if tokenEmail != u.Email {
				context.GetLogger(c).Info("JWT token email is not for the specified user", zap.Error(err))
				msg.Warning(c, "The link is either invalid or has expired. Please request a new one.")
				return c.Redirect(http.StatusFound, c.Echo().Reverse(pages.Names.User.Login))
			}

			logger := context.GetLogger(c).With(zap.Int("user_id", u.ID))

			if !u.Verified {
				err = db.C.SetUserVerified(c.Request().Context(), u.ID)
				if err != nil {
					logger.Warn("Unable to set user as verified", zap.Error(err))
				}
			}

			err = authClient.Login(c, u.ID)
			if err != nil {
				return echo.NewHTTPError(
					http.StatusInternalServerError,
					fmt.Sprintf("unable to log in user: %v", err),
				)
			}

			// Success. Log the user in.
			msg.Success(c, "You have been logged in!")

			err = db.C.IncrementLoginCount(c.Request().Context(), u.ID)
			if err != nil {
				logger.Warn("unable to increment login count",
					zap.Error(err))
			}

			c.Set(context.AuthenticatedUserKey, u)
			logger.Info("auth user loaded in to context via jwt token")

			groups, err := db.C.GetGroupsForUserID(c.Request().Context(), u.ID)
			if err != nil {
				return &echo.HTTPError{
					Code:     http.StatusInternalServerError,
					Message:  " Internal Server Error",
					Internal: fmt.Errorf("error querying for user groups %w", err),
				}
			}

			c.Set(context.AuthenticatedUserGroupsKey, groups)

			return next(c)
		}
	}
}

// RequireAuthentication requires that the user be authenticated in order to proceed
func RequireAuthentication() echo.MiddlewareFunc {
	return func(next echo.HandlerFunc) echo.HandlerFunc {
		return func(c echo.Context) error {
			if u := c.Get(context.AuthenticatedUserKey); u == nil {
				msg.Warning(c, "Please sign in to perform that action.")

				path := c.Path()
				if path != "" && path != "/" {
					// Preserve any present query params, and add one for returning to the desired page.
					values := c.Request().URL.Query()
					values.Add("return_to", c.Request().URL.Path)
					return c.Redirect(http.StatusFound, "/user/login?"+values.Encode())
				}
				return c.Redirect(http.StatusFound, "/user/login")
			}

			return next(c)
		}
	}
}

// RequireNoAuthentication requires that the user not be authenticated in order to proceed
func RequireNoAuthentication() echo.MiddlewareFunc {
	return func(next echo.HandlerFunc) echo.HandlerFunc {
		return func(c echo.Context) error {
			if u := c.Get(context.AuthenticatedUserKey); u != nil {
				return echo.NewHTTPError(http.StatusForbidden)
			}

			return next(c)
		}
	}
}

// RequireAdmin requires that the user be authenticated and have the
// necessary role to be a full admin in order to proceed.
func RequireAdmin(db *models.Client, enforcer *casbin.Enforcer) echo.MiddlewareFunc {
	return func(next echo.HandlerFunc) echo.HandlerFunc {
		return func(c echo.Context) error {
			enforceErr := enforceUser(c, enforcer, "site", "admin")
			if enforceErr != nil {
				return enforceErr
			}

			return next(c)
		}
	}
}

// RequirePostModerator requires that the user be authenticated and have the
// necessary role to moderate posts in order to proceed.
func RequirePostModerator(db *models.Client, enforcer *casbin.Enforcer) echo.MiddlewareFunc {
	return func(next echo.HandlerFunc) echo.HandlerFunc {
		return func(c echo.Context) error {
			enforceErr := enforceUser(c, enforcer, "post", "moderate")
			if enforceErr != nil {
				return enforceErr
			}

			return next(c)
		}
	}
}

// RecordAuthenticatedPageLoad increments the counter on the user record in the database
// which records how many pages this user has loaded.
func RecordAuthenticatedPageLoad(db *models.Client) echo.MiddlewareFunc {
	return func(next echo.HandlerFunc) echo.HandlerFunc {
		return func(c echo.Context) error {
			if u := c.Get(context.AuthenticatedUserKey); u != nil {
				userID := u.(sqlc.User).ID
				err := db.C.IncrementPageLoadCount(c.Request().Context(), userID)
				if err != nil {
					logger := context.GetLogger(c).With(zap.Int("user_id", userID))
					logger.Warn("unable to increment login count",
						zap.Error(err))
				}
			}

			return next(c)
		}
	}
}

// enforceUser checks if the user in the context has the permission to perform
// the specified action on the specified resource. Returns nil if allowed, or an
// error if disallowed.
func enforceUser(c echo.Context, enforcer *casbin.Enforcer, resource, action string) error {
	groups, ok := c.Get(context.AuthenticatedUserGroupsKey).([]string)
	if !ok {
		return echo.NewHTTPError(http.StatusForbidden)
	}

	for _, g := range groups {
		ok, err := enforcer.Enforce(g, resource, action)
		if err != nil {
			context.GetLogger(c).
				Warn("unable to enforce post moderator requirement", zap.Error(err))
			return echo.NewHTTPError(http.StatusInternalServerError)
		}

		if ok {
			return nil
		}
	}

	return echo.NewHTTPError(http.StatusForbidden)
}
