package middleware

import (
	"net/http"
	"strconv"
	"testing"

	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"

	"git.grosinger.net/tgrosinger/tonyslist/pkg/context"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/tests"
)

func TestLoadAuthenticatedUser(t *testing.T) {
	ctx, _ := tests.NewContext(c.Web, "/")
	tests.InitSession(ctx)
	mw := LoadAuthenticatedUser(c.Auth, c.DB)

	// Not authenticated
	_ = tests.ExecuteMiddleware(ctx, mw)
	assert.Nil(t, ctx.Get(context.AuthenticatedUserKey))

	// Login
	err := c.Auth.Login(ctx, usr.ID)
	require.NoError(t, err)

	// Verify the midldeware returns the authenticated user
	_ = tests.ExecuteMiddleware(ctx, mw)
	require.NotNil(t, ctx.Get(context.AuthenticatedUserKey))
	ctxUsr, ok := ctx.Get(context.AuthenticatedUserKey).(sqlc.User)
	require.True(t, ok)
	assert.Equal(t, usr.ID, ctxUsr.ID)
}

func TestRequireAuthentication(t *testing.T) {
	ctx, rec := tests.NewContext(c.Web, "/")
	tests.InitSession(ctx)

	// Not logged in
	err := tests.ExecuteMiddleware(ctx, RequireAuthentication())
	assert.NoError(t, err)
	assert.Equal(t, http.StatusFound, rec.Code)
	assert.Equal(t, "/user/login", rec.Result().Header.Get("Location"))

	// Login
	err = c.Auth.Login(ctx, usr.ID)
	require.NoError(t, err)
	_ = tests.ExecuteMiddleware(ctx, LoadAuthenticatedUser(c.Auth, c.DB))

	// Logged in
	err = tests.ExecuteMiddleware(ctx, RequireAuthentication())
	assert.Nil(t, err)
}

func TestRequireNoAuthentication(t *testing.T) {
	ctx, _ := tests.NewContext(c.Web, "/")
	tests.InitSession(ctx)

	// Not logged in
	err := tests.ExecuteMiddleware(ctx, RequireNoAuthentication())
	assert.Nil(t, err)

	// Login
	err = c.Auth.Login(ctx, usr.ID)
	require.NoError(t, err)
	_ = tests.ExecuteMiddleware(ctx, LoadAuthenticatedUser(c.Auth, c.DB))

	// Logged in
	err = tests.ExecuteMiddleware(ctx, RequireNoAuthentication())
	tests.AssertHTTPErrorCode(t, err, http.StatusForbidden)
}

func TestAuthenticateUserViaToken(t *testing.T) {
	ctx, recorder := tests.NewContext(c.Web, "/")
	tests.InitSession(ctx)

	// All values invalid
	ctx.SetParamNames("user", "token")
	ctx.SetParamValues("88", "bad")
	err := tests.ExecuteMiddleware(ctx, AuthenticateUserViaToken(c.Auth, c.DB))
	require.Error(t, err)

	// User is valid, but token invalid
	ctx.SetParamNames("user", "token")
	ctx.SetParamValues(strconv.Itoa(usr.ID), "bad")
	err = tests.ExecuteMiddleware(ctx, LoadUser(c.DB))
	require.NoError(t, err)
	err = tests.ExecuteMiddleware(ctx, AuthenticateUserViaToken(c.Auth, c.DB))
	require.NoError(t, err)
	require.Equal(t, recorder.Code, http.StatusFound)
	require.Equal(t, "", recorder.Header().Get("Location"))
	require.Nil(t, ctx.Get(context.AuthenticatedUserKey))

	// Create a valid token
	token, err := c.Auth.GenerateJWTEmailAuthToken(usr.Email)
	require.NoError(t, err)

	// User and token are valid
	ctx.SetParamNames("user", "token")
	ctx.SetParamValues(strconv.Itoa(usr.ID), token)
	err = tests.ExecuteMiddleware(ctx, LoadUser(c.DB))
	require.NoError(t, err)
	err = tests.ExecuteMiddleware(ctx, AuthenticateUserViaToken(c.Auth, c.DB))
	require.NoError(t, err)

	// Check that user is logged in
	authedUser, ok := ctx.Get(context.AuthenticatedUserKey).(sqlc.User)
	require.True(t, ok)
	require.Equal(t, usr.ID, authedUser.ID)
}
