package middleware

import (
	"time"

	"github.com/labstack/echo/v4"
	"go.uber.org/zap"

	"git.grosinger.net/tgrosinger/tonyslist/pkg/context"
)

// LogRequestResponse wraps the request handling in useful log messages and
// attaches a logger to the request context.
func LogRequestResponse(logger *zap.Logger) echo.MiddlewareFunc {
	return func(next echo.HandlerFunc) echo.HandlerFunc {
		return func(c echo.Context) error {
			start := time.Now()
			requestID := c.Response().Header().Get(echo.HeaderXRequestID)

			reqLogger := logger.With(
				zap.String("request_id", requestID),
				zap.String("method", c.Request().Method),
				zap.String("route", c.Path()),
				zap.String("path", c.Request().URL.Path))
			c.Set(context.LoggerKey, reqLogger)

			// Skip logging for the deliverability endpoint because it's very noisy
			if c.Path() != "/api/deliverability/login" {
				reqLogger.Info("handling incoming request",
					zap.Time("time", time.Now()))
			}

			if err := next(c); err != nil {
				c.Error(err)
				reqLogger.Warn("error while handling request",
					zap.Duration("process_time", time.Since(start)),
					zap.Int("response_code", c.Response().Status),
					zap.Error(err))
			} else {
				reqLogger.Info("handled incoming request",
					zap.Duration("process_time", time.Since(start)),
					zap.Int("response_code", c.Response().Status))
			}

			return nil
		}
	}
}
