package middleware

import (
	"database/sql"
	"net/http"
	"strconv"
	"time"

	"github.com/casbin/casbin/v2"
	"github.com/labstack/echo/v4"
	"go.uber.org/zap"

	"git.grosinger.net/tgrosinger/tonyslist/constants/pages"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/context"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/conv"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/errors"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/msg"
)

// WithEvent loads an event from the database and stores it in the request
// context. It will return a 404 if the event date was in the past.
func WithEvent(db *models.Client) echo.MiddlewareFunc {
	return func(next echo.HandlerFunc) echo.HandlerFunc {
		return func(c echo.Context) error {
			eventID, err := strconv.Atoi(c.Param("id"))
			if err != nil {
				msg.Warning(c, "The requested event could not be found")
				return c.Redirect(http.StatusFound,
					c.Echo().Reverse(pages.Names.Event.Browse))
			}

			event, err := db.C.GetEvent(c.Request().Context(), eventID)
			if err == sql.ErrNoRows {
				return errors.ErrNotFound
			} else if err != nil {
				return echo.NewHTTPError(http.StatusInternalServerError,
					"Unable to retrieve the requested event")
			}

			// If event is in the past, return a 404
			now := time.Now()
			if event.StartTime.Before(now) && !conv.DatesEqual(event.StartTime, now) {
				context.GetLogger(c).Info("Not returning event which is in the past",
					zap.Int("event_id", event.ID))
				return echo.NotFoundHandler(c)
			}

			c.Set(context.EventKey, event)

			return next(c)
		}
	}
}

// WithEventFullPermission checks that the event in this request context is owned
// by the user in the context. Must be called after LoadUser and WithEvent
// middleware.
func WithEventFullPermission(enforcer *casbin.Enforcer) echo.MiddlewareFunc {
	return func(next echo.HandlerFunc) echo.HandlerFunc {
		return func(c echo.Context) error {
			event, ok := c.Get(context.EventKey).(sqlc.Event)
			if !ok {
				context.GetLogger(c).Warn("unable to load event from context")
				return echo.NewHTTPError(
					http.StatusInternalServerError, "Unable to retrieve the requested event")
			}

			// Only the event creator may modify it
			user, ok := c.Get(context.AuthenticatedUserKey).(sqlc.User)
			if !ok {
				return echo.NewHTTPError(http.StatusInternalServerError,
					"failed to lookup authenticated user")
			}
			if event.Creator != user.ID {
				// Not the owner, check if they are a moderator
				enforceErr := enforceUser(c, enforcer, "event", "moderate")
				if enforceErr != nil {
					// Not a moderator either, so deny access
					return echo.NewHTTPError(http.StatusForbidden,
						"cannot modify event created by different user")
				}

				context.GetLogger(c).Info("Allowing moderator access to an event")
			}

			return next(c)
		}
	}
}
