package middleware

import (
	"database/sql"
	"net/http"
	"strconv"
	"time"

	"github.com/casbin/casbin/v2"
	"github.com/labstack/echo/v4"

	"git.grosinger.net/tgrosinger/tonyslist/constants"
	"git.grosinger.net/tgrosinger/tonyslist/constants/pages"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/context"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/errors"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/postbase"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/msg"
)

// WithPostFullPermission checks that the post cooresponding with this
// request exists, and that the logged in user has full permissions to modify
// the post. They may have permissions because they are the post owner or
// because they have moderation permissions on posts.
//
// Must be called after the LoadUser and WithPostBase middleware.
func WithPostFullPermission(db *models.Client, enforcer *casbin.Enforcer) echo.MiddlewareFunc {
	return func(next echo.HandlerFunc) echo.HandlerFunc {
		return func(c echo.Context) error {
			postBase, ok := c.Get(context.PostBaseKey).(*postbase.PostBase)
			if !ok {
				context.GetLogger(c).Warn("unable to load post base from context")
				return echo.NewHTTPError(
					http.StatusInternalServerError, "Unable to retrieve the requested post")
			}

			// Only the post creator or a moderator may modify it
			user, ok := c.Get(context.AuthenticatedUserKey).(sqlc.User)
			if !ok {
				return echo.NewHTTPError(http.StatusInternalServerError,
					"failed to lookup authenticated user")
			}
			if postBase.Creator != user.ID {
				// Not the owner, check if they are a moderator
				enforceErr := enforceUser(c, enforcer, "post", "moderate")
				if enforceErr != nil {
					// Not a moderator either, so deny access
					return echo.NewHTTPError(http.StatusForbidden,
						"cannot modify post created by different user")
				}

				context.GetLogger(c).Info("Allowing moderator access to a post")
			}

			return next(c)
		}
	}
}

// WithPost loads the postBase and post into the context.
func WithPost(db *models.Client) echo.MiddlewareFunc {
	return func(next echo.HandlerFunc) echo.HandlerFunc {
		return func(c echo.Context) error {
			postID, err := strconv.Atoi(c.Param("id"))
			if err != nil {
				msg.Warning(c, "The requested post could not be found")
				return c.Redirect(http.StatusFound,
					c.Echo().Reverse(pages.Names.Classified.ForSale.Browse))
			}

			postBase, err := db.PostBase.GetOne(c.Request().Context(), postID, true)
			if err == sql.ErrNoRows {
				return errors.ErrNotFound
			} else if err != nil {
				return echo.NewHTTPError(http.StatusInternalServerError,
					"Unable to retrieve the requested post")
			}

			// If expired or unpublished, only show to the post author
			if !postBase.PublishedAt.Valid || postBase.GetExpiration().Before(time.Now()) {
				u := c.Get(context.AuthenticatedUserKey)
				if u == nil || postBase.Creator != u.(sqlc.User).ID {
					return echo.NewHTTPError(http.StatusNotFound,
						"Unable to retrieve the requested post")
				}
			}

			// Load the full post too
			var post interface{}
			switch postBase.PostType {
			case constants.Classified:
				post, err = db.Classified.GetOneFromPostBase(c.Request().Context(), postBase)
			case constants.Housing:
				post, err = db.Housing.GetOneFromPostBase(c.Request().Context(), postBase)
			}
			if err != nil {
				return echo.NewHTTPError(http.StatusInternalServerError,
					"Unable to retrieve the requested post")
			}

			c.Set(context.PostBaseKey, postBase)
			c.Set(context.PostKey, post)
			return next(c)
		}
	}
}

// WithDigestMembership stores the value of the authenticated users digest
// subscription. It must be called after LoadAuthenticatedUser in order to have
// access to the authenticated user in the context. If there is no authenticated
// user then it will be empty.
func WithDigestMembership(db *models.Client) echo.MiddlewareFunc {
	return func(next echo.HandlerFunc) echo.HandlerFunc {
		return func(c echo.Context) error {
			u, ok := c.Get(context.AuthenticatedUserKey).(sqlc.User)
			if ok {
				digestSetting, err := db.C.GetUserSetting(c.Request().Context(),
					sqlc.GetUserSettingParams{
						UserID: u.ID,
						Key:    string(constants.EmailDigestSubscription),
					})
				if err == sql.ErrNoRows {
					c.Set(context.DigestMembershipKey, constants.EmailDigestSubscriptionTwiceWeekly)
				} else if err != nil {
					return err
				}

				if digestSetting.Valid {
					c.Set(context.DigestMembershipKey, digestSetting.String)
				}

			}

			return next(c)
		}
	}
}
