package models

import (
	"context"
	"database/sql"
	"fmt"
	"strings"

	sq "github.com/Masterminds/squirrel"

	"git.grosinger.net/tgrosinger/tonyslist/constants"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/conv"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/generic"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/classified"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/postbase"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
)

type ClassifiedClient struct {
	dbConfig
}

// TODO: Move this somewhere else
func withTx(ctx context.Context, client *Client, fn func(tx *sqlc.Queries) error) error {
	tx, err := client.db.BeginTx(ctx, &sql.TxOptions{})
	if err != nil {
		return err
	}

	defer func() {
		if err == nil {
			err = tx.Commit()
		} else {
			tx.Rollback()
		}
	}()

	sqlcTx := client.C.WithTx(tx)
	err = fn(sqlcTx)
	return err
}

func (c *ClassifiedClient) Create(ctx context.Context, input classified.NewPostInput) (int, error) {
	var postID int64
	err := withTx(ctx, c.parent, func(tx *sqlc.Queries) error {
		var err error
		postID, err = tx.CreatePost(ctx, input.CreatePostParams)
		if err != nil {
			return err
		}

		_, err = tx.CreateClassifiedPost(ctx, sqlc.CreateClassifiedPostParams{
			ID:             int(postID),
			ClassifiedType: input.ClassifiedType,
			Price:          input.Price,
		})

		return err
	})

	return int(postID), err
}

func (c *ClassifiedClient) Update(ctx context.Context, id int, updates classified.UpdatePostInput) error {
	price, err := conv.StringToNullFloat64("price", updates.Price)
	if err != nil {
		return err
	}

	err = withTx(ctx, c.parent, func(tx *sqlc.Queries) error {
		err = tx.UpdatePost(ctx, updates.UpdatePostParams)
		if err != nil {
			return err
		}

		return tx.UpdateClassifiedPost(ctx, sqlc.UpdateClassifiedPostParams{
			Price: price,
			ID:    id,
		})
	})

	return err
}

func (c *ClassifiedClient) GetOneFromPostBase(ctx context.Context, p *postbase.PostBase) (*classified.Post, error) {
	classifiedPost, err := c.parent.C.GetClassifiedPost(ctx, p.ID)
	if err != nil {
		return nil, err
	}

	p.ClassifiedType = classifiedPost.ClassifiedType
	result := classified.Post{
		PostBase: p,
		Price:    classifiedPost.Price,
	}

	return &result, nil
}

// Select returns the items from the database matching the query windowed by the
// pager as well as the total number of items matching the query.
func (c *ClassifiedClient) Select(ctx context.Context, q classified.SelectInput) ([]postbase.PostBase, int, error) {
	if q.SortColumn == "" {
		q.SortColumn = postbase.TableName + "." + postbase.CreatedAt
	}
	if q.SortDir == "" {
		q.SortDir = postbase.SortDesc
	}

	columns := fmt.Sprintf("%s, %s, %s",
		"classifieds.classified_type",
		"count(photos.post_id) AS count_photos",
		strings.Join(postbase.Fields(), ","))

	groupby := fmt.Sprintf("%s, %s, %s",
		"classifieds.classified_type",
		"photos.post_id",
		strings.Join(postbase.Fields(), ","))

	builder := sq.
		Select(columns).
		From(classified.TableName).
		Join(fmt.Sprintf("%s ON %[1]s.%s = %s.%[2]s",
			postbase.TableName, postbase.ID, classified.TableName)).
		LeftJoin(fmt.Sprintf("photos ON photos.post_id = %s.%s",
			postbase.TableName, postbase.ID)).
		GroupBy(groupby).
		OrderBy(fmt.Sprintf("%s %s", q.SortColumn, q.SortDir)).
		Where(sq.And{
			sq.Eq{
				classified.ClassifiedType:  q.PostType,
				postbase.ModeratorDelisted: false,
			},
			sq.Expr(fmt.Sprintf(`datetime(%s) > datetime('now', '-%d days')`,
				postbase.PublishedAt,
				constants.ClassifiedsDurationDays)),
		})

	if !q.IncludeUnlisted {
		builder = builder.Where(sq.Eq{
			postbase.Listed: true,
		})
	}

	if q.RequirePhotos {
		builder = builder.Having(sq.NotEq{
			"count_photos": 0,
		})
	}

	if q.ParsedQuery != nil {
		var err error
		builder, err = q.ParsedQuery.SetBuilder(builder, postbase.TableName, q.Fields)
		if err != nil {
			return nil, -1, err
		}
	}

	query, args, err := builder.ToSql()
	if err != nil {
		return nil, -1, err
	}

	var results []postbase.PostBase
	err = c.db.SelectContext(ctx, &results, query, args...)
	if err == sql.ErrNoRows {
		return results, 0, nil
	} else if err != nil {
		return nil, -1, err
	}

	totalCount := len(results)
	end := int(q.Offset + q.Limit)
	if totalCount < end {
		end = totalCount
	}
	results = results[q.Offset:end]

	return results, totalCount, nil
}

func (c *ClassifiedClient) MyPosts(ctx context.Context, postType constants.ClassifiedType, userID int) ([]postbase.PostBase, error) {
	results, err := c.parent.C.GetMyClassifiedPostIDs(ctx, sqlc.GetMyClassifiedPostIDsParams{
		Creator:        userID,
		PostType:       constants.Classified,
		ClassifiedType: postType,
	})
	if err != nil {
		return nil, err
	}

	converter := func(r sqlc.PostsBase) postbase.PostBase {
		return postbase.PostBase{PostsBase: r}
	}
	return generic.Map(results, converter), nil
}
