package models

import (
	"context"
	"database/sql"
	"testing"
	"time"

	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"

	"git.grosinger.net/tgrosinger/tonyslist/constants"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/conv"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/generic"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/classified"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/postbase"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/parser"
)

// TODO: Add unit tests for posts with pictures

func createClassifiedPost(t *testing.T, title, description string, published bool) int {
	t.Helper()

	// Create a post with all fields
	id, err := db.Classified.Create(context.Background(), classified.NewPostInput{
		CreatePostParams: sqlc.CreatePostParams{
			PostType:    constants.Classified,
			Title:       title,
			Description: description,
			PhoneCall:   conv.StringToNullString("123-456-7890"),
			PhoneText:   conv.StringToNullString("098-765-4321"),
			Creator:     usr.ID,
		},
		ClassifiedType: constants.WantToBuy,
		Price:          sql.NullFloat64{Valid: true, Float64: 10.25},
	})
	assert.NoError(t, err)

	if published {
		err = db.C.SetPostPublishedNow(context.Background(), id)
		assert.NoError(t, err)
	}

	return id
}

// TODO: Add a test that retrieving an expired post works when expected

func TestClassifiedPost(t *testing.T) {
	// Create a post with all fields
	id, err := db.Classified.Create(context.Background(), classified.NewPostInput{
		CreatePostParams: sqlc.CreatePostParams{
			PostType:       constants.Classified,
			Title:          "test-1-title",
			Description:    "test-1-description",
			PhoneCall:      conv.StringToNullString("123-456-7890"),
			PhoneText:      conv.StringToNullString("098-765-4321"),
			Creator:        usr.ID,
			AnonymousEmail: sql.NullString{Valid: true, String: "test"},
		},
		ClassifiedType: constants.WantToBuy,
		Price:          sql.NullFloat64{Valid: true, Float64: 10.25},
	})
	assert.NoError(t, err)

	// Retrieve the post and ensure all fields are the same
	postBase, err := db.PostBase.GetOne(context.Background(), id, false)
	require.NoError(t, err)
	p, err := db.Classified.GetOneFromPostBase(context.Background(), postBase)
	require.NoError(t, err)

	assert.Equal(t, id, p.ID)
	assert.True(t, time.Since(p.CreatedAt) < 1*time.Minute)
	assert.Equal(t, "test-1-title", p.Title)
	assert.Equal(t, "test-1-description", p.Description)
	assert.Equal(t, "123-456-7890", p.PhoneCall.String)
	assert.Equal(t, "098-765-4321", p.PhoneText.String)
	assert.True(t, p.AnonymousEmail.Valid)
	assert.Equal(t, "test", p.AnonymousEmail.String)
	assert.Equal(t, constants.WantToBuy, p.ClassifiedType)
	assert.Equal(t, 10.25, p.Price.Float64)

	selectInput := classified.SelectInput{
		SelectInput: postbase.SelectInput{
			Pagination: postbase.Pagination{
				Limit: 8,
			},
		},
		PostType: constants.WantToBuy,
	}

	// Check that unpublished post does not show up in search results.
	q, _, err := db.Classified.Select(context.Background(), selectInput)
	assert.NoError(t, err)
	for _, post := range q {
		assert.NotEqual(t, p.ID, post.ID)
	}

	// Publish the post so it shows up in search results.
	err = db.C.SetPostPublishedNow(context.Background(), p.ID)
	require.NoError(t, err)

	q, _, err = db.Classified.Select(context.Background(), selectInput)
	assert.NoError(t, err)
	idx := -1
	for i, p := range q {
		if p.ID == id {
			idx = i
			break
		}
	}
	assert.Greater(t, idx, -1)
	assert.Equal(t, id, q[idx].ID)
	assert.Equal(t, "test-1-title", q[idx].Title)

	postBase, err = db.PostBase.GetOne(context.Background(), id, false)
	require.NoError(t, err)
	p, err = db.Classified.GetOneFromPostBase(context.Background(), postBase)
	require.NoError(t, err)
	assert.NoError(t, err)
	assert.Equal(t, id, p.ID)
	assert.Equal(t, "test-1-title", p.Title)

	// Cleanup after ourselves
	err = db.C.DeletePost(context.Background(), id)
	assert.NoError(t, err)
}

func TestClassifiedPostSearch(t *testing.T) {
	// TODO: This test should not depend on a clean database state

	// Create some posts
	id1 := createClassifiedPost(t, "title cats", "description dogs", true)
	id2 := createClassifiedPost(t, "title house", "description mouse in a boat", true)
	id3 := createClassifiedPost(t, "title mouse", "description boat cats in a", true)

	selectInput := classified.SelectInput{
		SelectInput: postbase.SelectInput{
			Pagination: postbase.Pagination{
				Limit: 8,
			},
		},
		PostType: constants.WantToBuy,
	}
	p := parser.New()

	// Basic text search
	parsed, err := p.ParseSearchQuery("cats")
	require.NoError(t, err)
	selectInput.Fields = []string{postbase.Title, postbase.Description}
	selectInput.ParsedQuery = parsed
	results, _, err := db.Classified.Select(context.Background(), selectInput)
	require.NoError(t, err)
	assert.Len(t, results, 2)
	returnedIDs := generic.Map(results, func(p postbase.PostBase) int {
		return p.ID
	})
	assert.ElementsMatch(t, returnedIDs, []int{id1, id3})

	// Search titles only
	parsed, err = p.ParseSearchQuery("cats")
	require.NoError(t, err)
	selectInput.Fields = []string{postbase.Title}
	selectInput.ParsedQuery = parsed
	results, _, err = db.Classified.Select(context.Background(), selectInput)
	require.NoError(t, err)
	assert.Len(t, results, 1)
	returnedIDs = generic.Map(results, func(p postbase.PostBase) int {
		return p.ID
	})
	assert.ElementsMatch(t, returnedIDs, []int{id1})

	// Search exact string match
	parsed, err = p.ParseSearchQuery(`"in a boat"`)
	require.NoError(t, err)
	selectInput.Fields = []string{postbase.Title, postbase.Description}
	selectInput.ParsedQuery = parsed
	results, _, err = db.Classified.Select(context.Background(), selectInput)
	require.NoError(t, err)
	assert.Len(t, results, 1)
	returnedIDs = generic.Map(results, func(p postbase.PostBase) int {
		return p.ID
	})
	assert.ElementsMatch(t, returnedIDs, []int{id2})

	// Search multiple words, non-exact match
	parsed, err = p.ParseSearchQuery("in a boat")
	require.NoError(t, err)
	selectInput.Fields = []string{postbase.Title, postbase.Description}
	selectInput.ParsedQuery = parsed
	results, _, err = db.Classified.Select(context.Background(), selectInput)
	require.NoError(t, err)
	assert.Len(t, results, 2)
	returnedIDs = generic.Map(results, func(p postbase.PostBase) int {
		return p.ID
	})
	assert.ElementsMatch(t, returnedIDs, []int{id2, id3})

	// Search excluding a word
	parsed, err = p.ParseSearchQuery("boat -house")
	require.NoError(t, err)
	selectInput.Fields = []string{postbase.Title, postbase.Description}
	selectInput.ParsedQuery = parsed
	results, _, err = db.Classified.Select(context.Background(), selectInput)
	require.NoError(t, err)
	assert.Len(t, results, 1)
	returnedIDs = generic.Map(results, func(p postbase.PostBase) int {
		return p.ID
	})
	assert.ElementsMatch(t, returnedIDs, []int{id3})

	// Search with an OR statement
	// TODO

	// Search word prefix
	parsed, err = p.ParseSearchQuery("cat")
	require.NoError(t, err)
	selectInput.Fields = []string{postbase.Title, postbase.Description}
	selectInput.ParsedQuery = parsed
	results, _, err = db.Classified.Select(context.Background(), selectInput)
	require.NoError(t, err)
	assert.Len(t, results, 2)
	returnedIDs = generic.Map(results, func(p postbase.PostBase) int {
		return p.ID
	})
	assert.ElementsMatch(t, returnedIDs, []int{id1, id3})

	// Search word prefix multi-word
	parsed, err = p.ParseSearchQuery("(mous | boa)")
	require.NoError(t, err)
	selectInput.Fields = []string{postbase.Title, postbase.Description}
	selectInput.ParsedQuery = parsed
	results, _, err = db.Classified.Select(context.Background(), selectInput)
	require.NoError(t, err)
	assert.Len(t, results, 2)
	returnedIDs = generic.Map(results, func(p postbase.PostBase) int {
		return p.ID
	})
	assert.ElementsMatch(t, returnedIDs, []int{id2, id3})

	// Cleanup after ourselves
	err = db.C.DeletePost(context.Background(), id1)
	assert.NoError(t, err)
	err = db.C.DeletePost(context.Background(), id2)
	assert.NoError(t, err)
	err = db.C.DeletePost(context.Background(), id3)
	assert.NoError(t, err)
}

func TestClassified_SelectModeratorDelisted(t *testing.T) {
	postID := createClassifiedPost(t, "Test 1", "Test 1", true)
	t.Logf("Created post %d", postID)

	containsPost := func(posts []postbase.PostBase) bool {
		for _, p := range posts {
			t.Logf("Looking at post %d", p.ID)
			if p.ID == postID {
				return true
			}
		}
		return false
	}

	selectInput := classified.SelectInput{
		PostType: constants.WantToBuy,
		SelectInput: postbase.SelectInput{
			IncludeUnlisted: true,
			Sortable: postbase.Sortable{
				SortColumn: postbase.PublishedAt,
				SortDir:    postbase.SortDesc,
			},
			Pagination: postbase.Pagination{
				Limit: 5,
			},
		},
	}
	posts, _, err := db.Classified.Select(context.Background(), selectInput)
	require.NoError(t, err)
	res := containsPost(posts)
	require.True(t, res)

	err = db.C.SetPostModeratorDelisted(context.Background(), postID)
	require.NoError(t, err)

	posts, _, err = db.Classified.Select(context.Background(), selectInput)
	require.NoError(t, err)
	require.False(t, containsPost(posts))
}

func TestClassified_SelectPagination(t *testing.T) {
	_ = createClassifiedPost(t, "TestClassified_SelectPagination_1", "TestClassified_SelectPagination", true)
	_ = createClassifiedPost(t, "TestClassified_SelectPagination_2", "TestClassified_SelectPagination", true)
	_ = createClassifiedPost(t, "TestClassified_SelectPagination_3", "TestClassified_SelectPagination", true)

	selectInput := classified.SelectInput{
		PostType: constants.WantToBuy,
		SelectInput: postbase.SelectInput{
			Pagination: postbase.Pagination{
				Limit: 2,
			},
			SearchQuery: postbase.SearchQuery{
				Query: "TestClassified_SelectPagination",
			},
		},
	}

	results, count, err := db.Classified.Select(context.Background(), selectInput)
	require.NoError(t, err)
	assert.Equal(t, 3, count)
	assert.Equal(t, 2, len(results))
}
