package models

import (
	"database/sql"
	"fmt"
	"path/filepath"

	"github.com/golang-migrate/migrate/v4"
	"github.com/golang-migrate/migrate/v4/database/sqlite3"
	"github.com/golang-migrate/migrate/v4/source/file"
	"github.com/jmoiron/sqlx"
	"go.uber.org/zap"

	c "git.grosinger.net/tgrosinger/tonyslist/config"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
)

type dbConfig struct {
	parent *Client
	db     *sqlx.DB
	log    *zap.Logger
}

type Client struct {
	dbConfig

	C *sqlc.Queries

	User       *UserClient
	Group      *GroupClient
	PostBase   *PostBase
	Classified *ClassifiedClient
	Housing    *HousingClient
}

func New(log *zap.Logger, cfg *c.Config) (*Client, error) {
	dbFilepath := filepath.Join(cfg.Storage.RootDir, cfg.Storage.DatabaseFile)
	if cfg.App.Environment == c.EnvTest {
		// In memory only
		dbFilepath = ":memory:"
	}

	log.Info("Opening database file",
		zap.String("filepath", dbFilepath))

	fn := fmt.Sprintf("file:%s?_fk=1&cache=shared&_busy_timeout=5000", dbFilepath)
	db, err := sqlx.Connect("sqlite3", fn)
	if err != nil {
		return nil, err
	}

	client := Client{}
	client.dbConfig = dbConfig{
		parent: &client,
		db:     db,
		log:    log,
	}
	client.init()

	migrationsDirPath := c.MigrationsDir(cfg)
	log.Info("Loading schema migrations",
		zap.String("filepath", migrationsDirPath))
	err = client.initSchema(migrationsDirPath)
	if err != nil {
		return nil, err
	}

	client.C = sqlc.New(client.db)

	return &client, nil
}

// initSchema ensures that the database is current with the migrations contained
// in db/migrations.
func (db *Client) initSchema(migrationsDir string) error {
	driver, err := sqlite3.WithInstance(db.db.DB, &sqlite3.Config{})
	if err != nil {
		return err
	}

	fSrc, err := (&file.File{}).Open(migrationsDir)
	if err != nil {
		return err
	}

	m, err := migrate.NewWithInstance("file", fSrc, "sqlite", driver)
	if err != nil {
		return err
	}

	err = m.Up()
	if err == migrate.ErrNoChange {
		return nil
	}
	return err
}

// WithTx executes the provided callback with access to a database transaction.
// If the callback returns an error the transaction will be rolled back.
func (db *Client) WithTx(fn func(tx *sql.Tx) error) error {
	tx, err := db.db.Begin()
	if err != nil {
		return err
	}

	err = fn(tx)
	if err != nil {
		_ = tx.Rollback()
		return err
	}

	return tx.Commit()
}

func (db *Client) WithSqlcTx(fn func(*sqlc.Queries) error) error {
	return db.WithTx(
		func(tx *sql.Tx) error {
			return fn(db.C.WithTx(tx))
		},
	)
}

func (db *Client) Close() error {
	return db.db.Close()
}

func (db *Client) init() {
	db.User = &UserClient{db.dbConfig}
	db.Group = &GroupClient{db.dbConfig}
	db.PostBase = &PostBase{db.dbConfig}
	db.Classified = &ClassifiedClient{db.dbConfig}
	db.Housing = &HousingClient{db.dbConfig}
}
