package models

import (
	"context"
	"database/sql"
	"fmt"
	"strings"

	sq "github.com/Masterminds/squirrel"

	"git.grosinger.net/tgrosinger/tonyslist/constants"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/generic"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/housing"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/postbase"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
)

type HousingClient struct {
	dbConfig
}

func (c *HousingClient) Create(ctx context.Context, input housing.NewPostInput) (int, error) {
	var postID int64
	err := withTx(ctx, c.parent, func(tx *sqlc.Queries) error {
		var err error
		postID, err = tx.CreatePost(ctx, input.CreatePostParams)
		if err != nil {
			return err
		}

		err = tx.CreateHousingPost(ctx, sqlc.CreateHousingPostParams{
			ID:           int(postID),
			HousingType:  input.HousingType,
			Rent:         input.Rent,
			Street:       input.Street,
			Zipcode:      input.Zipcode,
			Pets:         input.Pets,
			Shared:       input.Shared,
			Bathrooms:    input.Bathrooms,
			Bedrooms:     input.Bedrooms,
			ParkingSpots: input.ParkingSpots,
		})

		return err
	})

	return int(postID), err
}

func (c *HousingClient) Update(ctx context.Context, id int, updates housing.UpdatePostInput) error {
	return withTx(ctx, c.parent, func(tx *sqlc.Queries) error {
		err := tx.UpdatePost(ctx, updates.UpdatePostParams)
		if err != nil {
			return err
		}

		return tx.UpdateHousingPost(ctx, sqlc.UpdateHousingPostParams{
			ID:           id,
			Rent:         updates.Rent,
			Street:       updates.Street,
			Zipcode:      updates.Zipcode,
			Pets:         updates.Pets,
			Shared:       updates.Shared,
			Bathrooms:    updates.Bathrooms,
			Bedrooms:     updates.Bedrooms,
			ParkingSpots: updates.ParkingSpots,
		})
	})
}

func (c *HousingClient) GetOneFromPostBase(ctx context.Context, p *postbase.PostBase) (*housing.Post, error) {
	hfrPost, err := c.parent.C.GetHousingPost(ctx, p.ID)
	if err != nil {
		return nil, err
	}

	p.HousingType = hfrPost.HousingType
	result := housing.Post{
		PostBase: p,
		EditableFields: housing.EditableFields{
			Bedrooms:     hfrPost.Bedrooms,
			Bathrooms:    hfrPost.Bathrooms,
			ParkingSpots: hfrPost.ParkingSpots,
			Pets:         hfrPost.Pets,
			Shared:       hfrPost.Shared,
			Rent:         hfrPost.Rent,
			Street:       hfrPost.Street,
			Zipcode:      hfrPost.Zipcode,
		},
	}

	return &result, nil
}

// Select returns the items from the database matching the query windowed by the
// pager as well as the total number of items matching the query.
func (c *HousingClient) Select(ctx context.Context, q housing.SelectInput) ([]postbase.PostBase, int, error) {
	if q.SortColumn == "" {
		// TODO: Does sort by created_at work?
		q.SortColumn = postbase.TableName + "." + postbase.CreatedAt
	}
	if q.SortDir == "" {
		q.SortDir = postbase.SortDesc
	}

	columns := fmt.Sprintf("%s, %s, %s",
		"housing.housing_type",
		"count(photos.post_id) AS count_photos",
		strings.Join(postbase.Fields(), ","))

	groupby := fmt.Sprintf("%s, %s, %s",
		"housing.housing_type",
		"photos.post_id",
		strings.Join(postbase.Fields(), ","))

	builder := sq.Select(columns).
		From(housing.TableName).
		Join(fmt.Sprintf("%s ON %[1]s.%s = %s.%[2]s",
			postbase.TableName, postbase.ID, housing.TableName)).
		// Omit maps from the count of photos
		LeftJoin(fmt.Sprintf("photos ON photos.post_id = %s.%s AND photos.is_map = false",
			postbase.TableName, postbase.ID)).
		GroupBy(groupby).
		Where(sq.And{
			sq.Eq{
				postbase.ModeratorDelisted: false,
				postbase.PostType:          constants.Housing,
			},
			sq.Expr(fmt.Sprintf(`datetime(%s) > datetime('now', '-%d days')`,
				postbase.PublishedAt,
				constants.ClassifiedsDurationDays)),
		}).
		OrderBy(fmt.Sprintf("%s %s", q.SortColumn, q.SortDir))

	if !q.IncludeUnlisted {
		builder = builder.Where(sq.Eq{
			postbase.Listed: true,
		})
	}

	if !q.IncludeForRent {
		builder = builder.Where(sq.NotEq{
			housing.HousingType: constants.ForRent,
		})
	}

	if !q.IncludeRequests {
		builder = builder.Where(sq.NotEq{
			housing.HousingType: constants.Request,
		})
	}

	if q.RequirePhotos {
		builder = builder.Having(sq.NotEq{
			"count_photos": 0,
		})
	}

	if q.ParsedQuery != nil {
		var err error
		builder, err = q.ParsedQuery.SetBuilder(builder, postbase.TableName, q.Fields)
		if err != nil {
			return nil, -1, err
		}
	}

	query, args, err := builder.ToSql()
	if err != nil {
		return nil, -1, err
	}

	var results []postbase.PostBase
	err = c.db.SelectContext(ctx, &results, query, args...)
	if err == sql.ErrNoRows {
		return results, 0, nil
	} else if err != nil {
		return nil, -1, err
	}

	totalCount := len(results)
	end := int(q.Offset + q.Limit)
	if totalCount < end {
		end = totalCount
	}
	results = results[q.Offset:end]

	return results, totalCount, nil
}

func (c *HousingClient) MyPosts(ctx context.Context, postType constants.HousingType, userID int) ([]postbase.PostBase, error) {
	results, err := c.parent.C.GetMyHousingPostIDs(ctx, sqlc.GetMyHousingPostIDsParams{
		Creator:     userID,
		PostType:    constants.Housing,
		HousingType: postType,
	})
	if err != nil {
		return nil, err
	}

	converter := func(r sqlc.PostsBase) postbase.PostBase {
		return postbase.PostBase{PostsBase: r}
	}
	return generic.Map(results, converter), nil
}
