package models

import (
	"context"
	"testing"

	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"

	"git.grosinger.net/tgrosinger/tonyslist/constants"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/conv"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/housing"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/postbase"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
)

func createHousingForRentPost(t *testing.T, title, description string, published bool) int {
	t.Helper()

	// Create a post with all fields
	id, err := db.Housing.Create(context.Background(), housing.NewPostInput{
		CreatePostParams: sqlc.CreatePostParams{
			PostType:    constants.Housing,
			Title:       title,
			Description: description,
			PhoneCall:   conv.StringToNullString("123-456-7890"),
			PhoneText:   conv.StringToNullString("098-765-4321"),
			Creator:     usr.ID,
		},
		HousingType: constants.ForRent,
	})
	assert.NoError(t, err)

	if published {
		db.C.SetPostPublishedNow(context.Background(), id)
		assert.NoError(t, err)
	}

	return id
}

func TestHousingForRent_SelectModeratorDelisted(t *testing.T) {
	postID := createHousingForRentPost(t, "Test 1", "Test 1", true)
	t.Logf("Created post %d", postID)

	containsPost := func(posts []postbase.PostBase) bool {
		for _, p := range posts {
			t.Logf("Looking at post %d", p.ID)
			if p.ID == postID {
				return true
			}
		}
		return false
	}

	selectInput := housing.SelectInput{
		SelectInput: postbase.SelectInput{
			IncludeUnlisted: true,
			Sortable: postbase.Sortable{
				SortColumn: postbase.PublishedAt,
				SortDir:    postbase.SortDesc,
			},
			Pagination: postbase.Pagination{
				Limit: 5,
			},
		},
		IncludeForRent:  true,
		IncludeRequests: false,
	}
	posts, _, err := db.Housing.Select(context.Background(), selectInput)
	require.NoError(t, err)
	res := containsPost(posts)
	require.True(t, res)

	err = db.C.SetPostModeratorDelisted(context.Background(), postID)
	require.NoError(t, err)

	posts, _, err = db.Housing.Select(context.Background(), selectInput)
	require.NoError(t, err)
	require.False(t, containsPost(posts))
}

func TestHousingForRent_SelectPagination(t *testing.T) {
	_ = createHousingForRentPost(t, "TestHousingForRent_SelectPagination_1", "TestHousingForRent_SelectPagination", true)
	_ = createHousingForRentPost(t, "TestHousingForRent_SelectPagination_2", "TestHousingForRent_SelectPagination", true)
	_ = createHousingForRentPost(t, "TestHousingForRent_SelectPagination_3", "TestHousingForRent_SelectPagination", true)

	selectInput := housing.SelectInput{
		SelectInput: postbase.SelectInput{
			Pagination: postbase.Pagination{
				Limit: 2,
			},
			SearchQuery: postbase.SearchQuery{
				Query: "TestHousingForRent_SelectPagination",
			},
		},
		IncludeForRent:  true,
		IncludeRequests: false,
	}

	results, count, err := db.Housing.Select(context.Background(), selectInput)
	require.NoError(t, err)
	assert.Equal(t, 3, count)
	assert.Equal(t, 2, len(results))
}
