package postbase

import (
	"time"

	opt "github.com/jordan-bonecutter/goption"
	"github.com/labstack/echo/v4"

	"git.grosinger.net/tgrosinger/tonyslist/constants"
	"git.grosinger.net/tgrosinger/tonyslist/constants/pages"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
)

type PostBase struct {
	sqlc.PostsBase

	PhotosCount int `db:"count_photos"`

	// ClassifiedType is only used when the PostType == constants.Classified
	ClassifiedType constants.ClassifiedType `db:"classified_type"`

	// HousingType is only used when the PostType == constants.Housing
	HousingType constants.HousingType `db:"housing_type"`
}

func (b PostBase) IsExpired(ttl time.Duration) bool {
	if !b.PublishedAt.Valid {
		return false
	}

	return b.PublishedAt.Time.Add(ttl).Before(time.Now())
}

// GetExpiration returns when the post expires.
func (b PostBase) GetExpiration() time.Time {
	if b.PublishedAt.Valid {
		return b.PublishedAt.Time.Add(constants.ClassifiedsDuration)
	}
	// TODO: It shouldn't be possible to have an empty expiration
	return time.Time{}
}

func (b PostBase) GetStatus() constants.PostStatus {
	if !b.PublishedAt.Valid {
		return constants.Draft
	} else if time.Now().After(b.GetExpiration()) {
		return constants.Expired
	} else if b.Listed {
		return constants.Active
	} else {
		return constants.Unlisted
	}
}

// GetFeaturedPhoto retrieves the featured photo for this post out of a list of photos.
// If the featured photo is not found it returns a None.
func (b PostBase) GetFeaturedPhoto(photos []sqlc.Photo) opt.Option[*sqlc.Photo] {
	if b.FeaturedPhoto.Valid {
		// A featured photo is specified. Check to make sure it exists, and return it.
		for _, p := range photos {
			photo := p
			if photo.ID != b.FeaturedPhoto.Int {
				continue
			}

			return opt.Some(&photo)
		}
	}

	return opt.None[*sqlc.Photo]()
}

func (r PostBase) GetViewHref(echo *echo.Echo) string {
	switch r.PostType {
	case constants.Housing:
		return echo.Reverse(pages.Names.Housing.One, r.ID)
	case constants.Classified:
		return echo.Reverse(pages.Names.Classified.One, r.ID)
	}
	return ""
}

func (r PostBase) GetAddPhotosHref(echo *echo.Echo) string {
	return echo.Reverse(pages.Names.Posts.NewPhotos, r.ID)
}

func (r PostBase) GetEditHref(echo *echo.Echo) string {
	switch r.PostType {
	case constants.Housing:
		return echo.Reverse(pages.Names.Housing.Edit, r.ID)
	case constants.Classified:
		return echo.Reverse(pages.Names.Classified.Edit, r.ID)
	}
	return ""
}

func (r PostBase) GetDelistHref(echo *echo.Echo) string {
	return echo.Reverse(pages.Names.Posts.Delist, r.ID)
}

func (r PostBase) GetRelistHref(echo *echo.Echo) string {
	return echo.Reverse(pages.Names.Posts.Relist, r.ID)
}

func (r PostBase) GetDeleteHref(echo *echo.Echo) string {
	return echo.Reverse(pages.Names.Posts.Delete, r.ID)
}

func (r PostBase) GetModeratorDeleteHref(echo *echo.Echo) string {
	return echo.Reverse(pages.Names.Posts.ModeratorDelete, r.ID)
}
