package parser

import (
	"fmt"
	"strings"

	sq "github.com/Masterminds/squirrel"
	"github.com/alecthomas/participle/v2"

	"git.grosinger.net/tgrosinger/tonyslist/pkg/generic"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/errors"
)

// A detailed description of valid search queriees can be found in the help templates.

const (
	FTSTableSuffix = "_fts"
)

type Parser struct {
	p *participle.Parser[SearchQuery]
}

func New() *Parser {
	p := participle.MustBuild[SearchQuery](participle.Unquote("String"))
	return &Parser{
		p: p,
	}
}

type SearchQuery struct {
	Statements []SearchQueryEntry `@@*`
}

type SearchQueryEntry struct {
	// Only Group or Statement should be populated
	Statement *SearchQueryStatement `  @@`
	Group     *SearchQueryGroup     `| @@`
}

type SearchQueryGroup struct {
	OrStatements []*SearchQueryOrStatement `"(" @@ ( "|" @@)* ")"`
}

type SearchQueryOrStatement struct {
	Str string `@(Ident | String)`
}

type SearchQueryStatement struct {
	Exclude bool   `@"-"?`
	Str     string `@(Ident | String)`
}

func (p *Parser) ParseSearchQuery(query string) (*SearchQuery, error) {
	return p.p.ParseString("", query)
}

func (s SearchQuery) SetBuilder(b sq.SelectBuilder, parentTable string,
	fields []string,
) (sq.SelectBuilder, error) {
	fieldPrefix := fmt.Sprintf("{%s}", strings.Join(fields, " "))
	table := parentTable + FTSTableSuffix

	pred, err := s.toPredicate()
	if err != nil {
		return b, err
	}
	arg := fmt.Sprintf("%s : %s", fieldPrefix, pred)
	whereClause := fmt.Sprintf("%s MATCH ?", table)

	query, args, err := sq.Select("rowid").
		From(table).
		Where(whereClause, arg).
		ToSql()
	if err != nil {
		return b, err
	}

	return b.Where(fmt.Sprintf("%s.%s in (%s)", parentTable, "id", query), args...), nil
}

// toPredicate turns this search query into a SQLite FTS5 query.
// https://www.sqlite.org/fts5.html#full_text_query_syntax
func (s SearchQuery) toPredicate() (string, error) {
	if len(s.Statements) < 1 {
		return "", errors.ErrEmptyQuery
	}

	builder := strings.Builder{}

	for i, stmt := range s.Statements {
		includeConj := false
		if i > 0 {
			includeConj = true
		}

		if stmt.Group != nil {
			builder.WriteString(stmt.Group.toPredicate(includeConj))
		} else if stmt.Statement != nil {
			p, err := stmt.Statement.toPredicate(includeConj)
			if err != nil {
				return "", err
			}
			builder.WriteString(p)
		}

		builder.WriteString(" ")
	}

	return builder.String(), nil
}

func (s SearchQueryGroup) toPredicate(includeConj bool) string {
	statements := s.OrStatements
	strs := generic.Map(statements, func(stmt *SearchQueryOrStatement) string {
		return stmt.Str + "*"
	})

	stmt := strings.Join(strs, " OR ")
	if includeConj {
		return "AND " + stmt
	}
	return stmt
}

func (s SearchQueryStatement) toPredicate(includeConj bool) (string, error) {
	if s.Exclude {
		if !includeConj {
			// Cannot use "NOT" as the first term, only to join a subsequent term.
			return "", errors.ErrInvalidQueryExclustion
		}

		return fmt.Sprintf("NOT %q*", s.Str), nil
	}

	if includeConj {
		return fmt.Sprintf("AND %q*", s.Str), nil
	}
	return fmt.Sprintf("%q*", s.Str), nil
}
