package parser

import (
	"testing"

	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
)

func TestSearchQueryParser(t *testing.T) {
	testcases := []struct {
		Name     string
		Query    string
		Expected *SearchQuery
	}{
		{
			Name:  "Simple single word string",
			Query: "hat",
			Expected: &SearchQuery{
				Statements: []SearchQueryEntry{
					{
						Statement: &SearchQueryStatement{
							Exclude: false,
							Str:     "hat",
						},
					},
				},
			},
		},
		{
			Name:  "Simple multi-word string",
			Query: "top hat",
			Expected: &SearchQuery{
				Statements: []SearchQueryEntry{
					{
						Statement: &SearchQueryStatement{
							Exclude: false,
							Str:     "top",
						},
					},
					{
						Statement: &SearchQueryStatement{
							Exclude: false,
							Str:     "hat",
						},
					},
				},
			},
		},
		{
			Name:  "Terms in quotes",
			Query: `"top hat"`,
			Expected: &SearchQuery{
				Statements: []SearchQueryEntry{
					{
						Statement: &SearchQueryStatement{
							Exclude: false,
							Str:     "top hat",
						},
					},
				},
			},
		},
		{
			Name:  "Multiple terms in quotes",
			Query: `"top hat" "baseball"`,
			Expected: &SearchQuery{
				Statements: []SearchQueryEntry{
					{
						Statement: &SearchQueryStatement{
							Exclude: false,
							Str:     "top hat",
						},
					},
					{
						Statement: &SearchQueryStatement{
							Exclude: false,
							Str:     "baseball",
						},
					},
				},
			},
		},
		{
			Name:  "Mix quoted terms",
			Query: `"top hat" baseball`,
			Expected: &SearchQuery{
				Statements: []SearchQueryEntry{
					{
						Statement: &SearchQueryStatement{
							Exclude: false,
							Str:     "top hat",
						},
					},
					{
						Statement: &SearchQueryStatement{
							Exclude: false,
							Str:     "baseball",
						},
					},
				},
			},
		},
		{
			Name:  "Mix quoted terms 2",
			Query: `baseball "top hat"`,
			Expected: &SearchQuery{
				Statements: []SearchQueryEntry{
					{
						Statement: &SearchQueryStatement{
							Exclude: false,
							Str:     "baseball",
						},
					},
					{
						Statement: &SearchQueryStatement{
							Exclude: false,
							Str:     "top hat",
						},
					},
				},
			},
		},
		{
			Name:  "Multi-word with exclusion",
			Query: "top hat -straw",
			Expected: &SearchQuery{
				Statements: []SearchQueryEntry{
					{
						Statement: &SearchQueryStatement{
							Exclude: false,
							Str:     "top",
						},
					},
					{
						Statement: &SearchQueryStatement{
							Exclude: false,
							Str:     "hat",
						},
					},
					{
						Statement: &SearchQueryStatement{
							Exclude: true,
							Str:     "straw",
						},
					},
				},
			},
		},
		{
			Name:  "Multi-word with multiple exclusions",
			Query: "top hat -straw -baseball",
			Expected: &SearchQuery{
				Statements: []SearchQueryEntry{
					{
						Statement: &SearchQueryStatement{
							Exclude: false,
							Str:     "top",
						},
					},
					{
						Statement: &SearchQueryStatement{
							Exclude: false,
							Str:     "hat",
						},
					},
					{
						Statement: &SearchQueryStatement{
							Exclude: true,
							Str:     "straw",
						},
					},
					{
						Statement: &SearchQueryStatement{
							Exclude: true,
							Str:     "baseball",
						},
					},
				},
			},
		},
		{
			Name:  "Multiple exclusions in mixed order",
			Query: "top -straw hat -baseball",
			Expected: &SearchQuery{
				Statements: []SearchQueryEntry{
					{
						Statement: &SearchQueryStatement{
							Exclude: false,
							Str:     "top",
						},
					},
					{
						Statement: &SearchQueryStatement{
							Exclude: true,
							Str:     "straw",
						},
					},
					{
						Statement: &SearchQueryStatement{
							Exclude: false,
							Str:     "hat",
						},
					},
					{
						Statement: &SearchQueryStatement{
							Exclude: true,
							Str:     "baseball",
						},
					},
				},
			},
		},
		{
			Name:  "quoted exclusion",
			Query: `baseball "top hat" -"straw cap"`,
			Expected: &SearchQuery{
				Statements: []SearchQueryEntry{
					{
						Statement: &SearchQueryStatement{
							Exclude: false,
							Str:     "baseball",
						},
					},
					{
						Statement: &SearchQueryStatement{
							Exclude: false,
							Str:     "top hat",
						},
					},
					{
						Statement: &SearchQueryStatement{
							Exclude: true,
							Str:     "straw cap",
						},
					},
				},
			},
		},
		{
			Name:  "or with group",
			Query: `(baseball | top) straw`,
			Expected: &SearchQuery{
				Statements: []SearchQueryEntry{
					{
						Group: &SearchQueryGroup{
							OrStatements: []*SearchQueryOrStatement{
								{Str: "baseball"},
								{Str: "top"},
							},
						},
					},
					{
						Statement: &SearchQueryStatement{
							Exclude: false,
							Str:     "straw",
						},
					},
				},
			},
		},
		{
			Name:  "or with group and exclusion",
			Query: `(baseball | top) -straw`,
			Expected: &SearchQuery{
				Statements: []SearchQueryEntry{
					{
						Group: &SearchQueryGroup{
							OrStatements: []*SearchQueryOrStatement{
								{Str: "baseball"},
								{Str: "top"},
							},
						},
					},
					{
						Statement: &SearchQueryStatement{
							Exclude: true,
							Str:     "straw",
						},
					},
				},
			},
		},
		{
			Name:  "or group without spaces",
			Query: `(baseball|top) straw`,
			Expected: &SearchQuery{
				Statements: []SearchQueryEntry{
					{
						Group: &SearchQueryGroup{
							OrStatements: []*SearchQueryOrStatement{
								{Str: "baseball"},
								{Str: "top"},
							},
						},
					},
					{
						Statement: &SearchQueryStatement{
							Exclude: false,
							Str:     "straw",
						},
					},
				},
			},
		},
	}

	p := New()

	for _, tc := range testcases {
		t.Run(tc.Name, func(t *testing.T) {
			result, err := p.ParseSearchQuery(tc.Query)
			require.NoError(t, err)
			assert.Equal(t, tc.Expected, result)
		})
	}
}

func ptr(s string) *string {
	return &s
}
