package postmark

import (
	"context"
	"fmt"

	postmarkSDK "github.com/mrz1836/postmark"
	"go.uber.org/zap"

	"git.grosinger.net/tgrosinger/tonyslist/config"
)

// NOTE: This would likely be better structured as a Terraform module, however I
// have not found a way to avoid having secrets in a .tfstate file that would
// need to be stored somewhere.

func Initialize(cfg *config.Config, logger *zap.Logger) error {
	if !cfg.Mail.ConfigurePostmark {
		logger.Info("Skipping Postmark configuration")
		return nil
	}

	client := postmarkSDK.NewClient(cfg.Mail.Token, "")

	mockMode := false

	// NOTE: This does not configure the server itself.
	// You must also set the webhook URL on the default inbound message stream.

	err := initMessageStream(client, logger, mockMode)
	if err != nil {
		return err
	}

	err = configureWebhooks(cfg, client, logger, mockMode)
	if err != nil {
		return err
	}

	return nil
}

func initMessageStream(client *postmarkSDK.Client, logger *zap.Logger, mockMode bool) error {
	logger.Info("Initializing message streams", zap.Bool("mock_mode", mockMode))

	expectedMessageStreams := []postmarkSDK.CreateMessageStreamRequest{
		{
			ID:                "account-management",
			Name:              "Account Management",
			MessageStreamType: postmarkSDK.TransactionalMessageStreamType,
			SubscriptionManagementConfiguration: postmarkSDK.MessageStreamSubscriptionManagementConfiguration{
				UnsubscribeHandlingType: postmarkSDK.NoneUnsubscribeHandlingType,
			},
		},
		{
			ID:                "admin-notifications",
			Name:              "Admin Notifications",
			MessageStreamType: postmarkSDK.TransactionalMessageStreamType,
			SubscriptionManagementConfiguration: postmarkSDK.MessageStreamSubscriptionManagementConfiguration{
				UnsubscribeHandlingType: postmarkSDK.NoneUnsubscribeHandlingType,
			},
		},
		{
			ID:                "events-transactional",
			Name:              "Events Transactional",
			MessageStreamType: postmarkSDK.TransactionalMessageStreamType,
			SubscriptionManagementConfiguration: postmarkSDK.MessageStreamSubscriptionManagementConfiguration{
				UnsubscribeHandlingType: postmarkSDK.NoneUnsubscribeHandlingType,
			},
		},
		{
			ID:                "grafana-alerts",
			Name:              "Grafana Alerts",
			MessageStreamType: postmarkSDK.TransactionalMessageStreamType,
			SubscriptionManagementConfiguration: postmarkSDK.MessageStreamSubscriptionManagementConfiguration{
				UnsubscribeHandlingType: postmarkSDK.NoneUnsubscribeHandlingType,
			},
		},
		{
			ID:                "user-messages",
			Name:              "User Messages",
			MessageStreamType: postmarkSDK.TransactionalMessageStreamType,
			SubscriptionManagementConfiguration: postmarkSDK.MessageStreamSubscriptionManagementConfiguration{
				UnsubscribeHandlingType: postmarkSDK.NoneUnsubscribeHandlingType,
			},
		},
		{
			ID:                "twice-weekly-digest",
			Name:              "Twice Weekly Digest",
			MessageStreamType: postmarkSDK.BroadcastMessageStreamType,
			SubscriptionManagementConfiguration: postmarkSDK.MessageStreamSubscriptionManagementConfiguration{
				UnsubscribeHandlingType: postmarkSDK.CustomUnsubscribeHandlingType,
			},
		},
	}

	res, err := client.ListMessageStreams(context.TODO(), "Any", false)
	if err != nil {
		logger.Warn("Unable to list message streams",
			zap.Error(err))
		return err
	}
	existingMsgStreams := make(map[string]postmarkSDK.MessageStream)
	for _, msgStream := range res {
		existingMsgStreams[msgStream.ID] = msgStream
	}

	for _, expected := range expectedMessageStreams {
		existing, ok := existingMsgStreams[expected.ID]
		if !ok {
			if mockMode {
				logger.Info("Would have created message stream", zap.String("id", expected.ID))
				continue
			}

			_, err = client.CreateMessageStream(context.TODO(), expected)
			if err != nil {
				logger.Warn("Message stream does not exist, and unable to create",
					zap.Error(err))
				return err
			}

			logger.Info("Created message stream", zap.String("id", expected.ID))
			continue
		}

		if existing.Name != expected.Name {
			logger.Warn("Message stream name does not match expected value",
				zap.String("id", expected.ID),
				zap.String("expected", expected.Name),
				zap.String("actual", existing.Name))
			return fmt.Errorf("Unexpected message stream configuration")
		}
		if existing.MessageStreamType != expected.MessageStreamType {
			logger.Warn("Message stream type does not match expected value",
				zap.String("id", expected.ID),
				zap.String("expected", string(expected.MessageStreamType)),
				zap.String("actual", string(existing.MessageStreamType)))
			return fmt.Errorf("Unexpected message stream configuration")
		}
		if existing.SubscriptionManagementConfiguration.UnsubscribeHandlingType != expected.SubscriptionManagementConfiguration.UnsubscribeHandlingType {
			logger.Warn("Message stream unsubscribe handling type does not match expected value",
				zap.String("id", expected.ID),
				zap.String("expected", string(expected.SubscriptionManagementConfiguration.UnsubscribeHandlingType)),
				zap.String("actual", string(existing.SubscriptionManagementConfiguration.UnsubscribeHandlingType)))
			return fmt.Errorf("Unexpected message stream configuration")
		}
	}

	return nil
}

func configureWebhooks(cfg *config.Config, client *postmarkSDK.Client, logger *zap.Logger, mockMode bool) error {
	logger.Info("Configuring webhooks", zap.Bool("mock_mode", mockMode))

	existingWebhooks, err := client.ListWebhooks(context.TODO(), "")
	if err != nil {
		logger.Warn("Unable to list existing webhooks", zap.Error(err))
	}
	webhooksByMessageStream := make(map[string]postmarkSDK.Webhook)
	for _, webhook := range existingWebhooks {
		webhooksByMessageStream[webhook.MessageStream] = webhook
	}

	url := fmt.Sprintf("%s/hooks/deliverability", cfg.App.PublicURL)

	auth := postmarkSDK.WebhookHTTPAuth{
		Username: cfg.App.HooksUsername,
		Password: cfg.App.HooksPassword,
	}

	triggers := postmarkSDK.WebhookTrigger{
		Bounce: postmarkSDK.WebhookTriggerIncContent{
			WebhookTriggerEnabled: postmarkSDK.WebhookTriggerEnabled{
				Enabled: true,
			},
			IncludeContent: false,
		},
		SpamComplaint: postmarkSDK.WebhookTriggerIncContent{
			WebhookTriggerEnabled: postmarkSDK.WebhookTriggerEnabled{
				Enabled: true,
			},
			IncludeContent: false,
		},
	}
	triggersWithDelivery := postmarkSDK.WebhookTrigger{
		Delivery: postmarkSDK.WebhookTriggerEnabled{
			Enabled: true,
		},
		Bounce: postmarkSDK.WebhookTriggerIncContent{
			WebhookTriggerEnabled: postmarkSDK.WebhookTriggerEnabled{
				Enabled: true,
			},
			IncludeContent: false,
		},
		SpamComplaint: postmarkSDK.WebhookTriggerIncContent{
			WebhookTriggerEnabled: postmarkSDK.WebhookTriggerEnabled{
				Enabled: true,
			},
			IncludeContent: false,
		},
	}

	expectedWebhooks := map[string]postmarkSDK.Webhook{
		"account-management": {
			URL:           url,
			MessageStream: "account-management",
			HTTPAuth:      &auth,
			Triggers:      triggersWithDelivery,
		},
		"events-transactional": {
			URL:           url,
			MessageStream: "events-transactional",
			HTTPAuth:      &auth,
			Triggers:      triggers,
		},
		"twice-weekly-digest": {
			URL:           url,
			MessageStream: "twice-weekly-digest",
			HTTPAuth:      &auth,
			Triggers:      triggers,
		},
	}

	for msgStreamID, expected := range expectedWebhooks {
		existing, ok := webhooksByMessageStream[msgStreamID]
		if !ok {
			if mockMode {
				logger.Info("Would have created webhook", zap.String("message_stream_id", msgStreamID))
				continue
			}

			// Create the webhook
			_, err := client.CreateWebhook(context.TODO(), expected)
			if err != nil {
				logger.Warn("Unable to create webhook",
					zap.String("message_stream_id", msgStreamID),
					zap.Error(err))
				return err
			}

			logger.Info("Created webhook", zap.String("message_stream_id", msgStreamID))
			continue
		}

		if existing.URL != expected.URL {
			logger.Warn("Webhook URL does not match expected value",
				zap.String("message_stream_id", msgStreamID),
				zap.String("expected", expected.URL),
				zap.String("actual", existing.URL))
			return fmt.Errorf("Unexpected webhook configuration")
		}
		if existing.MessageStream != expected.MessageStream {
			logger.Warn("Webhook message stream does not match expected value",
				zap.String("message_stream_id", msgStreamID),
				zap.String("expected", expected.MessageStream),
				zap.String("actual", existing.MessageStream))
			return fmt.Errorf("Unexpected webhook configuration")
		}
		if (existing.HTTPAuth == nil && expected.HTTPAuth != nil) ||
			(existing.HTTPAuth != nil && expected.HTTPAuth == nil) ||
			(expected.HTTPAuth != nil && *existing.HTTPAuth != *expected.HTTPAuth) {

			logger.Warn("Webhook HTTP auth does not match expected value",
				zap.String("message_stream_id", msgStreamID),
				zap.String("expected", fmt.Sprintf("%+v", expected.HTTPAuth)),
				zap.String("actual", fmt.Sprintf("%+v", existing.HTTPAuth)))
			return fmt.Errorf("Unexpected webhook configuration")
		}
		if existing.Triggers != expected.Triggers {
			logger.Warn("Webhook triggers do not match expected value",
				zap.String("message_stream_id", msgStreamID),
				zap.String("expected", fmt.Sprintf("%+v", expected.Triggers)),
				zap.String("actual", fmt.Sprintf("%+v", existing.Triggers)))
			return fmt.Errorf("Unexpected webhook configuration")
		}
	}

	return nil
}
