package routes

import (
	"database/sql"
	"errors"
	"fmt"

	"github.com/labstack/echo/v4"
	"go.uber.org/zap"

	"git.grosinger.net/tgrosinger/tonyslist/constants"
	"git.grosinger.net/tgrosinger/tonyslist/constants/pages"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/context"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/controller"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/msg"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/views"
)

type account struct {
	controller.Controller
}

// Get displays a page with account information and actions. This handler is
// protected by middleware so that only logged-in users can access it.
func (c *account) Get(ctx echo.Context) error {
	user, ok := ctx.Get(context.AuthenticatedUserKey).(sqlc.User)
	if !ok {
		return c.BadReq("Missing authenticated user")
	}

	billingPortalURL := ""
	if user.StripeCustomerID.Valid {
		var err error
		returnURL := c.Container.Config.App.PublicURL + ctx.Echo().Reverse(pages.Names.User.Account)
		billingPortalURL, err = c.Container.Stripe.GetSession(user.StripeCustomerID.String, returnURL)
		if err != nil {
			context.GetLogger(ctx).Warn("Failed to create Stripe billing portal session",
				zap.Int("user_id", user.ID),
				zap.Error(err))
		}
	}

	digestSetting, err := getEmailDigestSetting(c.Container.DB, user, ctx)
	if err != nil {
		return c.Fail(err, "Unable to retrieve user digest settings")
	}

	page := controller.NewPage(ctx, c.Container)
	page.Title = "Account"
	page.Metatags.Description = "Account settings."

	form := views.AccountUpdateForm{
		EmailDigests: digestSetting,
	}
	if f := ctx.Get(context.FormKey); f != nil {
		form = *f.(*views.AccountUpdateForm)
	}

	v := views.Account{
		Form:             form,
		BillingPortalURL: billingPortalURL,
		Page:             page,
	}
	return c.RenderHTML(ctx, page, &v)
}

func (c *account) Post(ctx echo.Context) error {
	var form views.AccountUpdateForm
	ctx.Set(context.FormKey, &form)

	if err := ctx.Bind(&form); err != nil {
		return c.Fail(err, "unable to bind form")
	}

	if err := form.Submission.Process(ctx, form); err != nil {
		return c.Fail(err, "unable to process form submission")
	}

	if form.Submission.HasErrors() {
		return c.Get(ctx)
	}

	var newDigestSetting string
	switch form.EmailDigests {
	case constants.EmailDigestSubscriptionTwiceWeekly,
		constants.EmailDigestSubscriptionUnsubscribed:
		newDigestSetting = form.EmailDigests
	default:
		// Not using a form error here because this is a select field and if
		// they submit an incorrect value then they are doing something strange.
		return c.BadReq(fmt.Sprintf("Invalid value %q for email digest setting", form.EmailDigests))
	}

	user, ok := ctx.Get(context.AuthenticatedUserKey).(sqlc.User)
	if !ok {
		return c.Fail(errors.New("failed to lookup authenticated user"), "unable to lookup authenticated user")
	}

	err := c.Container.DB.C.SetUserSetting(ctx.Request().Context(),
		sqlc.SetUserSettingParams{
			UserID: user.ID,
			Key:    string(constants.EmailDigestSubscription),
			Value:  sql.NullString{Valid: true, String: newDigestSetting},
		})
	if err != nil {
		return c.Fail(err, "unable to store user settings")
	}
	return c.Get(ctx)
}

// GetDelete displays a confirmation before deleting an account. This handler is
// protected by middleware so that only logged-in users can access it.
func (c *account) GetDelete(ctx echo.Context) error {
	page := controller.NewPage(ctx, c.Container)
	page.Title = "Delete Account"
	page.Metatags.Description = "Account settings."

	v := views.ActionConfirmation{
		Heading:     "Delete Account",
		Prompt:      "Are you sure you want to delete your account? This action cannot be undone.",
		ButtonText:  "Delete account",
		Page:        page,
		Destructive: true,
	}
	return c.RenderHTML(ctx, page, &v)
}

// PostDelete processes an account deletion confirmation. This handler is
// protected by middleware so that only logged-in users can access it.
func (c *account) PostDelete(ctx echo.Context) error {
	var form views.ActionConfirmationForm
	ctx.Set(context.FormKey, &form)

	// Parse the form values
	if err := ctx.Bind(&form); err != nil {
		return c.Fail(err, "unable to parse account form")
	}

	if err := form.Submission.Process(ctx, form); err != nil {
		return c.Fail(err, "unable to process form submission")
	}

	if form.Submission.HasErrors() {
		return c.Get(ctx)
	}

	usr := ctx.Get(context.AuthenticatedUserKey).(sqlc.User)
	err := c.Container.DB.C.DeleteUser(ctx.Request().Context(), usr.ID)
	if err != nil {
		context.GetLogger(ctx).Warn("Failed to delete account",
			zap.Int("user_id", usr.ID),
			zap.Error(err))
		msg.Danger(ctx, "Unable to delete your account. Please contact support.")
		return c.Redirect(ctx, pages.Names.User.Account)
	}

	msg.Success(ctx, "Your account has been deleted")
	return c.Redirect(ctx, pages.Names.Home)
}

func (c *account) GetJoinDigest(ctx echo.Context) error {
	usr := ctx.Get(context.AuthenticatedUserKey).(sqlc.User)

	digestSetting, err := getEmailDigestSetting(c.Container.DB, usr, ctx)
	if err != nil {
		return c.Fail(err, "Unable to retrieve user digest settings")
	}
	if digestSetting == constants.EmailDigestSubscriptionTwiceWeekly {
		msg.Info(ctx, "You are already receiving digest emails")
		return c.Redirect(ctx, pages.Names.Home)
	}

	page := controller.NewPage(ctx, c.Container)
	page.Title = "Join Digest"
	page.Metatags.Description = "Join the twice-weekly email digest."

	v := views.ActionConfirmation{
		Heading:     "Join Email Digest",
		Prompt:      "Would you like to receive the twice-weekly email digest?",
		ButtonText:  "Join",
		Page:        page,
		Destructive: false,
	}
	return c.RenderHTML(ctx, page, &v)
}

func (c *account) PostJoinDigest(ctx echo.Context) error {
	var form views.ActionConfirmationForm
	ctx.Set(context.FormKey, &form)

	// Parse the form values
	if err := ctx.Bind(&form); err != nil {
		return c.Fail(err, "unable to parse account form")
	}

	if err := form.Submission.Process(ctx, form); err != nil {
		return c.Fail(err, "unable to process form submission")
	}

	if form.Submission.HasErrors() {
		return c.Get(ctx)
	}

	usr := ctx.Get(context.AuthenticatedUserKey).(sqlc.User)

	err := c.Container.DB.C.SetUserSetting(ctx.Request().Context(),
		sqlc.SetUserSettingParams{
			UserID: usr.ID,
			Key:    string(constants.EmailDigestSubscription),
			Value:  sql.NullString{Valid: true, String: constants.EmailDigestSubscriptionTwiceWeekly},
		})
	if err != nil {
		context.GetLogger(ctx).Warn("Failed to add user to digest email",
			zap.Int("user_id", usr.ID),
			zap.Error(err))
		msg.Danger(ctx, "Unable to add you to the digest email.")
		return c.Redirect(ctx, pages.Names.User.Account)
	}

	msg.Success(ctx, "Successfully joined digest emails!")
	return c.Redirect(ctx, pages.Names.Home)
}

func getEmailDigestSetting(db *models.Client, user sqlc.User, ctx echo.Context) (string, error) {
	digestSetting, err := db.C.GetUserSetting(ctx.Request().Context(),
		sqlc.GetUserSettingParams{
			UserID: user.ID,
			Key:    string(constants.EmailDigestSubscription),
		})
	if err == sql.ErrNoRows {
		return constants.EmailDigestSubscriptionTwiceWeekly, nil
	} else if err != nil {
		return "", err
	}

	if digestSetting.Valid {
		return digestSetting.String, nil
	}

	return constants.EmailDigestSubscriptionTwiceWeekly, nil
}
