package admin

import (
	"fmt"
	"math"
	"net/http"
	"strconv"

	"git.grosinger.net/tgrosinger/tonyslist/constants/pages"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/context"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/controller"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/conv"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/views"
	"github.com/labstack/echo/v4"
)

func (c *adminCtr) GetUsers(ctx echo.Context) error {
	htmxReq := ctx.Request().Header.Get("HX-Request")
	page := controller.NewPage(ctx, c.Container)
	page.Pager = controller.NewPager(ctx, 50)
	page.Title = "Admin - Users"

	var users []sqlc.User
	var err error
	query := ctx.QueryParam("search")
	if query != "" {
		dbQuery := "%" + query + "%"
		users, err = c.Container.DB.C.SearchUsers(ctx.Request().Context(), sqlc.SearchUsersParams{
			Name:  conv.StringToNullString(dbQuery),
			Email: dbQuery,
		})
	} else {
		users, err = c.Container.DB.C.GetUsers(ctx.Request().Context())
	}

	if err != nil {
		return c.Fail(err, "unable to retrieve users")
	}

	// We could add the limit and offset to the sql query, but this is fine for now too.
	page.Pager.Items = len(users)
	if len(users) > 0 {
		start := math.Max(0, float64(page.Pager.GetOffset()))
		end := math.Min(float64(len(users)), start+float64(page.Pager.ItemsPerPage))
		users = users[int(start):int(end)]
	}

	v := views.AdminUsers{
		Users:        users,
		UserListOnly: htmxReq == "true",
		Page:         page,
	}

	return c.RenderHTML(ctx, page, &v)
}

func (c *adminCtr) GetUserDetails(ctx echo.Context) error {
	userID, err := strconv.Atoi(ctx.Param("id"))
	if err != nil {
		return c.Fail(err, "unable to parse requested user")
	}

	user, err := c.Container.DB.C.GetUserByID(ctx.Request().Context(), userID)
	if err != nil {
		return c.Fail(err, "unable to retrieve requested user")
	}

	page := controller.NewPage(ctx, c.Container)
	page.Title = "Admin - User"

	activePostCount, err := c.Container.DB.C.GetActivePostsForUserCount(ctx.Request().Context(), user.ID)
	if err != nil {
		return c.Fail(err, "unable to retrieve active posts count for requested user")
	}

	v := views.AdminUserDetails{
		User:        user,
		PostsActive: activePostCount,
		HTMXReq:     ctx.QueryParam("field"),
		Page:        page,
	}

	return c.RenderHTML(ctx, page, &v)
}

func (c *adminCtr) PatchUserDetails(ctx echo.Context) error {
	userID, err := strconv.Atoi(ctx.Param("id"))
	if err != nil {
		return c.Fail(err, "unable to parse requested user")
	}

	page := controller.NewPage(ctx, c.Container)
	page.Title = "Admin - User"

	name := ctx.FormValue("name")
	if name != "" {
		err = c.Container.DB.C.SetUserName(ctx.Request().Context(), sqlc.SetUserNameParams{
			Name: conv.StringToNullString(name),
			ID:   userID,
		})
		if err != nil {
			return c.Fail(err, "failed to update user name")
		}

		// Be sure to get user after updating so it has the new value.
		user, err := c.Container.DB.C.GetUserByID(ctx.Request().Context(), userID)
		if err != nil {
			return c.Fail(err, "unable to retrieve requested user")
		}

		v := views.AdminUserDetails{
			User:    user,
			HTMXReq: "name-static",
			Page:    page,
		}

		return c.RenderHTML(ctx, page, &v)
	}

	return c.Fail(nil, "unrecognized user patch operation")
}

func (c *adminCtr) GetDeleteUser(ctx echo.Context) error {
	userID, err := strconv.Atoi(ctx.Param("id"))
	if err != nil {
		return c.Fail(err, "unable to parse requested user")
	}

	user, err := c.Container.DB.C.GetUserByID(ctx.Request().Context(), userID)
	if err != nil {
		return c.Fail(err, "unable to retrieve requested user")
	}

	page := controller.NewPage(ctx, c.Container)
	page.Title = "Delete User"

	v := views.ActionConfirmation{
		Heading:     "Delete user",
		Prompt:      fmt.Sprintf("Are you sure you want to delete user with email %s? This action cannot be undone.", user.Email),
		ButtonText:  "Delete",
		Destructive: true,

		Page: page,
	}
	return c.RenderHTML(ctx, page, &v)
}

func (c *adminCtr) PostDeleteUser(ctx echo.Context) error {
	userID, err := strconv.Atoi(ctx.Param("id"))
	if err != nil {
		return c.Fail(err, "unable to parse requested user")
	}

	var form views.ActionConfirmationForm
	ctx.Set(context.FormKey, &form)

	if err := ctx.Bind(&form); err != nil {
		return c.Fail(err, "unable to bind form")
	}

	if err := form.Submission.Process(ctx, form); err != nil {
		return c.Fail(err, "unable to process form submission")
	}

	if form.Submission.HasErrors() {
		return c.GetDeleteUser(ctx)
	}

	// NOTE: Follows foreign keys to clean up posts created by this user too.
	err = c.Container.DB.C.DeleteUser(ctx.Request().Context(), userID)
	if err != nil {
		return c.Fail(err, "unable to delete the post")
	}

	return ctx.Redirect(http.StatusFound, ctx.Echo().Reverse(pages.Names.Admin.Users))
}
