package classified

import (
	"database/sql"
	"errors"
	"fmt"
	"strings"

	"github.com/labstack/echo/v4"
	"go.uber.org/zap"

	"git.grosinger.net/tgrosinger/tonyslist/constants"
	"git.grosinger.net/tgrosinger/tonyslist/constants/pages"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/context"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/controller"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/conv"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/classified"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/views"
)

type create struct {
	controller.Controller
}

// Get displays a page for users to create a new post in the classifieds section.
// This handler is protected by middleware so that only logged-in users can
// access it.
func (c *create) Get(ctx echo.Context) error {
	page := controller.NewPage(ctx, c.Container)
	page.Title = "New Post"
	page.Metatags.Description = fmt.Sprintf("Create new classifieds post on %s.", c.Container.Config.App.Name)
	page.Metatags.Keywords = keywords

	classifiedType := string(constants.ForSale)
	switch ctx.QueryParam("src_classified_type") {
	case string(constants.ForFree):
		classifiedType = string(constants.ForFree)
	case string(constants.WantToBuy):
		classifiedType = string(constants.WantToBuy)
	case string(constants.WantToBorrow):
		classifiedType = string(constants.WantToBorrow)
	}

	form := views.ClassifiedCreateForm{
		IncludeEmail:   true,
		ClassifiedType: classifiedType,
	}
	if f := ctx.Get(context.FormKey); f != nil {
		form = *f.(*views.ClassifiedCreateForm)
	}

	v := views.ClassifiedCreate{
		IsCreate:   true,
		ActionHref: ctx.Echo().Reverse(pages.Names.Classified.New),
		CancelHref: ctx.Echo().Reverse(pages.Names.Classified.ForSale.Browse),
		SubmitText: "Next",
		Form:       form,
		Page:       page,
	}
	return c.RenderHTML(ctx, page, &v)
}

// Post processes the submitted form to create a new post in the classifieds
// section.  This handler is protected by middleware so that only logged-in
// users can access it
func (c *create) Post(ctx echo.Context) error {
	var form views.ClassifiedCreateForm
	ctx.Set(context.FormKey, &form)

	if err := ctx.Bind(&form); err != nil {
		return c.Fail(err, "unable to bind form")
	}

	if err := form.Submission.Process(ctx, form); err != nil {
		return c.Fail(err, "unable to process form submission")
	}

	form.ProcessCreate(context.GetLogger(ctx))

	// Price is only considered in ForSale posts.
	var price sql.NullFloat64
	if form.ClassifiedType == string(constants.ForSale) {
		var err error

		// Strip any commas from the price field
		form.Price = strings.ReplaceAll(form.Price, ",", "")

		price, err = conv.StringToNullFloat64("price", form.Price)
		if err != nil {
			form.Submission.SetFieldError("price", "Invalid value")
		}

		// If the price is empty or 0, change to a ForFree post.
		if !price.Valid || price.Float64 == 0 {
			form.ClassifiedType = string(constants.ForFree)
			price.Valid = false
		}
	}

	if form.Submission.HasErrors() {
		return c.Get(ctx)
	}

	user, ok := ctx.Get(context.AuthenticatedUserKey).(sqlc.User)
	if !ok {
		return c.Fail(errors.New("failed to lookup authenticated user"), "unable to lookup authenticated user")
	}

	form.Sanitize(c.Container.Sanitizer)

	i := classified.NewPostInput{
		CreatePostParams: sqlc.CreatePostParams{
			PostType:       constants.Classified,
			Title:          form.Title,
			Description:    form.Description,
			Creator:        user.ID,
			AnonymousEmail: c.Container.DB.PostBase.CreateAnonymousEmail(form.IncludeEmail),
			PhoneCall:      conv.StringToNullString(form.PhoneNumber),
			PhoneText:      conv.StringToNullString(form.SMSNumber),
		},
		ClassifiedType: constants.ClassifiedType(form.ClassifiedType),
		Price:          price,
	}

	id, err := c.Container.DB.Classified.Create(ctx.Request().Context(), i)
	if err != nil {
		return c.Fail(err, "unable to save a new post")
	}

	statName := constants.ForSaleCreated
	switch form.ClassifiedType {
	case string(constants.ForFree):
		statName = constants.ForFreeCreated
	case string(constants.WantToBuy):
		statName = constants.WantToBuyCreated
	case string(constants.WantToBorrow):
		statName = constants.WantToBorrowCreated
	}
	err = c.Container.DB.C.IncrementStat(ctx.Request().Context(), statName)
	if err != nil {
		context.GetLogger(ctx).Warn("unable to increment stat for create post",
			zap.Error(err))
	}

	c.Container.DB.C.IncrementPostsCreatedCount(ctx.Request().Context(), user.ID)
	if err != nil {
		context.GetLogger(ctx).Warn("unable to increment posts created stat",
			zap.Error(err))
	}

	return c.Redirect(ctx, pages.Names.Posts.NewPhotos, id)
}
