package classified

import (
	"fmt"
	"strconv"

	"github.com/labstack/echo/v4"

	"git.grosinger.net/tgrosinger/tonyslist/constants/pages"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/context"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/controller"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/classified"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/msg"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/views"
)

type edit struct {
	controller.Controller
}

// Get displays a page for users to edit an existing post in the housing
// section. This handler is protected by middleware so that only the post owner
// can access it.
// NOTE: We use the term "post" because it could be a for_rent or a request.
func (c *edit) Get(ctx echo.Context) error {
	// Middleware protects against posts not existing and being expired.
	p, ok := ctx.Get(context.PostKey).(*classified.Post)
	if !ok {
		msg.Warning(ctx, "The requested post could not be found.")
		return c.RedirectToPostBrowser(ctx)
	}

	page := controller.NewPage(ctx, c.Container)
	page.Title = "Edit Post"
	page.Metatags.Description = fmt.Sprintf("Edit a post on %s.", c.Container.Config.App.Name)
	page.Metatags.Keywords = keywords

	var form views.ClassifiedCreateForm
	if f := ctx.Get(context.FormKey); f != nil {
		form = *f.(*views.ClassifiedCreateForm)
	} else {
		form = views.ClassifiedCreateForm{
			ClassifiedUpdateForm: views.ClassifiedUpdateForm{
				Title:       p.Title,
				Description: p.Description,
			},
			ClassifiedType: string(p.ClassifiedType),
		}
		if p.Price.Valid {
			form.Price = strconv.FormatFloat(p.Price.Float64, 'f', 2, 64)
		}
	}

	v := views.ClassifiedCreate{
		IsCreate:   false,
		ActionHref: ctx.Echo().Reverse(pages.Names.Classified.Edit, p.ID),
		CancelHref: ctx.Echo().Reverse(pages.Names.Classified.One, p.ID),
		SubmitText: "Update",
		Form:       form,
		Page:       page,
	}
	return c.RenderHTML(ctx, page, &v)
}

// Post processes the submitted form to edit a post in the classified section. This
// handler is protected by middleware so that only the post owner can access it.
func (c *edit) Post(ctx echo.Context) error {
	postID, err := strconv.Atoi(ctx.Param("id"))
	if err != nil {
		msg.Warning(ctx, "The requested post could not be found.")
		return c.Redirect(ctx, pages.Names.Classified.ForSale.Browse)
	}

	var form views.ClassifiedCreateForm
	ctx.Set(context.FormKey, &form)

	if err := ctx.Bind(&form); err != nil {
		return c.Fail(err, "unable to bind form")
	}

	// Only validate the nested PostUpdateForm. Validating the entire form will
	// result in a failure because the outer form does not currently have all
	// the fields filled out.
	if err := form.Submission.Process(ctx, form.ClassifiedUpdateForm); err != nil {
		return c.Fail(err, "unable to process form submission")
	}

	if form.Submission.HasErrors() {
		return c.Get(ctx)
	}

	form.Sanitize(c.Container.Sanitizer)

	i := classified.UpdatePostInput{
		UpdatePostParams: sqlc.UpdatePostParams{
			ID:          postID,
			Title:       form.Title,
			Description: form.Description,
		},
		Price: form.Price,
	}

	err = c.Container.DB.Classified.Update(ctx.Request().Context(), postID, i)
	if err != nil {
		return c.Fail(err, "unable to save a new post")
	}

	msg.Success(ctx, "Post updated successfully")
	return c.Redirect(ctx, pages.Names.Classified.One, postID)
}
