package classified

import (
	"fmt"
	"strings"

	"github.com/labstack/echo/v4"
	"go.uber.org/zap"

	"git.grosinger.net/tgrosinger/tonyslist/constants"
	"git.grosinger.net/tgrosinger/tonyslist/constants/pages"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/context"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/controller"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/postbase"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/views"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/views/components"
)

type forSale struct {
	controller.Controller
}

func (c *forSale) Get(ctx echo.Context) error {
	htmxReq := ctx.Request().Header.Get("HX-Request")
	pager := controller.NewPager(ctx, controller.DefaultItemsPerPage)
	query := postQueryFromContext(c.Container, ctx, &pager, constants.ForSale)

	data, count, err := c.Container.DB.Classified.Select(ctx.Request().Context(), query)
	if err != nil {
		if he, ok := err.(*echo.HTTPError); ok {
			return he
		}
		return c.Fail(err, "unable to retrieve requested post")
	}
	pager.SetItems(count)

	featurePhotos, err := c.Container.DB.PostBase.GetFeaturedPhotos(ctx.Request().Context(), data)
	if err != nil {
		c.Container.Logger.Warn("unable to load featured photos when loading post browser",
			zap.Error(err))
	}

	var myForSale []postbase.PostBase
	if user, ok := ctx.Get(context.AuthenticatedUserKey).(sqlc.User); ok {
		myForSale, err = c.Container.DB.Classified.MyPosts(ctx.Request().Context(), constants.ForSale, user.ID)
		if err != nil {
			if he, ok := err.(*echo.HTTPError); ok {
				return he
			}
			return c.Fail(err, "unable to retrieve requested post")
		}
	}

	page := controller.NewPage(ctx, c.Container)
	page.Title = "For Sale"
	page.Metatags.Description = fmt.Sprintf("Items for sale on %s Classifieds.", c.Container.Config.App.Name)
	page.Metatags.Keywords = keywords
	page.Metatags.HasFeed = true
	page.Pager = pager

	if strings.HasSuffix(ctx.Path(), ".rss") {
		v := views.RSS{
			HomeHref: c.Container.Config.App.PublicURL +
				ctx.Echo().Reverse(pages.Names.Classified.ForSale.Browse),
			SupportEmail: c.Container.Config.Mail.SupportAddress,
			Items:        postsToRSS(ctx, c.Container.Config, data),
			Page:         page,
			Sanitizer:    c.Container.Sanitizer,
		}
		return c.RenderXML(ctx, page, &v)
	}

	digest := ctx.Get(context.DigestMembershipKey)
	page.DisplayEmailDigestCTA = digest != constants.EmailDigestSubscriptionTwiceWeekly

	v := views.PostBrowse{
		Posts:              data,
		FeaturePhotos:      featurePhotos,
		MyPosts:            myForSale,
		DisplayType:        views.GridDisplayType,
		SortOptions:        constants.DefaultSortOptions,
		CategorySelections: categorySelection,
		Filters:            constants.DefaultFilters,
		NewPostLink:        fmt.Sprintf("%s?src_classified_type=%s", ctx.Echo().Reverse(pages.Names.Classified.New), constants.ForSale),
		PostsListOnly:      htmxReq == "true",

		Headers: components.ClassifiedHeader(page, constants.ForSale),
		Page:    page,
	}
	return c.RenderHTML(ctx, page, &v)
}
