package classified

import (
	"net/url"
	"strings"

	"github.com/labstack/echo/v4"
	"go.uber.org/zap"

	"git.grosinger.net/tgrosinger/tonyslist/config"
	"git.grosinger.net/tgrosinger/tonyslist/constants"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/context"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/controller"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/classified"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/postbase"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/services"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/views"
)

// categorySelection is a placeholder for the quick filter options on the
// sidebar of the post-browser template.
var categorySelection = []string{}

type postQueryArgs struct {
	search  string
	sort    string
	filters string
}

func postQueryFromContext(c *services.Container, ctx echo.Context, pager *controller.Pager, postType constants.ClassifiedType) classified.SelectInput {
	var args postQueryArgs
	htmxReq := ctx.Request().Header.Get("HX-Request")
	if htmxReq == "true" {
		args = postQueryArgs{
			search: ctx.QueryParam("search"),
		}

		currentURL := ctx.Request().Header.Get("HX-Current-URL")
		parsedURL, err := url.Parse(currentURL)
		if err == nil {
			args.sort = parsedURL.Query().Get("sort")
			args.filters = parsedURL.Query().Get("filters")
		}
	} else {
		args = postQueryArgs{
			search:  ctx.QueryParam("search"),
			sort:    ctx.QueryParam("sort"),
			filters: ctx.QueryParam("filters"),
		}
	}

	q := classified.SelectInput{
		SelectInput: postbase.SelectInput{
			Pagination: postbase.Pagination{
				Limit:  uint64(pager.ItemsPerPage),
				Offset: uint64(pager.GetOffset()),
			},
			IncludeUnlisted: false,
		},
		PostType: postType,
	}

	switch args.sort {
	case "high_to_low":
		q.SortColumn = classified.Price
		q.SortDir = postbase.SortDesc
	case "low_to_high":
		q.SortColumn = classified.Price
		q.SortDir = postbase.SortAsc
	default: // "newest"
		q.SortColumn = postbase.PublishedAt
		q.SortDir = postbase.SortDesc
	}

	if strings.HasSuffix(ctx.Path(), ".rss") {
		// Override the sort and limit parameters for rss feeds
		q.SortColumn = postbase.PublishedAt
		q.SortDir = postbase.SortDesc
		q.Offset = 0
		q.Limit = 30
	}

	q.Fields = []string{postbase.Title, postbase.Description}
	if filters := args.filters; filters != "" {
		filters = strings.TrimSpace(filters)
		for _, filter := range strings.Split(filters, ",") {
			switch filter {
			case "basic_photos":
				q.RequirePhotos = true
			case "basic_title-only":
				q.Fields = []string{postbase.Title}
			}
		}
	}

	if query := args.search; query != "" {
		q.Query = query
		parsed, err := c.Parser.ParseSearchQuery(query)
		if err != nil {
			context.GetLogger(ctx).Error("Unable to parse query",
				zap.Error(err),
				zap.String("query", query))
		} else {
			q.ParsedQuery = parsed
		}
	}

	return q
}

func postsToRSS(ctx echo.Context, cfg *config.Config, data []postbase.PostBase) []views.RSSItem {
	items := make([]views.RSSItem, 0, len(data))
	for _, p := range data {
		items = append(items, views.RSSItem{
			Title:         p.Title,
			Description:   p.Description,
			Href:          cfg.App.PublicURL + p.GetViewHref(ctx.Echo()),
			PublishedDate: p.CreatedAt,
		})
	}

	return items
}
