package routes

import (
	"context"
	"database/sql"
	"mime/multipart"
	"net/http"
	"net/url"
	"testing"
	"time"

	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"

	"git.grosinger.net/tgrosinger/tonyslist/constants"
	"git.grosinger.net/tgrosinger/tonyslist/constants/pages"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/classified"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
)

func TestClassified_GetExpired(t *testing.T) {
	// Ensure that the two users have been instantiated
	_ = request(t).withUnverifiedUser()
	_ = request(t).withVerifiedUser()

	postID, err := c.DB.Classified.Create(context.Background(), classified.NewPostInput{
		CreatePostParams: sqlc.CreatePostParams{
			PostType:    constants.Classified,
			Title:       "Test post that is expired",
			Description: "Test description",
			PhoneCall:   sql.NullString{Valid: false},
			PhoneText:   sql.NullString{Valid: false},
			Creator:     verifiedUser.id,
		},
		ClassifiedType: constants.ForSale,
		Price:          sql.NullFloat64{Valid: true, Float64: 12.23},
	})
	require.NoError(t, err)

	// Manually set the post to have been published 1 month ago
	oldTime := time.Now().Add(-1 * 24 * 30 * time.Hour)
	err = c.DB.C.SetPostPublishedAt(context.Background(), sqlc.SetPostPublishedAtParams{
		ID:          postID,
		PublishedAt: sql.NullTime{Valid: true, Time: oldTime},
	})
	require.NoError(t, err)

	// Verify that non-creator cannot retrieve the post
	request(t).
		withUnverifiedUser().
		setRoute(pages.Names.Classified.One, postID).
		get().
		requireStatusCode(http.StatusNotFound)

	// Verify that the owner can retrieve the post even though it's expired.
	request(t).
		withVerifiedUser().
		setRoute(pages.Names.Classified.One, postID).
		get().
		requireStatusCode(http.StatusOK)
}

func TestClassified_SignedOutNewPostGet(t *testing.T) {
	expectedPath := c.Web.Reverse(pages.Names.User.Login) + "?return_to=" + url.QueryEscape(c.Web.Reverse(pages.Names.Classified.New))
	request(t).
		setRoute(pages.Names.Classified.New).
		disableFollowRedirects().
		get().
		requireStatusCode(http.StatusFound).
		assertRedirectPath(expectedPath)
}

func TestClassified_SignedOutNewPostPost(t *testing.T) {
	expectedPath := c.Web.Reverse(pages.Names.User.Login) + "?return_to=" + url.QueryEscape(c.Web.Reverse(pages.Names.Classified.New))
	request(t).
		setRoute(pages.Names.Classified.New).
		disableFollowRedirects().
		postMultipartForm(func(w *multipart.Writer) {
			setMultipartField(t, w, "classified_type", "for-sale")
			setMultipartField(t, w, "title", "Created from TestClassified")
			setMultipartField(t, w, "description", "The Description")
		}).
		requireStatusCode(http.StatusFound).
		assertRedirectPath(expectedPath)
}

func TestClassified_InvalidPostType(t *testing.T) {
	fieldErrorEl := request(t).
		withVerifiedUser().
		setRoute(pages.Names.Classified.New).
		disableFollowRedirects().
		postMultipartForm(func(w *multipart.Writer) {
			setMultipartField(t, w, "classified_type", "invalid")
			setMultipartField(t, w, "title", "Created from TestClassified")
			setMultipartField(t, w, "description", "The Description")
		}).
		requireStatusCode(http.StatusOK).
		toDoc().
		Find(`.text-red-600`).
		First()

	html, err := fieldErrorEl.Html()
	require.NoError(t, err)
	assert.Equal(t, "Invalid classified type in request", html)
}

func TestClassified_CreateSparse(t *testing.T) {
	postID := request(t).
		withVerifiedUser().
		setRoute(pages.Names.Classified.New).
		disableFollowRedirects().
		postMultipartForm(func(w *multipart.Writer) {
			setMultipartField(t, w, "classified_type", "for-sale")
			setMultipartField(t, w, "title", "Created from TestClassified")
			setMultipartField(t, w, "description", "The Description")
			setMultipartField(t, w, "price", "4.50")
			setMultipartField(t, w, "include_email", "true")
		}).
		requireStatusCode(http.StatusFound).
		getLocationID()

	postBase, err := c.DB.PostBase.GetOne(context.Background(), postID, false)
	require.NoError(t, err)
	post, err := c.DB.Classified.GetOneFromPostBase(context.Background(), postBase)
	require.NoError(t, err)

	assert.Equal(t, postID, post.ID)
	assert.Equal(t, "Created from TestClassified", post.Title)
	assert.Equal(t, "The Description", post.Description)
	assert.Equal(t, constants.ForSale, post.ClassifiedType)
	assert.Equal(t, 4.50, post.Price.Float64)
	assert.Less(t, float64(time.Until(post.CreatedAt)*-1), float64(time.Second))
	assert.Less(t, float64(time.Until(post.PublishedAt.Time)*-1), float64(time.Second))
	assert.False(t, post.FeaturedPhoto.Valid)
	assert.False(t, post.PhoneCall.Valid)
	assert.False(t, post.PhoneText.Valid)
	assert.True(t, post.Listed)
	assert.True(t, post.AnonymousEmail.Valid)
	assert.Len(t, post.AnonymousEmail.String, constants.AnonymousEmailLen+len(constants.AnonymousEmailPrefix))
}

func TestClassified_CreateWithoutEmail(t *testing.T) {
	postID := request(t).
		withVerifiedUser().
		setRoute(pages.Names.Classified.New).
		disableFollowRedirects().
		postMultipartForm(func(w *multipart.Writer) {
			setMultipartField(t, w, "classified_type", "for-sale")
			setMultipartField(t, w, "title", "Created from TestClassified")
			setMultipartField(t, w, "description", "The Description")
			setMultipartField(t, w, "price", "4.50")
			setMultipartField(t, w, "include_email", "false")
			setMultipartField(t, w, "sms_number", "222-222-2222")
		}).
		requireStatusCode(http.StatusFound).
		getLocationID()

	postBase, err := c.DB.PostBase.GetOne(context.Background(), postID, false)
	require.NoError(t, err)
	post, err := c.DB.Classified.GetOneFromPostBase(context.Background(), postBase)
	require.NoError(t, err)

	assert.Equal(t, postID, post.ID)
	assert.Equal(t, "Created from TestClassified", post.Title)
	assert.Equal(t, "The Description", post.Description)
	assert.Equal(t, constants.ForSale, post.ClassifiedType)
	assert.Equal(t, 4.50, post.Price.Float64)
	assert.Less(t, float64(time.Until(post.CreatedAt)*-1), float64(time.Second))
	assert.Less(t, float64(time.Until(post.PublishedAt.Time)*-1), float64(time.Second))
	assert.False(t, post.FeaturedPhoto.Valid)
	assert.False(t, post.PhoneCall.Valid)
	assert.Equal(t, "222-222-2222", post.PhoneText.String)
	assert.True(t, post.Listed)
	assert.False(t, post.AnonymousEmail.Valid)
}

func TestClassified_Create(t *testing.T) {
	postID := request(t).
		withVerifiedUser().
		setRoute(pages.Names.Classified.New).
		disableFollowRedirects().
		postMultipartForm(func(w *multipart.Writer) {
			setMultipartField(t, w, "classified_type", "to-buy")
			setMultipartField(t, w, "title", "Created from TestClassified")
			setMultipartField(t, w, "description", "The Description")
			setMultipartField(t, w, "price", "1.25")
			setMultipartField(t, w, "include_email", "true")
			setMultipartField(t, w, "phone_number", "111-111-1111")
			setMultipartField(t, w, "sms_number", "222-222-2222")
		}).
		requireStatusCode(http.StatusFound).
		getLocationID()

	postBase, err := c.DB.PostBase.GetOne(context.Background(), postID, false)
	require.NoError(t, err)
	post, err := c.DB.Classified.GetOneFromPostBase(context.Background(), postBase)
	require.NoError(t, err)

	assert.Equal(t, postID, post.ID)
	assert.Equal(t, "Created from TestClassified", post.Title)
	assert.Equal(t, "The Description", post.Description)
	assert.Equal(t, constants.WantToBuy, post.ClassifiedType)
	assert.Equal(t, "111-111-1111", post.PhoneCall.String)
	assert.Equal(t, "222-222-2222", post.PhoneText.String)
	assert.True(t, post.AnonymousEmail.Valid)
	assert.Len(t, post.AnonymousEmail.String, constants.AnonymousEmailLen+len(constants.AnonymousEmailPrefix))
	assert.False(t, post.Price.Valid) // To-Buy posts do not have a price

	// Check that post is not retrievable by other users until published
	status := request(t).
		withUnverifiedUser().
		disableFollowRedirects().
		setRoute(pages.Names.Classified.One, postID).
		get().
		StatusCode
	require.Equal(t, http.StatusNotFound, status)

	// Check that the owner can retrieve it
	resp := request(t).
		withVerifiedUser().
		disableFollowRedirects().
		setRoute(pages.Names.Classified.One, postID).
		get()
	require.Equal(t, http.StatusOK, resp.StatusCode)

	// Mark published and check that other users can retrieve it
	err = c.DB.C.SetPostPublishedNow(context.Background(), postID)
	require.NoError(t, err)

	resp = request(t).
		withUnverifiedUser().
		disableFollowRedirects().
		setRoute(pages.Names.Classified.One, postID).
		get()
	require.Equal(t, http.StatusOK, resp.StatusCode)
}

func TestClassified_ZeroPrice(t *testing.T) {
	postID := request(t).
		withVerifiedUser().
		setRoute(pages.Names.Classified.New).
		disableFollowRedirects().
		postMultipartForm(func(w *multipart.Writer) {
			setMultipartField(t, w, "classified_type", "for-sale")
			setMultipartField(t, w, "title", "Created from TestClassified")
			setMultipartField(t, w, "description", "The Description")
			setMultipartField(t, w, "price", "0")
			setMultipartField(t, w, "include_email", "true")
		}).
		requireStatusCode(http.StatusFound).
		getLocationID()

	postBase, err := c.DB.PostBase.GetOne(context.Background(), postID, false)
	require.NoError(t, err)
	post, err := c.DB.Classified.GetOneFromPostBase(context.Background(), postBase)
	require.NoError(t, err)

	assert.Equal(t, postID, post.ID)
	assert.Equal(t, "Created from TestClassified", post.Title)
	assert.Equal(t, "The Description", post.Description)
	assert.Equal(t, constants.ForFree, post.ClassifiedType)
	assert.False(t, post.Price.Valid)
	assert.False(t, post.FeaturedPhoto.Valid)
	assert.False(t, post.PhoneCall.Valid)
	assert.False(t, post.PhoneText.Valid)
	assert.True(t, post.Listed)
	assert.True(t, post.AnonymousEmail.Valid)
	assert.Len(t, post.AnonymousEmail.String, constants.AnonymousEmailLen+len(constants.AnonymousEmailPrefix))

	postID = request(t).
		withVerifiedUser().
		setRoute(pages.Names.Classified.New).
		disableFollowRedirects().
		postMultipartForm(func(w *multipart.Writer) {
			setMultipartField(t, w, "classified_type", "for-sale")
			setMultipartField(t, w, "title", "Created from TestClassified")
			setMultipartField(t, w, "description", "The Description")
			setMultipartField(t, w, "include_email", "true")
		}).
		requireStatusCode(http.StatusFound).
		getLocationID()

	postBase, err = c.DB.PostBase.GetOne(context.Background(), postID, false)
	require.NoError(t, err)
	post, err = c.DB.Classified.GetOneFromPostBase(context.Background(), postBase)
	require.NoError(t, err)

	assert.Equal(t, postID, post.ID)
	assert.Equal(t, "Created from TestClassified", post.Title)
	assert.Equal(t, "The Description", post.Description)
	assert.Equal(t, constants.ForFree, post.ClassifiedType)
	assert.False(t, post.Price.Valid)
	assert.False(t, post.FeaturedPhoto.Valid)
	assert.False(t, post.PhoneCall.Valid)
	assert.False(t, post.PhoneText.Valid)
	assert.True(t, post.Listed)
	assert.True(t, post.AnonymousEmail.Valid)
	assert.Len(t, post.AnonymousEmail.String, constants.AnonymousEmailLen+len(constants.AnonymousEmailPrefix))
}
