package routes

import (
	"database/sql"
	"net/http"
	"strconv"

	"github.com/labstack/echo/v4"
	"github.com/stripe/stripe-go/v78"
	"github.com/stripe/stripe-go/v78/paymentintent"
	"github.com/stripe/stripe-go/v78/subscription"
	"go.uber.org/zap"

	"git.grosinger.net/tgrosinger/tonyslist/constants/pages"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/context"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/controller"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/conv"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/metrics"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/msg"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/views"
)

type (
	donate struct {
		controller.Controller
	}
)

func (c *donate) Get(ctx echo.Context) error {
	customerID, ok := ctx.Get(context.StripeCustomerIDKey).(string)
	stripeCustomerID := sql.NullString{String: customerID, Valid: ok}

	page := controller.NewPage(ctx, c.Container)
	page.Title = "Donate"
	page.Metatags.Description = "Donate to support OrcasHub"

	v := views.Donate{
		CustomerID:     stripeCustomerID,
		PublishableKey: c.Container.Config.Stripe.PublishableKey,
		ReturnURL:      c.Container.Config.App.PublicURL + ctx.Echo().Reverse(pages.Names.DonateComplete),
		Page:           page,
	}
	return c.RenderHTML(ctx, page, &v)
}

func (c *donate) GetComplete(ctx echo.Context) error {
	status := ctx.QueryParam("redirect_status")
	if status != "" && status != "succeeded" {
		metrics.DonateErrorCount.
			WithLabelValues(metrics.MetricDonateCompleteUnsuccessful).
			Inc()
		msg.Danger(ctx, "Your donation was not completed. Please try again.")
		return c.Redirect(ctx, pages.Names.Donate)
	}

	page := controller.NewPage(ctx, c.Container)
	page.Title = "Thank you"
	page.Metatags.Description = "Thank you for donating to support OrcasHub"

	donateCompleteContent, err := c.LoadMarkdown(ctx, "donate-complete.md")
	if err != nil {
		return err
	}

	v := views.MarkdownView{
		Content: donateCompleteContent,
		Page:    page,
	}
	return c.RenderHTML(ctx, page, &v)
}

func (c *donate) CreateCustomer(ctx echo.Context) error {
	email := ctx.QueryParam("email")
	customerID, err := c.Container.Stripe.GetStripeCustomerIDForEmail(ctx.Request().Context(), email, sql.NullString{})
	if err != nil {
		context.GetLogger(ctx).Warn("unable to retrieve Stripe customer ID",
			zap.Error(err))
		return ctx.JSON(http.StatusInternalServerError, map[string]any{
			"status":  "error",
			"message": "unable to retrieve Stripe customer ID",
		})
	}

	return ctx.JSON(http.StatusOK, map[string]any{
		"status":      "ok",
		"customer_id": customerID,
	})
}

// CreatePaymentIntent is used to create a payment intent for a one-time donation.
// Based on the following docs:
// - https://docs.stripe.com/payments/quickstart?client=html
// - https://docs.stripe.com/payments/accept-a-payment-deferred
func (c *donate) CreatePaymentIntent(ctx echo.Context) error {
	customerID := ctx.FormValue("customer-id")
	if customerID == "" {
		return ctx.JSON(http.StatusBadRequest, map[string]any{
			"status":  "error",
			"message": "Could not find your Stripe customer ID.",
		})
	}

	amountStr := ctx.FormValue("amount")
	if amountStr == "" {
		return ctx.JSON(http.StatusBadRequest, map[string]any{
			"status":  "error",
			"message": "Invalid payment amount",
		})
	}
	amount, err := strconv.ParseInt(amountStr, 10, 64)
	if err != nil {
		return ctx.JSON(http.StatusBadRequest, map[string]any{
			"status":  "error",
			"message": "Invalid payment amount",
		})
	}

	name := ctx.FormValue("name")
	c.Container.Stripe.UpdateCustomerName(ctx.Request().Context(), customerID, name)
	user, ok := ctx.Get(context.AuthenticatedUserKey).(sqlc.User)
	if ok {
		c.Container.DB.C.SetUserName(ctx.Request().Context(), sqlc.SetUserNameParams{
			Name: conv.StringToNullString(name),
			ID:   user.ID,
		})
	}

	params := &stripe.PaymentIntentParams{
		Amount:   stripe.Int64(amount),
		Customer: stripe.String(customerID),
		Currency: stripe.String(string(stripe.CurrencyUSD)),
		// In the latest version of the API, specifying the `automatic_payment_methods` parameter is optional because Stripe enables its functionality by default.
		AutomaticPaymentMethods: &stripe.PaymentIntentAutomaticPaymentMethodsParams{
			Enabled: stripe.Bool(true),
		},
		SetupFutureUsage: stripe.String("on_session"),
	}

	pi, err := paymentintent.New(params)
	if err != nil {
		metrics.DonateErrorCount.
			WithLabelValues(metrics.MetricDonatePaymentIntent).
			Inc()
		context.GetLogger(ctx).Warn("Failed to create payment intent", zap.Error(err))
		return ctx.JSON(http.StatusInternalServerError, map[string]any{
			"status":  "error",
			"message": "Invalid payment amount",
		})
	}

	return ctx.JSON(http.StatusOK, map[string]any{
		"status":       "ok",
		"clientSecret": pi.ClientSecret,
	})
}

// CreateSubscription is used to create a subscription for a recurring donation.
// Based on the following docs:
//   - https://docs.stripe.com/payments/accept-a-payment-deferred?platform=web&type=subscription&lang=go
//   - https://docs.stripe.com/products-prices/pricing-models#variable-pricing
func (c *donate) CreateSubscription(ctx echo.Context) error {
	customerID := ctx.FormValue("customer-id")
	if customerID == "" {
		return ctx.JSON(http.StatusBadRequest, map[string]any{
			"status":  "error",
			"message": "Could not find your Stripe customer ID.",
		})
	}

	amountStr := ctx.FormValue("amount")
	if amountStr == "" {
		return ctx.JSON(http.StatusBadRequest, map[string]any{
			"status":  "error",
			"message": "Invalid payment amount",
		})
	}
	amount, err := strconv.ParseInt(amountStr, 10, 64)
	if err != nil {
		return ctx.JSON(http.StatusBadRequest, map[string]any{
			"status":  "error",
			"message": "Invalid payment amount",
		})
	}

	name := ctx.FormValue("name")
	c.Container.Stripe.UpdateCustomerName(ctx.Request().Context(), customerID, name)
	user, ok := ctx.Get(context.AuthenticatedUserKey).(sqlc.User)
	if ok {
		c.Container.DB.C.SetUserName(ctx.Request().Context(), sqlc.SetUserNameParams{
			Name: conv.StringToNullString(name),
			ID:   user.ID,
		})
	}

	var item *stripe.SubscriptionItemsParams
	switch amount {
	case 500:
		item = &stripe.SubscriptionItemsParams{
			Price: stripe.String(c.Container.Config.Stripe.SubscriptionID5),
		}
	case 1000:
		item = &stripe.SubscriptionItemsParams{
			Price: stripe.String(c.Container.Config.Stripe.SubscriptionID10),
		}
	case 2500:
		item = &stripe.SubscriptionItemsParams{
			Price: stripe.String(c.Container.Config.Stripe.SubscriptionID25),
		}
	default:
		item = &stripe.SubscriptionItemsParams{
			PriceData: &stripe.SubscriptionItemPriceDataParams{
				Product:    stripe.String(c.Container.Config.Stripe.SubscriptionIDProduct),
				UnitAmount: stripe.Int64(amount),
				Currency:   stripe.String(string(stripe.CurrencyUSD)),
				Recurring: &stripe.SubscriptionItemPriceDataRecurringParams{
					Interval: stripe.String("year"),
				},
			},
		}
	}

	subscriptionParams := &stripe.SubscriptionParams{
		Customer: stripe.String(customerID),
		Items: []*stripe.SubscriptionItemsParams{
			item,
		},
		PaymentSettings: &stripe.SubscriptionPaymentSettingsParams{
			// Automatically save the payment method to the subscription
			// when the first payment is successful.
			SaveDefaultPaymentMethod: stripe.String("on_subscription"),
		},
		PaymentBehavior: stripe.String("default_incomplete"),
	}

	// Expand the Subscription's latest invoice and that invoice's
	// payment_intent so we can pass it to the front end to confirm the payment.
	subscriptionParams.AddExpand("latest_invoice.payment_intent")
	subscriptionParams.AddExpand("pending_setup_intent")

	s, err := subscription.New(subscriptionParams)
	if err != nil {
		metrics.DonateErrorCount.
			WithLabelValues(metrics.MetricDonateCreateSubscription).
			Inc()
		context.GetLogger(ctx).Warn("unable to create new subscription",
			zap.Error(err))
		return ctx.JSON(http.StatusInternalServerError, map[string]any{
			"status":  "error",
			"message": "unable to create new subscription",
		})
	}

	var clientSecret string
	var txType string
	if s.PendingSetupIntent != nil {
		txType = "setup"
		clientSecret = s.PendingSetupIntent.ClientSecret
	} else {
		txType = "payment"
		clientSecret = s.LatestInvoice.PaymentIntent.ClientSecret
	}

	return ctx.JSON(http.StatusOK, map[string]any{
		"status":         "ok",
		"subscriptionID": s.ID,
		"type":           txType,
		"clientSecret":   clientSecret,
	})
}
