package event

import (
	"database/sql"
	"encoding/base64"
	"errors"
	"fmt"
	"time"

	ics "github.com/arran4/golang-ical"
	"github.com/labstack/echo/v4"
	"go.uber.org/zap"

	"git.grosinger.net/tgrosinger/tonyslist/constants/pages"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/context"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/controller"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/msg"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/services/mail"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/services/mail/types"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/views"
)

// These properties are a workaround for all-day events
// https://github.com/arran4/golang-ical/issues/55
const (
	dateFormatUtc = "20060102"

	propertyDtStart ics.Property = "DTSTART;VALUE=DATE"
	propertyDtEnd   ics.Property = "DTEND;VALUE=DATE"

	componentPropertyDtStart = ics.ComponentProperty(propertyDtStart)
	componentPropertyDtEnd   = ics.ComponentProperty(propertyDtEnd)
)

type details struct {
	controller.Controller
}

func (c *details) Get(ctx echo.Context) error {
	event, ok := ctx.Get(context.EventKey).(sqlc.Event)
	if !ok {
		context.GetLogger(ctx).Warn("Unable to retrieve event from context unexpectedly")
		return echo.ErrInternalServerError
	}

	photo, err := c.Container.DB.C.GetEventPhoto(ctx.Request().Context(), event.ID)
	var photoPtr *sqlc.EventPhoto
	if err == nil {
		photoPtr = &photo
	} else if err != nil && err != sql.ErrNoRows {
		context.GetLogger(ctx).Warn("Unable to retrieve event photo", zap.Error(err))
		return echo.ErrInternalServerError
	}

	page := controller.NewPage(ctx, c.Container)
	page.Title = "Events - " + event.Title
	page.Metatags.Description = c.Container.Config.App.Name + " Events Calendar"
	page.Metatags.Keywords = keywords

	// If the viewer is a moderator, retrieve the post creator user info to display in the moderator toolbar.
	var info sqlc.GetUserInfoRow
	if page.Enforce("event", "moderate") {
		var err error
		info, err = c.Container.DB.C.GetUserInfo(ctx.Request().Context(), event.Creator)
		if err != nil {
			msg.Warning(ctx, "Unable to retrieve post creator info")
			context.GetLogger(ctx).Warn("Unable to retrieve post creator info", zap.Error(err))
		}
	}

	v := views.EventDetails{
		Event:            event,
		Photo:            photoPtr,
		CreatorInfo:      &info,
		MarkdownRenderer: c.Container.MarkdownRenderer,
		Page:             page,
	}
	return c.RenderHTML(ctx, page, &v)
}

func (c *details) AddToCalendar(ctx echo.Context) error {
	event, ok := ctx.Get(context.EventKey).(sqlc.Event)
	if !ok {
		context.GetLogger(ctx).Warn("Unable to retrieve event from context unexpectedly")
		return echo.ErrInternalServerError
	}

	user, ok := ctx.Get(context.AuthenticatedUserKey).(sqlc.User)
	if !ok {
		return c.Fail(errors.New("failed to lookup authenticated user"), "unable to lookup authenticated user")
	}

	err := c.emailICal(ctx, event, user)
	if err != nil {
		context.GetLogger(ctx).Error("Unable to send event ical", zap.Error(err))
		msg.Danger(ctx, "Unable to send event invitation.")
	} else {
		msg.Success(ctx, "An invitation has been emailed to you.")
	}

	return c.Redirect(ctx, pages.Names.Event.One, event.ID)
}

func (c *details) emailICal(ctx echo.Context, event sqlc.Event, user sqlc.User) error {
	cal := ics.NewCalendar()
	icalEvent := ics.NewEvent(fmt.Sprintf("%d@%s", event.ID, c.Container.Config.App.PublicURL))
	icalEvent.SetLocation(event.Location)
	icalEvent.SetSummary(event.Title)

	icalEvent.SetOrganizer(c.Container.Config.Mail.EventsAddress,
		ics.WithCN(c.Container.Config.App.Name))

	// TODO: Handle RSVP responses when I start tracking who is or is not attending the event.
	// See support-email.go
	icalEvent.AddAttendee(user.Email, ics.WithRSVP(false),
		ics.ParticipationRoleOptParticipant,
		ics.CalendarUserTypeIndividual)

	url := c.Container.Config.App.PublicURL + c.Container.Web.Reverse(pages.Names.Event.One, event.ID)
	icalEvent.SetDescription(fmt.Sprintf("%s\n\n%s", event.Description, url))
	icalEvent.SetURL(url)

	startTime := ""
	if event.HasTime {
		startTime = event.StartTime.In(c.Container.Config.App.Loc).Format(time.Kitchen)
		icalEvent.SetStartAt(event.StartTime.In(c.Container.Config.App.Loc))
	} else {
		// Set all-day event
		// https://github.com/arran4/golang-ical/issues/55
		icalEvent.SetProperty(componentPropertyDtStart, event.StartTime.UTC().Format(dateFormatUtc))
		icalEvent.SetProperty(componentPropertyDtEnd, event.StartTime.UTC().AddDate(0, 0, 1).Format(dateFormatUtc))
	}

	if event.EndTime.Valid {
		icalEvent.SetEndAt(event.EndTime.Time.In(c.Container.Config.App.Loc))
	} else {
		icalEvent.SetDuration(time.Hour)
	}

	cal.AddVEvent(icalEvent)
	icalStr := cal.Serialize()
	icalB64 := base64.StdEncoding.EncodeToString([]byte(icalStr))

	templateData := map[string]interface{}{
		"title":       event.Title,
		"date":        event.StartTime.Format("Mon Jan _2"),
		"start_time":  startTime,
		"description": event.Description,
	}

	m, err := c.Container.Mail.Template(mail.TemplateArgs{
		MessageStream: types.EventsTransactional,
		To:            user.Email,
		From:          c.Container.Config.Mail.EventsAddress,
		Subject:       fmt.Sprintf("[%s Event] %s", c.Container.Config.App.Name, event.Title),
		TextTemplate:  "event_ical_text",
		TemplateData:  templateData,
		Attachments: []types.Attachment{
			{
				Name:        "calendar_invite.ics",
				ContentType: types.Calendar,
				Disposition: "attachment",
				Content:     icalB64,
			},
		},
	})
	if err != nil {
		return err
	}

	c.Container.Mail.Send(context.GetLogger(ctx), []*types.Mail{m}, nil)
	return nil
}
