package event

import (
	"fmt"
	"strconv"
	"time"

	"github.com/labstack/echo/v4"

	"git.grosinger.net/tgrosinger/tonyslist/constants/pages"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/context"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/controller"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/msg"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/views"
)

func (c *update) EditGet(ctx echo.Context) error {
	event, ok := ctx.Get(context.EventKey).(sqlc.Event)
	if !ok {
		context.GetLogger(ctx).Warn("Unable to retrieve event from context unexpectedly")
		return echo.ErrInternalServerError
	}

	page := controller.NewPage(ctx, c.Container)
	page.Title = "Edit Event"
	page.Metatags.Description = fmt.Sprintf("Edit event on %s.", c.Container.Config.App.Name)
	page.Metatags.Keywords = keywords

	var form views.EventCreateForm
	if f := ctx.Get(context.FormKey); f != nil {
		form = *f.(*views.EventCreateForm)
	} else {
		startTime := ""
		if event.HasTime {
			startTime = event.StartTime.Format("15:04")
		}

		form = views.EventCreateForm{
			Title:       event.Title,
			Description: event.Description,
			Location:    event.Location,
			StartTime:   startTime,
			Date:        event.StartTime.Format(time.DateOnly),
		}
		if event.EndTime.Valid {
			form.EndTime = event.EndTime.Time.Format("15:04")
		}
	}

	v := views.EventCreate{
		Location:   c.Container.Config.App.Loc,
		ActionHref: ctx.Echo().Reverse(pages.Names.Event.Edit, event.ID),
		CancelHref: ctx.Echo().Reverse(pages.Names.Event.One, event.ID),
		SubmitText: "Update",
		Form:       form,
		Page:       page,
	}
	return c.RenderHTML(ctx, page, &v)
}

func (c *update) EditPost(ctx echo.Context) error {
	eventID, err := strconv.Atoi(ctx.Param("id"))
	if err != nil {
		msg.Warning(ctx, "The requested event could not be found.")
		return c.Redirect(ctx, pages.Names.Event.Browse)
	}

	var form views.EventCreateForm
	ctx.Set(context.FormKey, &form)

	if err := ctx.Bind(&form); err != nil {
		return c.Fail(err, "unable to bind form")
	}

	if err := form.Submission.Process(ctx, form); err != nil {
		return c.Fail(err, "unable to process form submission")
	}

	form.Process(context.GetLogger(ctx), c.Container.Config.App.Loc)

	if form.Submission.HasErrors() {
		return c.EditGet(ctx)
	}

	form.Sanitize(c.Container.Sanitizer)

	err = c.Container.DB.C.UpdateEvent(ctx.Request().Context(), sqlc.UpdateEventParams{
		Title:       form.Title,
		Description: form.Description,
		Location:    form.Location,
		HasTime:     form.StartTime != "",
		StartTime:   form.ProcessedStartTime, // Valid is checked above
		EndTime:     form.ProcessedEndTime,
		ID:          eventID,
	})
	if err != nil {
		return c.Fail(err, "unable to update event")
	}

	return c.Redirect(ctx, pages.Names.Event.One, eventID)
}
