package routes

import (
	"fmt"
	"html/template"

	"github.com/labstack/echo/v4"

	"git.grosinger.net/tgrosinger/tonyslist/pkg/context"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/controller"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/msg"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/services/mail"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/services/mail/types"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/views"
)

type (
	home struct {
		controller.Controller
	}

	contactForm struct {
		Name       string `form:"name" valid:"required"`
		Email      string `form:"email" valid:"required,email"`
		Message    string `form:"message" valid:"required"`
		Empty      string `form:"maiden_name" valid:"optional"`
		Submission controller.FormSubmission
	}
)

func (c *home) Get(ctx echo.Context) error {
	page := controller.NewPage(ctx, c.Container)
	page.Metatags.Description = "The community of Orcas Island. Classifieds, housing, and events posted by Orcasonians."
	page.Metatags.Keywords = []string{"Classifieds", "Mailing List", "Events", "Community"}

	leftMarkdown, err := c.LoadMarkdown(ctx, "homepage-left.md")
	if err != nil {
		return err
	}
	rightMarkdown, err := c.LoadMarkdown(ctx, "homepage-right.md")
	if err != nil {
		return err
	}

	v := views.Home{
		LeftSectionContent:  leftMarkdown,
		RightSectionContent: rightMarkdown,
		Page:                page,
	}
	return c.RenderHTML(ctx, page, &v)
}

func (c *home) PostContact(ctx echo.Context) error {
	var form contactForm
	ctx.Set(context.FormKey, &form)

	// Parse the form values
	if err := ctx.Bind(&form); err != nil {
		return c.Fail(err, "unable to bind form")
	}

	if err := form.Submission.Process(ctx, form); err != nil {
		return c.Fail(err, "unable to process form submission")
	}

	if !form.Submission.HasErrors() {
		if form.Empty != "" {
			msg.Danger(ctx, "Form contained an unexpected value")
			return c.Get(ctx)
		}

		templateData := map[string]interface{}{
			"name":  form.Name,
			"email": form.Email,
			// NOTE: Not actually safe, but we're just sending to ourselves here, not members.
			"message": template.HTML(form.Message),
		}

		m, err := c.Container.Mail.Template(mail.TemplateArgs{
			MessageStream: types.AdminNotifications,
			To:            c.Container.Config.Mail.AdminAddress,
			ReplyTo:       fmt.Sprintf("%s<%s>", form.Name, form.Email),
			Subject:       fmt.Sprintf("%s Contact Form", c.Container.Config.App.Name),
			TextTemplate:  "contact_form_text",
			TemplateData:  templateData,
		})
		if err != nil {
			return c.Fail(err, "unable to send email")
		}

		c.Container.Mail.Send(context.GetLogger(ctx), []*types.Mail{m}, nil)
		msg.Success(ctx, "Message sent")
	}

	return c.Get(ctx)
}

func (c *home) CommunityResources(ctx echo.Context) error {
	page := controller.NewPage(ctx, c.Container)
	page.Metatags.Description = "The community of Orcas Island."

	v := views.CommunityResources{
		Page: page,
	}
	return c.RenderHTML(ctx, page, &v)
}

func (c *home) Editor(ctx echo.Context) error {
	page := controller.NewPage(ctx, c.Container)
	page.Metatags.Description = "The community of Orcas Island."

	becomeAnEditorContent, err := c.LoadMarkdown(ctx, "become-an-editor.md")
	if err != nil {
		return err
	}

	v := views.Editor{
		Page:           page,
		BecomeAnEditor: becomeAnEditorContent,
	}
	return c.RenderHTML(ctx, page, &v)
}
