package hooks

import (
	"database/sql"
	"encoding/json"
	"time"

	"github.com/labstack/echo/v4"
	"github.com/prometheus/client_golang/prometheus"
	"go.uber.org/zap"

	"git.grosinger.net/tgrosinger/tonyslist/constants"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/context"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/controller"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/metrics"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
)

/*

Test Success:

curl http://localhost:8000/hooks/deliverability \
  -X POST \
  -H "Content-Type: application/json" \
  -d '{
  "MessageID": "883953f4-6105-42a2-a16a-77a8eac79483",
  "Recipient": "tony@grosinger.net",
  "DeliveredAt": "2014-08-01T13:28:10.2735393-04:00",
  "Details": "Test delivery webhook details",
  "Tag": "welcome-email",
  "ServerID": 23,
  "Metadata": {
    "a_key": "a_value",
    "b_key": "b_value"
  },
  "RecordType": "Delivery",
  "MessageStream": "account-management"
}'

Test Failure:

curl http://localhost:8000/hooks/deliverability \
  -X POST \
  -H "Content-Type: application/json" \
  -d '{
  "ID": 4323372036854775807,
  "Type": "HardBounce",
  "TypeCode": 1,

  "Name": "Hard bounce",
  "Tag": "Test",
  "MessageID": "883953f4-6105-42a2-a16a-77a8eac79483",
  "MessageStream": "account-management",
  "ServerID": 23,
  "Description": "The server was unable to deliver your message (ex: unknown user, mailbox not found).",
  "Details": "Test bounce details",
  "Email": "tony@grosinger.net",
  "From": "sender@example.com",
  "BouncedAt": "2014-08-01T13:28:10.2735393-04:00",
  "DumpAvailable": true,
  "Inactive": true,
  "CanActivate": true,
  "RecordType": "Bounce",
  "Subject": "Test subject"
}'

*/

type deliverability struct {
	controller.Controller
}

func (c *deliverability) Post(ctx echo.Context) error {
	var msg Bounce
	err := json.NewDecoder(ctx.Request().Body).Decode(&msg)
	ctx.Request().Body.Close()
	if err != nil {
		return c.Fail(err, "Failed to decode incoming bounce report")
	}

	logger := context.GetLogger(ctx).With(
		zap.String("record_type", msg.RecordType),
		zap.String("type", msg.Type),
		zap.String("message_stream", msg.MessageStream),
		zap.String("email", msg.Email),
		zap.String("recipient", msg.Recipient))
	logger.Info("Received deliverability report")

	// Bounce types documentation:
	//     https://postmarkapp.com/developer/api/bounce-api#bounce-types

	switch msg.RecordType {
	case "SpamComplaint":
		metrics.SpamCount.With(prometheus.Labels{
			metrics.LabelMessageStream: msg.MessageStream,
		}).Inc()

		switch msg.MessageStream {
		case "account-management":
			err = c.Container.DB.C.SetAdminMarkedSpam(ctx.Request().Context(), msg.Email)
			if err != nil {
				logger.Warn("unable to set admin marked spam for user",
					zap.Error(err),
					zap.String("email", msg.Email))
			}
		case "twice-weekly-digest":
			err = c.Container.DB.C.SetDigestMarkedSpam(ctx.Request().Context(), msg.Email)
			if err != nil {
				logger.Warn("unable to set digests marked spam for user",
					zap.Error(err),
					zap.String("email", msg.Email))
			}

			// If we can find the user for this email, unsubscribe them from digest emails.
			user, err := c.Container.DB.C.GetUserByEmail(ctx.Request().Context(), msg.Email)
			if err == nil {
				err = c.Container.DB.C.SetUserSetting(ctx.Request().Context(),
					sqlc.SetUserSettingParams{
						UserID: user.ID,
						Key:    string(constants.EmailDigestSubscription),
						Value: sql.NullString{
							Valid:  true,
							String: constants.EmailDigestSubscriptionUnsubscribed,
						},
					})
				if err != nil {
					logger.Warn("unable to unsubscribe user from digest emails after marking spam",
						zap.Error(err),
						zap.String("email", msg.Email))
				}
			} else {
				logger.Warn("unable to find user account for email that marked digest as spam",
					zap.Error(err),
					zap.String("email", msg.Email))
			}
		case "user-messages":
			err = c.Container.DB.C.SetUserMessageMarkedSpam(ctx.Request().Context(), msg.Email)
			if err != nil {
				logger.Warn("unable to set user messages marked spam for user",
					zap.Error(err),
					zap.String("email", msg.Email))
			}
		default:
			logger.Info("user maked unknown message type as spam",
				zap.String("message-stream", msg.MessageStream),
				zap.String("email", msg.Email))
		}

	case "Bounce":
		metrics.BounceCount.With(prometheus.Labels{
			metrics.LabelMessageStream: msg.MessageStream,
			metrics.LabelEventType:     msg.Type,
		}).Inc()
		if msg.MessageStream == "account-management" {
			ok := c.Container.Cache.Set().
				Group(constants.CachedDeliverabilityGroup).
				Key(msg.Email).
				Data("bounce").
				Expiration(5 * time.Minute).
				Save()
			if !ok {
				logger.Error("unable to save deliverability report to cache")
			}
		}

		c.Container.DB.C.IncrementBouncedEmailCount(ctx.Request().Context(), msg.Email)
	case "Delivery":
		if msg.MessageStream == "account-management" {
			ok := c.Container.Cache.Set().
				Group(constants.CachedDeliverabilityGroup).
				Key(msg.Recipient).
				Data("delivered").
				Expiration(5 * time.Minute).
				Save()
			if !ok {
				logger.Error("unable to save deliverability report to cache")
			}
		}
	default:
		logger.Error("deliverability report has unexpected record type")
	}

	return nil
}
