package hooks

import (
	"database/sql"
	"encoding/json"
	"fmt"
	"html/template"
	"net/http"
	"strconv"
	"strings"

	"github.com/labstack/echo/v4"
	"github.com/prometheus/client_golang/prometheus"
	"go.uber.org/zap"

	"git.grosinger.net/tgrosinger/tonyslist/constants"
	"git.grosinger.net/tgrosinger/tonyslist/constants/pages"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/context"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/controller"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/conv"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/metrics"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/services/mail"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/services/mail/types"
)

type incomingEmail struct {
	controller.Controller
}

func (c *incomingEmail) Post(ctx echo.Context) error {
	var email IncomingEmail
	err := json.NewDecoder(ctx.Request().Body).Decode(&email)
	ctx.Request().Body.Close()
	if err != nil {
		return c.Fail(err, "Failed to decode incoming support email")
	}

	name := c.fromNameForEmail(email)
	if name != "" {
		err = c.Container.DB.C.SetUserNameIfEmpty(ctx.Request().Context(), sqlc.SetUserNameIfEmptyParams{
			Name:  conv.StringToNullString(name),
			Email: c.fromAddressForEmail(email),
		})
		if err != nil {
			logger := c.loggerForEmail(ctx, email)
			logger.Warn("unable to set user name from received email",
				zap.Error(err))
		}
	}

	toAddr := c.toAddressForEmail(email)

	if strings.HasPrefix(toAddr, constants.AnonymousEmailPrefix) {
		return c.handleClassifiedEmail(ctx, email)
	}

	// We check for classified posts first, so anything after this is an "other_incoming_email"
	err = c.Container.DB.C.IncrementUserEmailOtherCount(ctx.Request().Context(), c.fromAddressForEmail(email))
	if err != nil {
		c.Container.Logger.Warn("unable to increment user email other count",
			zap.Error(err))
	}

	if strings.HasPrefix(toAddr, constants.MailingListPrefix) {
		return c.handleMailingListEmail(ctx, email)
	}

	switch toAddr {
	case c.Container.Config.Mail.EventsAddress:
		return c.handleEventEmail(ctx, email)
	case c.Container.Config.Mail.SupportAddress:
		return c.forwardEmailToAdmin(ctx, email)
	default:
		return c.forwardEmailToAdmin(ctx, email)
	}
}

// forwardEmailToAdmin takes an email with a To address that we don't recognize
// and forwards it unmodified to the configured support email.
func (c *incomingEmail) forwardEmailToAdmin(ctx echo.Context, email IncomingEmail) error {
	logger := c.loggerForEmail(ctx, email)
	logger.Info("Received email for unknown destination")

	c.Container.Mail.Send(logger, []*types.Mail{
		{
			MessageStream: types.AdminNotifications,
			To:            c.Container.Config.Mail.AdminAddress,
			From:          fmt.Sprintf("%s (%s)<%s>", c.fromNameForEmail(email), c.Container.Config.App.Name, c.Container.Config.Mail.SupportAddress),
			ReplyTo:       c.fromAddressForEmail(email),
			Subject:       email.Subject,
			TextBody:      email.TextBody,
			HTMLBody:      email.HTMLBody,
		},
	}, func(r types.Response) {
		logger.Info("Forwarde email for unknown destination to support.",
			zap.Any("response", r))
	})
	return ctx.NoContent(http.StatusOK)
}

func (c *incomingEmail) handleEventEmail(ctx echo.Context, email IncomingEmail) error {
	logger := c.loggerForEmail(ctx, email)
	logger.Info("Received and ignoring event RSVP email")

	metrics.IncomingEmail.With(prometheus.Labels{
		metrics.LabelMessageStream: string(types.EventsTransactional),
		metrics.LabelID:            "-1", // TODO: Replace with actual event once we parse these messages
	}).Inc()

	// TODO: Track RSVP responses once recording who is attending events.

	return ctx.NoContent(http.StatusOK)
}

// handleClassifiedEmail looks up the To address against the mailinglists and
// classified anonymous addresses and forwards the email to the correct
// recipients.
func (c *incomingEmail) handleClassifiedEmail(ctx echo.Context, email IncomingEmail) error {
	logger := c.loggerForEmail(ctx, email)

	to := c.toAddressForEmail(email)
	localPart := strings.Split(to, "@")[0]

	// First check if it's a classified post.
	row, err := c.Container.DB.C.GetPostForAnonymousEmail(ctx.Request().Context(),
		sqlc.GetPostForAnonymousEmailParams{
			AnonymousEmail: conv.StringToNullString(localPart),
			Datetime:       fmt.Sprintf("-%d days", constants.ClassifiedsDurationDays),
		})
	if err == sql.ErrNoRows {
		// There is not an active classifieds post for this anonymous email address.
		return c.sendUnknownClassifiedEmail(ctx, logger, email)
	} else if err != nil {
		logger.Error("Failed to check if incoming email is for active post", zap.Error(err))
		return c.Fail(err, "Failed to check if incoming email is for active post")
	}

	logger.Info("Found an active classified post for anonymous email address")

	err = c.Container.DB.C.IncrementEmailResponseCount(ctx.Request().Context(), row.ID)
	if err != nil {
		c.Container.Logger.Warn("unable to increment email response count",
			zap.Error(err))
	}

	err = c.Container.DB.C.IncrementUserEmailResponseCount(ctx.Request().Context(), c.fromAddressForEmail(email))
	if err != nil {
		c.Container.Logger.Warn("unable to increment user email response count",
			zap.Error(err))
	}

	return c.forwardClassifiedEmail(ctx, logger, email, row)
}

func (c *incomingEmail) sendUnknownClassifiedEmail(ctx echo.Context, logger *zap.Logger, email IncomingEmail) error {
	logger.Info("Received email for unknown classified post")

	metrics.IncomingEmailUnhandled.With(prometheus.Labels{
		metrics.LabelError: "unknown-classified-post",
	}).Inc()

	quotedMessage := c.quoteMessageTextBody(email)

	templateData := map[string]interface{}{
		"date":           email.Date,
		"sender_address": email.From,

		// quotedMessage starts with `>` characters which we do not want to be escaped.
		"quoted_original_message": template.HTML(quotedMessage),
	}

	m, err := c.Container.Mail.Template(mail.TemplateArgs{
		MessageStream: types.AccountManagement,
		To:            c.fromAddressForEmail(email),
		Subject:       "Re: " + email.Subject,
		TextTemplate:  "unknown_classified_anonymous_text",
		TemplateData:  templateData,
	})
	if err != nil {
		return c.Fail(err, "Failed to respond to unknown classified post anonymous email")
	}

	c.Container.Mail.Send(logger, []*types.Mail{m}, nil)
	return ctx.NoContent(http.StatusOK)
}

func (c *incomingEmail) forwardClassifiedEmail(ctx echo.Context, logger *zap.Logger, email IncomingEmail, p sqlc.GetPostForAnonymousEmailRow) error {
	metrics.IncomingEmail.With(prometheus.Labels{
		metrics.LabelMessageStream: string(types.UserMessages),
		metrics.LabelID:            strconv.Itoa(p.ID),
	}).Inc()

	textTemplate := ""
	if email.TextBody != "" {
		textTemplate = "forward_classified_anonymous_text"
	}

	originalHTMLMessage := template.HTML(email.HTMLBody)
	if originalHTMLMessage == "" {
		originalHTMLMessage = template.HTML(email.TextBody)
	}

	var viewHref string
	switch p.PostType {
	case constants.Classified:
		viewHref = c.Container.Web.Reverse(pages.Names.Classified.One, p.ID)
	case constants.Housing:
		viewHref = c.Container.Web.Reverse(pages.Names.Housing.One, p.ID)
	}

	logger.Info("Forwarding classified anonymous email",
		zap.String("recipient", p.Email),
		zap.Int("post_id", p.ID))

	templateData := map[string]interface{}{
		"domain":                c.Container.Config.App.PublicURL,
		"view_href":             viewHref,
		"support_email":         c.Container.Config.Mail.SupportAddress,
		"original_message_text": email.TextBody,
		"original_message_html": originalHTMLMessage,
		"post_title":            p.Title,
	}

	m, err := c.Container.Mail.Template(mail.TemplateArgs{
		ReplyTo:       email.From,
		To:            p.Email,
		Subject:       fmt.Sprintf("[%s] %s", c.Container.Config.App.Name, email.Subject),
		HTMLTemplate:  "forward_classified_anonymous_html",
		TextTemplate:  textTemplate,
		MessageStream: types.UserMessages,
		TemplateData:  templateData,
	})
	if err != nil {
		logger.Error("Failed to forward classified anonymous email",
			zap.Error(err))
		return c.Fail(err, "Failed to forward classified anonymous email")
	}

	c.Container.Mail.Send(logger, []*types.Mail{m}, nil)
	return ctx.NoContent(http.StatusOK)
}

func (c *incomingEmail) handleMailingListEmail(ctx echo.Context, email IncomingEmail) error {
	logger := c.loggerForEmail(ctx, email)

	to := c.toAddressForEmail(email)
	localPart := strings.Split(to, "@")[0]
	shortName := strings.TrimPrefix(localPart, constants.MailingListPrefix)
	logger = logger.With(zap.String("short_name", shortName))

	if shortName == "wsntn" {
		return c.sendWrongWSNTNAddress(ctx, logger, email)
	}
	return c.forwardEmailToAdmin(ctx, email)
}

func (c *incomingEmail) sendWrongWSNTNAddress(ctx echo.Context, logger *zap.Logger, email IncomingEmail) error {
	logger.Info("Sending notice of incorrect wsntn email address")

	quotedMessage := c.quoteMessageTextBody(email)

	templateData := map[string]interface{}{
		"date":           email.Date,
		"sender_address": email.From,

		// quotedMessage starts with `>` characters which we do not want to be escaped.
		"quoted_original_message": template.HTML(quotedMessage),
	}

	m, err := c.Container.Mail.Template(mail.TemplateArgs{
		MessageStream: types.AccountManagement,
		To:            c.fromAddressForEmail(email),
		Subject:       "Re: " + email.Subject,
		TextTemplate:  "wrong_wsntn_address_text",
		TemplateData:  templateData,
	})
	if err != nil {
		return c.Fail(err, "Failed to respond to email to wrong wsntn email address")
	}

	c.Container.Mail.Send(logger, []*types.Mail{m}, nil)
	return ctx.NoContent(http.StatusOK)
}

func (c *incomingEmail) loggerForEmail(ctx echo.Context, email IncomingEmail) *zap.Logger {
	return context.GetLogger(ctx).With(
		zap.String("from", email.From),
		zap.String("to", c.toAddressForEmail(email)),
		zap.String("subject", email.Subject))
}

func (c *incomingEmail) toAddressForEmail(email IncomingEmail) string {
	for _, to := range email.ToFull {
		if to.Email != "" {
			return strings.ToLower(to.Email)
		}
	}

	return strings.ToLower(email.To)
}

func (c *incomingEmail) fromNameForEmail(email IncomingEmail) string {
	if email.FromFull.Name != "" {
		return email.FromFull.Name
	}
	return email.FromName
}

func (c *incomingEmail) fromAddressForEmail(email IncomingEmail) string {
	if email.FromFull.Email != "" {
		return email.FromFull.Email
	}
	return email.From
}

func (c *incomingEmail) quoteMessageTextBody(email IncomingEmail) string {
	return "> " + strings.ReplaceAll(email.TextBody, "\n", "\n> ")
}
