package housing

import (
	"errors"
	"fmt"

	"github.com/labstack/echo/v4"
	"go.uber.org/zap"

	"git.grosinger.net/tgrosinger/tonyslist/constants"
	"git.grosinger.net/tgrosinger/tonyslist/constants/pages"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/context"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/controller"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/conv"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/housing"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/tasks"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/views"
)

type create struct {
	controller.Controller
}

// Get displays a page for users to create a new post in the housing section.
// This handler is protected by middleware so that only logged-in users can
// access it.
// NOTE: We use the term "post" because it could be a for_rent or a request.
func (c *create) Get(ctx echo.Context) error {
	page := controller.NewPage(ctx, c.Container)
	page.Title = "New Post"
	page.Metatags.Description = fmt.Sprintf("Create new housing post on %s.", c.Container.Config.App.Name)
	page.Metatags.Keywords = keywords

	housingType := string(constants.ForRent)
	if ctx.QueryParam("src_housing_type") == string(constants.Request) {
		housingType = string(constants.Request)
	}

	form := views.HousingCreateForm{
		IncludeEmail: true,
		HousingType:  housingType,
	}
	if f := ctx.Get(context.FormKey); f != nil {
		form = *f.(*views.HousingCreateForm)
	}

	v := views.HousingCreate{
		IsCreate:   true,
		ActionHref: ctx.Echo().Reverse(pages.Names.Housing.New),
		CancelHref: ctx.Echo().Reverse(pages.Names.Housing.Browse),
		// TODO: Submit text needs to change to "Publish" when housing_type==='request'
		SubmitText: "Next",
		Form:       form,
		Page:       page,
	}
	return c.RenderHTML(ctx, page, &v)
}

// Post processes the submitted form to create a new post in the housing
// section.  This handler is protected by middleware so that only logged-in
// users can access it.
func (c *create) Post(ctx echo.Context) error {
	var form views.HousingCreateForm
	ctx.Set(context.FormKey, &form)

	if err := ctx.Bind(&form); err != nil {
		return c.Fail(err, "unable to bind form")
	}

	if err := form.Submission.Process(ctx, form); err != nil {
		return c.Fail(err, "unable to process form submission")
	}

	form.ProcessCreate(context.GetLogger(ctx))

	// Ignoring ok value because the process function will also set errors on the form.
	editableFields, _ := form.ProcessEditableFields()

	if form.Submission.HasErrors() {
		return c.Get(ctx)
	}

	user, ok := ctx.Get(context.AuthenticatedUserKey).(sqlc.User)
	if !ok {
		return c.Fail(errors.New("failed to lookup authenticated user"), "unable to lookup authenticated user")
	}

	form.Sanitize(c.Container.Sanitizer)

	var err error
	switch form.HousingType {
	case string(constants.ForRent):
		err = c.storeNewForRent(ctx, user, form, editableFields)
	case string(constants.Request):
		err = c.storeNewRequest(ctx, user, form, editableFields)
	default:
		// Shouldn't get here because form validation will return a field error.
		return c.BadReq(fmt.Sprintf("invalid housing_type %q in request", form.HousingType))
	}

	// TODO: Is this being used?
	if _, ok := err.(controller.RecoverableError); ok {
		return c.Get(ctx)
	}

	return err
}

func (c *create) storeNewForRent(ctx echo.Context, user sqlc.User, form views.HousingCreateForm, editableFields housing.EditableFields) error {
	i := housing.NewPostInput{
		CreatePostParams: sqlc.CreatePostParams{
			PostType:       constants.Housing,
			Title:          form.Title,
			Description:    form.Description,
			Creator:        user.ID,
			AnonymousEmail: c.Container.DB.PostBase.CreateAnonymousEmail(form.IncludeEmail),
			PhoneCall:      conv.StringToNullString(form.PhoneNumber),
			PhoneText:      conv.StringToNullString(form.SMSNumber),
		},
		EditableFields: editableFields,
		HousingType:    constants.ForRent,
	}

	id, err := c.Container.DB.Housing.Create(ctx.Request().Context(), i)
	if err != nil {
		return c.Fail(err, "unable to save a new post")
	}

	c.Container.WorkerPool.Submit(func() {
		logger := context.GetLogger(ctx)
		tasks.DownloadHousingMap(c.Container.Config, logger, c.Container.DB, form.Street, form.Zipcode, id)
	})

	err = c.Container.DB.C.IncrementStat(ctx.Request().Context(), constants.ForRentCreated)
	if err != nil {
		context.GetLogger(ctx).Error("unable to increment stat for create post",
			zap.Error(err))
	}

	c.Container.DB.C.IncrementPostsCreatedCount(ctx.Request().Context(), user.ID)
	if err != nil {
		context.GetLogger(ctx).Warn("unable to increment posts created stat",
			zap.Error(err))
	}

	return c.Redirect(ctx, pages.Names.Posts.NewPhotos, id)
}

func (c *create) storeNewRequest(ctx echo.Context, user sqlc.User, form views.HousingCreateForm, editableFields housing.EditableFields) error {
	i := housing.NewPostInput{
		CreatePostParams: sqlc.CreatePostParams{
			PostType:       constants.Housing,
			Title:          form.Title,
			Description:    form.Description,
			Creator:        user.ID,
			AnonymousEmail: c.Container.DB.PostBase.CreateAnonymousEmail(form.IncludeEmail),
			PhoneCall:      conv.StringToNullString(form.PhoneNumber),
			PhoneText:      conv.StringToNullString(form.SMSNumber),
		},
		EditableFields: editableFields,
		HousingType:    constants.Request,
	}

	id, err := c.Container.DB.Housing.Create(ctx.Request().Context(), i)
	if err != nil {
		return c.Fail(err, "unable to save a new post")
	}

	err = c.Container.DB.C.SetPostPublishedNow(ctx.Request().Context(), id)
	if err != nil {
		return c.Fail(err, "unable to publish the post")
	}

	err = c.Container.DB.C.IncrementStat(ctx.Request().Context(), constants.RequestCreated)
	if err != nil {
		context.GetLogger(ctx).Warn("unable to increment stat for create post",
			zap.Error(err))
	}

	c.Container.DB.C.IncrementPostsCreatedCount(ctx.Request().Context(), user.ID)
	if err != nil {
		context.GetLogger(ctx).Warn("unable to increment posts created stat",
			zap.Error(err))
	}

	return c.Redirect(ctx, pages.Names.Housing.One, id)
}
