package housing

import (
	"fmt"
	"strconv"

	"github.com/labstack/echo/v4"

	"git.grosinger.net/tgrosinger/tonyslist/constants/pages"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/context"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/controller"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/housing"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/msg"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/views"
)

type edit struct {
	controller.Controller
}

// Get displays a page for users to edit an existing post in the housing
// section. This handler is protected by middleware so that only the post owner
// can access it.
// NOTE: We use the term "post" because it could be a for_rent or a request.
func (c *edit) Get(ctx echo.Context) error {
	p, ok := ctx.Get(context.PostKey).(*housing.Post)
	if !ok {
		msg.Warning(ctx, "The requested post could not be found.")
		return c.Redirect(ctx, pages.Names.Housing.Browse)
	}

	page := controller.NewPage(ctx, c.Container)
	page.Title = "Edit Post"
	page.Metatags.Description = fmt.Sprintf("Edit a housing post on %s.", c.Container.Config.App.Name)
	page.Metatags.Keywords = keywords

	var form views.HousingCreateForm
	if f := ctx.Get(context.FormKey); f != nil {
		form = *f.(*views.HousingCreateForm)
	} else {
		form = loadHousingUpdateForm(p)
	}

	v := views.HousingCreate{
		IsCreate:   false,
		ActionHref: ctx.Echo().Reverse(pages.Names.Housing.Edit, p.ID),
		CancelHref: ctx.Echo().Reverse(pages.Names.Housing.One, p.ID),
		SubmitText: "Update",
		Form:       form,
		Page:       page,
	}
	return c.RenderHTML(ctx, page, &v)
}

// Post processes the submitted form to edit a post in the housing section. This
// handler is protected by middleware so that only the post owner can access it.
func (c *edit) Post(ctx echo.Context) error {
	postID, err := strconv.Atoi(ctx.Param("id"))
	if err != nil {
		msg.Warning(ctx, "The requested post could not be found.")
		return c.Redirect(ctx, pages.Names.Housing.Browse)
	}

	var form views.HousingCreateForm
	ctx.Set(context.FormKey, &form)

	if err := ctx.Bind(&form); err != nil {
		return c.Fail(err, "unable to bind form")
	}

	// Only validate the nested HousingUpdateForRentForm. Validating the entire
	// form will result in a failure because the outer form does not currently
	// have all the fields filled out.
	if err := form.Submission.Process(ctx, form.HousingUpdateForm); err != nil {
		return c.Fail(err, "unable to process form submission")
	}

	// Ignoring ok value because the process function will also set errors on the form.
	editableFields, _ := form.ProcessEditableFields()

	if form.Submission.HasErrors() {
		return c.Get(ctx)
	}

	form.Sanitize(c.Container.Sanitizer)

	i := housing.UpdatePostInput{
		UpdatePostParams: sqlc.UpdatePostParams{
			ID:          postID,
			Title:       form.Title,
			Description: form.Description,
		},
		EditableFields: editableFields,
	}

	err = c.Container.DB.Housing.Update(ctx.Request().Context(), postID, i)
	if err != nil {
		return c.Fail(err, "unable to save a new post")
	}

	msg.Success(ctx, "Post updated successfully")
	return c.Redirect(ctx, pages.Names.Housing.One, postID)
}

func loadHousingUpdateForm(post *housing.Post) views.HousingCreateForm {
	form := views.HousingCreateForm{
		HousingUpdateForm: views.HousingUpdateForm{
			Title:       post.Title,
			Description: post.Description,
			Pets:        post.Pets,
			Shared:      post.Shared,
		},
		HousingType: string(post.HousingType),
	}

	if post.Bedrooms.Valid {
		form.Bedrooms = strconv.Itoa(post.Bedrooms.Int)
	}
	if post.Bathrooms.Valid {
		form.Bathrooms = strconv.FormatFloat(post.Bathrooms.Float64, 'f', 2, 64)
	}
	if post.ParkingSpots.Valid {
		form.Parking = strconv.Itoa(post.ParkingSpots.Int)
	}
	if post.Rent.Valid {
		form.Rent = strconv.FormatFloat(post.Rent.Float64, 'f', 2, 64)
	}

	return form
}
