package housing

import (
	"fmt"
	"net/url"
	"strings"

	"github.com/labstack/echo/v4"
	"go.uber.org/zap"

	"git.grosinger.net/tgrosinger/tonyslist/constants"
	"git.grosinger.net/tgrosinger/tonyslist/constants/pages"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/context"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/controller"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/housing"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/postbase"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/msg"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/services"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/views"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/views/components"
)

type housingCtr struct {
	controller.Controller
}

func (c *housingCtr) RedirectBrowse(ctx echo.Context) error {
	return c.Redirect(ctx, pages.Names.Housing.Browse)
}

func (c *housingCtr) Get(ctx echo.Context) error {
	htmxReq := ctx.Request().Header.Get("HX-Request")
	pager := controller.NewPager(ctx, controller.DefaultItemsPerPage)

	q := housingQueryFromContext(c.Container, ctx, &pager)
	data, count, err := c.Container.DB.Housing.Select(ctx.Request().Context(), q)
	if err != nil {
		c.Fail(err, "unable to retrieve requested posts")
	}
	pager.SetItems(count)

	var myHousing []postbase.PostBase
	if user, ok := ctx.Get(context.AuthenticatedUserKey).(sqlc.User); ok {
		myHousing, err = c.Container.DB.PostBase.GetMyPosts(ctx.Request().Context(),
			sqlc.GetMyPostsParams{
				Creator:  user.ID,
				PostType: constants.Housing,
			})
		if err != nil {
			if he, ok := err.(*echo.HTTPError); ok {
				return he
			}
			return c.Fail(err, "unable to retrieve requested post")
		}
	}

	page := controller.NewPage(ctx, c.Container)
	page.Title = "Housing"
	page.Metatags.Description = "Housing on Orcas."
	page.Metatags.Keywords = keywords
	page.Metatags.HasFeed = true

	if strings.HasSuffix(ctx.Path(), ".rss") {
		items := make([]views.RSSItem, 0, len(data))
		for _, p := range data {
			items = append(items, views.RSSItem{
				Title:         p.Title,
				Description:   p.Description,
				Href:          c.Container.Config.App.PublicURL + p.GetViewHref(ctx.Echo()),
				PublishedDate: p.CreatedAt,
			})
		}

		v := views.RSS{
			HomeHref: c.Container.Config.App.PublicURL +
				ctx.Echo().Reverse(pages.Names.Housing.Browse),
			SupportEmail: c.Container.Config.Mail.SupportAddress,
			Items:        items,
			Page:         page,
			Sanitizer:    c.Container.Sanitizer,
		}
		return c.RenderXML(ctx, page, &v)
	}

	digest := ctx.Get(context.DigestMembershipKey)
	page.DisplayEmailDigestCTA = digest != constants.EmailDigestSubscriptionTwiceWeekly

	filters := []constants.Filter{
		{
			SectionDisplay: "Filters",
			SectionName:    "basic",
			Options: []constants.FilterOption{
				{
					Value:   "for-rent",
					Display: "For Rent",
				},
				{
					Value:   "request",
					Display: "Requests",
				},
				{
					Value:   "photos",
					Display: "Has Photos",
				},
				{
					Value:   "title-only",
					Display: "Search Title Only",
				},
			},
		},
	}

	// NOTE: The housing browse page does not display photos so we are not loading feature photos.

	v := views.PostBrowse{
		Posts:              data,
		MyPosts:            myHousing,
		DisplayType:        views.ListDisplayType,
		SortOptions:        constants.DefaultSortOptions,
		CategorySelections: categorySelection,
		Filters:            filters,
		NewPostLink:        ctx.Echo().Reverse(pages.Names.Housing.New),
		PostsListOnly:      htmxReq == "true",

		Headers: components.HousingHeader(page),
		Page:    page,
	}
	return c.RenderHTML(ctx, page, &v)
}

func (c *housingCtr) GetOne(ctx echo.Context) error {
	// Middleware protects against posts not existing and being expired.
	p, ok := ctx.Get(context.PostKey).(*housing.Post)
	if !ok {
		msg.Warning(ctx, "The requested post could not be found.")
		return c.RedirectToPostBrowser(ctx)
	}

	photos, err := c.Container.DB.C.GetPhotos(ctx.Request().Context(), p.ID)
	if err != nil {
		c.Container.Logger.Warn("unable to load photos for post",
			zap.Error(err))
	}

	page := controller.NewPage(ctx, c.Container)
	page.Title = "For Rent"
	page.Metatags.Description = fmt.Sprintf("Housing for rent on %s.", c.Container.Config.App.Name)
	page.Metatags.Keywords = keywords

	// If the viewer is a moderator, retrieve the post creator user info to display in the moderator toolbar.
	var info sqlc.GetUserInfoRow
	if page.Enforce("post", "moderate") {
		var err error
		info, err = c.Container.DB.C.GetUserInfo(ctx.Request().Context(), p.Creator)
		if err != nil {
			msg.Warning(ctx, "Unable to retrieve post creator info")
			context.GetLogger(ctx).Warn("Unable to retrieve post creator info", zap.Error(err))
		}
	}

	if !page.IsAuth || page.AuthUser.ID != p.Creator {
		err := c.Container.DB.C.IncrementNonOwnerViews(ctx.Request().Context(), p.ID)
		if err != nil {
			c.Container.Logger.Warn("unable to increment post view count",
				zap.Error(err))
		}
	}

	v := views.PostDetails{
		PostBase:         *p.PostBase,
		Price:            p.Rent,
		Bedrooms:         p.Bedrooms,
		Bathrooms:        p.Bathrooms,
		ParkingSpots:     p.ParkingSpots,
		Pets:             p.Pets,
		Shared:           p.Shared,
		Photos:           photos,
		CreatorInfo:      &info,
		MarkdownRenderer: c.Container.MarkdownRenderer,
		Headers:          components.HousingHeader(page),
		Page:             page,
	}
	return c.RenderHTML(ctx, page, &v)
}

// categorySelection is a placeholder for the quick filter options on the
// sidebar of the post-browser template.
var categorySelection = []string{}

type postQueryArgs struct {
	search  string
	sort    string
	filters string
}

func housingQueryFromContext(c *services.Container, ctx echo.Context, pager *controller.Pager) housing.SelectInput {
	var args postQueryArgs
	htmxReq := ctx.Request().Header.Get("HX-Request")
	if htmxReq == "true" {
		args = postQueryArgs{
			search: ctx.QueryParam("search"),
		}

		currentURL := ctx.Request().Header.Get("HX-Current-URL")
		parsedURL, err := url.Parse(currentURL)
		if err == nil {
			args.sort = parsedURL.Query().Get("sort")
			args.filters = parsedURL.Query().Get("filters")
		}
	} else {
		args = postQueryArgs{
			search:  ctx.QueryParam("search"),
			sort:    ctx.QueryParam("sort"),
			filters: ctx.QueryParam("filters"),
		}
	}

	q := housing.SelectInput{
		SelectInput: postbase.SelectInput{
			Pagination: postbase.Pagination{
				Limit:  uint64(pager.ItemsPerPage),
				Offset: uint64(pager.GetOffset()),
			},
			IncludeUnlisted: false,
		},
	}

	// TODO: Add another filter for just for-rent or request posts.

	switch args.sort {
	case "high_to_low":
		q.SortColumn = housing.Rent
		q.SortDir = postbase.SortDesc
	case "low_to_high":
		q.SortColumn = housing.Rent
		q.SortDir = postbase.SortAsc
	default: // "newest"
		q.SortColumn = postbase.PublishedAt
		q.SortDir = postbase.SortDesc
	}

	if strings.HasSuffix(ctx.Path(), ".rss") {
		// Override the sort and limit parameters for rss feeds
		q.SortColumn = postbase.PublishedAt
		q.SortDir = postbase.SortDesc
		q.Offset = 0
		q.Limit = 30
	}

	q.Fields = []string{postbase.Title, postbase.Description}
	if filters := args.filters; filters != "" {
		filters = strings.TrimSpace(filters)
		for _, filter := range strings.Split(filters, ",") {
			switch filter {
			case "basic_photos":
				q.RequirePhotos = true
			case "basic_title-only":
				q.Fields = []string{postbase.Title}
			case "basic_for-rent":
				q.IncludeForRent = true
			case "basic_request":
				q.IncludeRequests = true
			}
		}
	}

	if !q.IncludeForRent && !q.IncludeRequests {
		q.IncludeForRent = true
		q.IncludeRequests = true
	}

	if query := args.search; query != "" {
		q.Query = query
		parsed, err := c.Parser.ParseSearchQuery(query)
		if err != nil {
			context.GetLogger(ctx).Error("Unable to parse query",
				zap.Error(err),
				zap.String("query", query))
		} else {
			q.ParsedQuery = parsed
		}
	}

	return q
}
