package routes

import (
	"context"
	"database/sql"
	"mime/multipart"
	"net/http"
	"net/url"
	"testing"
	"time"

	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"

	"git.grosinger.net/tgrosinger/tonyslist/constants"
	"git.grosinger.net/tgrosinger/tonyslist/constants/pages"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/housing"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
)

func TestHousingForRent_GetExpired(t *testing.T) {
	// Ensure that the two users have been instantiated
	_ = request(t).withUnverifiedUser()
	_ = request(t).withVerifiedUser()

	postID, err := c.DB.Housing.Create(context.Background(), housing.NewPostInput{
		CreatePostParams: sqlc.CreatePostParams{
			PostType:    constants.Housing,
			Title:       "Test post that is expired",
			Description: "Test description",
			PhoneCall:   sql.NullString{Valid: false},
			PhoneText:   sql.NullString{Valid: false},
			Creator:     verifiedUser.id,
		},
		EditableFields: housing.EditableFields{
			Rent: sql.NullFloat64{Valid: true, Float64: 12.23},
		},
		HousingType: constants.ForRent,
	})
	require.NoError(t, err)

	// Manually set the post to have been published 1 month ago
	oldTime := time.Now().Add(-1 * 24 * 30 * time.Hour)
	err = c.DB.C.SetPostPublishedAt(context.Background(), sqlc.SetPostPublishedAtParams{
		ID:          postID,
		PublishedAt: sql.NullTime{Valid: true, Time: oldTime},
	})
	require.NoError(t, err)

	// Verify that non-creator cannot retrieve the post
	request(t).
		withUnverifiedUser().
		setRoute(pages.Names.Housing.One, postID).
		get().
		requireStatusCode(http.StatusNotFound)

	// Verify that the owner can retrieve the post even though it's expired.
	request(t).
		withVerifiedUser().
		setRoute(pages.Names.Housing.One, postID).
		get().
		requireStatusCode(http.StatusOK)
}

func TestHousingForRent_SignedOutNewPostGet(t *testing.T) {
	expectedPath := c.Web.Reverse(pages.Names.User.Login) + "?return_to=" + url.QueryEscape(c.Web.Reverse(pages.Names.Housing.New))
	request(t).
		setRoute(pages.Names.Housing.New).
		disableFollowRedirects().
		get().
		requireStatusCode(http.StatusFound).
		assertRedirectPath(expectedPath)
}

func TestHousingForRent_SignedOutNewPostPost(t *testing.T) {
	expectedPath := c.Web.Reverse(pages.Names.User.Login) + "?return_to=" + url.QueryEscape(c.Web.Reverse(pages.Names.Housing.New))
	request(t).
		setRoute(pages.Names.Housing.New).
		disableFollowRedirects().
		postMultipartForm(func(w *multipart.Writer) {
			setMultipartField(t, w, "housing_type", "for-rent")
			setMultipartField(t, w, "title", "Created from TestHousing")
			setMultipartField(t, w, "description", "The Description")
		}).
		requireStatusCode(http.StatusFound).
		assertRedirectPath(expectedPath)
}

func TestHousingForRent_InvalidPostType(t *testing.T) {
	fieldErrorEl := request(t).
		withVerifiedUser().
		setRoute(pages.Names.Housing.New).
		disableFollowRedirects().
		postMultipartForm(func(w *multipart.Writer) {
			setMultipartField(t, w, "housing_type", "invalid")
			setMultipartField(t, w, "title", "Created from TestHousing")
			setMultipartField(t, w, "description", "The Description")
			setMultipartField(t, w, "include_email", "true")
		}).
		// A field error will be set, so we get an OK instead of a bad req or a redirect
		requireStatusCode(http.StatusOK).
		toDoc().
		Find(`.text-red-600`).
		First()

	html, err := fieldErrorEl.Html()
	require.NoError(t, err)
	assert.Equal(t, "Invalid housing type in request", html)
}

func TestHousingForRent_CreateSparseNoPhotos(t *testing.T) {
	postID := request(t).
		withVerifiedUser().
		setRoute(pages.Names.Housing.New).
		disableFollowRedirects().
		postMultipartForm(func(w *multipart.Writer) {
			setMultipartField(t, w, "housing_type", "for-rent")
			setMultipartField(t, w, "title", "Created from TestHousing")
			setMultipartField(t, w, "description", "The Description")
			setMultipartField(t, w, "include_email", "true")
		}).
		requireStatusCode(http.StatusFound).
		getLocationID()

	postBase, err := c.DB.PostBase.GetOne(context.Background(), postID, false)
	require.NoError(t, err)
	post, err := c.DB.Housing.GetOneFromPostBase(context.Background(), postBase)
	require.NoError(t, err)

	assert.Equal(t, postID, post.ID)
	assert.Equal(t, "Created from TestHousing", post.Title)
	assert.Equal(t, "The Description", post.Description)
	assert.False(t, post.ParkingSpots.Valid)
	assert.False(t, post.Bedrooms.Valid)
	assert.False(t, post.Bathrooms.Valid)
	assert.False(t, post.PhoneCall.Valid)
	assert.False(t, post.PhoneText.Valid)
	assert.True(t, post.Listed)
	assert.False(t, post.Shared)
	assert.False(t, post.Pets)
	assert.Len(t, post.AnonymousEmail.String, constants.AnonymousEmailLen+len(constants.AnonymousEmailPrefix))
	assert.True(t, post.AnonymousEmail.Valid)
}

func TestHousingForRent_CreateSparseNoEmail(t *testing.T) {
	postID := request(t).
		withVerifiedUser().
		setRoute(pages.Names.Housing.New).
		disableFollowRedirects().
		postMultipartForm(func(w *multipart.Writer) {
			setMultipartField(t, w, "housing_type", "for-rent")
			setMultipartField(t, w, "title", "Created from TestHousing")
			setMultipartField(t, w, "description", "The Description")
			setMultipartField(t, w, "include_email", "false")
			setMultipartField(t, w, "sms_number", "222-222-2222")
		}).
		requireStatusCode(http.StatusFound).
		getLocationID()

	postBase, err := c.DB.PostBase.GetOne(context.Background(), postID, false)
	require.NoError(t, err)
	post, err := c.DB.Housing.GetOneFromPostBase(context.Background(), postBase)
	require.NoError(t, err)

	assert.Equal(t, postID, post.ID)
	assert.Equal(t, "Created from TestHousing", post.Title)
	assert.Equal(t, "The Description", post.Description)
	assert.False(t, post.ParkingSpots.Valid)
	assert.False(t, post.Bedrooms.Valid)
	assert.False(t, post.Bathrooms.Valid)
	assert.False(t, post.PhoneCall.Valid)
	assert.Equal(t, "222-222-2222", post.PhoneText.String)
	assert.True(t, post.Listed)
	assert.False(t, post.Shared)
	assert.False(t, post.Pets)
	assert.False(t, post.AnonymousEmail.Valid)
}

func TestHousingForRent_Create(t *testing.T) {
	postID := request(t).
		withVerifiedUser().
		setRoute(pages.Names.Housing.New).
		disableFollowRedirects().
		postMultipartForm(func(w *multipart.Writer) {
			setMultipartField(t, w, "housing_type", "for-rent")
			setMultipartField(t, w, "title", "Created from TestHousing")
			setMultipartField(t, w, "description", "The Description")
			setMultipartField(t, w, "parking", "1")
			setMultipartField(t, w, "bathrooms", "1.25")
			setMultipartField(t, w, "phone_number", "111-111-1111")
			setMultipartField(t, w, "sms_number", "222-222-2222")
			setMultipartField(t, w, "pets_allowed", "true")
			setMultipartField(t, w, "shared", "true")
		}).
		requireStatusCode(http.StatusFound).
		getLocationID()

	postBase, err := c.DB.PostBase.GetOne(context.Background(), postID, false)
	require.NoError(t, err)
	post, err := c.DB.Housing.GetOneFromPostBase(context.Background(), postBase)
	require.NoError(t, err)

	assert.Equal(t, postID, post.ID)
	assert.Equal(t, "Created from TestHousing", post.Title)
	assert.Equal(t, "The Description", post.Description)
	assert.Equal(t, 1, post.ParkingSpots.Int)
	assert.Equal(t, 1.25, post.Bathrooms.Float64)
	assert.Equal(t, "111-111-1111", post.PhoneCall.String)
	assert.Equal(t, "222-222-2222", post.PhoneText.String)
	assert.False(t, post.Bedrooms.Valid)
	assert.True(t, post.Listed)
	assert.True(t, post.Shared)
	assert.True(t, post.Pets)
	assert.False(t, post.AnonymousEmail.Valid)
}
