package routes

import (
	"context"
	"database/sql"
	"net/http"
	"testing"
	"time"

	"github.com/stretchr/testify/require"

	"git.grosinger.net/tgrosinger/tonyslist/constants"
	"git.grosinger.net/tgrosinger/tonyslist/constants/pages"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/housing"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
)

func TestHousingRequest_GetExpired(t *testing.T) {
	// Ensure that the two users have been instantiated
	_ = request(t).withUnverifiedUser()
	_ = request(t).withVerifiedUser()

	postID, err := c.DB.Housing.Create(context.Background(), housing.NewPostInput{
		CreatePostParams: sqlc.CreatePostParams{
			PostType:    constants.Housing,
			Title:       "Test post that is expired",
			Description: "Test description",
			PhoneCall:   sql.NullString{Valid: false},
			PhoneText:   sql.NullString{Valid: false},
			Creator:     verifiedUser.id,
		},
		HousingType: constants.Request,
	})
	require.NoError(t, err)

	// Manually set the post to have been published 1 month ago
	oldTime := time.Now().Add(-1 * 24 * 30 * time.Hour)
	err = c.DB.C.SetPostPublishedAt(context.Background(), sqlc.SetPostPublishedAtParams{
		ID:          postID,
		PublishedAt: sql.NullTime{Valid: true, Time: oldTime},
	})
	require.NoError(t, err)

	// Verify that non-creator cannot retrieve the post
	request(t).
		withUnverifiedUser().
		setRoute(pages.Names.Housing.One, postID).
		get().
		requireStatusCode(http.StatusNotFound)

	// Verify that the owner can retrieve the post even though it's expired.
	request(t).
		withVerifiedUser().
		setRoute(pages.Names.Housing.One, postID).
		get().
		requireStatusCode(http.StatusOK)
}
