package routes

import (
	"bytes"
	"context"
	"database/sql"
	"encoding/json"
	"fmt"
	"net/http"
	"testing"
	"time"

	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/mock"
	"github.com/stretchr/testify/require"

	"git.grosinger.net/tgrosinger/tonyslist/constants"
	"git.grosinger.net/tgrosinger/tonyslist/constants/pages"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/classified"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/routes/hooks"
	mailmock "git.grosinger.net/tgrosinger/tonyslist/pkg/services/mail/mock"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/services/mail/types"
)

func TestIncomingUnknownEmail(t *testing.T) {
	providerMock := mailmock.New()
	c.Mail.SetProvider(providerMock)
	providerMock.On("Send", mock.Anything).Return([]types.Response{}, nil)

	data := hooks.IncomingEmail{
		From:     "testing@example.com",
		To:       "anonymous@" + c.Config.Mail.Domain,
		Date:     "July 1st 2010",
		Subject:  "test subject",
		TextBody: "test text body",
		HTMLBody: "<p>test html body</p>",
	}
	marshalled, err := json.Marshal(data)
	require.NoError(t, err)

	request(t).
		setRoute(pages.Names.Hooks.IncomingEmail).
		withBasicAuth(c.Config.App.HooksUsername, c.Config.App.HooksPassword).
		post("application/json", bytes.NewBufferString(string(marshalled))).
		requireStatusCode(http.StatusOK)

	// Allow a sleep for the mail processing task to be executed
	time.Sleep(5 * time.Millisecond)

	require.Len(t, providerMock.Calls, 1)
	call := providerMock.Calls[0]
	require.Len(t, call.Arguments, 1)
	arg := call.Arguments.Get(0).([]*types.Mail)
	require.Len(t, arg, 1)

	sentMessage := arg[0]
	assert.Equal(t, &types.Mail{
		From:          fmt.Sprintf(" (OrcasHub)<%s>", c.Config.Mail.SupportAddress),
		To:            c.Config.Mail.AdminAddress,
		ReplyTo:       "testing@example.com",
		Subject:       data.Subject,
		MessageStream: types.AdminNotifications,
		TextBody:      data.TextBody,
		HTMLBody:      data.HTMLBody,
	}, sentMessage)
}

func TestIncomingClassifiedEmailInvalidCredentials(t *testing.T) {
	data := hooks.IncomingEmail{
		From:     "testing@example.com",
		To:       "anonymous@" + c.Config.Mail.Domain,
		Subject:  "test subject",
		TextBody: "test text body",
		HTMLBody: "<p>test html body</p>",
	}
	marshalled, err := json.Marshal(data)
	require.NoError(t, err)

	request(t).
		setRoute(pages.Names.Hooks.IncomingEmail).
		withBasicAuth(c.Config.App.HooksUsername, "bad-password").
		post("application/json", bytes.NewBufferString(string(marshalled))).
		requireStatusCode(http.StatusUnauthorized)
}

func TestIncomingClassifiedEmail(t *testing.T) {
	// Not parallel-safe because we depend on the shared mail mock

	// Ensure the user has been instantiated
	_ = request(t).withVerifiedUser()

	email := c.DB.PostBase.CreateAnonymousEmail(true)

	// Create a post and retrive the generated fields
	postID, err := c.DB.Classified.Create(context.Background(),
		classified.NewPostInput{
			CreatePostParams: sqlc.CreatePostParams{
				PostType:       constants.Classified,
				Title:          "Testing",
				Description:    "Testing",
				Creator:        verifiedUser.id,
				AnonymousEmail: email,
			},
			ClassifiedType: constants.ForFree,
		})
	require.NoError(t, err)

	// callHook sends a POST request to the hook endpoint
	callHook := func() {
		data := hooks.IncomingEmail{
			From:     "testing@example.com",
			To:       email.String + "@" + c.Config.Mail.Domain,
			Date:     "July 1st 2010",
			Subject:  "test subject",
			TextBody: "test text body",
			HTMLBody: "<p>test html body</p>",
		}
		marshalled, err := json.Marshal(data)
		require.NoError(t, err)

		request(t).
			setRoute(pages.Names.Hooks.IncomingEmail).
			withBasicAuth(c.Config.App.HooksUsername, c.Config.App.HooksPassword).
			post("application/json", bytes.NewBufferString(string(marshalled))).
			requireStatusCode(http.StatusOK)

		// Allow a sleep for the mail processing task to be executed
		time.Sleep(5 * time.Millisecond)
	}

	t.Run("valid post", func(t *testing.T) {
		err = c.DB.C.SetPostPublishedAt(context.Background(),
			sqlc.SetPostPublishedAtParams{
				PublishedAt: sql.NullTime{Valid: true, Time: time.Now().AddDate(0, 0, -1)},
				ID:          postID,
			})
		require.NoError(t, err)

		providerMock := mailmock.New()
		c.Mail.SetProvider(providerMock)
		providerMock.On("Send", mock.Anything).Return([]types.Response{}, nil)

		callHook()

		// Verify that message was forwarded as expected
		require.Len(t, providerMock.Calls, 1)
		call := providerMock.Calls[0]
		require.Len(t, call.Arguments, 1)
		arg := call.Arguments.Get(0).([]*types.Mail)
		require.Len(t, arg, 1)

		sentMessage := arg[0]
		assert.Equal(t, &types.Mail{
			From:          c.Config.Mail.FromAddress,
			ReplyTo:       "testing@example.com",
			Subject:       "[OrcasHub] test subject",
			To:            verifiedUser.email,
			MessageStream: types.UserMessages,
			TextBody:      fmt.Sprintf("Hello,\nYou have been sent an email in response to your OrcasHub post titled \"Testing\".\n\n    https://example.com/classified/%d\n\nPlease note, replying to this message will allow the original sender to see your actual email address.\nIf you have any questions or concerns, please feel free to email us at (support@orcashub.org).\n\nOriginal Message:\n\ntest text body", postID),
			HTMLBody:      fmt.Sprintf("<p>\n    Hello,\n</p>\n<p>\n    You have been sent an email in response to\n    your <a href=\"https://example.com/classified/%d\">OrcasHub post</a> titled \"Testing\".\n</p>\n<p>\n    Please note, replying to this message will allow the original sender to see your actual email address.\n</p>\n<p>\n    If you have any questions or concerns, please feel free to <a href=\"mailto:support@orcashub.org\">email us</a>.\n</p>\n<p>---------- Original message ---------</p>\n\n<p>test html body</p>", postID),
		}, sentMessage)
	})

	t.Run("expired post", func(t *testing.T) {
		err = c.DB.C.SetPostPublishedAt(context.Background(),
			sqlc.SetPostPublishedAtParams{
				PublishedAt: sql.NullTime{Valid: true, Time: time.Now().AddDate(0, -1, 0)},
				ID:          postID,
			})
		require.NoError(t, err)

		providerMock := mailmock.New()
		c.Mail.SetProvider(providerMock)
		providerMock.On("Send", mock.Anything).Return([]types.Response{}, nil)

		callHook()

		require.Len(t, providerMock.Calls, 1)
		call := providerMock.Calls[0]
		require.Len(t, call.Arguments, 1)
		arg := call.Arguments.Get(0).([]*types.Mail)
		require.Len(t, arg, 1)

		sentMessage := arg[0]
		assert.Equal(t, &types.Mail{
			From:          c.Config.Mail.FromAddress,
			To:            "testing@example.com",
			Subject:       "Re: test subject",
			MessageStream: types.AccountManagement,
			TextBody:      "Hello,\nThe OrcasHub post you are trying to email could not be found.\nIt may have already been sold or deleted by the poster.\n\nThank you,\n  - OrcasHub\n\n> On July 1st 2010, testing@example.com wrote:\n> test text body\n",
		}, sentMessage)
	})

	t.Run("unpublished post", func(t *testing.T) {
		err = c.DB.C.SetPostPublishedAt(context.Background(),
			sqlc.SetPostPublishedAtParams{
				PublishedAt: sql.NullTime{Valid: false},
				ID:          postID,
			})
		require.NoError(t, err)

		providerMock := mailmock.New()
		c.Mail.SetProvider(providerMock)
		providerMock.On("Send", mock.Anything).Return([]types.Response{}, nil)

		callHook()

		require.Len(t, providerMock.Calls, 1)
		call := providerMock.Calls[0]
		require.Len(t, call.Arguments, 1)
		arg := call.Arguments.Get(0).([]*types.Mail)
		require.Len(t, arg, 1)

		sentMessage := arg[0]
		assert.Equal(t, &types.Mail{
			From:          c.Config.Mail.FromAddress,
			To:            "testing@example.com",
			Subject:       "Re: test subject",
			MessageStream: types.AccountManagement,
			TextBody:      "Hello,\nThe OrcasHub post you are trying to email could not be found.\nIt may have already been sold or deleted by the poster.\n\nThank you,\n  - OrcasHub\n\n> On July 1st 2010, testing@example.com wrote:\n> test text body\n",
		}, sentMessage)

		err = c.DB.C.SetPostUnlisted(context.Background(), postID)
		require.NoError(t, err)
	})

	t.Run("delisted post", func(t *testing.T) {
		err = c.DB.C.SetPostUnlisted(context.Background(), postID)
		require.NoError(t, err)

		err = c.DB.C.SetPostPublishedAt(context.Background(),
			sqlc.SetPostPublishedAtParams{
				PublishedAt: sql.NullTime{Valid: true, Time: time.Now().AddDate(0, 0, -1)},
				ID:          postID,
			})
		require.NoError(t, err)

		providerMock := mailmock.New()
		c.Mail.SetProvider(providerMock)
		providerMock.On("Send", mock.Anything).Return([]types.Response{}, nil)

		callHook()

		require.Len(t, providerMock.Calls, 1)
		call := providerMock.Calls[0]
		require.Len(t, call.Arguments, 1)
		arg := call.Arguments.Get(0).([]*types.Mail)
		require.Len(t, arg, 1)

		sentMessage := arg[0]
		assert.Equal(t, &types.Mail{
			From:          c.Config.Mail.FromAddress,
			To:            "testing@example.com",
			Subject:       "Re: test subject",
			MessageStream: types.AccountManagement,
			TextBody:      "Hello,\nThe OrcasHub post you are trying to email could not be found.\nIt may have already been sold or deleted by the poster.\n\nThank you,\n  - OrcasHub\n\n> On July 1st 2010, testing@example.com wrote:\n> test text body\n",
		}, sentMessage)
	})
}
