package routes

import (
	"net/http"
	"regexp"
	"strings"

	"github.com/labstack/echo/v4"

	"git.grosinger.net/tgrosinger/tonyslist/constants/pages"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/context"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/controller"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/conv"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/views"
)

type login struct {
	controller.Controller
}

func (c *login) Get(ctx echo.Context) error {
	// This is guarded by a middleware which performs the auth.

	dest := ctx.QueryParam("return_to")
	if dest == "" {
		return c.Redirect(ctx, pages.Names.Home)
	}

	// Preserve any additionally query params
	values := ctx.Request().URL.Query()
	values.Del("return_to")

	query := values.Encode()
	if query != "" {
		return ctx.Redirect(http.StatusFound, dest+"?"+query)
	}
	return ctx.Redirect(http.StatusFound, dest)
}

func (c *login) GetRequireName(ctx echo.Context) error {
	user, ok := ctx.Get(context.AuthenticatedUserKey).(sqlc.User)
	if !ok {
		return c.BadReq("Missing authenticated user")
	}

	if user.Username.Valid && user.Name.Valid {
		url := ctx.Echo().Reverse(pages.Names.OAuth.Authorize) + "?" + ctx.Request().URL.RawQuery
		return ctx.Redirect(http.StatusFound, url)
	}

	page := controller.NewPage(ctx, c.Container)
	page.Title = "Register"

	form := views.RequireNameForm{}
	if f := ctx.Get(context.FormKey); f != nil {
		form = *f.(*views.RequireNameForm)
	}

	v := views.RequireName{
		Form:             form,
		NameRequired:     !user.Name.Valid,
		UsernameRequired: !user.Username.Valid,
		Page:             page,
	}
	return c.RenderHTML(ctx, page, &v)
}

func (c *login) PostRequireName(ctx echo.Context) error {
	var form views.RequireNameForm
	ctx.Set(context.FormKey, &form)

	// Parse the form values
	err := ctx.Bind(&form)
	if err != nil {
		return c.Fail(err, "unable to parse the form")
	}

	if err := form.Submission.Process(ctx, form); err != nil {
		return c.Fail(err, "unable to process form submission")
	}

	user, ok := ctx.Get(context.AuthenticatedUserKey).(sqlc.User)
	if !ok {
		return c.BadReq("Missing authenticated user")
	}

	if !user.Name.Valid && form.Name == "" {
		form.Submission.SetFieldError("Name", "Name is required")
	}
	if !user.Username.Valid && form.Username == "" {
		form.Submission.SetFieldError("Username", "Username is required")
	}
	if match, err := regexp.MatchString("^[a-zA-Z0-9-_]+$", form.Username); !match || err != nil {
		form.Submission.SetFieldError("Username", "Username can only contain letters, numbers, dashes, and underscores")
	}

	if form.Submission.HasErrors() {
		return c.GetRequireName(ctx)
	}

	if form.Name != "" {
		err := c.Container.DB.C.SetUserName(ctx.Request().Context(), sqlc.SetUserNameParams{
			Name: conv.StringToNullString(form.Name),
			ID:   user.ID,
		})
		if err != nil {
			return c.Fail(err, "unable to set user name")
		}
	}

	if form.Username != "" {
		err := c.Container.DB.C.SetUserUsername(ctx.Request().Context(), sqlc.SetUserUsernameParams{
			Username: conv.StringToNullString(form.Username),
			ID:       user.ID,
		})
		if err != nil {
			if strings.HasPrefix(err.Error(), "UNIQUE constraint failed") {
				form.Submission.SetFieldError("Username", "Username is unavailable")
				return c.GetRequireName(ctx)
			}

			return c.Fail(err, "unable to set user name")
		}
	}

	url := ctx.Echo().Reverse(pages.Names.OAuth.Authorize) + "?" + ctx.Request().URL.RawQuery
	return ctx.Redirect(http.StatusFound, url)
}
