package routes

import (
	"net/http"
	"strconv"
	"strings"

	"github.com/labstack/echo/v4"

	"git.grosinger.net/tgrosinger/tonyslist/constants/pages"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/context"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/controller"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
)

type Oauth2 struct {
	controller.Controller
}

// Authorize is protected by auth required middleware. It performs Oauth
// authorization for logged-in users.
//
// NOTE: Users created by the authorize process will be automatically unsubscribed from digest emails.
// See register.go:Post()
func (c *Oauth2) Authorize(ctx echo.Context) error {
	user, ok := ctx.Get(context.AuthenticatedUserKey).(sqlc.User)
	if !ok {
		return c.BadReq("Missing authenticated user")
	}

	if !user.Username.Valid || !user.Name.Valid {
		// Prompt the user to set their username and name before authorizing.
		url := ctx.Echo().Reverse(pages.Names.User.RequireName) + "?" + ctx.Request().URL.RawQuery
		return ctx.Redirect(http.StatusFound, url)
	}

	r := ctx.Request()
	updatedContext := context.WithValue(r.Context(), "oauth-user-id", strconv.Itoa(user.ID))
	updatedRequest := r.WithContext(updatedContext)
	ctx.SetRequest(updatedRequest)

	err := c.Controller.Container.OAuth2.OAuth2Server.
		HandleAuthorizeRequest(ctx.Response(), ctx.Request())
	if err != nil {
		echoErr := echo.NewHTTPError(http.StatusBadRequest, "Bad Request")
		echoErr.SetInternal(err)
		return echoErr
	}

	return nil
}

func (c *Oauth2) Token(ctx echo.Context) error {
	err := c.Controller.Container.OAuth2.OAuth2Server.
		HandleTokenRequest(ctx.Response(), ctx.Request())
	if err != nil {
		return c.BadReq(err.Error())
	}

	return nil
}

// Lookup returns additional information used by the Wiki when performing an
// OAuth2 login. Right now this is only the users email address.
func (c *Oauth2) Lookup(ctx echo.Context) error {
	token := ""

	// Check header first
	tokenHeader := ctx.Request().Header.Get("Authorization")
	if strings.HasPrefix(tokenHeader, "token ") {
		token = strings.TrimPrefix(tokenHeader, "token ")
	}

	// Check query param next
	if token == "" {
		token = ctx.QueryParam("oauth_token")
	}

	if token == "" {
		return c.BadReq("Missing token")
	}

	ti, err := c.Controller.Container.OAuth2.OAuth2Server.Manager.LoadAccessToken(ctx.Request().Context(), token)
	if err != nil {
		return c.BadReq(err.Error())
	}

	userID := ti.GetUserID()

	userIDParsed, err := strconv.Atoi(userID)
	if err != nil {
		return c.BadReq("Invalid userID")
	}

	user, err := c.Container.DB.C.GetUserByID(ctx.Request().Context(), userIDParsed)
	if err != nil {
		return c.BadReq("Invalid userID")
	}

	// We'll prompt the user for their username before returning a successful oauth response,
	// so the username will always be valid by the time we get here.

	return ctx.JSON(http.StatusOK, map[string]string{
		"user_id":  userID,
		"name":     user.Name.String,
		"username": user.Username.String,
		"email":    user.Email,
	})
}
