package routes

import (
	"context"
	"database/sql"
	"mime/multipart"
	"net/http"
	"testing"
	"time"

	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/mock"
	"github.com/stretchr/testify/require"

	"git.grosinger.net/tgrosinger/tonyslist/constants"
	"git.grosinger.net/tgrosinger/tonyslist/constants/pages"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/classified"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
	mailmock "git.grosinger.net/tgrosinger/tonyslist/pkg/services/mail/mock"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/services/mail/types"
)

func TestModeratorToolbarVisibility(t *testing.T) {
	// Ensure that the needed users have been instantiated
	_ = request(t).withUnverifiedUser()
	_ = request(t).withVerifiedUser()
	_ = request(t).withAdminUser()

	postID, err := c.DB.Classified.Create(context.Background(), classified.NewPostInput{
		CreatePostParams: sqlc.CreatePostParams{
			PostType:    constants.Classified,
			Title:       "Test post that is expired",
			Description: "Test description",
			PhoneCall:   sql.NullString{Valid: false},
			PhoneText:   sql.NullString{Valid: false},
			Creator:     unverifiedUser.id,
		},
		ClassifiedType: constants.ForSale,
		Price:          sql.NullFloat64{Valid: true, Float64: 12.23},
	})
	require.NoError(t, err)

	err = c.DB.C.SetPostPublishedNow(context.Background(), postID)
	require.NoError(t, err)

	// Moderator can see the mod toolbar but not the owner.
	doc := request(t).
		withAdminUser().
		setRoute(pages.Names.Classified.One, postID).
		get().
		toDoc()
	assert.Equal(t, 1, doc.Find("#moderator-toolbar").Length())
	assert.Equal(t, 0, doc.Find("#owner-toolbar").Length())

	// Poster cannot see the mod toolbar, but can see the owner toolbar.
	doc = request(t).
		withUnverifiedUser().
		setRoute(pages.Names.Classified.One, postID).
		get().
		toDoc()
	assert.Equal(t, 0, doc.Find("#moderator-toolbar").Length())
	assert.Equal(t, 1, doc.Find("#owner-toolbar").Length())

	// Random user cannot see either
	doc = request(t).
		withVerifiedUser().
		setRoute(pages.Names.Classified.One, postID).
		get().
		toDoc()
	assert.Equal(t, 0, doc.Find("#moderator-toolbar").Length())
	assert.Equal(t, 0, doc.Find("#owner-toolbar").Length())

	// Logged out viewer cannot see either
	doc = request(t).
		setRoute(pages.Names.Classified.One, postID).
		get().
		toDoc()
	assert.Equal(t, 0, doc.Find("#moderator-toolbar").Length())
	assert.Equal(t, 0, doc.Find("#owner-toolbar").Length())
}

func TestModerateDelete(t *testing.T) {
	// Ensure that the needed users have been instantiated
	_ = request(t).withUnverifiedUser()
	_ = request(t).withAdminUser()

	postID, err := c.DB.Classified.Create(context.Background(), classified.NewPostInput{
		CreatePostParams: sqlc.CreatePostParams{
			PostType:    constants.Classified,
			Title:       "Test post that is expired",
			Description: "Test description",
			PhoneCall:   sql.NullString{Valid: false},
			PhoneText:   sql.NullString{Valid: false},
			Creator:     unverifiedUser.id,
		},
		ClassifiedType: constants.ForSale,
		Price:          sql.NullFloat64{Valid: true, Float64: 12.23},
	})
	require.NoError(t, err)

	err = c.DB.C.SetPostPublishedNow(context.Background(), postID)
	require.NoError(t, err)

	// Moderator can load the delete post form.
	request(t).
		withAdminUser().
		setRoute(pages.Names.Posts.ModeratorDelete, postID).
		get().
		requireStatusCode(http.StatusOK)

	// Unverified user cannot load the delete post form.
	request(t).
		withUnverifiedUser().
		disableFollowRedirects().
		setRoute(pages.Names.Posts.ModeratorDelete, postID).
		get().
		requireStatusCode(http.StatusForbidden)

	// Unverified user cannot delete the post.
	request(t).
		withUnverifiedUser().
		disableFollowRedirects().
		setRoute(pages.Names.Posts.ModeratorDelete, postID).
		postMultipartForm(func(w *multipart.Writer) {
			setMultipartField(t, w, "reason", "unverified user testing")
		}).
		requireStatusCode(http.StatusForbidden)

	// Moderator can delete the post.
	providerMock := mailmock.New()
	c.Mail.SetProvider(providerMock)
	providerMock.On("Send", mock.Anything).Return([]types.Response{}, nil)

	resp := request(t).
		withAdminUser().
		disableFollowRedirects().
		setRoute(pages.Names.Posts.ModeratorDelete, postID).
		postMultipartForm(func(w *multipart.Writer) {
			setMultipartField(t, w, "reason", "unverified user testing")
		}).
		requireStatusCode(http.StatusFound)
	loc, err := resp.Location()
	require.NoError(t, err)
	require.NotEmpty(t, loc)
	require.Equal(t, c.Web.Reverse(pages.Names.Classified.ForSale.Browse), loc.Path)

	// Allow the mail sending Goroutine to take over and process the messages.
	time.Sleep(100 * time.Millisecond)

	// Calls are batched, so this one call should have two emails.
	require.Len(t, providerMock.Calls, 1)
	call := providerMock.Calls[0]
	sentMail, ok := call.Arguments.Get(0).([]*types.Mail)
	require.True(t, ok)
	require.Len(t, sentMail, 2)

	// After deleting, it can't be deleted again by moderator.
	request(t).
		withAdminUser().
		setRoute(pages.Names.Posts.ModeratorDelete, postID).
		get().
		requireStatusCode(http.StatusNotFound)

	// After deleting, the post cannot be retrieved by users.
	request(t).
		withUnverifiedUser().
		disableFollowRedirects().
		setRoute(pages.Names.Classified.One, postID).
		get().
		requireStatusCode(http.StatusNotFound)
}
