package postbase

import (
	"fmt"
	"html/template"

	"github.com/labstack/echo/v4"
	"go.uber.org/zap"

	"git.grosinger.net/tgrosinger/tonyslist/constants/pages"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/context"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/controller"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/postbase"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/msg"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/services/mail"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/services/mail/types"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/views"
)

func (c *commonController) ModeratorDeleteGet(ctx echo.Context) error {
	page := controller.NewPage(ctx, c.Container)
	page.Metatags.Keywords = keywords
	page.Title = "Delete Post"
	page.Metatags.Description = fmt.Sprintf("Delete a post as a moderator on %s.", c.Container.Config.App.Name)

	form := views.PostModeratorDeleteForm{}
	if f := ctx.Get(context.FormKey); f != nil {
		form = *f.(*views.PostModeratorDeleteForm)
	}

	v := views.PostModeratorDelete{
		Form: form,
		Page: page,
	}
	return c.RenderHTML(ctx, page, &v)
}

func (c *commonController) ModeratorDeletePost(ctx echo.Context) error {
	post, ok := ctx.Get(context.PostBaseKey).(*postbase.PostBase)
	if !ok {
		msg.Warning(ctx, "The requested post could not be found.")
		return c.RedirectToPostBrowser(ctx)
	}

	usr, ok := ctx.Get(context.AuthenticatedUserKey).(sqlc.User)
	if !ok {
		return c.BadReq("unable to look up the current authenticated user")
	}

	var form views.PostModeratorDeleteForm
	ctx.Set(context.FormKey, &form)

	if err := ctx.Bind(&form); err != nil {
		return c.Fail(err, "unable to bind form")
	}

	if err := form.Submission.Process(ctx, form); err != nil {
		return c.Fail(err, "unable to process form submission")
	}

	if form.Submission.HasErrors() {
		return c.DeleteGet(ctx)
	}

	// NOTE: We are not sanitizing the input,
	// However we are only use the input in text templates for emails.
	// This content must be sanitized before doing anything else with it where it could be used
	// in a database or HTML.

	err := c.Container.DB.C.SetPostModeratorDelisted(ctx.Request().Context(), post.ID)
	if err != nil {
		return c.Fail(err, "unable to delete the post")
	}

	toSend := make([]*types.Mail, 0, 2)

	logger := context.GetLogger(ctx)
	m, err := c.Container.Mail.Template(mail.TemplateArgs{
		MessageStream: types.AdminNotifications,
		To:            c.Container.Config.Mail.AdminAddress,
		TextTemplate:  "moderator_action_text",
		Subject:       c.Container.Config.App.Name + " Moderator Action",
		TemplateData: map[string]interface{}{
			"moderator_name": usr.Email,
			"action":         fmt.Sprintf("Delist post %d", post.ID),
			"reason":         form.Reason,
			"custom_message": template.HTML(form.CustomMessage),
		},
	})
	if err != nil {
		logger.Error("unable to send email to admin for moderator action", zap.Error(err))
	} else {
		toSend = append(toSend, m)
	}

	creator, err := c.Container.DB.C.GetUserByID(ctx.Request().Context(), post.Creator)
	if err != nil {
		logger.Error("unable to retrieve creator for post to send moderation notification",
			zap.Error(err))
	} else {
		usrMsg, err := c.Container.Mail.Template(mail.TemplateArgs{
			MessageStream: types.AccountManagement,
			To:            creator.Email,
			Subject:       c.Container.Config.App.Name + " Moderator Action",
			TextTemplate:  "moderation_notification_text",
			TemplateData: map[string]interface{}{
				"post_title":     post.Title,
				"custom_message": template.HTML(form.CustomMessage),
				"domain":         c.Container.Config.App.PublicURL,
				"toc_path":       ctx.Echo().Reverse(pages.Names.TermsOfService),
				"support_email":  c.Container.Config.Mail.SupportAddress,
			},
		})
		if err != nil {
			logger.Error("unable to send email to user for moderator action", zap.Error(err))
		} else {
			toSend = append(toSend, usrMsg)
		}
	}

	if len(toSend) > 0 {
		c.Container.Mail.Send(logger, toSend, nil)
	}

	msg.Success(ctx, "The post was successfully delisted.")
	return c.RedirectToPostBrowser(ctx)
}
