package postbase

import (
	"database/sql"
	"fmt"
	"strconv"

	"github.com/labstack/echo/v4"

	"git.grosinger.net/tgrosinger/tonyslist/constants/pages"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/context"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/controller"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/conv"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/postbase"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/msg"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/views"
)

func (c *commonController) PhotosGet(ctx echo.Context) error {
	// Middleware protects against posts not existing and being expired.
	p, ok := ctx.Get(context.PostBaseKey).(*postbase.PostBase)
	if !ok {
		msg.Warning(ctx, "The requested post could not be found.")
		return c.RedirectToPostBrowser(ctx)
	}

	publishDisplay := "Publish"
	if p.PublishedAt.Valid {
		// Editing photos on published post
		publishDisplay = "Done"
	}

	photos, err := c.Container.DB.C.GetPhotos(ctx.Request().Context(), p.ID)
	if err == sql.ErrNoRows {
		photos = make([]sqlc.Photo, 0)
	} else if err != nil {
		return c.Fail(err, "unable to get existing photos for post")
	}

	token, err := c.Container.Auth.GenerateJWTAuthToken(map[string]interface{}{
		"post_id": p.ID,
	}, &c.Container.Config.App.EmailVerificationTokenExpiration)
	if err != nil {
		return c.Fail(err, "unable to generate jwt token")
	}

	page := controller.NewPage(ctx, c.Container)
	page.Title = "New Post"
	if p.PublishedAt.Valid {
		// Editing photos on published post
		page.Title = "Edit Post"
	}
	page.Metatags.Description = fmt.Sprintf("Create new post on %s.", c.Container.Config.App.Name)
	page.Metatags.Keywords = keywords

	v := views.PostPhotoUpload{
		PublishDisplay:           publishDisplay,
		PublishAction:            ctx.Echo().Reverse(pages.Names.Posts.Publish, p.ID),
		PreviouslyUploadedPhotos: photos,
		PostType:                 p.PostType,
		PostID:                   p.ID,
		Token:                    token,

		Page: page,
	}
	return c.RenderHTML(ctx, page, &v)
}

func (c *commonController) PhotoDelete(ctx echo.Context) error {
	// Middleware protects against posts not existing and being expired.
	post, ok := ctx.Get(context.PostBaseKey).(*postbase.PostBase)
	if !ok {
		msg.Warning(ctx, "The requested post could not be found.")
		return c.RedirectToPostBrowser(ctx)
	}

	photoID, err := strconv.Atoi(ctx.Param("photo-id"))
	if err != nil {
		c.Fail(err, "invalid photo id")
	}

	err = c.Container.DB.WithTx(func(tx *sql.Tx) error {
		txDB := c.Container.DB.C.WithTx(tx)

		p, err := txDB.GetPhoto(ctx.Request().Context(), photoID)
		if err != nil {
			c.Fail(err, "invalid photo id")
		}

		if post.FeaturedPhoto.Valid && post.FeaturedPhoto.Int == p.ID {
			photos, err := txDB.GetPhotos(ctx.Request().Context(), post.ID)
			if err != nil {
				return c.Fail(err, "internal server error")
			}

			foundNewFeaturedPhoto := false
			for _, otherPhoto := range photos {
				// This should skip the first photo and then assign the next photo if there is one
				if otherPhoto.ID == p.ID {
					continue
				}

				err = txDB.SetPostFeaturedPhotoID(ctx.Request().Context(), sqlc.SetPostFeaturedPhotoIDParams{
					FeaturedPhoto: conv.IntToNullInt(otherPhoto.ID),
					ID:            post.ID,
				})
				if err != nil {
					return c.Fail(err, "internal server error")
				}

				foundNewFeaturedPhoto = true
				break
			}

			if !foundNewFeaturedPhoto {
				err = txDB.SetPostFeaturedPhotoID(ctx.Request().Context(), sqlc.SetPostFeaturedPhotoIDParams{
					FeaturedPhoto: conv.NullInt{Valid: false},
					ID:            post.ID,
				})
				if err != nil {
					return c.Fail(err, "internal server error")
				}
			}
		}

		err = txDB.DeletePhoto(ctx.Request().Context(), sqlc.DeletePhotoParams{
			ID:     p.ID,
			PostID: post.ID,
		})
		if err != nil {
			c.Fail(err, "invalid photo id")
		}

		return nil
	})
	if err != nil {
		return err
	}

	return c.Redirect(ctx, pages.Names.Posts.NewPhotos, post.ID)
}
