package routes

import (
	"database/sql"
	"net/http"
	"strings"

	"github.com/labstack/echo/v4"
	"go.uber.org/zap"

	"git.grosinger.net/tgrosinger/tonyslist/constants"
	"git.grosinger.net/tgrosinger/tonyslist/constants/pages"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/context"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/controller"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/msg"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/services/mail"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/services/mail/types"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/views"
)

type register struct {
	controller.Controller
}

// Get handles signing up or logging in to existing accounts by prompting users
// for their email address.
func (c *register) Get(ctx echo.Context) error {
	_, ok := ctx.Get(context.AuthenticatedUserKey).(sqlc.User)
	if ok {
		msg.Info(ctx, "You are already signed in")
		dest := ctx.QueryParam("return_to")
		if dest == "" {
			return c.Redirect(ctx, pages.Names.Home)
		}

		// Preserve any additionally query params
		values := ctx.Request().URL.Query()
		values.Del("return_to")

		query := values.Encode()
		if query != "" {
			return ctx.Redirect(http.StatusFound, dest+"?"+query)
		}
		return ctx.Redirect(http.StatusFound, dest)

	}

	page := controller.NewPage(ctx, c.Container)
	page.Title = "Sign In"

	form := views.RegisterForm{}
	if f := ctx.Get(context.FormKey); f != nil {
		form = *f.(*views.RegisterForm)
	}

	v := views.Register{
		Form: form,
		Page: page,
	}
	return c.RenderHTML(ctx, page, &v)
}

// Post handles submissions from the email-only signin and register forms
// including from the home page. It should check if the email exists, and
// present a helpful message accordingly.
func (c *register) Post(ctx echo.Context) error {
	var form views.RegisterForm
	ctx.Set(context.FormKey, &form)

	// Parse the form values
	err := ctx.Bind(&form)
	if err != nil {
		return c.Fail(err, "unable to parse the form")
	}

	if err := form.Submission.Process(ctx, form); err != nil {
		return c.Fail(err, "unable to process form submission")
	}

	if form.Submission.HasErrors() {
		return c.Get(ctx)
	}

	if form.Empty != "" {
		msg.Danger(ctx, "Form contained an unexpected value")
		return c.Get(ctx)
	}

	// Always treat email addresses as lowercase
	form.Email = strings.ToLower(form.Email)

	if form.Action == "login" {
		_, err := c.Container.DB.C.GetUserByEmail(ctx.Request().Context(), form.Email)
		if err != nil && err != sql.ErrNoRows {
			return c.Fail(err, "unable to look up the email in the database")
		} else if err == sql.ErrNoRows {
			form.Submission.SetFieldError("Email", "An account for that address could not be found")
			return c.Get(ctx)
		}
	}

	newUser, userID, err := c.getOrCreateUser(ctx.Request().Context(), form.Email)
	if err != nil {
		return err
	}

	returnTo := ctx.Request().URL.Query().Get("return_to")
	if newUser && returnTo == "/oauth2/authorize" {
		// If this user was created as part of an oauth flow, unsubscribe them from digest emails.
		err = c.Container.DB.C.SetUserSetting(ctx.Request().Context(),
			sqlc.SetUserSettingParams{
				UserID: userID,
				Key:    string(constants.EmailDigestSubscription),
				Value: sql.NullString{
					Valid:  true,
					String: constants.EmailDigestSubscriptionUnsubscribed,
				},
			})
		if err != nil {
			c.Container.Logger.Error("unable to unsubscribe new oauth user from digest emails", zap.Error(err))
		}
	}

	err = controller.UserController{Controller: c.Controller}.
		SendLoginEmail(ctx, userID, form.Email, newUser, ctx.Request().URL.Query())
	if err != nil {
		return err
	}

	page := controller.NewPage(ctx, c.Container)
	page.Title = "Sign In"

	message := "We're emailing you a magic link!"
	if newUser {
		message = "You're on the list!"
	}

	v := views.MagicSigninSent{
		EmailAddress: form.Email,
		Message:      message,

		Page: page,
	}
	return c.RenderHTML(ctx, page, &v)
}

func (c *register) getOrCreateUser(ctx context.Context, email string) (newUser bool, userID int, err error) {
	user, err := c.Container.DB.C.GetUserByEmail(ctx, email)
	if err == sql.ErrNoRows {
		newUser = true
		var newUserID int64
		newUserID, err = c.Container.DB.C.CreateUser(ctx,
			sqlc.CreateUserParams{
				Email: email,
			})
		if err != nil {
			err = c.Fail(err, "unable to create a new user for email")
			return
		}

		userID = int(newUserID)
		c.emailAdminUserSignup(c.Container.Logger, email)
	} else if err != nil {
		err = c.Fail(err, "unable to look up the email in the database")
		return
	} else {
		userID = user.ID
	}

	return
}

// emailAdminUserSignup notifies the configured administrator of a user signup.
func (c *register) emailAdminUserSignup(logger *zap.Logger, userEmail string) {
	templateData := map[string]string{
		"email": userEmail,
	}

	m, err := c.Container.Mail.Template(mail.TemplateArgs{
		MessageStream: types.AdminNotifications,
		To:            c.Container.Config.Mail.AdminAddress,
		Subject:       c.Container.Config.App.Name + " User Signup",
		TextTemplate:  "user_signup_notification_text",
		TemplateData:  templateData,
	})
	if err != nil {
		logger.Error("unable to send email to admin for user signup", zap.Error(err))
		return
	}

	c.Container.Mail.Send(logger, []*types.Mail{m}, nil)
}
