package routes

import (
	"net/http"
	"strings"

	"github.com/gorilla/sessions"
	"github.com/labstack/echo-contrib/session"
	"github.com/labstack/echo/v4"
	echomw "github.com/labstack/echo/v4/middleware"

	"git.grosinger.net/tgrosinger/tonyslist/config"
	"git.grosinger.net/tgrosinger/tonyslist/constants/pages"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/context"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/controller"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/middleware"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/routes/admin"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/routes/api"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/routes/classified"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/routes/event"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/routes/hooks"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/routes/housing"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/routes/postbase"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/services"
)

// BuildRouter builds the router
func BuildRouter(c *services.Container) {
	// Non static file route group
	g := c.Web.Group("")

	// Force HTTPS, if enabled
	if c.Config.HTTP.TLS.Enabled {
		g.Use(echomw.HTTPSRedirect())
	}

	g.Use(
		echomw.RemoveTrailingSlashWithConfig(echomw.TrailingSlashConfig{
			RedirectCode: http.StatusMovedPermanently,
		}),
		echomw.Recover(),
		echomw.Secure(),
		echomw.RequestID(),
		echomw.Gzip(),
		middleware.RequestMetrics(),
		middleware.LogRequestResponse(c.Logger),
		echomw.TimeoutWithConfig(echomw.TimeoutConfig{
			Timeout: c.Config.App.Timeout,
			Skipper: func(ctx echo.Context) bool {
				path := ctx.Request().URL.Path
				if strings.HasPrefix(path, "/uploads") {
					context.GetLogger(ctx).Debug("Skipping timeout middleware for tus endpoint")
					return true
				}
				return false
			},
		}),
		session.Middleware(sessions.NewCookieStore([]byte(c.Config.App.EncryptionKey))),
		middleware.LoadAuthenticatedUser(c.Auth, c.DB),
		echomw.CSRFWithConfig(echomw.CSRFConfig{
			CookieSameSite: http.SameSiteStrictMode,
			CookieSecure:   c.Config.App.Environment != config.EnvTest,
			TokenLookup:    "form:csrf",
			CookiePath:     "/",
			Skipper: func(ctx echo.Context) bool {
				path := ctx.Request().URL.Path
				if strings.HasPrefix(path, "/hooks") {
					context.GetLogger(ctx).Debug("Skipping CSRF middleware for hooks endpoint")
					return true
				}
				if strings.HasPrefix(path, "/uploads") {
					context.GetLogger(ctx).Debug("Skipping CSRF middleware for tus endpoint")
					return true
				}
				if path == "/oauth2/token" {
					context.GetLogger(ctx).Debug("Skipping CSRF middleware for oauth2 endpoint")
					return true
				}
				return false
			},
		}),
	)

	// Base controller
	ctr := controller.NewController(c)

	// Static files with proper cache control
	// helpers.File() should be used in templates to append a cache key to the URL in order to break cache
	// after each server restart
	staticCacheGroup := g.Group("", middleware.CacheControl(c.Config.Cache.Expiration.StaticFile))
	staticCacheGroup.Static(config.StaticPrefix, config.StaticDir(c.Config))

	static := static{Controller: ctr}
	staticCacheGroup.GET(config.UserContentPrefix+"/:filename", static.Get)

	// Error handler
	err := errorHandler{Controller: ctr}
	c.Web.HTTPErrorHandler = err.Get

	wellKnownRoutes(c, g, ctr)
	oauth2Routes(c, g, ctr)
	hooks.BuildRoutes(c, g, ctr)
	api.BuildRoutes(c, g, ctr)

	// Router which records page loads
	statsG := g.Group("", middleware.RecordAuthenticatedPageLoad(c.DB))

	donateRoutes(c, statsG, ctr)
	photoUploadRoutes(c, statsG, ctr)
	navRoutes(c, statsG, ctr)
	userRoutes(c, statsG, ctr)
	helpRoutes(c, statsG, ctr)
	admin.BuildRoutes(c, statsG, ctr)
	event.BuildRoutes(c, statsG, ctr)
	classified.BuildRoutes(c, statsG, ctr)
	housing.BuildRoutes(c, statsG, ctr)
	postbase.BuildRoutes(c, statsG, ctr)

	robots := robots{Controller: ctr}
	g.GET("/robots.txt", robots.Get)
}

func wellKnownRoutes(c *services.Container, g *echo.Group, ctr controller.Controller) {
	sg := g.Group("/.well-known")

	wellKnown := wellknown{Controller: ctr}
	sg.GET("/apple-developer-merchantid-domain-association", wellKnown.GetAppleVerification)
}

func donateRoutes(c *services.Container, g *echo.Group, ctr controller.Controller) {
	sg := g.Group("", middleware.WithStripeCustomer(c.Stripe, c.DB, c.Config.Stripe))

	donate := donate{Controller: ctr}
	sg.GET("/donate", donate.Get).Name = pages.Names.Donate
	sg.GET("/donate/complete", donate.GetComplete).Name = pages.Names.DonateComplete

	sg.POST("/api/stripe/customer", donate.CreateCustomer).Name = pages.Names.API.Stripe.CreateCustomer
	g.POST("/api/stripe/payment-intent", donate.CreatePaymentIntent).Name = pages.Names.API.Stripe.CreatePaymentIntent
	g.POST("/api/stripe/subscription", donate.CreateSubscription).Name = pages.Names.API.Stripe.CreateSubscription
}

func photoUploadRoutes(c *services.Container, g *echo.Group, ctr controller.Controller) {
	photoUploads := Tus{Controller: ctr}
	g.Match([]string{http.MethodHead, http.MethodPost, http.MethodPatch},
		"/uploads", photoUploads.Handle,
		middleware.RequireAuthentication())
}

func oauth2Routes(c *services.Container, g *echo.Group, ctr controller.Controller) {
	oauth2 := Oauth2{Controller: ctr}

	g.GET("/oauth2/authorize", oauth2.Authorize, middleware.RequireAuthentication()).Name = pages.Names.OAuth.Authorize
	g.POST("/oauth2/token", oauth2.Token)
	g.GET("/oauth2/lookup", oauth2.Lookup)
}

func navRoutes(c *services.Container, g *echo.Group, ctr controller.Controller) {
	home := home{Controller: ctr}
	g.GET("/", home.Get).Name = pages.Names.Home
	g.POST("/", home.PostContact)
	g.GET("/community-resources", home.CommunityResources).Name = pages.Names.CommunityResources
	g.GET("/editor", home.Editor).Name = pages.Names.Editor

	policies := policies{Controller: ctr}
	g.GET("/faq", policies.GetFAQ).Name = pages.Names.FAQ
	g.GET("/terms-of-service", policies.GetTermsOfService).Name = pages.Names.TermsOfService
	g.GET("/privacy-policy", policies.GetPrivacyPolicy).Name = pages.Names.PrivacyPolicy
}

func userRoutes(c *services.Container, g *echo.Group, ctr controller.Controller) {
	tokenAuthMW := middleware.AuthenticateUserViaToken(c.Auth, c.DB)

	unsubscribe := unsubscribe{Controller: ctr}
	g.GET("/unsubscribe", unsubscribe.Get).Name = pages.Names.User.Unsubscribe

	logout := logout{Controller: ctr}
	g.GET("/logout", logout.Get, middleware.RequireAuthentication()).Name = pages.Names.User.Logout

	register := register{Controller: ctr}
	g.GET("/user/login", register.Get).Name = pages.Names.User.Login
	g.POST("/user/login", register.Post, middleware.RequireNoAuthentication())

	login := login{Controller: ctr}
	g.GET("/user/login/token/:user/:token", login.Get,
		middleware.LoadUser(c.DB),
		tokenAuthMW,
	).Name = pages.Names.User.LoginToken
	g.GET("/user/login/require-name", login.GetRequireName).Name = pages.Names.User.RequireName
	g.POST("/user/login/require-name", login.PostRequireName)

	withAuth := g.Group("/user", middleware.RequireAuthentication())

	account := account{Controller: ctr}
	withAuth.GET("/account", account.Get).Name = pages.Names.User.Account
	withAuth.POST("/account", account.Post)

	withAuth.GET("/account/delete", account.GetDelete).Name = pages.Names.User.DeleteAccount
	withAuth.POST("/account/delete", account.PostDelete).Name = pages.Names.User.DeleteAccount

	withAuth.GET("/account/digest", account.GetJoinDigest).Name = pages.Names.User.JoinDigest
	withAuth.POST("/account/digest", account.PostJoinDigest)
}

func helpRoutes(c *services.Container, g *echo.Group, ctr controller.Controller) {
	help := help{Controller: ctr}
	g.GET("/help", help.Index).Name = pages.Names.Help.Classified.Help
	g.GET("/help/classified/rss", help.ClassifiedRSS).Name = pages.Names.Help.Classified.RSS
	g.GET("/help/classified/search", help.ClassifiedSearch).Name = pages.Names.Help.Classified.Search
	g.GET("/help/classified/markdown", help.ClassifiedMarkdown).Name = pages.Names.Help.Classified.Markdown
}
