package routes

import (
	"bytes"
	"context"
	"io"
	"mime/multipart"
	"net/http"
	"net/http/cookiejar"
	"net/http/httptest"
	"net/url"
	"os"
	"regexp"
	"strconv"
	"strings"
	"sync"
	"testing"

	"github.com/PuerkitoBio/goquery"
	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
	"go.uber.org/zap"

	"git.grosinger.net/tgrosinger/tonyslist/constants/pages"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/services"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/tests"
)

// Mirrors value in pkg/services/auth.go
const authSessionName = "ua"

type userAuth struct {
	id            int
	name          string
	email         string
	sessionCookie *http.Cookie
}

var (
	srv *httptest.Server
	c   *services.Container

	adminUser     userAuth
	adminUserOnce sync.Once

	verifiedUser     userAuth
	verifiedUserOnce sync.Once

	unverifiedUser     userAuth
	unverifiedUserOnce sync.Once
)

func TestMain(m *testing.M) {
	// Start a new container
	configPath := tests.LoadConfigPath()
	logger, err := zap.NewDevelopment()
	if err != nil {
		panic(err)
	}
	c = services.NewContainer(configPath, logger)

	// Start a test HTTP server
	BuildRouter(c)
	srv = httptest.NewServer(c.Web)

	// Run tests
	exitVal := m.Run()

	// Shutdown the container and test server
	if err := c.Shutdown(); err != nil {
		panic(err)
	}
	srv.Close()

	os.Exit(exitVal)
}

type httpRequest struct {
	route             string
	client            http.Client
	basicAuthUsername string
	basicAuthPassword string
	t                 *testing.T
}

func request(t *testing.T) *httpRequest {
	jar, err := cookiejar.New(nil)
	require.NoError(t, err)
	r := httpRequest{
		t: t,
		client: http.Client{
			Jar: jar,
		},
	}
	return &r
}

func (h *httpRequest) setClient(client http.Client) *httpRequest {
	h.client = client
	return h
}

// disableFollowRedirects will prevent this request from following any
// redirects. This value is set on the client and will be lost if setClient is
// called.
func (h *httpRequest) disableFollowRedirects() *httpRequest {
	h.client.CheckRedirect = func(_ *http.Request, _ []*http.Request) error {
		return http.ErrUseLastResponse
	}
	return h
}

func (h *httpRequest) withAdminUser() *httpRequest {
	u, err := url.Parse(srv.URL)
	require.NoError(h.t, err)

	adminUserOnce.Do(func() {
		adminUser = userAuth{}

		// Create a test user
		usr, err := c.DB.User.CreateMockUser()
		require.NoError(h.t, err)
		adminUser.id = usr.ID
		adminUser.name = usr.Name.String
		adminUser.email = usr.Email

		// Mark user as verified
		err = c.DB.C.SetUserVerified(context.Background(), usr.ID)
		require.NoError(h.t, err)

		// Mark user as admin
		err = c.DB.User.SetAdmin(context.Background(), usr.ID)
		require.NoError(h.t, err)

		// Login and store session Cookie
		adminUser.sessionCookie = getSessionCookie(h.t, usr.ID, usr.Email)
	})

	h.client.Jar.SetCookies(u, []*http.Cookie{adminUser.sessionCookie})
	return h
}

func (h *httpRequest) withVerifiedUser() *httpRequest {
	u, err := url.Parse(srv.URL)
	require.NoError(h.t, err)

	verifiedUserOnce.Do(func() {
		verifiedUser = userAuth{}

		// Create a test user
		usr, err := c.DB.User.CreateMockUser()
		require.NoError(h.t, err)
		verifiedUser.id = usr.ID
		verifiedUser.name = usr.Name.String
		verifiedUser.email = usr.Email

		require.NotEmpty(h.t, verifiedUser.name)

		// Mark user as verified
		err = c.DB.C.SetUserVerified(context.Background(), usr.ID)
		require.NoError(h.t, err)

		// Login and store session Cookie
		verifiedUser.sessionCookie = getSessionCookie(h.t, usr.ID, usr.Email)
	})

	h.client.Jar.SetCookies(u, []*http.Cookie{verifiedUser.sessionCookie})
	return h
}

func (h *httpRequest) withUnverifiedUser() *httpRequest {
	u, err := url.Parse(srv.URL)
	require.NoError(h.t, err)

	unverifiedUserOnce.Do(func() {
		unverifiedUser = userAuth{}

		// Create a test user
		usr, err := c.DB.User.CreateMockUser()
		require.NoError(h.t, err)
		unverifiedUser.id = usr.ID
		unverifiedUser.name = usr.Name.String
		unverifiedUser.email = usr.Email

		// Login and store session Cookie
		unverifiedUser.sessionCookie = getSessionCookie(h.t, usr.ID, usr.Email)
	})

	h.client.Jar.SetCookies(u, []*http.Cookie{unverifiedUser.sessionCookie})
	return h
}

func (h *httpRequest) withUser(userID int, email string) *httpRequest {
	u, err := url.Parse(srv.URL)
	require.NoError(h.t, err)

	// Login and store session Cookie
	sessionCookie := getSessionCookie(h.t, userID, email)

	h.client.Jar.SetCookies(u, []*http.Cookie{sessionCookie})
	return h
}

func (h *httpRequest) withBasicAuth(username, password string) *httpRequest {
	h.basicAuthUsername = username
	h.basicAuthPassword = password
	return h
}

func (h *httpRequest) setRoute(route string, params ...interface{}) *httpRequest {
	h.route = srv.URL + c.Web.Reverse(route, params...)
	return h
}

// setQueryString appends the query string to the route. Must be called after
// setRoute. Example query: `?param=value&key=value2`
func (h *httpRequest) setQueryString(query string) *httpRequest {
	h.route += query
	return h
}

func (h *httpRequest) get() *httpResponse {
	req, err := http.NewRequest(http.MethodGet, h.route, nil)
	require.NoError(h.t, err)

	if h.basicAuthUsername != "" {
		req.SetBasicAuth(h.basicAuthUsername, h.basicAuthPassword)
	}

	resp, err := h.client.Do(req)
	require.NoError(h.t, err)
	r := httpResponse{
		t:        h.t,
		Response: resp,
	}
	return &r
}

func (h *httpRequest) postForm(body url.Values) *httpResponse {
	body["csrf"] = []string{h.getCSRF()}

	req, err := http.NewRequest(http.MethodPost, h.route, strings.NewReader(body.Encode()))
	require.NoError(h.t, err)
	req.Header.Set("Content-Type", "application/x-www-form-urlencoded")

	if h.basicAuthUsername != "" {
		req.SetBasicAuth(h.basicAuthUsername, h.basicAuthPassword)
	}

	// Make the POST requests
	resp, err := h.client.Do(req)
	require.NoError(h.t, err)
	r := httpResponse{
		t:        h.t,
		Response: resp,
	}
	return &r
}

// postMulitpartForm helps with the process of creating and sending a
// multipart/form-data request. The provided callback will be called with a
// writer, allowing the addition of any necessary fields.
// This function will add the csrf field and close the writer.
func (h *httpRequest) postMultipartForm(setValuesFn func(*multipart.Writer)) *httpResponse {
	body := &bytes.Buffer{}
	writer := multipart.NewWriter(body)
	setValuesFn(writer)

	csrfWriter, err := writer.CreateFormField("csrf")
	require.NoError(h.t, err)
	_, err = csrfWriter.Write([]byte(h.getCSRF()))
	require.NoError(h.t, err)

	writer.Close()

	return h.post("multipart/form-data; boundary="+writer.Boundary(), body)
}

// post sends a post request to the test webserver. Unlike postForm, this method
// will not set the CSRF form value for you. You should call getCSRF and set it
// manually in the body.
func (h *httpRequest) post(contentType string, body io.Reader) *httpResponse {
	req, err := http.NewRequest(http.MethodPost, h.route, body)
	require.NoError(h.t, err)
	req.Header.Set("Content-Type", contentType)

	if h.basicAuthUsername != "" {
		req.SetBasicAuth(h.basicAuthUsername, h.basicAuthPassword)
	}

	resp, err := h.client.Do(req)
	require.NoError(h.t, err)
	r := httpResponse{
		t:        h.t,
		Response: resp,
	}
	return &r
}

// getCSRF makes a GET request to the test server to retrieve the CSRF token.
// Temporarily sets the route back to the homepage since all users can load it.
func (h *httpRequest) getCSRF() string {
	originalRoute := h.route
	defer func() {
		h.route = originalRoute
	}()

	h.setRoute(pages.Names.Home)

	// Make a get request to get the CSRF token
	doc := h.get().
		requireStatusCode(http.StatusOK).
		toDoc()

	// Extract the CSRF and include it in the POST request body
	csrf := doc.Find(`input[name="csrf"]`).First()
	token, exists := csrf.Attr("value")
	assert.True(h.t, exists)
	return token
}

type httpResponse struct {
	*http.Response
	t *testing.T
}

func (h *httpResponse) requireStatusCode(code int) *httpResponse {
	require.Equal(h.t, code, h.Response.StatusCode)
	return h
}

func (h *httpResponse) assertRedirect(route string, params ...interface{}) *httpResponse {
	return h.assertRedirectPath(c.Web.Reverse(route, params...))
}

func (h *httpResponse) assertRedirectPath(path string) *httpResponse {
	assert.Equal(h.t, path, h.Header.Get("Location"))
	return h
}

// getLocationID extracts the last segment in the URL path from the Location
// header, converts it to a number and returns it.
func (h *httpResponse) getLocationID() int {
	loc, err := h.Location()
	require.NoError(h.t, err)
	require.NotEmpty(h.t, loc)
	path := loc.Path

	locationRe := regexp.MustCompile(`/post/(\d+)/photos`)
	matches := locationRe.FindStringSubmatch(path)
	if len(matches) == 2 {
		id, err := strconv.Atoi(string(matches[1]))
		require.NoError(h.t, err)
		return id
	}

	id, err := strconv.Atoi(path[strings.LastIndex(path, "/")+1:])
	require.NoError(h.t, err)
	return id
}

func (h *httpResponse) body() string {
	body, err := io.ReadAll(h.Body)
	require.NoError(h.t, err)

	err = h.Body.Close()
	assert.NoError(h.t, err)

	return string(body)
}

func (h *httpResponse) toDoc() *goquery.Document {
	doc, err := goquery.NewDocumentFromReader(h.Body)
	require.NoError(h.t, err)
	err = h.Body.Close()
	assert.NoError(h.t, err)
	return doc
}

// getSessionCookie performs a login on the test server with the provided
// credentials and returns the resulting session cookie. If no cookie is found
// or if the login fails then the test will be failed.
func getSessionCookie(t *testing.T, userID int, userEmail string) *http.Cookie {
	r := request(t)
	jar := r.client.Jar

	token, err := c.Auth.GenerateJWTEmailAuthToken(userEmail)
	require.NoError(t, err)

	r.setRoute(pages.Names.User.LoginToken, userID, token).
		get().
		requireStatusCode(http.StatusOK)

	u, err := url.Parse(srv.URL)
	require.NoError(t, err)

	for _, cookie := range jar.Cookies(u) {
		if cookie.Name == authSessionName {
			return cookie
		}
	}

	t.Fatal("Unable to retrieve session cookie from cookie jar")
	return nil
}

func setMultipartField(t *testing.T, w *multipart.Writer, field, value string) {
	fw, err := w.CreateFormField(field)
	require.NoError(t, err)
	_, err = fw.Write([]byte(value))
	require.NoError(t, err)
}
