package routes

import (
	"bytes"
	"net/http"
	"path/filepath"
	"regexp"
	"strconv"
	"time"

	"github.com/labstack/echo/v4"

	"git.grosinger.net/tgrosinger/tonyslist/pkg/controller"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/errors"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
)

type (
	static struct {
		controller.Controller
	}
)

// postID-photoID_size.jpg
var postMatcher = regexp.MustCompile(`(\d+)-(\d+)_(\w).\w{3}`)

// event-eventID-photoID.jpg
var eventMatcher = regexp.MustCompile(`event-(\d+)-(\d+).\w{3}`)

func (c *static) Get(ctx echo.Context) error {
	filename := ctx.Param("filename")
	matches := postMatcher.FindStringSubmatch(filename)
	if len(matches) == 4 {
		return c.ServePostPhoto(ctx, matches[1], matches[2], matches[3])
	}

	matches = eventMatcher.FindStringSubmatch(filename)
	if len(matches) == 3 {
		return c.ServeEventPhoto(ctx, matches[1], matches[2])
	}

	return errors.ErrNotFound
}

func (c *static) ServeEventPhoto(ctx echo.Context, eventIDStr, photoIDStr string) error {
	eventID, err := strconv.Atoi(eventIDStr)
	if err != nil {
		return errors.ErrNotFound
	}

	photoID, err := strconv.Atoi(photoIDStr)
	if err != nil {
		return errors.ErrNotFound
	}

	photo, err := c.Container.DB.C.GetEventPhoto(ctx.Request().Context(), eventID)
	if err != nil {
		return errors.ErrNotFound
	}

	if photo.ID != photoID {
		return errors.ErrNotFound
	}

	sendImageBytes(ctx, photo.PreviewJpg, "jpg")
	return nil
}

func (c *static) ServePostPhoto(ctx echo.Context, postIDStr, photoIDStr, requestedSize string) error {
	postID, err := strconv.Atoi(postIDStr)
	if err != nil {
		return errors.ErrNotFound
	}

	photoID, err := strconv.Atoi(photoIDStr)
	if err != nil {
		return errors.ErrNotFound
	}

	photo, err := c.Container.DB.C.GetPhoto(ctx.Request().Context(), photoID)
	if err != nil {
		return errors.ErrNotFound
	}

	if photo.PostID != postID {
		return errors.ErrNotFound
	}

	// For each size, if we requested a smaller size but it was not available,
	// then serve the current size if it is available.
	if requestedSize == "t" && photo.ThumbnailJpg != nil {
		sendImageBytes(ctx, photo.ThumbnailJpg, getExtension(photo))
		return nil
	}
	if (requestedSize == "t" || requestedSize == "p") && photo.PreviewJpg != nil {
		sendImageBytes(ctx, photo.PreviewJpg, getExtension(photo))
		return nil
	}
	if photo.FullsizeJpg != nil {
		sendImageBytes(ctx, photo.FullsizeJpg, getExtension(photo))
		return nil
	}

	// No resizes are available yet, serve the original uploaded image if it exists.
	if photo.OriginalFilename.Valid {
		c.sendImageFile(ctx, photo.OriginalFilename.String)
		return nil
	}

	return errors.ErrNotFound
}

func getExtension(p sqlc.Photo) string {
	if p.AlternativeExt.Valid {
		return p.AlternativeExt.String
	}
	return "jpg"
}

func sendImageBytes(ctx echo.Context, data []byte, extension string) {
	w := ctx.Response()
	w.Header().Add("Content-Type", "image/"+extension)
	b := bytes.NewReader(data)
	http.ServeContent(w, ctx.Request(), "", time.Time{}, b)
}

func (c *static) sendImageFile(ctx echo.Context, path string) {
	dir := filepath.Join(c.Container.Config.Storage.RootDir,
		c.Container.Config.Storage.PhotoStorageDir)
	fullPath := filepath.Join(dir, path)
	http.ServeFile(ctx.Response().Writer, ctx.Request(), fullPath)
}
