package services

import (
	"fmt"
	"time"

	"git.grosinger.net/tgrosinger/tonyslist/config"
	"github.com/dgraph-io/ristretto"
)

type (
	// CacheClient is the client that allows you to interact with the cache
	CacheClient struct {
		cache *ristretto.Cache
	}

	// cacheSet handles chaining a set operation
	cacheSet struct {
		client     *CacheClient
		key        string
		group      string
		data       interface{}
		expiration time.Duration
	}

	// cacheGet handles chaining a get operation
	cacheGet struct {
		client *CacheClient
		key    string
		group  string
	}

	// cacheFlush handles chaining a flush operation
	cacheFlush struct {
		client *CacheClient
		key    string
		group  string
	}
)

// NewCacheClient creates a new cache client
func NewCacheClient(cfg *config.Config) (*CacheClient, error) {
	ristrettoCache, err := ristretto.NewCache(&ristretto.Config{
		NumCounters: cfg.Cache.NumCounters,
		MaxCost:     cfg.Cache.MaxCost,
		BufferItems: cfg.Cache.BufferItems,
	})
	if err != nil {
		return nil, err
	}

	return &CacheClient{
		cache: ristrettoCache,
	}, nil
}

// Close closes the connection to the cache
func (c *CacheClient) Close() error {
	c.cache.Close()
	return nil
}

// Set creates a cache set operation
func (c *CacheClient) Set() *cacheSet {
	return &cacheSet{
		client: c,
	}
}

// Get creates a cache get operation
func (c *CacheClient) Get() *cacheGet {
	return &cacheGet{
		client: c,
	}
}

// Wait blocks until the cache set buffers have cleared. This is useful for unit
// tests to give up control and allow the cache to process before checking that
// it stored properly. It should not need to be used in production.
func (c *CacheClient) Wait() {
	c.cache.Wait()
}

// Flush creates a cache flush operation
func (c *CacheClient) Flush() *cacheFlush {
	return &cacheFlush{
		client: c,
	}
}

// cacheKey formats a cache key with an optional group
func (c *CacheClient) cacheKey(group, key string) string {
	if group != "" {
		return fmt.Sprintf("%s::%s", group, key)
	}
	return key
}

// Key sets the cache key
func (c *cacheSet) Key(key string) *cacheSet {
	c.key = key
	return c
}

// Group sets the cache group
func (c *cacheSet) Group(group string) *cacheSet {
	c.group = group
	return c
}

// Data sets the data to cache
func (c *cacheSet) Data(data interface{}) *cacheSet {
	c.data = data
	return c
}

// Expiration sets the expiration duration of the cached data
func (c *cacheSet) Expiration(expiration time.Duration) *cacheSet {
	c.expiration = expiration
	return c
}

// Save saves the data in the cache
func (c *cacheSet) Save() bool {
	if c.key == "" {
		return false
	}

	// TODO: Use cache cost more intelligently

	if c.expiration != 0 {
		return c.client.cache.SetWithTTL(c.client.cacheKey(c.group, c.key), c.data, 1, c.expiration)
	} else {
		return c.client.cache.Set(c.client.cacheKey(c.group, c.key), c.data, 1)
	}
}

// Key sets the cache key
func (c *cacheGet) Key(key string) *cacheGet {
	c.key = key
	return c
}

// Group sets the cache group
func (c *cacheGet) Group(group string) *cacheGet {
	c.group = group
	return c
}

// Fetch fetches the data from the cache
func (c *cacheGet) Fetch() (interface{}, bool) {
	if c.key == "" {
		return nil, false
	}

	return c.client.cache.Get(c.client.cacheKey(c.group, c.key))
}

// Key sets the cache key
func (c *cacheFlush) Key(key string) *cacheFlush {
	c.key = key
	return c
}

// Group sets the cache group
func (c *cacheFlush) Group(group string) *cacheFlush {
	c.group = group
	return c
}

// Execute flushes the data from the cache
func (c *cacheFlush) Execute() {
	if c.key != "" {
		c.client.cache.Del(c.client.cacheKey(c.group, c.key))
	}
}
