package services

import (
	"context"
	"fmt"
	"path/filepath"

	"github.com/casbin/casbin/v2"
	"github.com/casbin/casbin/v2/model"
	fileadapter "github.com/casbin/casbin/v2/persist/file-adapter"
	"github.com/gammazero/workerpool"
	"github.com/labstack/echo/v4"
	"github.com/labstack/gommon/log"
	_ "github.com/mattn/go-sqlite3"
	"github.com/yuin/goldmark"
	"go.uber.org/zap"

	"git.grosinger.net/tgrosinger/tonyslist/config"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/auth"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/metrics"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/parser"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/services/mail"
	mailmock "git.grosinger.net/tgrosinger/tonyslist/pkg/services/mail/mock"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/services/mail/postmark"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/services/mail/types"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/services/oauth2"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/services/stripe"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/services/template"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/tus"
)

// Container contains all services used by the application and provides an easy way to handle dependency
// injection including within tests
type Container struct {
	Logger *zap.Logger

	// Validator stores a validator
	Validator *Validator

	// Sanitizer sanitizes user input before storing in the database
	Sanitizer *Sanitizer

	// Parser can be used to parse an input into different types used in the app
	Parser *parser.Parser

	// Web stores the web framework
	Web *echo.Echo

	// Config stores the application configuration
	Config *config.Config

	// Cache contains the cache client
	Cache *CacheClient

	DB *models.Client

	// Mail stores an email sending client
	Mail mail.Client

	Stripe stripe.Client

	// Auth stores an authentication client
	Auth *auth.Client

	// TemplateRenderer stores a service to easily render and cache templates
	TemplateRenderer *template.Renderer

	// MarkdownRenderer can render markdown in a safe manner. It will not render
	// any input HTML, instead just passing it through as text.
	// TODO: Should this also run through BlueMonday?
	MarkdownRenderer goldmark.Markdown

	// Tus is the wrapper for the Tus upload handler
	// https://github.com/tus/tusd
	Tus *tus.Client

	OAuth2 *oauth2.Server

	Metrics *metrics.Server

	// WorkerPool stores workerpool where jobs can be submitted.
	WorkerPool *workerpool.WorkerPool

	// Enforcer makes policy decisions based on the loaded config.
	Enforcer *casbin.Enforcer

	// Cron is the cron runner which executes periodic jobs.
	Cron *CronRunner
}

// NewContainer creates and initializes a new Container
func NewContainer(configPath string, logger *zap.Logger) *Container {
	c := new(Container)
	c.Logger = logger
	c.initConfig(configPath)
	c.initMetrics()
	c.initValidator()
	c.initSanitizer()
	c.initParser()
	c.initWeb()
	c.initCache()
	c.initDB()
	c.initAuth()
	c.initTemplateRenderer()
	c.initMarkdownRenderer()
	c.initWorkerPool()
	c.initTus()
	c.initOauth2()
	c.initMail()
	c.initStripe()
	c.initRBAC()
	c.initCron()
	return c
}

// Shutdown shuts the Container down and disconnects all connections
func (c *Container) Shutdown() error {
	c.WorkerPool.StopWait()

	c.Cron.Shutdown()

	c.Tus.Shutdown()

	if err := c.DB.Close(); err != nil {
		return err
	}
	if err := c.Cache.Close(); err != nil {
		return err
	}
	if err := c.Metrics.Shutdown(context.Background()); err != nil {
		return err
	}

	return nil
}

// initConfig initializes configuration
func (c *Container) initConfig(configPath string) {
	cfg, err := config.GetConfig(configPath)
	if err != nil {
		panic(fmt.Sprintf("failed to load config: %v", err))
	}
	c.Config = &cfg
}

func (c *Container) initMetrics() {
	c.Metrics = metrics.NewMetricsServer(c.Config)
	go c.Metrics.Run()
}

// initValidator initializes the validator
func (c *Container) initValidator() {
	c.Validator = NewValidator()
}

func (c *Container) initSanitizer() {
	c.Sanitizer = NewSanitizer()
}

func (c *Container) initParser() {
	c.Parser = parser.New()
}

// initWeb initializes the web framework
func (c *Container) initWeb() {
	c.Web = echo.New()
	c.Web.Validator = c.Validator

	c.Web.Logger.SetLevel(log.OFF)
}

// initCache initializes the cache
func (c *Container) initCache() {
	var err error
	if c.Cache, err = NewCacheClient(c.Config); err != nil {
		panic(err)
	}
}

func (c *Container) initDB() {
	var err error
	c.DB, err = models.New(c.Logger, c.Config)
	if err != nil {
		panic(err)
	}
}

// initAuth initializes the authentication client
func (c *Container) initAuth() {
	c.Auth = auth.New(c.Config, c.DB)
}

// initTemplateRenderer initializes the template renderer
func (c *Container) initTemplateRenderer() {
	c.TemplateRenderer = template.NewRenderer(c.Config)
}

func (c *Container) initMarkdownRenderer() {
	c.MarkdownRenderer = goldmark.New()
}

// initWorkerPool creates the worker pool
func (c *Container) initWorkerPool() {
	c.WorkerPool = workerpool.New(c.Config.App.MaxWorkers)
}

func (c *Container) initTus() {
	t, err := tus.New(c.Config, c.Auth, c.Logger, c.DB, c.WorkerPool)
	if err != nil {
		panic(err)
	}
	c.Tus = t
}

func (c *Container) initOauth2() {
	server, err := oauth2.New(c.Config, c.Logger)
	if err != nil {
		panic(err)
	}
	c.OAuth2 = server
}

// initMail initialize the mail client
func (c *Container) initMail() {
	var provider types.Provider
	if c.Config.App.Environment == config.EnvTest {
		provider = mailmock.New()
	} else {
		provider = postmark.New(c.Config.Mail.Token)
	}

	var err error
	c.Mail, err = mail.NewClient(c.Config, c.TemplateRenderer, c.WorkerPool, provider)
	if err != nil {
		panic(fmt.Sprintf("failed to create mail client: %v", err))
	}
}

func (c *Container) initStripe() {
	client := stripe.New(*c.DB.C, c.Config.Stripe.SecretKey)
	c.Stripe = client
}

// initRBAC creates the RBAC policy enforcer.
func (c *Container) initRBAC() {
	m := model.NewModel()
	err := m.LoadModelFromText(`
[request_definition]
r = sub, obj, act

[policy_definition]
p = sub, obj, act

[policy_effect]
e = some(where (p.eft == allow))

[matchers]
m = r.sub == p.sub && r.obj == p.obj && r.act == p.act
`)
	if err != nil {
		panic(fmt.Sprintf("failed to load policy model: %v", err))
	}

	policyPath := filepath.Join(c.Config.Storage.RootDir, c.Config.Storage.RBACPolicyPath)
	policy := fileadapter.NewAdapter(policyPath)

	e, err := casbin.NewEnforcer(m, policy)
	if err != nil {
		panic(fmt.Sprintf("failed to create policy enforcer: %v", err))
	}

	c.Enforcer = e
}

// initCron creates and launches the Cron runner in a goroutine.
func (c *Container) initCron() {
	c.Cron = &CronRunner{
		container: c,
	}
	go c.Cron.Run()
}
