package services

import (
	"sync"
	"time"

	"github.com/Unity-Technologies/go-cticker"
	"go.uber.org/zap"
)

// CronJob is a function that should be run on a schedule
type CronJob func(c *Container)

// CronRunner runs the registered jobs on a set frequency.
type CronRunner struct {
	metricsJobs []CronJob
	hourlyJobs  []CronJob
	dailyJobs   []CronJob

	container *Container

	wg         sync.WaitGroup
	stopSignal chan struct{}
}

// RegisterMetricsJobs adds a job that will be run on the metric refresh interval.
// Note however that it does not actually make an attempt to align with the
// metric refresh.
func (r *CronRunner) RegisterMetricsJobs(jobs ...CronJob) {
	r.metricsJobs = append(r.metricsJobs, jobs...)
}

// RegisterHourlyJobs adds a job that will be run at the top of every hour.
func (r *CronRunner) RegisterHourlyJobs(jobs ...CronJob) {
	r.hourlyJobs = append(r.hourlyJobs, jobs...)
}

// RegisterDailyJobs add a job that will be run sometime once a day, but not
// exactly at midnight.
func (r *CronRunner) RegisterDailyJobs(jobs ...CronJob) {
	r.dailyJobs = append(r.dailyJobs, jobs...)
}

// Run starts executing the registered jobs on their defined frequencies until
// Shutdown is called. Jobs added after Run is called will be executed at the
// next iteration of their defined frequency.
func (r *CronRunner) Run() {
	r.stopSignal = make(chan struct{})

	metrics := time.NewTicker(5 * time.Minute)
	daily := time.NewTicker(24 * time.Hour)
	defer daily.Stop()
	topOfTheHour := cticker.New(time.Hour, time.Minute)
	defer topOfTheHour.Stop()

	for {
		select {
		case <-r.stopSignal:
			return
		case <-metrics.C:
			r.runJobs("metrics", r.metricsJobs)
		case <-topOfTheHour.C:
			r.runJobs("hourly", r.hourlyJobs)
		case <-daily.C:
			r.runJobs("daily", r.dailyJobs)
		}
	}
}

// Shutdown stops executing new jobs and blocks until currently executing jobs
// complete.
func (r *CronRunner) Shutdown() {
	close(r.stopSignal)
	r.wg.Wait()
}

func (r *CronRunner) runJobs(name string, jobs []CronJob) {
	r.container.Logger.Info("Running cron jobs", zap.String("name", name))
	for _, job := range jobs {
		// Avoid for-loop gotcha
		// https://github.com/golang/go/wiki/CommonMistakes#using-goroutines-on-loop-iterator-variables
		j := job

		r.wg.Add(1)
		go func(j CronJob) {
			defer r.wg.Done()
			j(r.container)
		}(j)
	}
}
