package mail

import (
	"errors"
	"fmt"

	"github.com/gammazero/workerpool"
	"go.uber.org/zap"

	"git.grosinger.net/tgrosinger/tonyslist/config"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/services/mail/types"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/services/template"
)

type Client interface {
	SetProvider(types.Provider)
	Template(TemplateArgs) (*types.Mail, error)
	Send(logger *zap.Logger, mail []*types.Mail, callback func(types.Response))
}

type (
	// MailClient provides a client for sending email
	MailClient struct {
		// config stores application configuration
		config *config.Config

		provider types.Provider

		// templates stores the template renderer
		templates *template.Renderer

		workerpool *workerpool.WorkerPool
	}
)

// NewClient creates a new MailClient
func NewClient(cfg *config.Config, templates *template.Renderer,
	wp *workerpool.WorkerPool, provider types.Provider,
) (*MailClient, error) {
	return &MailClient{
		config:     cfg,
		templates:  templates,
		provider:   provider,
		workerpool: wp,
	}, nil
}

// SetProvider should only be used by unit tests.
func (m *MailClient) SetProvider(p types.Provider) {
	m.provider = p
}

type TemplateArgs struct {
	MessageStream types.MessageStream
	From          string
	To            string
	// ReplyTo is a comma separated list of emails that sets who the message
	// should be replied to.
	ReplyTo string
	Subject string
	// HTMLTemplate sets the template to be used to produce the body of the email
	// The template name should only include the filename without the extension or directory.
	// The template must reside within the emails sub-directory.
	// The funcmap will be automatically added to the template.
	// Use TemplateData() to supply the data that will be passed in to the template.
	HTMLTemplate string
	TextTemplate string
	TemplateData interface{}
	Attachments  []types.Attachment
	Headers      []types.Header
}

func (m *MailClient) Template(args TemplateArgs) (*types.Mail, error) {
	if args.MessageStream == "" {
		return nil, errors.New("email cannot be sent without a message stream")
	}
	if args.HTMLTemplate == "" && args.TextTemplate == "" {
		return nil, errors.New("email cannot be sent without a template")
	}
	if args.To == "" {
		return nil, errors.New("email must specify a recipient")
	}
	if args.Subject == "" {
		return nil, errors.New("email must specify a subject")
	}

	if args.From == "" {
		args.From = m.config.Mail.FromAddress
	}

	var htmlBody string
	if args.HTMLTemplate != "" {
		buf, err := m.templates.Parse().
			Group("mail").
			Key(args.HTMLTemplate).
			Base(args.HTMLTemplate).
			Files(fmt.Sprintf("emails/%s", args.HTMLTemplate)).
			Execute(args.TemplateData)
		if err != nil {
			return nil, err
		}
		htmlBody = buf.String()
	}

	var textBody string
	if args.TextTemplate != "" {
		buf, err := m.templates.Parse().
			Group("mail").
			Key(args.TextTemplate).
			Base(args.TextTemplate).
			Files(fmt.Sprintf("emails/%s", args.TextTemplate)).
			Execute(args.TemplateData)
		if err != nil {
			return nil, err
		}
		textBody = buf.String()
	}

	return &types.Mail{
		MessageStream: args.MessageStream,
		From:          args.From,
		To:            args.To,
		ReplyTo:       args.ReplyTo,
		Subject:       args.Subject,
		HTMLBody:      htmlBody,
		TextBody:      textBody,
		Attachments:   args.Attachments,
		Headers:       args.Headers,
	}, nil
}

func (m *MailClient) Send(logger *zap.Logger, mail []*types.Mail, callback func(types.Response)) {
	// Check if mail sending should be skipped
	if !m.config.Mail.Enable {
		for _, m := range mail {
			logger.Debug("skipping email send",
				zap.String("subject", m.Subject),
				zap.String("to", m.To))
		}
		return
	}

	logger.Debug("scheduling email to be sent by worker")
	m.workerpool.Submit(func() {
		response, err := m.provider.Send(mail)
		if err != nil {
			logger.Error("failed sending email", zap.Error(err))
		} else {
			logger.Debug("successfully sent email")
		}

		if callback != nil {
			for _, r := range response {
				callback(r)
			}
		}
	})
}
