package postmark

import (
	"bytes"
	"encoding/json"
	"fmt"
	"net/http"

	"git.grosinger.net/tgrosinger/tonyslist/pkg/generic"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/metrics"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/services/mail/types"
)

const (
	postmarkURL = "https://api.postmarkapp.com/email/batch"
)

type Provider struct {
	token string
}

func New(token string) *Provider {
	return &Provider{
		token: token,
	}
}

type txReqBody struct {
	From          string             `json:"From"`
	To            string             `json:"To"`
	ReplyTo       string             `json:"ReplyTo,omitempty"`
	Subject       string             `json:"Subject"`
	HtmlBody      string             `json:"HtmlBody,omitempty"`
	TextBody      string             `json:"TextBody,omitempty"`
	MessageStream string             `json:"MessageStream"`
	Attachments   []types.Attachment `json:"Attachments,omitempty"`
	Headers       []types.Header     `json:"Headers,omitempty"`
}

// txRespBody is the struct which will be sent back from Postmark when sending
// an email.
type txRespBody struct {
	ErrorCode   int    `json:"ErrorCode"`
	Message     string `json:"Message"`
	MessageID   string `json:"MessageID"`
	SubmittedAt string `json:"SubmittedAt"`
	To          string `json:"To"`
}

func (c *Provider) Send(m []*types.Mail) ([]types.Response, error) {
	responses := make([]types.Response, 0, len(m))
	for _, batch := range generic.Batches(m, 100) {
		resp, err := c.doBatchSend(batch)
		if err != nil {
			metrics.MailSendFailureCount.Inc()
			return nil, err
		}

		for _, r := range resp {
			responses = append(responses, types.Response{
				ErrorCode: r.ErrorCode,
				Message:   r.Message,
				MessageID: r.MessageID,
				To:        r.To,
			})
		}
	}
	return responses, nil
}

func (c *Provider) doBatchSend(m []*types.Mail) ([]txRespBody, error) {
	body := generic.Map(m, func(m *types.Mail) txReqBody {
		return txReqBody{
			From:          m.From,
			To:            m.To,
			ReplyTo:       m.ReplyTo,
			Subject:       m.Subject,
			HtmlBody:      m.HTMLBody,
			TextBody:      m.TextBody,
			MessageStream: string(m.MessageStream),
			Attachments:   m.Attachments,
			Headers:       m.Headers,
		}
	})

	bodyBytes, err := json.Marshal(body)
	if err != nil {
		return nil, err
	}
	b := bytes.NewBuffer(bodyBytes)

	req, err := http.NewRequest(http.MethodPost, postmarkURL, b)
	if err != nil {
		return nil, err
	}

	req.Header.Set("Accept", "application/json")
	req.Header.Set("Content-Type", "application/json")
	req.Header.Set("X-Postmark-Server-Token", c.token)

	client := http.Client{}
	resp, err := client.Do(req)
	if err != nil {
		return nil, err
	}

	var respParsed []txRespBody
	err = json.NewDecoder(resp.Body).Decode(&respParsed)
	resp.Body.Close()
	if err != nil {
		return nil, err
	}

	if resp.StatusCode != http.StatusOK {
		return nil, fmt.Errorf("unable to send email: %d - %+v", resp.StatusCode, respParsed)
	}

	return respParsed, nil
}
