package oauth2

import (
	"net/http"

	"github.com/go-oauth2/oauth2/v4"
	"github.com/go-oauth2/oauth2/v4/errors"
	"github.com/go-oauth2/oauth2/v4/manage"
	"github.com/go-oauth2/oauth2/v4/models"
	"github.com/go-oauth2/oauth2/v4/server"
	"github.com/go-oauth2/oauth2/v4/store"
	"go.uber.org/zap"

	"git.grosinger.net/tgrosinger/tonyslist/config"
)

type Server struct {
	OAuth2Server *server.Server
}

func New(cfg *config.Config, logger *zap.Logger) (*Server, error) {
	manager := manage.NewDefaultManager()

	// Tokens are stored in memory only
	manager.MustTokenStorage(store.NewMemoryTokenStore())

	// client memory store
	clientStore := store.NewClientStore()
	clientStore.Set(cfg.Discourse.ClientID, &models.Client{
		ID:     cfg.Discourse.ClientID,
		Secret: cfg.Discourse.ClientSecret,
		Domain: cfg.Discourse.Domain,
	})
	clientStore.Set(cfg.Wiki.ClientID, &models.Client{
		ID:     cfg.Wiki.ClientID,
		Secret: cfg.Wiki.ClientSecret,
		Domain: cfg.Wiki.Domain,
	})
	manager.MapClientStorage(clientStore)

	s := &Server{
		OAuth2Server: server.NewDefaultServer(manager),
	}

	s.OAuth2Server.SetAllowGetAccessRequest(true)
	s.OAuth2Server.SetClientInfoHandler(server.ClientFormHandler)
	s.OAuth2Server.SetUserAuthorizationHandler(s.userAuthorizationHandler)
	s.OAuth2Server.SetExtensionFieldsHandler(s.extensionFieldsHandler)

	s.OAuth2Server.SetInternalErrorHandler(func(err error) (re *errors.Response) {
		logger.Warn("internal oauth2 error", zap.Error(err))
		return
	})

	s.OAuth2Server.SetResponseErrorHandler(func(re *errors.Response) {
		logger.Warn("oauth response error", zap.Error(re.Error))
	})

	return s, nil
}

func (s *Server) userAuthorizationHandler(w http.ResponseWriter, r *http.Request) (string, error) {
	userID, ok := r.Context().Value("oauth-user-id").(string)
	if !ok {
		return "", errors.New("Missing authenticated user")
	}

	return userID, nil
}

func (s *Server) extensionFieldsHandler(ti oauth2.TokenInfo) map[string]interface{} {
	return map[string]interface{}{
		"user_id": ti.GetUserID(),
	}
}
