package stripe

import (
	"context"
	"database/sql"
	"fmt"

	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
	s "github.com/stripe/stripe-go/v78"
	"github.com/stripe/stripe-go/v78/billingportal/session"
	"github.com/stripe/stripe-go/v78/customer"
)

type stripe struct {
	db sqlc.Queries
}

type Client interface {
	GetStripeCustomerIDForEmail(ctx context.Context, email string, name sql.NullString) (string, error)
	UpdateCustomerName(ctx context.Context, customerID, name string) error
	GetSession(customerID, returnURL string) (string, error)
}

func New(db sqlc.Queries, secretKey string) Client {
	s.Key = secretKey

	return stripe{
		db: db,
	}
}

func (s stripe) GetStripeCustomerIDForEmail(ctx context.Context, email string, name sql.NullString) (string, error) {
	if email == "" {
		return "", fmt.Errorf("email is required")
	}

	user, err := s.db.GetUserByEmail(ctx, email)
	if err != nil && err != sql.ErrNoRows {
		return "", err
	} else if err == nil && user.StripeCustomerID.Valid {
		return user.StripeCustomerID.String, nil
	}

	customerID, err := lookupCustomer(email)
	if err != nil {
		return "", err
	}

	if customerID == "" {
		customerID, err = createCustomer(email, name)
		if err != nil {
			return "", err
		}
	}

	err = s.db.SetUserStripeCustomerID(ctx, sqlc.SetUserStripeCustomerIDParams{
		Email:            email,
		StripeCustomerID: sql.NullString{String: customerID, Valid: true},
	})
	return customerID, err
}

func (st stripe) UpdateCustomerName(ctx context.Context, customerID, name string) error {
	params := &s.CustomerParams{
		Name: s.String(name),
	}
	_, err := customer.Update(customerID, params)
	return err
}

func (st stripe) GetSession(customerID, returnURL string) (string, error) {
	params := &s.BillingPortalSessionParams{
		Customer:  s.String(customerID),
		ReturnURL: s.String(returnURL),
	}
	sess, err := session.New(params)
	if err != nil {
		return "", err
	}
	return sess.URL, nil
}

func lookupCustomer(email string) (string, error) {
	params := &s.CustomerSearchParams{
		SearchParams: s.SearchParams{
			Query: fmt.Sprintf("email:%q", email),
		},
	}
	result := customer.Search(params)
	if err := result.Err(); err != nil {
		return "", err
	}
	for result.Next() {
		return result.Customer().ID, nil
	}
	return "", nil
}

func createCustomer(email string, name sql.NullString) (string, error) {
	params := &s.CustomerParams{
		Email: s.String(email),
	}
	if name.Valid {
		params.Name = s.String(name.String)
	}
	result, err := customer.New(params)
	if err != nil {
		return "", err
	}

	return result.ID, err
}
