package template

import (
	"bytes"
	"errors"
	"fmt"
	"html"
	"html/template"
	"reflect"
	"sync"

	"git.grosinger.net/tgrosinger/tonyslist/config"
)

type (
	// Renderer provides a flexible and easy to use method of rendering simple templates or complex sets of
	// templates while also providing caching and/or hot-reloading depending on your current environment
	Renderer struct {
		// templateCache stores a cache of parsed page templates
		templateCache sync.Map

		// templatePath stores the complete path to the templates directory
		templatesPath string

		// config stores application configuration
		config *config.Config
	}

	// Parsed is a wrapper around parsed templates which are stored in the Renderer cache
	Parsed struct {
		// Template is the parsed template
		Template *template.Template

		// build stores the build data used to parse the template
		build *templateBuild
	}

	// templateBuild stores the build data used to parse a template
	templateBuild struct {
		group       string
		key         string
		base        string
		files       []string
		directories []string
	}

	// templateBuilder handles chaining a template parse operation
	templateBuilder struct {
		build    *templateBuild
		renderer *Renderer
	}
)

func FuncMap() template.FuncMap {
	return template.FuncMap{
		"empty":    empty,
		"unescape": unescape,
	}
}

// Retrieved from Sprig
// https://github.com/Masterminds/sprig/blob/581758eb7d96ae4d113649668fa96acc74d46e7f/defaults.go#L35
func empty(given interface{}) bool {
	g := reflect.ValueOf(given)
	if !g.IsValid() {
		return true
	}

	// Basically adapted from text/template.isTrue
	switch g.Kind() {
	default:
		return g.IsNil()
	case reflect.Array, reflect.Slice, reflect.Map, reflect.String:
		return g.Len() == 0
	case reflect.Bool:
		return !g.Bool()
	case reflect.Complex64, reflect.Complex128:
		return g.Complex() == 0
	case reflect.Int, reflect.Int8, reflect.Int16, reflect.Int32, reflect.Int64:
		return g.Int() == 0
	case reflect.Uint, reflect.Uint8, reflect.Uint16, reflect.Uint32, reflect.Uint64, reflect.Uintptr:
		return g.Uint() == 0
	case reflect.Float32, reflect.Float64:
		return g.Float() == 0
	case reflect.Struct:
		return false
	}
}

// unescape will unescape the input text. This is safe to use in templates
// because we sanitize the inputs from users before storing to the database, and
// we escape strings again when templating. Essentially this prevents double
// escaping.
func unescape(s string) string {
	return html.UnescapeString(s)
}

// NewRenderer creates a new TemplateRenderer
func NewRenderer(cfg *config.Config) *Renderer {
	t := &Renderer{
		templateCache: sync.Map{},
		config:        cfg,
		templatesPath: config.TemplateDir(cfg),
	}

	return t
}

// Parse creates a template build operation
func (t *Renderer) Parse() *templateBuilder {
	return &templateBuilder{
		renderer: t,
		build:    &templateBuild{},
	}
}

// GetTemplatesPath gets the complete path to the templates directory
func (t *Renderer) GetTemplatesPath() string {
	return t.templatesPath
}

// getCacheKey gets a cache key for a given group and ID
func (t *Renderer) getCacheKey(group, key string) string {
	if group != "" {
		return fmt.Sprintf("%s:%s", group, key)
	}
	return key
}

// parse parses a set of templates and caches them for quick execution
// If the application environment is set to local, the cache will be bypassed and templates will be
// parsed upon each request so hot-reloading is possible without restarts.
// Also included will be the function map provided by the funcmap package.
func (t *Renderer) parse(build *templateBuild) (*Parsed, error) {
	switch {
	case len(build.files) == 0 && len(build.directories) == 0:
		return nil, errors.New("cannot parse template without files or directories")
	case build.base == "":
		return nil, errors.New("cannot parse template without base")
	}

	if t.config.App.Environment != config.EnvLocal && build.key != "" {
		tp, err := t.Load(build.group, build.key)
		if err == nil {
			// Successfully loaded cached template
			return tp, nil
		}
	}

	// Could not load from template, so we'll generate now
	var err error

	// Initialize the parsed template with the function map
	parsed := template.New(build.base + config.TemplateExt).
		Funcs(FuncMap())

	// Parse all files provided
	if len(build.files) > 0 {
		for k, v := range build.files {
			build.files[k] = fmt.Sprintf("%s/%s%s", t.templatesPath, v, config.TemplateExt)
		}

		parsed, err = parsed.ParseFiles(build.files...)
		if err != nil {
			return nil, err
		}
	}

	// Parse all templates within the provided directories
	for _, dir := range build.directories {
		dir = fmt.Sprintf("%s/%s/*%s", t.templatesPath, dir, config.TemplateExt)
		parsed, err = parsed.ParseGlob(dir)
		if err != nil {
			return nil, err
		}
	}

	// Store the template so this process only happens once
	tp := &Parsed{
		Template: parsed,
		build:    build,
	}

	// Cache the parsed template if we have a key to cache under
	if build.key != "" {
		cacheKey := t.getCacheKey(build.group, build.key)
		t.templateCache.Store(cacheKey, tp)
	}

	return tp, nil
}

// Load loads a template from the cache
func (t *Renderer) Load(group, key string) (*Parsed, error) {
	load, ok := t.templateCache.Load(t.getCacheKey(group, key))
	if !ok {
		return nil, errors.New("uncached page template requested")
	}

	tmpl, ok := load.(*Parsed)
	if !ok {
		return nil, errors.New("unable to cast cached template")
	}

	return tmpl, nil
}

// Execute executes a template with the given data and provides the output
func (t *Parsed) Execute(data interface{}) (*bytes.Buffer, error) {
	if t.Template == nil {
		return nil, errors.New("cannot execute template: template not initialized")
	}

	buf := new(bytes.Buffer)
	err := t.Template.ExecuteTemplate(buf, t.build.base+config.TemplateExt, data)
	if err != nil {
		return nil, err
	}

	return buf, nil
}

// Group sets the cache group for the template being built
func (t *templateBuilder) Group(group string) *templateBuilder {
	t.build.group = group
	return t
}

// Key sets the cache key for the template being built
func (t *templateBuilder) Key(key string) *templateBuilder {
	t.build.key = key
	return t
}

// Base sets the name of the base template to be used during template parsing and execution.
// This should be only the file name without a directory or extension.
func (t *templateBuilder) Base(base string) *templateBuilder {
	t.build.base = base
	return t
}

// Files sets a list of template files to include in the parse.
// This should not include the file extension and the paths should be relative to the templates directory.
func (t *templateBuilder) Files(files ...string) *templateBuilder {
	t.build.files = files
	return t
}

// Directories sets a list of directories that all template files within will be parsed.
// The paths should be relative to the templates directory.
func (t *templateBuilder) Directories(directories ...string) *templateBuilder {
	t.build.directories = directories
	return t
}

// Store parsed the templates and stores them in the cache
func (t *templateBuilder) Store() (*Parsed, error) {
	return t.renderer.parse(t.build)
}

// Execute executes the template with the given data.
// If the template has not already been cached, this will parse and cache the template
func (t *templateBuilder) Execute(data interface{}) (*bytes.Buffer, error) {
	tp, err := t.Store()
	if err != nil {
		return nil, err
	}

	return tp.Execute(data)
}
