package services

import (
	"errors"
	"testing"

	"github.com/asaskevich/govalidator"
	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
)

func TestValidator(t *testing.T) {
	testcases := []struct {
		Name           string
		Input          interface{}
		HasErrors      bool
		ExpectedErrors []govalidator.Error
	}{
		{
			Name: "Missing required field",
			Input: struct {
				Value string `valid:"required"`
			}{},
			HasErrors: true,
			ExpectedErrors: []govalidator.Error{
				{
					Name:      "Value",
					Err:       errors.New("non zero value required"),
					Validator: "required",
				},
			},
		},
		{
			Name: "Has required field",
			Input: struct {
				Value string `valid:"required"`
			}{
				Value: "a",
			},
			HasErrors: false,
		},
		{
			Name: "Has phone number",
			Input: struct {
				Value string `valid:"required,phone"`
			}{
				Value: "123-456-7890",
			},
			HasErrors: false,
		},
		{
			Name: "Optional phone field not set",
			Input: struct {
				Value string `valid:"optional,phone"`
			}{},
			HasErrors: false,
		},
		{
			Name: "Has invalid phone number",
			Input: struct {
				Value string `valid:"required,phone"`
			}{
				Value: "123-abc-7890",
			},
			HasErrors: true,
			ExpectedErrors: []govalidator.Error{
				{
					Name:      "Value",
					Err:       errors.New("123-abc-7890 does not validate as phone"),
					Validator: "phone",
				},
			},
		},
		{
			Name: "Has money field",
			Input: struct {
				Value string `valid:"required,money"`
			}{
				Value: "123.12",
			},
			HasErrors: false,
		},
		{
			Name: "Has invalid money field",
			Input: struct {
				Value string `valid:"required,money"`
			}{
				Value: "12.test",
			},
			HasErrors: true,
			ExpectedErrors: []govalidator.Error{
				{
					Name:      "Value",
					Err:       errors.New("12.test does not validate as money"),
					Validator: "money",
				},
			},
		},
		{
			Name: "Has money field with too many decimals",
			Input: struct {
				Value string `valid:"required,money"`
			}{
				Value: "12.123",
			},
			HasErrors: true,
			ExpectedErrors: []govalidator.Error{
				{
					Name:      "Value",
					Err:       errors.New("12.123 does not validate as money"),
					Validator: "money",
				},
			},
		},
		{
			Name: "Has money field with comma",
			Input: struct {
				Value string `valid:"required,money"`
			}{
				Value: "1,200.54",
			},
			HasErrors: false,
		},
		{
			Name: "Has 1.25 bathrooms",
			Input: struct {
				Value string `valid:"bathroomCount,optional"`
			}{
				Value: "1.25",
			},
			HasErrors: false,
		},
		{
			Name: "Has 1.5 bathrooms",
			Input: struct {
				Value string `valid:"bathroomCount,optional"`
			}{
				Value: "1.5",
			},
			HasErrors: false,
		},
		{
			Name: "Has 1.75 bathrooms",
			Input: struct {
				Value string `valid:"bathroomCount,optional"`
			}{
				Value: "1.75",
			},
			HasErrors: false,
		},
		{
			Name: "Has 20.5 bathrooms",
			Input: struct {
				Value string `valid:"bathroomCount,optional"`
			}{
				Value: "20.5",
			},
			HasErrors: false,
		},
		{
			Name: "Has 0.25 bathrooms",
			Input: struct {
				Value string `valid:"bathroomCount,optional"`
			}{
				Value: "0.25",
			},
			HasErrors: false,
		},
		{
			Name: "Has .25 bathrooms",
			Input: struct {
				Value string `valid:"bathroomCount,optional"`
			}{
				Value: ".25",
			},
			HasErrors: false,
		},
		{
			Name: "Has invalid .23 bathrooms",
			Input: struct {
				Value string `valid:"bathroomCount,optional"`
			}{
				Value: ".23",
			},
			HasErrors: true,
			ExpectedErrors: []govalidator.Error{
				{
					Name:      "Value",
					Err:       errors.New(".23 does not validate as bathroomCount"),
					Validator: "bathroomCount",
				},
			},
		},
		{
			Name: "Has invalid 1.73 bathrooms",
			Input: struct {
				Value string `valid:"bathroomCount,optional"`
			}{
				Value: "1.73",
			},
			HasErrors: true,
			ExpectedErrors: []govalidator.Error{
				{
					Name:      "Value",
					Err:       errors.New("1.73 does not validate as bathroomCount"),
					Validator: "bathroomCount",
				},
			},
		},
	}

	for _, tc := range testcases {
		t.Run(tc.Name, func(t *testing.T) {
			resultErr := c.Validator.Validate(tc.Input)
			if tc.HasErrors {
				require.Error(t, resultErr)
				errs, ok := resultErr.(govalidator.Errors)
				require.True(t, ok)
				require.Len(t, errs, len(tc.ExpectedErrors))
				for i, err := range errs {
					// There could be nested govalidator.Errors
					if ve, ok := err.(govalidator.Errors); ok {
						assert.Len(t, ve.Errors(), 1)
						err = ve.Errors()[0]
					}

					ve, ok := err.(govalidator.Error)
					if ok {
						require.Equal(t, tc.ExpectedErrors[i].Name, ve.Name)
						require.Equal(t, tc.ExpectedErrors[i].Error(), ve.Error())
						require.Equal(t, tc.ExpectedErrors[i].Validator, ve.Validator)
					} else {
						require.Equal(t, tc.ExpectedErrors[i].Error(), err)
					}
				}
			} else {
				require.NoError(t, resultErr)
			}
		})
	}
}
