package tasks

import (
	"context"
	"encoding/json"
	"errors"
	"fmt"
	"io"
	"math/rand"
	"net/http"
	"net/url"

	"go.uber.org/zap"

	"git.grosinger.net/tgrosinger/tonyslist/config"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/conv"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
)

// TODO: Consider using OpenCage instead of MapBox
// https://opencagedata.com/

type Location struct {
	Latitude  float64
	Longitude float64
}

// Format returns the location formatted as "longitude,latitude"
func (l Location) Format() string {
	return fmt.Sprintf("%f,%f", l.Longitude, l.Latitude)
}

// Jitter returns a new location which has been offset a random amount within
// approximately a 0.75mi radius. This is accomplished by adding a random amount
// between -0.01 and 0.01 to both coordinates. This in effect creates a square
// area in which the location can be moved to.
func (l Location) Jitter() Location {
	latRand := (rand.Float64() - 0.5) / 50
	longRand := (rand.Float64() - 0.5) / 50
	return Location{
		Latitude:  l.Latitude + latRand,
		Longitude: l.Longitude + longRand,
	}
}

// ForwardGeocodingResults is a small subset of the response from a request to
// the Mapbox Forward Geocoding API.
type ForwardGeocodingResults struct {
	Features []struct {
		Center []float64 `json:"center"`
	} `json:"features"`
}

func lookupAddress(cfg *config.Config, address string) (Location, error) {
	url := fmt.Sprintf(
		"https://api.mapbox.com/geocoding/v5/mapbox.places/%s.json?bbox=%s&types=address&access_token=%s",
		url.PathEscape(address),
		cfg.Mapbox.BoundingBox,
		cfg.Mapbox.Token,
	)

	resp, err := http.Get(url)
	if err != nil {
		return Location{}, fmt.Errorf("failed to lookup address: %w", err)
	}

	var data ForwardGeocodingResults
	err = json.NewDecoder(resp.Body).Decode(&data)
	resp.Body.Close()
	if err != nil {
		return Location{}, fmt.Errorf("unable to decode response from geo-provider: %w", err)
	}

	// Features are ordered in decreasing relevance.
	// Only consider the first one that has all the required information.
	for _, feature := range data.Features {
		if len(feature.Center) != 2 {
			// malformed location
			continue
		}

		return Location{
			Longitude: feature.Center[0],
			Latitude:  feature.Center[1],
		}, nil
	}

	return Location{}, errors.New("did not find a matching location for this address")
}

// getMapURL returns an authenticated URL to retrieve a static image of the
// location using the bounding box and resolution from the config.
func getMapURL(cfg *config.Config, loc Location) string {
	return fmt.Sprintf(
		"https://api.mapbox.com/styles/v1/mapbox/streets-v11/static/pin-s+555555(%s)/[%s]/%s?logo=false&access_token=%s",
		loc.Format(),
		cfg.Mapbox.BoundingBox,
		cfg.Mapbox.Resolution,
		cfg.Mapbox.Token,
	)
}

// downloadMapImage retrieves the image at the given location and stores it in
// the filesystem along side the other photos. A reference to the map file is
// stored in the database with the post.
func downloadMapImage(cfg *config.Config, logger *zap.Logger, db *models.Client,
	ctx context.Context, location Location, postID int,
) {
	mapURL := getMapURL(cfg, location)
	logger.Info("requesting map image", zap.String("map_url", mapURL))
	resp, err := http.Get(mapURL)
	if err != nil {
		logger.Warn("unable to request map of location", zap.Error(err))
		return
	}
	defer resp.Body.Close()

	if resp.StatusCode != http.StatusOK {
		body, _ := io.ReadAll(resp.Body)
		logger.Warn("unable to retrieve image for housing post",
			zap.String("response_body", string(body)),
			zap.Int("response_code", resp.StatusCode))
		return
	}

	mapBytes, err := io.ReadAll(resp.Body)
	if err != nil {
		logger.Warn("unable to read image for housing map download",
			zap.Error(err))
		return
	}

	err = db.C.CreateHousingMap(ctx, sqlc.CreateHousingMapParams{
		PostID:         postID,
		FullsizeJpg:    mapBytes,
		AlternativeExt: conv.StringToNullString("png"),
	})
	if err != nil {
		logger.Warn("unable to store housing map in database",
			zap.Int("post_id", postID),
			zap.Error(err))
		return
	}

	logger.Info("map downloaded successfully")
}

func DownloadHousingMap(cfg *config.Config, logger *zap.Logger, db *models.Client,
	street, zipcode string, housingForRentID int,
) {
	ctx := context.TODO()
	logger = logger.With(zap.Int("housing_for_rent_id", housingForRentID))
	logger.Info("downloading map of housing post")

	if street == "" || zipcode == "" {
		logger.Debug("missing address parts, not looking up map")
		return
	}

	address := fmt.Sprintf("%s, %s", street, zipcode)
	logger = logger.With(zap.String("address", address))

	location, err := lookupAddress(cfg, address)
	if err != nil {
		logger.Warn("unable to lookup address for retrieving map", zap.Error(err))
		return
	}
	location = location.Jitter()

	downloadMapImage(cfg, logger, db, ctx, location, housingForRentID)
}
