package tasks

import (
	"context"
	"fmt"
	"os"
	"os/exec"
	"path/filepath"
	"time"

	"github.com/prometheus/client_golang/prometheus"
	"go.uber.org/zap"

	"git.grosinger.net/tgrosinger/tonyslist/config"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/metrics"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
)

func ResizeEventImage(cfg *config.Config, logger *zap.Logger, db *models.Client, eventID int, inputFilename string) {
	start := time.Now()

	logger = logger.With(
		zap.String("input_filename", inputFilename),
	)
	logger.Info("Beginning photo resize")

	outputDir := filepath.Join(cfg.Storage.RootDir, cfg.Storage.PhotoStorageDir)
	inputFilepath := filepath.Join(outputDir, inputFilename)

	outFilepath := filepath.Join(outputDir, fmt.Sprintf("event-%d-p.jpg", eventID))
	defer func() {
		err := os.Remove(outFilepath)
		if err != nil {
			logger.Warn("failed to cleanup temp file when converting photo",
				zap.String("filepath", outFilepath),
				zap.Error(err))
		}
	}()

	/* #nosec */
	err := exec.Command(vipsthumbnail, inputFilepath, "--size="+maxPreviewImageDimension, "-o", outFilepath+"[optimize_coding,strip]").Run()
	if err != nil {
		logger.Warn("failed to vips preview resize", zap.Error(err))
		metrics.TasksImageResizeFailureCount.
			With(prometheus.Labels{metrics.LabelError: metricErrVipsResize}).
			Inc()
		return
	}

	data, err := os.ReadFile(outFilepath)
	if err != nil {
		logger.Warn("failed to read bytes of resized photo",
			zap.String("filepath", outFilepath),
			zap.Error(err))
		metrics.TasksImageResizeFailureCount.
			With(prometheus.Labels{metrics.LabelError: metricErrVipsResize}).
			Inc()
		return
	}

	err = db.C.DeleteEventPhoto(context.Background(), eventID)
	if err != nil {
		logger.Warn("failed to delete previous even photos",
			zap.Error(err))
	}
	err = db.C.SetEventPhoto(context.Background(), sqlc.SetEventPhotoParams{
		EventID:    eventID,
		PreviewJpg: data,
	})
	if err != nil {
		logger.Warn("failed to store resized preview size photo to database",
			zap.Error(err))
		metrics.TasksImageResizeFailureCount.
			With(prometheus.Labels{metrics.LabelError: metricErrVipsResize}).
			Inc()
		return
	}

	err = os.Remove(inputFilepath)
	if err != nil {
		logger.Warn("failed to delete the input photo",
			zap.Error(err))
		metrics.TasksImageResizeFailureCount.
			With(prometheus.Labels{metrics.LabelError: metricErrFilesystem}).
			Inc()
	}

	logger.Info("Finished photo resize")

	ms := time.Since(start).Milliseconds()
	metrics.TasksImageResizeTiming.Observe(float64(ms))
}
