package tasks

import (
	"context"
	"fmt"
	"os"
	"os/exec"
	"path/filepath"
	"strconv"
	"strings"
	"sync"
	"time"

	"github.com/prometheus/client_golang/prometheus"
	"go.uber.org/zap"

	"git.grosinger.net/tgrosinger/tonyslist/config"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/metrics"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
)

const (
	exiftool      = "exiftool"
	vips          = "vips"
	vipsthumbnail = "vipsthumbnail"
	vipsheader    = "vipsheader"

	maxPreviewImageDimension  = "560"
	maxFullSizeImageDimension = 2000

	metricErrFilesystem = "filesystem_error"
	metricErrDatabase   = "database_error"
	metricErrVipsMeta   = "vips_metadata_error"
	metricErrVipsResize = "vips_resize_error"
)

// ProcessTask converts the images associated with the post in the payload into
// more appropriate sizes.
func ResizePostImages(cfg *config.Config, logger *zap.Logger,
	db *models.Client, photoID int, inputFilename string,
) {
	start := time.Now()

	logger = logger.With(
		zap.Int("photo_id", photoID),
		zap.String("input_filename", inputFilename),
	)
	logger.Info("Beginning photo resize")

	outputDir := filepath.Join(cfg.Storage.RootDir, cfg.Storage.PhotoStorageDir)
	inputFilepath := filepath.Join(outputDir, inputFilename)

	wg := sync.WaitGroup{}
	wg.Add(3)

	go func() {
		generateFullSize(logger, db.C, photoID, inputFilepath, outputDir)
		wg.Done()
	}()
	go func() {
		generatePreviewSize(logger, db.C, photoID, inputFilepath, outputDir)
		wg.Done()
	}()
	go func() {
		generateThumbnailSize(logger, db.C, photoID, inputFilepath, outputDir)
		wg.Done()
	}()

	wg.Wait()

	err := os.Remove(inputFilepath)
	if err != nil {
		logger.Warn("failed to delete the input photo",
			zap.Int("photo_id", photoID), zap.Error(err))
		metrics.TasksImageResizeFailureCount.
			With(prometheus.Labels{metrics.LabelError: metricErrFilesystem}).
			Inc()
	}

	logger.Info("Finished photo resize")

	ms := time.Since(start).Milliseconds()
	metrics.TasksImageResizeTiming.Observe(float64(ms))
}

// generateFullSize resizes the input image to a maximum of 2000px in width and
// saves the result as a jpg.
func generateFullSize(logger *zap.Logger, db *sqlc.Queries, photoID int, inputFilepath, outputDir string) {
	height, err := getHeight(inputFilepath)
	if err != nil {
		logger.Warn("failed to find image height", zap.Error(err))
		metrics.TasksImageResizeFailureCount.
			With(prometheus.Labels{metrics.LabelError: metricErrVipsMeta}).
			Inc()
		return
	}

	width, err := getWidth(inputFilepath)
	if err != nil {
		logger.Warn("failed to find image width", zap.Error(err))
		metrics.TasksImageResizeFailureCount.
			With(prometheus.Labels{metrics.LabelError: metricErrVipsMeta}).
			Inc()
		return
	}

	scaleFactor := strconv.FormatFloat(calcScale(width, maxFullSizeImageDimension), 'f', 2, 64)

	logger.Info("Resizing image to full size",
		zap.Int("original_width", width),
		zap.Int("original_height", height),
		zap.String("scale_factor", scaleFactor))

	outFilepath := filepath.Join(outputDir, fmt.Sprintf("%d-f.jpg", photoID))
	defer func() {
		err = os.Remove(outFilepath)
		if err != nil {
			logger.Warn("failed to cleanup temp file when converting photo",
				zap.Int("photo_id", photoID),
				zap.String("filepath", outFilepath),
				zap.Error(err))
		}
	}()

	err = exec.Command(vips, "resize", inputFilepath, outFilepath, scaleFactor).Run()
	if err != nil {
		logger.Warn("failed to vips fullsize resize",
			zap.Int("photo_id", photoID),
			zap.Error(err))
		metrics.TasksImageResizeFailureCount.
			With(prometheus.Labels{metrics.LabelError: metricErrVipsResize}).
			Inc()
		return
	}

	stripMetadata(logger, outFilepath)

	data, err := os.ReadFile(outFilepath)
	if err != nil {
		logger.Warn("failed to read bytes of resized photo",
			zap.Int("photo_id", photoID),
			zap.String("filepath", outFilepath),
			zap.Error(err))
		metrics.TasksImageResizeFailureCount.
			With(prometheus.Labels{metrics.LabelError: metricErrVipsResize}).
			Inc()
		return
	}

	err = db.SetPhotoFullsize(context.Background(), sqlc.SetPhotoFullsizeParams{
		ID:          photoID,
		FullsizeJpg: data,
	})
	if err != nil {
		logger.Warn("failed to store resized full size photo to database",
			zap.Int("photo_id", photoID),
			zap.Error(err))
		metrics.TasksImageResizeFailureCount.
			With(prometheus.Labels{metrics.LabelError: metricErrVipsResize}).
			Inc()
		return
	}
}

// generatePreviewSize resizes the input image to a maximum of 560px in width and
// saves the result as a jpg.
func generatePreviewSize(logger *zap.Logger, db *sqlc.Queries, photoID int, inputFilepath, outputDir string) {
	logger.Info("Resizing image to preview size")

	outFilepath := filepath.Join(outputDir, fmt.Sprintf("%d-p.jpg", photoID))
	defer func() {
		err := os.Remove(outFilepath)
		if err != nil {
			logger.Warn("failed to cleanup temp file when converting photo",
				zap.Int("photo_id", photoID),
				zap.String("filepath", outFilepath),
				zap.Error(err))
		}
	}()

	/* #nosec */
	err := exec.Command(vipsthumbnail, inputFilepath, "--size="+maxPreviewImageDimension, "-o", outFilepath+"[optimize_coding,strip]").Run()
	if err != nil {
		logger.Warn("failed to vips preview resize", zap.Error(err))
		metrics.TasksImageResizeFailureCount.
			With(prometheus.Labels{metrics.LabelError: metricErrVipsResize}).
			Inc()
		return
	}

	data, err := os.ReadFile(outFilepath)
	if err != nil {
		logger.Warn("failed to read bytes of resized photo",
			zap.Int("photo_id", photoID),
			zap.String("filepath", outFilepath),
			zap.Error(err))
		metrics.TasksImageResizeFailureCount.
			With(prometheus.Labels{metrics.LabelError: metricErrVipsResize}).
			Inc()
		return
	}

	err = db.SetPhotoPreview(context.Background(), sqlc.SetPhotoPreviewParams{
		ID:         photoID,
		PreviewJpg: data,
	})
	if err != nil {
		logger.Warn("failed to store resized preview size photo to database",
			zap.Int("photo_id", photoID),
			zap.Error(err))
		metrics.TasksImageResizeFailureCount.
			With(prometheus.Labels{metrics.LabelError: metricErrVipsResize}).
			Inc()
		return
	}
}

// generateThumbnailSize resizes the input image to a maximum of 200px width and
// saves the result as a jpg.
func generateThumbnailSize(logger *zap.Logger, db *sqlc.Queries, photoID int, inputFilepath, outputDir string) {
	logger.Info("Resizing image to thumbnail size")

	outFilepath := filepath.Join(outputDir, fmt.Sprintf("%d-t.jpg", photoID))
	defer func() {
		err := os.Remove(outFilepath)
		if err != nil {
			logger.Warn("failed to cleanup temp file when converting photo",
				zap.Int("photo_id", photoID),
				zap.String("filepath", outFilepath),
				zap.Error(err))
		}
	}()

	/* #nosec */
	err := exec.Command(vipsthumbnail, inputFilepath, "--smartcrop=attention", "--size=200", "-o", outFilepath+"[optimize_coding,strip]").Run()
	if err != nil {
		logger.Warn("failed to vips thumbnail resize", zap.Error(err))
		metrics.TasksImageResizeFailureCount.
			With(prometheus.Labels{metrics.LabelError: metricErrVipsResize}).
			Inc()
		return
	}

	data, err := os.ReadFile(outFilepath)
	if err != nil {
		logger.Warn("failed to read bytes of resized photo",
			zap.Int("photo_id", photoID),
			zap.String("filepath", outFilepath),
			zap.Error(err))
		metrics.TasksImageResizeFailureCount.
			With(prometheus.Labels{metrics.LabelError: metricErrVipsResize}).
			Inc()
		return
	}

	err = db.SetPhotoThumbnail(context.Background(), sqlc.SetPhotoThumbnailParams{
		ID:           photoID,
		ThumbnailJpg: data,
	})
	if err != nil {
		logger.Warn("failed to store resized thumbnail photo to database",
			zap.Int("photo_id", photoID),
			zap.Error(err))
		metrics.TasksImageResizeFailureCount.
			With(prometheus.Labels{metrics.LabelError: metricErrVipsResize}).
			Inc()
		return
	}
}

func stripMetadata(logger *zap.Logger, filepath string) {
	err := exec.Command(exiftool, "-overwrite_original", "-All=", "-tagsfromfile", "@", "-exif:Orientation", filepath).Run()
	if err != nil {
		logger.Error("unable to strip image metadata",
			zap.String("filepath", filepath), zap.Error(err))
	}
}

func getWidth(filepath string) (int, error) {
	width, err := exec.Command(vipsheader, "-f", "width", filepath).Output()
	if err != nil {
		return -1, err
	}

	return strconv.Atoi(strings.TrimSpace(string(width)))
}

func getHeight(filepath string) (int, error) {
	width, err := exec.Command(vipsheader, "-f", "height", filepath).Output()
	if err != nil {
		return -1, err
	}

	return strconv.Atoi(strings.TrimSpace(string(width)))
}

func calcScale(width, maxWidth int) float64 {
	return float64(maxWidth) / float64(width)
}

func removeExtension(filepath string) string {
	idx := strings.LastIndex(filepath, ".")
	if idx == -1 {
		return filepath
	}

	return filepath[:idx]
}
