package tus

import (
	"context"
	"errors"
	"fmt"
	"os"
	"path/filepath"
	"strconv"

	"github.com/gammazero/workerpool"
	"github.com/tus/tusd/pkg/filestore"
	tusd "github.com/tus/tusd/pkg/handler"
	"go.uber.org/zap"

	"git.grosinger.net/tgrosinger/tonyslist/config"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/auth"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/conv"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/metrics"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/tasks"
)

type Client struct {
	Handler *tusd.Handler

	cfg        *config.Config
	auth       *auth.Client
	logger     *zap.Logger
	db         *models.Client
	workerpool *workerpool.WorkerPool

	storagePath string

	stopSignal chan struct{}
}

func New(cfg *config.Config, auth *auth.Client, logger *zap.Logger,
	db *models.Client, workerpool *workerpool.WorkerPool,
) (*Client, error) {
	storagePath := filepath.Join(cfg.Storage.RootDir, cfg.Storage.PhotoStorageDir)
	store := filestore.FileStore{
		Path: storagePath,
	}
	composer := tusd.NewStoreComposer()
	store.UseIn(composer)

	t := &Client{
		cfg:         cfg,
		auth:        auth,
		logger:      logger,
		db:          db,
		workerpool:  workerpool,
		storagePath: storagePath,
		stopSignal:  make(chan struct{}),
	}

	handler, err := tusd.NewHandler(tusd.Config{
		BasePath:              "/uploads",
		StoreComposer:         composer,
		NotifyCompleteUploads: true,
		MaxSize:               32 << 20, // 32MB

		PreUploadCreateCallback: t.PreUploadCreateCallback,
		DisableDownload:         true,
	})
	if err != nil {
		return nil, err
	}

	t.Handler = handler
	go t.Run()

	return t, nil
}

func (t *Client) Run() {
	func() {
		for {
			select {
			case <-t.stopSignal:
				return
			case event := <-t.Handler.CompleteUploads:
				t.handleCompleted(event)
			}
		}
	}()
}

func (t *Client) Shutdown() {
	close(t.stopSignal)
}

// PreUploadCreateCallback ensures that the incoming upload request contains a
// post_id and valid JWT token.
// NOTE: The /uploads endpoint is already protected by the RequireAuthentication middleware.
func (t *Client) PreUploadCreateCallback(hook tusd.HookEvent) error {
	token := hook.Upload.MetaData["token"]

	if postID, ok := hook.Upload.MetaData["post_id"]; ok {
		// Classifieds post
		if err := t.validateTokenClaimID(postID, "post_id", token); err != nil {
			t.logger.Warn("Uploaded post failed token validation")
			return err
		}
		return nil
	}

	if eventID, ok := hook.Upload.MetaData["event_id"]; ok {
		// Event
		if err := t.validateTokenClaimID(eventID, "event_id", token); err != nil {
			t.logger.Warn("Uploaded event failed token validation")
			return err
		}
		return nil
	}

	// Unknown upload type
	t.logger.Warn("Unknown upload type in tus handler")
	return errors.New("failed token validation")
}

// handleCompleted processes the completed upload by renaming the file, cleaning
// up the .info file, storing the photo metadata in the database, and submitting
// it to a task worker for resizing.
func (t *Client) handleCompleted(event tusd.HookEvent) {
	logger := t.logger.With(zap.String("tus_id", event.Upload.ID))
	logger.Info("Upload finished")

	var err error
	filename := event.Upload.ID

	defer func() {
		if err != nil {
			// There was an error, attempt to clean up the file
			removeErr := os.Remove(filepath.Join(t.storagePath, filename))
			if removeErr != nil {
				logger.Warn("Failed to remove uploaded file for orphaned upload",
					zap.Error(removeErr))
			}
		}
	}()

	// Clean up the .info file
	err = os.Remove(filepath.Join(t.storagePath, event.Upload.ID+".info"))
	if err != nil {
		logger.Warn("Failed to remove tus upload info file", zap.Error(err))
		err = nil
	}

	_, ok := event.Upload.MetaData["post_id"]
	if ok {
		err = t.handleCompletedPostPhoto(logger, event)
		return
	}

	_, ok = event.Upload.MetaData["event_id"]
	if ok {
		err = t.handleCompletedEventPhoto(logger, event)
		return
	}

	err = errors.New("Unexpected upload type")
}

func (t *Client) handleCompletedEventPhoto(logger *zap.Logger, event tusd.HookEvent) error {
	metrics.PhotoUploadCount.Add(1)

	var err error

	defer func() {
		if err != nil {
			metrics.PhotoUploadFailureCount.Add(1)
		}
	}()

	token := event.Upload.MetaData["token"]
	eventID, err := strconv.Atoi(event.Upload.MetaData["event_id"])
	if err != nil {
		logger.Warn("Failed to retrieve event ID from metadata", zap.Error(err))
		return err
	}

	logger = logger.With(
		zap.Int("event_id", eventID))

	err = t.validateTokenClaimID(event.Upload.MetaData["event_id"], "event_id", token)
	if err != nil {
		logger.Warn("Unable to validate upload token", zap.Error(err))
		return err
	}

	// Rename the file properly
	extension := event.Upload.MetaData["ext"]
	filename := fmt.Sprintf("event-%d-%s.%s", eventID, event.Upload.ID, extension)
	err = os.Rename(
		filepath.Join(t.storagePath, event.Upload.ID),
		filepath.Join(t.storagePath, filename))
	if err != nil {
		logger.Warn("Failed to rename uploaded file", zap.Error(err))
		return err
	}

	// Send the file for resizing
	t.workerpool.Submit(func() {
		tasks.ResizeEventImage(t.cfg, logger, t.db, int(eventID), filename)
	})

	return nil
}

func (t *Client) handleCompletedPostPhoto(logger *zap.Logger, event tusd.HookEvent) error {
	metrics.PhotoUploadCount.Add(1)

	var err error

	defer func() {
		if err != nil {
			metrics.PhotoUploadFailureCount.Add(1)
		}
	}()

	token := event.Upload.MetaData["token"]
	postType := event.Upload.MetaData["post_type"]
	postID, err := strconv.Atoi(event.Upload.MetaData["post_id"])
	if err != nil {
		logger.Warn("Failed to retrieve post ID from metadata", zap.Error(err))
		return err
	}

	logger = logger.With(
		zap.String("post_type", postType),
		zap.Int("post_id", postID))

	err = t.validateTokenClaimID(event.Upload.MetaData["post_id"], "post_id", token)
	if err != nil {
		logger.Warn("Unable to validate upload token", zap.Error(err))
		return err
	}

	// Rename the file properly
	extension := event.Upload.MetaData["ext"]
	filename := fmt.Sprintf("%s-%d-%s.%s", postType, postID, event.Upload.ID, extension)
	err = os.Rename(
		filepath.Join(t.storagePath, event.Upload.ID),
		filepath.Join(t.storagePath, filename))
	if err != nil {
		logger.Warn("Failed to rename uploaded file", zap.Error(err))
		return err
	}

	// Store the file in the database
	photoID, err := t.db.C.CreatePhoto(context.Background(),
		sqlc.CreatePhotoParams{
			PostID:           postID,
			OriginalFilename: conv.StringToNullString(filename),
		})
	if err != nil {
		logger.Warn("failed to store the photo to the database", zap.Error(err))
		return err
	}

	// Errors checking for and setting featured photo do not block continuing to resizing
	// so they should be set back to nil after logging.
	post, err := t.db.C.GetPost(context.Background(), postID)
	if err != nil {
		metrics.PhotoUploadFailureCount.Add(1)
		logger.Warn("failed to retrieve existing post to check for featured photo", zap.Error(err))
		err = nil
	} else if !post.PostsBase.FeaturedPhoto.Valid {
		err = t.db.C.SetPostFeaturedPhotoID(context.Background(), sqlc.SetPostFeaturedPhotoIDParams{
			FeaturedPhoto: conv.IntToNullInt(int(photoID)),
			ID:            post.PostsBase.ID,
		})
		if err != nil {
			metrics.PhotoUploadFailureCount.Add(1)
			logger.Warn("failed to set post featured photo", zap.Error(err))
			err = nil
		}
	}

	// Send the file for resizing
	t.workerpool.Submit(func() {
		tasks.ResizePostImages(t.cfg, logger, t.db, int(photoID), filename)
	})

	return nil
}

// validateTokenClaimID validates the provided JWT token and ensures that the
// contained post_id claim matches the provided postID.
func (t *Client) validateTokenClaimID(id string, claimField string, token string) error {
	claims, err := t.auth.ValidateJWTAuthTokenMap(token)
	if err != nil {
		return err
	}

	claimID, ok := claims[claimField].(float64)
	if !ok {
		return errors.New("unexpected post_id type in claim")
	}

	idInt, err := strconv.Atoi(id)
	if err != nil {
		t.logger.Warn("Failed to retrieve ID from metadata", zap.Error(err))
		return errors.New("failed token validation")
	}

	if int(claimID) != idInt {
		t.logger.Warn("Claim ID did not match expected value", zap.Error(err))
		return errors.New("failed token validation")
	}

	return nil
}
