package views

import (
	"io"
	"strconv"

	g "maragu.dev/gomponents"
	h "maragu.dev/gomponents/html"

	"git.grosinger.net/tgrosinger/tonyslist/constants/pages"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/controller"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
	c "git.grosinger.net/tgrosinger/tonyslist/pkg/views/components"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/views/layouts"
)

const (
	timeFormat = "Jan 02 2006"
)

type AdminUserDetails struct {
	User        sqlc.User
	PostsActive int

	HTMXReq string
	Page    controller.Page
}

// TODO: Add a last-page-load metric to user?
// TODO: Add button to reset bounces.

func (v *AdminUserDetails) Render(w io.Writer) error {
	switch v.HTMXReq {
	case "name-input":
		// changes the user name field into an input field.
		return h.Form(
			h.ID("name-form"),
			c.CSRF(v.Page),
			h.Input(
				h.Name("name"),
				h.Value(v.User.Name.String),
				h.AutoFocus(),
				c.HxTrigger("blur"),
				c.HxPatch(v.Page.URL),
				c.HxTarget("#name-form"),
				c.HxSwap("outerHTML"),
				h.Class("block w-full max-w-lg appearance-none rounded-md border border-gray-300 px-3 py-2 placeholder-gray-400 shadow-sm focus:border-indigo-500 focus:ring-indigo-500 focus:outline-none sm:text-sm"),
			)).Render(w)
	case "name-static":
		return v.nameCellValue().Render(w)
	}

	content := c.NarrowContainer(
		dividedSections(
			g.Group{
				c.PrimaryHeader("User Details"),
				c.DescriptionList(
					h.Class("mt-2"),
					c.MakeTwoColumnRow(
						c.MakeCell("Name", v.nameCellValue()),
						c.MakeCell("Email", g.Text(v.User.Email))),
				),
			},
			g.Group{
				c.TextLarge("Visits"),
				c.StatRow(
					c.NewStat(c.StatOptions{Label: "Page loads", Value: strconv.Itoa(v.User.CountPageLoads)}),
					c.NewStat(c.StatOptions{Label: "Logins", Value: strconv.Itoa(v.User.CountLogins)}),
					c.NewStat(c.StatOptions{
						Label: "Created",
						Value: v.User.CreatedAt.Format(timeFormat),
						Size:  c.StatSmall,
					}),
					c.LazyIfElse(v.User.LastLogin.Valid,
						func() g.Node {
							return c.NewStat(c.StatOptions{
								Label: "Last login",
								Value: v.User.LastLogin.Time.Format(timeFormat),
								Size:  c.StatSmall,
							})
						},
						func() g.Node {
							return c.NewStat(c.StatOptions{
								Label: "Last login",
								Value: "Never",
								Size:  c.StatSmall,
							})
						}),
				),
			},
			g.Group{
				c.TextLarge("Emails"),
				c.StatRow(
					c.NewStat(c.StatOptions{Label: "Bounced", Value: strconv.Itoa(v.User.CountBouncedEmails)}),
					c.NewStat(c.StatOptions{Label: "Post responses", Value: strconv.Itoa(v.User.CountPostsResponseEmails)}),
					c.NewStat(c.StatOptions{Label: "Other received", Value: strconv.Itoa(v.User.CountOtherIncomingEmail)}),
				),
			},
			g.Group{
				c.TextLarge("Classifieds"),
				c.StatRow(
					c.NewStat(c.StatOptions{Label: "Posts created", Value: strconv.Itoa(v.User.CountPostsCreated)}),
					c.NewStat(c.StatOptions{Label: "Posts active", Value: strconv.Itoa(v.PostsActive)}),
				),
			},
			g.Group{
				c.ButtonDestructive(v.Page.ToURL(pages.Names.Admin.UserDelete, v.User.ID), g.Text("Delete")),
			},
		),
	)

	l := layouts.Primary{
		Header: c.HeaderContainer(
			c.Breadcrumbs([]c.BreadcrumbPart{
				{
					Display: "Admin",
					Href:    v.Page.ToURL(pages.Names.Admin.Home),
				},
				{
					Display: "Users",
					Href:    v.Page.ToURL(pages.Names.Admin.Users),
				},
				{
					Display: v.User.Email,
				},
			},
			),
			nil,
		),
		Content: content,
		Page:    v.Page,
	}

	return l.Render(w)
}

func (v *AdminUserDetails) nameCellValue() g.Node {
	name := "unknown"
	if v.User.Name.Valid {
		name = v.User.Name.String
	}

	return h.Span(
		c.HxTrigger("click"),
		c.HxGet(v.Page.ToURL(pages.Names.Admin.UserDetails, v.User.ID)+"?field=name-input"),
		c.HxSwap("outerHTML"),
		g.Text(name))
}

func dividedSections(sections ...g.Node) g.Node {
	return h.Ul(
		h.Class("divide-y divide-gray-200"),
		g.Map(sections, func(section g.Node) g.Node {
			return h.Li(
				h.Class("px-4 py-4 sm:px-0"),
				section)
		}),
	)
}
