package views

import (
	"fmt"
	"html"
	"io"
	"net/http"
	"strings"

	"go.uber.org/zap"
	g "maragu.dev/gomponents"
	h "maragu.dev/gomponents/html"

	"git.grosinger.net/tgrosinger/tonyslist/constants"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/controller"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/services"
	c "git.grosinger.net/tgrosinger/tonyslist/pkg/views/components"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/views/layouts"
)

// ClassifiedUpdateForm is the subset of fields in a Post which are updatable. Fields
// not in this struct may not be updated after the post has been created.
type ClassifiedUpdateForm struct {
	Title string `form:"title" valid:"required"`
	// Price is optional because it is not required in for_free
	Price       string `form:"price" valid:"money,optional"`
	Description string `form:"description" valid:"required"`

	Submission controller.FormSubmission
}

type ClassifiedCreateForm struct {
	// Embed the PostUpdateForm as these fields are needed when creating a post
	// too. The Submission controller from the PostUpdateForm will be used for
	// this form too.
	ClassifiedUpdateForm

	ClassifiedType string `form:"classified_type" valid:"required"`
	IncludeEmail   bool   `form:"include_email"`
	PhoneNumber    string `form:"phone_number" valid:"phone,optional"`
	SMSNumber      string `form:"sms_number" valid:"phone,optional"`
}

// ProcessCreate verifies fields which are settable when creating a post. This
// process function should not be called when updating a post.
func (f *ClassifiedCreateForm) ProcessCreate(logger *zap.Logger) {
	if f.ClassifiedType != string(constants.ForSale) &&
		f.ClassifiedType != string(constants.ForFree) &&
		f.ClassifiedType != string(constants.WantToBuy) &&
		f.ClassifiedType != string(constants.WantToBorrow) {
		f.Submission.SetFieldError("ClassifiedType", "Invalid classified type in request")
	}

	if !f.IncludeEmail && f.PhoneNumber == "" && f.SMSNumber == "" {
		f.Submission.SetFieldError("IncludeEmail", "Must include at least one contact method")
	}
}

// Sanitize should call the sanitize function on any field which is not
// validated to be of a specific format.
func (f *ClassifiedUpdateForm) Sanitize(s *services.Sanitizer) {
	f.Title = s.Sanitize(f.Title)
	f.Description = s.SanitizeHTML(f.Description)
	f.Price = strings.ReplaceAll(f.Price, ",", "")
}

type ClassifiedCreate struct {
	IsCreate   bool
	ActionHref string
	CancelHref string
	SubmitText string
	Form       ClassifiedCreateForm
	Page       controller.Page
}

func (v *ClassifiedCreate) Render(w io.Writer) error {
	content := c.WideContainer(
		c.CenteredX(
			c.FullScreenForm(
				http.MethodPost,
				v.ActionHref,
				"",
				fmt.Sprintf("{classified_type: '%s'}", v.Form.ClassifiedType),
				c.FormFooter(
					h.Div(
						h.Class("flex flex-col grow"),
						c.LazyIf(v.IsCreate, func() g.Node {
							return h.Span(
								h.Class("grow text-gray-500"),
								g.Text("Add photos on the following page after clicking next."))
						}),
						h.Span(
							h.Class("grow text-gray-500"),
							g.Text("Posts are active for 14 days, after which they may be renewed."))),
					v.CancelHref, v.SubmitText, false),
				c.FormSection(
					c.FormHeading("Item Information", ""),
					c.CSRF(v.Page),
					c.LazyIf(v.IsCreate, func() g.Node {
						return h.Div(
							// If the Housing option is selected, redirect to the new housing form.
							g.Attr("x-effect", "if(classified_type === 'housing') window.location = '/housing/new'"),
							c.InputSelect(c.FormSelectParams{
								FormInputParams: c.FormInputParams{
									Name:     "classified_type",
									Display:  "Post Type",
									Required: true,
									Value:    v.Form.ClassifiedType,
									Errors:   v.Form.Submission.GetFieldErrors("ClassifiedType"),
								},
								XModel: "classified_type",
								Options: []c.FormSelectOption{
									{Display: "For Sale", Value: string(constants.ForSale)},
									{Display: "For Free", Value: string(constants.ForFree)},
									{Display: "Looking For", Value: string(constants.WantToBuy)},
									{Display: "Want to Borrow", Value: string(constants.WantToBorrow)},
									{Display: "Housing", Value: string(constants.Housing)},
								},
							}))
					}),
					c.InputText(c.FormInputParams{
						Name:     "title",
						Display:  "Title",
						Required: true,
						Value:    html.UnescapeString(v.Form.Title),
						Errors:   v.Form.Submission.GetFieldErrors("Title"),
					}),
					h.Div(
						c.XShow("classified_type === 'for-sale'"),
						c.InputDollar(c.FormInputParams{
							Name:    "price",
							Display: "Price",
							Value:   v.Form.Price,
							Errors:  v.Form.Submission.GetFieldErrors("Price"),
						})),
					c.InputRichTextEditor(c.FormInputParams{
						Name:    "description",
						Display: "Description",
						HintNode: g.Group([]g.Node{
							h.P(
								h.Class("text-sm text-gray-500"),
								g.Text("Write a few sentences about the item and the condition it is in.")),
						}),
						Required: true,
						Value:    html.UnescapeString(v.Form.Description),
						Errors:   v.Form.Submission.GetFieldErrors("Description"),
					})),
				c.LazyIf(v.IsCreate, func() g.Node {
					return c.FormSection(
						c.FormHeading("Contact Information",
							"How would you like those interested to contact you? Leave a field blank if you do not want it to be used. Contact information entered will be displayed to all site visitors."),
						c.InputCheckbox(v.Form.IncludeEmail, c.FormInputParams{
							Name:    "include_email",
							Display: "Include Anonymous Email",
							Hint:    "Display an anonymous email address. Messages will forward to the email associated with your account.",
							Errors:  v.Form.Submission.GetFieldErrors("IncludeEmail"),
						}),
						c.InputPhone(c.FormInputParams{
							Name:    "phone_number",
							Display: "Phone Call",
							Value:   v.Form.PhoneNumber,
							Errors:  v.Form.Submission.GetFieldErrors("PhoneNumber"),
						}),
						c.InputPhone(c.FormInputParams{
							Name:    "sms_number",
							Display: "Text/SMS Message",
							Value:   v.Form.SMSNumber,
							Errors:  v.Form.Submission.GetFieldErrors("SMSNumber"),
						}))
				}))),
	)

	l := layouts.Primary{
		Header:  c.ClassifiedHeader(v.Page, ""),
		Content: content,
		Page:    v.Page,
	}
	return l.Render(w)
}
