package components

import (
	"fmt"

	g "maragu.dev/gomponents"
)

func XRef(name string) g.Node {
	return RawAttr("x-ref", name)
}

// XInit is shorthand for the Alpine.js x-init directive. It allows you to hook
// into the initialization of Alpine.js.
// https://alpinejs.dev/directives/init
func XInit(action string) g.Node {
	return RawAttr("x-init", action)
}

// XOn is shorthand for the Alpine.js x-on directive.
// https://alpinejs.dev/directives/on
func XOn(action, result string) g.Node {
	return RawAttr("x-on:"+action, result)
}

// XModel is a shorthand for the Apline.js x-model directive.
// https://alpinejs.dev/directives/model
func XModel(variable string) g.Node {
	return RawAttr("x-model", variable)
}

// XModelDebounce is a shorthand for the Alpine.js x-model directive with the
// debounce modifier. Model will bind the the value of an input to a variable.
// https://alpinejs.dev/directives/model#debounce
func XModelDebounce(variable, delay string) g.Node {
	return RawAttr("x-model.debounce."+delay, variable)
}

// XClick is a special case of XOn specifically for click events.
func XClick(result string) g.Node {
	return XOn("click", result)
}

// XClickStop is a special case of XOn specifically for click events. This
// version also includes the stop propagation modifier.
func XClickStop(result string) g.Node {
	return XOn("click.stop", result)
}

// XData is shorthand for the Alpine.js x-data directive.
// https://alpinejs.dev/directives/data
func XData(format string, a ...any) g.Node {
	return RawAttr("x-data", fmt.Sprintf(format, a...))
}

// XBind is shorthand for the Alpine.js x-bind directive.
// https://alpinejs.dev/directives/bind
func XBind(target, expr string) g.Node {
	return RawAttr("x-bind:"+target, expr)
}

func XShow(expr string) g.Node {
	return RawAttr("x-show", expr)
}

func XIfTemplate(expression string, children ...g.Node) g.Node {
	return g.El("template",
		append([]g.Node{RawAttr("x-if", expression)}, children...)...)
}

// XText is shorthand for the Alpine.js x-text directive.
// https://alpinejs.dev/directives/text
func XText(expr string) g.Node {
	return RawAttr("x-text", expr)
}

// XCloak adds the Alpine cloak directive to prevent a node from showing until
// Alpine is fully loaded.
// https://alpinejs.dev/directives/cloak
func XCloak() g.Node {
	return RawAttr("x-cloak")
}

func XMask(expr string) g.Node {
	return RawAttr("x-mask:dynamic", expr)
}

type TransitionOptions struct {
	DurationEnter  string
	DurationLeave  string
	OperationEnter string
	OperationLeave string
	OpacityStart   string
	OpacityEnd     string
	ScaleStart     string
	ScaleEnd       string
}

func XTransition(opt TransitionOptions) g.Node {
	return g.Group([]g.Node{
		g.Attr("x-transition:enter", fmt.Sprintf("transition %s %s", opt.OperationEnter, opt.DurationEnter)),
		g.Attr("x-transition:enter-start", fmt.Sprintf("transform %s %s", opt.OpacityStart, opt.ScaleStart)),
		g.Attr("x-transition:enter-end", fmt.Sprintf("transform %s %s", opt.OpacityEnd, opt.ScaleEnd)),
		g.Attr("x-transition:leave", fmt.Sprintf("transition %s %s", opt.OperationLeave, opt.DurationLeave)),
		g.Attr("x-transition:leave-start", fmt.Sprintf("transform %s %s", opt.OpacityEnd, opt.ScaleEnd)),
		g.Attr("x-transition:leave-end", fmt.Sprintf("transform %s %s", opt.OpacityStart, opt.ScaleStart)),
	})
}
