package components

import (
	g "maragu.dev/gomponents"
	h "maragu.dev/gomponents/html"
)

func FullWidthContainer(children ...g.Node) g.Node {
	return h.Div(
		h.Class("mx-auto max-w-7xl sm:px-6 sm:py-6 lg:px-8"),
		h.Div(
			h.Class("bg-white shadow sm:rounded-lg"),
			g.Group(children)))
}

func NarrowContainer(children ...g.Node) g.Node {
	return FullWidthContainer(
		h.Div(
			h.Class("mx-auto max-w-2xl py-8 sm:py-12 px-4 sm:px-6 lg:px-8"),
			g.Group(children)))
}

// WideContainer is a large, centered container with a light background.
func WideContainer(children ...g.Node) g.Node {
	return FullWidthContainer(
		h.Div(
			h.Class("py-8 sm:py-12 px-4 sm:px-6 lg:px-8"),
			g.Group(children)))
}

// InsetContainer is a smaller container with a differentiated background.
func InsetContainer(children ...g.Node) g.Node {
	return h.Div(
		h.Class("my-4 bg-gray-50 shadow sm:rounded-lg"),
		g.Group(children))
}

func DividedList(children ...g.Node) g.Node {
	return h.Div(
		h.Class("divide-y divide-gray-400"),
		g.Group(
			g.Map(children, func(child g.Node) g.Node {
				if child == nil {
					return nil
				}
				return h.Div(
					h.Class("py-4 first:pt-0 last:pb-0"),
					child,
				)
			}),
		),
	)
}

// MarkdownRenderingContainer applies styles to rendered markdown as defined in
// tailwind.css
func MarkdownRenderingContainer(content string) g.Node {
	return h.Div(
		h.Class("markdown-container"),
		g.Raw(content))
}

// DescriptionContainer is used to display pre-wrapped content in a
// lower-contrast text color.
func DescriptionContainer(text string) g.Node {
	return h.P(
		h.Class("text-base text-gray-500 whitespace-pre-wrap"),
		g.Text(text),
	)
}

// CenteredX centers the provided children horizontally in the parent.
func CenteredX(children ...g.Node) g.Node {
	return h.Div(
		h.Class("mx-auto max-w-3xl py-12 px-4 sm:px-6 lg:px-8"),
		g.Group(children),
	)
}

// SpacedYSmall adds a medium amount of space vertically between children.
func SpacedYSmall(children ...g.Node) g.Node {
	return h.Div(
		h.Class("space-y-2"),
		g.Group(children),
	)
}

// SpacedYLarge adds a medium amount of space vertically between children.
func SpacedYMedium(children ...g.Node) g.Node {
	return h.Div(
		h.Class("space-y-6"),
		g.Group(children),
	)
}

// SpacedYLarge adds a large amount of space vertically between children.
func SpacedYLarge(children ...g.Node) g.Node {
	return h.Div(
		h.Class("space-y-10"),
		g.Group(children),
	)
}

// FloatingToolbar contains a text message on the left side and buttons on the
// far right. Fits within whatever container it's put into.
func FloatingToolbar(id, color string, left g.Node, buttons ...g.Node) g.Node {
	return h.Div(
		h.ID(id),
		h.Class("mx-auto max-w-7xl px-2 pt-4 sm:px-6 lg:px-8"),
		h.Div(
			h.Class("rounded-lg p-2 shadow-lg sm-3 "+color),
			h.Div(
				h.Class("flex flex-wrap items-center gap-y-2 justify-between"),
				h.Div(
					h.Class("flex flex-1 min-w-fit items-center justify-start"),
					left),
				h.Div(
					h.Class("flex w-auto flex-shrink-0 sm:mt-0"),
					g.Group(buttons)))))
}

// Tooltip creates a tool tip when the target element is hovered over.
// Position must be one of [top, left, right, bottom].
func Tooltip(msg string, position string, target g.Node) g.Node {
	containerClass := ""
	arrowClassOuter := ""
	arrowClassInner := ""
	switch position {
	case "top":
		containerClass = "top-0 left-1/2 -translate-x-1/2 -mt-0.5 -translate-y-full"
		arrowClassOuter = "bottom-0 -translate-x-1/2 left-1/2 w-2.5 translate-y-full"
		arrowClassInner = "origin-top-left -rotate-45"
	case "left":
		containerClass = "top-1/2 -translate-y-1/2 -ml-0.5 left-0 -translate-x-full"
		arrowClassOuter = "right-0 -translate-y-1/2 top-1/2 h-2.5 -mt-px translate-x-full"
		arrowClassInner = "origin-top-left rotate-45"
	case "bottom":
		containerClass = "bottom-0 left-1/2 -translate-x-1/2 -mb-0.5 translate-y-full"
		arrowClassOuter = "top-0 -translate-x-1/2 left-1/2 w-2.5 -translate-y-full"
		arrowClassInner = "origin-bottom-left rotate-45"
	case "right":
		containerClass = "top-1/2 -translate-y-1/2 -mr-0.5 right-0 translate-x-full"
		arrowClassOuter = "left-0 -translate-y-1/2 top-1/2 h-2.5 -mt-px -translate-x-full"
		arrowClassInner = "origin-top-right -rotate-45"
	}

	return h.Div(
		XData("{'visible': false}"),
		XInit("$refs.content.addEventListener('mouseenter', () => { visible = true; }); $refs.content.addEventListener('mouseleave', () => { visible = false; });"),
		h.Class("relative"),
		h.Div(
			XShow("visible"),
			h.Class("absolute w-auto text-sm "+containerClass),
			XCloak(),
			h.Div(
				h.Class("relative px-2 py-1 text-white bg-black rounded bg-opacity-90"),
				h.P(
					h.Class("flex-shrink-0 block text-xs whitespace-nowrap"),
					g.Text(msg)),
				h.Div(
					h.Class("absolute inline-flex items-center justify-center overflow-hidden "+arrowClassOuter),
					h.Div(
						h.Class("w-1.5 h-1.5 transform bg-black bg-opacity-90 "+arrowClassInner))))),
		h.Div(
			XRef("content"),
			target))
}

// Template creates an HTML template.
// https://developer.mozilla.org/en-US/docs/Web/HTML/Element/template
func Template(id string, template g.Node) g.Node {
	return g.El("template",
		h.ID(id),
		template)
}
