package components

import (
	"git.grosinger.net/tgrosinger/tonyslist/pkg/controller"

	g "maragu.dev/gomponents"
	c "maragu.dev/gomponents/components"
	h "maragu.dev/gomponents/html"
)

// Creates a form intended for embedding in a CenteredX container.
// If actions, encType, or xData are empty they will be omitted.
func FullScreenForm(method, action, encType, xData string, footer g.Node, sections ...g.Node) g.Node {
	return h.Form(
		h.Class("space-y-8 divide-y divide-gray-200"),
		LazyIf(action != "", func() g.Node { return h.Action(action) }),
		LazyIf(method != "", func() g.Node { return h.Method(method) }),
		LazyIf(encType != "", func() g.Node { return h.EncType(encType) }),
		LazyIf(xData != "", func() g.Node { return g.Attr("x-data", xData) }),
		h.Div(
			h.Class("space-y-8 divide-y divide-gray-200 sm:space-y-5"),
			g.Group(sections)),
		footer)
}

func FormSection(heading g.Node, fields ...g.Node) g.Node {
	return h.Div(
		h.Class("space-y-6 sm:space-y-5"),
		heading,
		h.Div(
			h.Class("space-y-6 sm:space-y-5"),
			g.Group(fields)))
}

func FormHeading(title, description string) g.Node {
	return h.Div(
		LazyIf(title != "", func() g.Node {
			return ListHeader(title)
		}),
		LazyIf(description != "", func() g.Node {
			return TextDescription(description, "mt-1")
		}))
}

func FormFooterNodes(footerHint g.Node, buttons []g.Node) g.Node {
	return h.Div(
		h.Class("pt-5 flex flex-col sm:flex-row items-center justify-end gap-x-2 gap-y-2"),
		// For some reason the ButtonCancelBack() does not work without
		// this empty XData.
		XData("{}"),
		footerHint,
		h.Div(
			h.Class("flex items-center justify-end gap-x-2"),
			g.Group(buttons)))
}

func FormFooter(footerHint g.Node, cancelHref, submitText string, destructiveAction bool) g.Node {
	buttons := make([]g.Node, 0, 2)

	if cancelHref == "back" {
		buttons = append(buttons, ButtonCancelBack())
	} else if cancelHref != "" {
		buttons = append(buttons, ButtonCancel(cancelHref))
	}

	if destructiveAction {
		buttons = append(buttons, FormSubmitRed(submitText))
	} else {
		buttons = append(buttons, FormSubmit(submitText))
	}

	return FormFooterNodes(footerHint, buttons)
}

// FormStaticText allows setting a key=value in the form which follows the flow
// of normal form inputs, but instead is used to display data, not request it.
func FormStaticText(label, content string) g.Node {
	return formFieldWrapper(
		FormInputParams{
			Display: label,
		},
		h.Div(
			h.Class("block w-full max-w-lg appearance-none rounded-md px-3 py-2 shadow-sm sm:text-sm"),
			g.Text(content)),
	)
}

type FormInputParams struct {
	Name      string
	Display   string
	Value     string
	Autofocus bool
	Required  bool
	Disabled  bool
	Hint      string
	HintNode  g.Node
	Errors    []string
}

func innerFormField(params FormInputParams, attrs ...g.Node) g.Node {
	return h.Input(
		h.Name(params.Name),
		h.ID(params.Name),
		g.Group(attrs),
		LazyIf(params.Value != "", func() g.Node {
			return h.Value(params.Value)
		}),
		LazyIf(params.Autofocus, h.AutoFocus),
		LazyIf(params.Disabled, h.Disabled),
		LazyIf(params.Required, h.Required),
		c.Classes{
			"block w-full max-w-lg appearance-none rounded-md border px-3 py-2 placeholder-gray-400 shadow-sm focus:border-indigo-500 focus:ring-indigo-500 focus:outline-none sm:text-sm": true,
			"disabled:cursor-not-allowed disabled:border-gray-200 disabled:bg-gray-50 disabled:text-gray-500":                                                                              params.Disabled,
			"border-gray-300": len(params.Errors) == 0,
			"border-red-300":  len(params.Errors) > 0,
		},
	)
}

func formFieldWrapper(params FormInputParams, input g.Node) g.Node {
	return h.Div(
		h.Class("sm:grid sm:grid-cols-3 sm:items-start sm:gap-4 sm:border-t sm:border-gray-200 sm:pt-5"),
		LazyIf(params.Display != "", func() g.Node {
			return h.Label(
				h.Class("block text-sm font-medium text-gray-700 sm:mt-px sm:pt-2"),
				h.For(params.Name),
				g.Text(params.Display))
		}),
		h.Div(
			c.Classes{
				"mt-1  sm:mt-0": true,
				"sm:col-span-2": params.Display != "",
				"sm:col-span-3": params.Display == "",
			},
			input,
			LazyIf(params.Hint != "", func() g.Node {
				return TextDescription(params.Hint)
			}),
			LazyIf(params.HintNode != nil, func() g.Node {
				return params.HintNode
			}),
			LazyIf(len(params.Errors) > 0, func() g.Node {
				return g.Group(
					g.Map(params.Errors, func(err string) g.Node {
						return h.P(
							h.Class("mt-2 text-sm text-red-600"),
							g.Text(err))
					}))
			})))
}

func formInput(params FormInputParams, attrs ...g.Node) g.Node {
	return formFieldWrapper(params,
		innerFormField(params, attrs...))
}

type FormSelectOption struct {
	Display string
	Value   string
}

type FormSelectParams struct {
	FormInputParams

	XModel  string
	Options []FormSelectOption
}

func CustomFormField(params FormInputParams, el g.Node) g.Node {
	return formFieldWrapper(params, el)
}

func InputSelect(params FormSelectParams) g.Node {
	return formFieldWrapper(params.FormInputParams,
		h.Select(
			h.Name(params.Name),
			h.ID(params.Name),
			XModel(params.XModel),
			h.Class("block w-full max-w-lg rounded-md border-gray-300 shadow-sm focus:border-indigo-500 focus:ring-indigo-500 sm:text-sm"),
			g.Group(g.Map(params.Options, func(opt FormSelectOption) g.Node {
				return h.Option(
					h.Value(opt.Value),
					g.Text(opt.Display),
					LazyIf(opt.Value == params.Value, h.Selected))
			})),
		))
}

func InputTextFullWidth(params FormInputParams, attrs ...g.Node) g.Node {
	return h.Div(
		h.Class("w-full"),
		innerFormField(params,
			h.Type("text"),
			g.Group(attrs)),
		LazyIf(len(params.Errors) > 0, func() g.Node {
			return g.Group(
				g.Map(params.Errors, func(err string) g.Node {
					return h.P(
						h.Class("mt-2 text-sm text-red-600"),
						g.Text(err))
				}))
		}))
}

// InputEmailFullWidth creates an email input field. The "Display" value from the params
// is displayed only for screen readers.
// NOTE: This form field does not show hints. It doesn't use the normal form wrappers.
func InputEmailFullWidth(params FormInputParams, attrs ...g.Node) g.Node {
	return h.Div(
		h.Class("w-full"),
		SRLabel("email", "Email Address"),
		innerFormField(params,
			h.Type("email"),
			h.AutoComplete("email"),
			h.Placeholder("name@email.com"),
			g.Group(attrs)),
		LazyIf(len(params.Errors) > 0, func() g.Node {
			return g.Group(
				g.Map(params.Errors, func(err string) g.Node {
					return h.P(
						h.Class("mt-2 text-sm text-red-600"),
						g.Text(err))
				}))
		}))
}

func InputEmail(params FormInputParams) g.Node {
	return formInput(params,
		h.Type("email"),
		h.AutoComplete("email"),
		h.Placeholder("name@email.com"))
}

func InputText(params FormInputParams) g.Node {
	return formInput(params,
		h.Type("text"),
	)
}

func InputAutoCompleteText(autocomplete string, params FormInputParams) g.Node {
	return formInput(params,
		h.Type("text"),
		h.AutoComplete(autocomplete))
}

func InputDollar(params FormInputParams) g.Node {
	return formFieldWrapper(params,
		h.Div(
			h.Class("relative"),
			h.Div(
				h.Class("pointer-events-none absolute inset-y-0 left-0 top-0 flex items-center pl-3"),
				h.Span(
					h.Class("text-gray-500 sm:text-sm"),
					g.Text("$"))),

			h.Input(
				h.Name(params.Name),
				h.ID(params.Name),
				h.Type("text"),
				XMask("$money($input)"),
				LazyIf(params.Value != "", func() g.Node {
					return h.Value(params.Value)
				}),
				LazyIf(params.Required, h.Required),
				c.Classes{
					"block w-full min-w-0 max-w-lg appearance-none rounded-md border pl-7 px-3 py-2 placeholder-gray-400 shadow-sm focus:border-indigo-500 focus:ring-indigo-500 focus:outline-none sm:text-sm": true,
					"border-gray-300": len(params.Errors) == 0,
					"border-red-300":  len(params.Errors) > 0,
				},
			)))
}

// InputStreetAndZip combines the street and zip input fields.
func InputStreetAndZip(display, streetValue, zipValue, hint string, streetErrors, zipErrors []string) g.Node {
	dummyParams := FormInputParams{
		Display: display,
		Hint:    hint,
		Errors:  append(streetErrors, zipErrors...),
	}

	return formFieldWrapper(dummyParams,
		h.Div(
			h.Class("flex max-w-lg -space-x-px"),
			h.Div(
				h.Class("w-2/3 min-w-0"),
				SRLabel("street", "Street Address"),
				h.Input(
					h.Type("text"),
					h.Name("street"),
					h.ID("street"),
					h.Value(streetValue),
					h.Class("relative block w-full rounded-none rounded-bl-md border-gray-300 bg-transparent focus:z-10 focus:border-indigo-500 focus:ring-indigo-500 sm:text-sm"),
					h.Placeholder("999 Prune Alley"))),
			h.Div(
				h.Class("min-w-0"),
				SRLabel("zipcode", "Zipcode"),
				h.Input(
					h.Type("text"),
					h.Name("zipcode"),
					h.ID("zipcode"),
					h.Value(zipValue),
					h.Class("relative block w-full rounded-none rounded-br-md border-gray-300 bg-transparent focus:z-10 focus:border-indigo-500 focus:ring-indigo-500 sm:text-sm"),
					h.Placeholder("98XXX")))))
}

func InputPhone(params FormInputParams) g.Node {
	return formInput(params,
		h.Type("tel"),
		g.Attr("x-mask", "999-999-9999"),
		h.Placeholder("000-000-0000"),
		h.AutoComplete("tel"),
	)
}

func InputSteppedNumber(min, step string, params FormInputParams) g.Node {
	return formInput(params,
		h.Type("number"),
		h.Min(min),
		h.Step(step))
}

func InputRichTextEditor(params FormInputParams) g.Node {
	return formFieldWrapper(params,
		g.Group([]g.Node{
			h.Input(
				h.Name(params.Name),
				h.ID(params.Name),
				h.Type("hidden"),
				h.Value(params.Value)),
			g.El("trix-editor",
				// Not actually markdown, but we want to use the same styles.
				h.Class("markdown-container"),
				g.Attr("input", params.Name)),
		}),
	)
}

func InputTextArea(params FormInputParams) g.Node {
	return formFieldWrapper(params,
		h.Textarea(
			h.Name(params.Name),
			h.ID(params.Name),
			h.Rows("3"),
			c.Classes{
				"block w-full max-w-lg md:max-w-full rounded-md border shadow-sm focus:border-indigo-500 focus:ring-indigo-500 sm:text-sm": true,
				"border-gray-300": len(params.Errors) == 0,
				"border-red-300":  len(params.Errors) > 0,
			},
			g.Text(params.Value)))
}

func InputCheckbox(checked bool, params FormInputParams) g.Node {
	return h.Div(
		h.Class("sm:grid sm:grid-cols-3 sm:items-start sm:gap-4 sm:border-t sm:border-gray-200 sm:pt-5"),
		h.Div(
			h.Class("relative flex items-start sm:col-span-2 sm:col-start-2"),
			h.Div(
				h.Class("flex h-5 items-center"),
				h.Input(
					h.Name(params.Name),
					h.ID(params.Name),
					h.Type("checkbox"),
					h.Value("true"),
					h.Class("h-4 w-4 rounded border-gray-300 text-indigo-600 focus:ring-indigo-500"),
					LazyIf(checked, func() g.Node { return g.Attr("checked") }))),
			h.Div(
				h.Class("ml-3 text-sm"),
				h.Label(
					h.Class("font-medium text-gray-700"),
					h.For(params.Name),
					g.Text(params.Display)),
				LazyIf(params.Hint != "", func() g.Node {
					return TextDescription(params.Hint)
				}),
				LazyIf(len(params.Errors) > 0, func() g.Node {
					return g.Group(
						g.Map(params.Errors, func(err string) g.Node {
							return h.P(
								h.Class("mt-2 text-sm text-red-600"),
								g.Text(err))
						}))
				}))))
}

// InputData is a native date input field with graceful degredation in
// unsupported browsers by providing a pattern attribute. Use the attrs param to
// set the Min date. Value should always be validated server-side.
func InputDate(params FormInputParams, attrs ...g.Node) g.Node {
	return formInput(params,
		h.Type("date"),
		h.Pattern(`\d{4}-\d{2}-\d{2}`), // Support graceful degredation
		g.Group(attrs))
}

// InputTime is a native time input field with graceful degredation in
// unsupported browser by providing a pattern attribute. Value should always be
// validated server-side.
func InputTime(params FormInputParams) g.Node {
	return formInput(params,
		h.Type("time"),
		h.Pattern(`[0-9]{2}:[0-9]{2}`)) // Support graceful degredation
}

func CSRF(page controller.Page) g.Node {
	return h.Input(
		h.Type("hidden"),
		h.Name("csrf"),
		h.Value(page.CSRF))
}

// MothersMaidenName is a honeypot field which can be used in a form to help
// detect spammers. It is contained in a hidden div.
func MothersMaidenName() g.Node {
	return h.Div(
		h.Class("hidden"),
		InputText(FormInputParams{
			Name:    "maiden_name",
			Display: "Mother's Maiden Name",
		}))
}

// SRLabel creates a label for a form field which is only displayed to screen readers.
func SRLabel(labelFor, display string) g.Node {
	return h.Label(
		h.For(labelFor),
		h.Class("sr-only"),
		g.Text(display))
}

// FormSubmit is a form button of the default color to indicate the default action.
func FormSubmit(text string) g.Node {
	return h.Button(
		h.Type("submit"),
		h.Class("inline-flex items-center justify-center rounded-md border border-transparent bg-indigo-600 px-3 py-2 font-medium leading-4 text-white shadow-sm hover:bg-indigo-700 focus:outline-none focus:ring-2 focus:ring-indigo-500 focus:ring-offset-2 sm:text-sm"),
		g.Text(text))
}

// FormSubmitRed is a form button which is red, to indicate a potentially destructive action.
func FormSubmitRed(text string) g.Node {
	return h.Button(
		h.Type("submit"),
		h.Class("inline-flex items-center justify-center rounded-md border border-transparent bg-red-100 px-4 py-2 font-medium text-red-700 hover:bg-red-200 focus:outline-none focus:ring-2 focus:ring-red-500 focus:ring-offset-2 sm:text-sm"),
		g.Text(text))
}
