package rss

import (
	"io"

	g "maragu.dev/gomponents"
)

// statefulWriter only writes if no errors have occurred earlier in its lifetime.
//
// Copied from gomponents.
type statefulWriter struct {
	w   io.Writer
	err error
}

// Copied from gomponents.
func (w *statefulWriter) Write(p []byte) {
	if w.err != nil {
		return
	}
	_, w.err = w.w.Write(p)
}

// nodeTypeDescriber can be implemented by Nodes to let callers know whether the Node is
// an ElementType or an AttributeType. This is used for rendering.
//
// Copied from gomponents.
type nodeTypeDescriber interface {
	Type() g.NodeType
}

// renderChild c to the given writer w if the node type is t.
//
// Copied from gomponents, with the Group functionality removed because it used an internal API.
func renderChild(w *statefulWriter, c g.Node, t g.NodeType) {
	if w.err != nil || c == nil {
		return
	}

	switch t {
	case g.ElementType:
		if p, ok := c.(nodeTypeDescriber); !ok || p.Type() == g.ElementType {
			w.err = c.Render(w.w)
		}
	case g.AttributeType:
		if p, ok := c.(nodeTypeDescriber); ok && p.Type() == g.AttributeType {
			w.err = c.Render(w.w)
		}
	}
}

// xmlEl behaves identically to El with the exception that checking for void
// elements is disabled. This allows use of elements such as "link" which need
// to have children in RSS feeds, for example.
//
// NOTE: Groups are not supported because they were removed from the copied version of renderChild.
func xmlEl(name string, children ...g.Node) g.Node {
	return g.NodeFunc(func(w2 io.Writer) error {
		w := &statefulWriter{w: w2}

		w.Write([]byte("<" + name))

		for _, c := range children {
			renderChild(w, c, g.AttributeType)
		}

		w.Write([]byte(">"))

		for _, c := range children {
			renderChild(w, c, g.ElementType)
		}

		w.Write([]byte("</" + name + ">"))
		return w.err
	})
}

func RSS(children ...g.Node) g.Node {
	return xmlEl("rss", children...)
}

func Channel(children ...g.Node) g.Node {
	return xmlEl("channel", children...)
}

func Title(v string) g.Node {
	return xmlEl("title",
		g.Raw(v))
}

func Link(v string) g.Node {
	return xmlEl("link",
		g.Text(v))
}

func Description(v string) g.Node {
	return xmlEl("description",
		g.Raw(v))
}

func WebMaster(v string) g.Node {
	return xmlEl("webMaster",
		g.Text(v))
}

func Item(children ...g.Node) g.Node {
	return xmlEl("item", children...)
}

func PubDate(v string) g.Node {
	return xmlEl("pubDate",
		g.Text(v))
}
