package components

import (
	"fmt"
	"strings"

	g "maragu.dev/gomponents"
	h "maragu.dev/gomponents/html"
)

// PrimaryHeader is the largest of the header sizes, often used at the top of the page.
func PrimaryHeader(text string) g.Node {
	return h.H1(
		h.Class("text-4xl font-bold tracking-tight text-gray-900"),
		g.Raw(text),
	)
}

// SecondaryHeader is the second largest header sizes. It can denote major
// sections within a page. This size sometimes may be skipped.
func SecondaryHeader(text string) g.Node {
	return h.H2(
		h.Class("text-3xl font-bold leading-6 text-gray-900 sm:text-4xl"),
		g.Raw(text),
	)
}

// TertiaryHeader is a small header for use within sections.
func TertiaryHeader(text string) g.Node {
	return h.H3(
		h.Class("text-xl sm:text-2xl font-medium truncate text-gray-900"),
		g.Raw(text),
	)
}

// ListHeader is even smaller than a TertiaryHeader and is often used in a
// divided list to highlight the next list item.
func ListHeader(text string) g.Node {
	return h.H4(
		h.Class("leading6 text-xl font-medium text-gray-900"),
		g.Raw(text),
	)
}

func TextLarge(text string) g.Node {
	return h.P(
		h.Class("text-lg text-gray-900 sm:text-xl"),
		g.Raw(text))
}

// TextDescription is used for deemphasized text which is describing something.
// More prominent than a hint, but less than normal text.
func TextDescription(text string, extraClasses ...string) g.Node {
	classes := strings.Join(extraClasses, " ")
	return h.P(
		h.Class("text-sm text-gray-500 "+classes),
		g.Raw(text),
	)
}

// TextNormal is used for standard text.
func TextNormal(text string, extraClasses ...string) g.Node {
	classes := strings.Join(extraClasses, " ")
	return h.P(
		h.Class("text-base text-gray-700 "+classes),
		g.Raw(text),
	)
}

func TextLink(href string, children ...g.Node) g.Node {
	return h.A(
		h.Class("underline hover:text-gray-800"),
		h.Href(href),
		g.Group(children))
}

// TextButtonLink is a button without a colored border.
func TextButtonLink(href, display string) g.Node {
	return h.A(
		h.Class("text-indigo-600 hover:text-indigo-900"),
		h.Href(href),
		g.Text(display))
}

// TextCurrency ouputs a pointer to a float as 2.00.
func TextCurrency(f *float64) string {
	return fmt.Sprintf("%.2f", *f)
}

// TextPill returns a rounded pill with the provided text.
// Colors should be similar to "bg-indigo-100 text-indigo-800"
func TextPill(display, colors string) g.Node {
	return h.Span(
		h.Class("inline-flex items-center rounded-full px-3 py-0.5 text-sm font-medium "+colors),
		g.Text(display))
}

// TextSROnly returns a span with the given text, only visible to screen
// readers.
func TextSROnly(text string) g.Node {
	return h.Span(
		h.Class("sr-only"),
		g.Text(text))
}
