package views

import (
	"fmt"
	"io"
	"strconv"
	"time"

	g "maragu.dev/gomponents"
	gc "maragu.dev/gomponents/components"
	h "maragu.dev/gomponents/html"

	"git.grosinger.net/tgrosinger/tonyslist/constants/pages"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/controller"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/conv"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
	c "git.grosinger.net/tgrosinger/tonyslist/pkg/views/components"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/views/layouts"
)

type EventBrowse struct {
	UpcomingEvents         []sqlc.Event
	SelectedDate           time.Time
	FirstDayOfSelectedWeek time.Time
	LastDayOfSelectedWeek  time.Time

	Page controller.Page
}

func (v *EventBrowse) Render(w io.Writer) error {
	content := c.WideContainer(
		h.Div(
			// TODO: Standardize this container
			h.Class("mx-auto max-w-2xl py-4 px-4 sm:py-6 sm:px-6 lg:max-w-7xl lg:px-8"),
			h.Div(
				h.Class("lg:pl-6"),
				c.PrimaryHeader("Events the Week of "+v.FirstDayOfSelectedWeek.Format("Jan _2"))),
			h.Div(
				h.Class("lg:grid lg:grid-cols-12 lg:gap-x-16"),
				v.MiniCalendar(),
				v.EventList(),
			)),
	)

	l := layouts.Primary{
		Header:  c.EventHeader(v.Page),
		Content: content,
		Page:    v.Page,
	}
	return l.Render(w)
}

func (v *EventBrowse) EventList() g.Node {
	return h.Div(
		h.Class("mt-4 flex flex-col justify-between min-h-full lg:col-span-7 xl:col-span-8"),
		c.LazyIfElse(len(v.UpcomingEvents) == 0,
			func() g.Node {
				return h.Div(
					h.Class("flex flex-col items-center justify-center py-6 lg:mt-24"),
					h.Div(
						h.Class("text-gray-500"),
						c.IconCalendarDays()),
					c.TextLarge("There are no events for the selected week."))
			},
			func() g.Node {
				return h.Ol(
					h.Class("divide-y divide-gray-100 text-sm leading-6"),
					g.Group(g.Map(v.UpcomingEvents, v.EventListItem)))
			}),
		v.Pager())
}

func (v *EventBrowse) EventListItem(e sqlc.Event) g.Node {
	return h.Li(
		h.Class("relative flex space-x-6 py-6 px-6 hover:bg-gray-50 xl:static"),
		h.A(
			h.Href(v.Page.ToURL(pages.Names.Event.One, e.ID)),
			h.Class("flex justify-between items-center min-w-full"),
			h.Div(
				h.Class("flex-auto overflow-hidden"),
				c.TertiaryHeader(e.Title),
				c.LazyIf(v.Page.IsAuth, func() g.Node {
					dateTime := e.StartTime.In(v.Page.Cfg.App.Loc).Format("Mon Jan _2")
					if e.HasTime {
						dateTime += " at " + e.StartTime.In(v.Page.Cfg.App.Loc).Format(time.Kitchen)
					}

					return h.Dl(
						h.Class("mt-2 flex flex-col text-gray-500 xl:flex-row"),
						h.Div(
							h.Class("flex items-start space-x-3"),
							h.Dt(
								h.Class("mt-0.5"),
								c.TextSROnly("Date"),
								c.IconCalendar()),
							h.Dd(
								g.Text(dateTime))),
						h.Div(
							h.Class("mt-2 flex items-start space-x-3 xl:ml-3.5 xl:mt-0 xl:border-l xl:border-gray-400 xl:border-opacity-50 xl:pl-3.5"),
							h.Dt(
								h.Class("mt-0.5"),
								c.TextSROnly("Location"),
								c.IconMapPin()),
							h.Dd(
								g.Text(e.Location))))
				})),
			c.IconChevronRight(h.Class("text-gray-500 h-8 w-8"))))
}

func (v *EventBrowse) Pager() g.Node {
	showPrevious := !v.FirstDayOfSelectedWeek.Before(time.Now())
	weekNext := v.FirstDayOfSelectedWeek.AddDate(0, 0, 7).Format(time.DateOnly)

	return h.Nav(
		h.Class("flex items-center justify-between border-t border-gray-200 px-4 sm:px-0"),
		h.Div(
			h.Class("-mt-px flex w-0 flex-1"),
			h.A(
				gc.Classes{
					"border-t-2 border-transparent pt-4 pr-1 inline-flex items-center text-sm font-medium": true,
					"text-gray-300": !showPrevious,
					"text-gray-500 hover:text-gray-700 hover:border-gray-300": showPrevious,
				},
				c.LazyIfElse(showPrevious,
					func() g.Node {
						weekPrevious := v.FirstDayOfSelectedWeek.AddDate(0, 0, -7).Format(time.DateOnly)
						return h.Href("?selectedDate=" + weekPrevious)
					},
					h.Disabled,
				),
				g.Text("Previous Week"))),
		h.Div(
			h.Class("-mt-px flex w-0 flex-1 justify-end"),
			h.A(
				h.Class("border-t-2 border-transparent pt-4 pl-1 inline-flex items-center text-sm font-medium text-gray-500 hover:text-gray-700 hover:border-gray-300"),
				h.Href("?selectedDate="+weekNext),
				g.Text("Next Week"))))
}

func (v *EventBrowse) MiniCalendar() g.Node {
	today := time.Now()
	if v.SelectedDate.Before(today) {
		v.SelectedDate = today
	}

	selectedYear, selectedMonth, _ := v.SelectedDate.Date()
	rangeStart := conv.SundayBeforeFirstOfMonth(v.SelectedDate)
	rangeEnd := conv.SaturdayAfterEndOfMonth(v.SelectedDate)
	dates := conv.DateRange(rangeStart, rangeEnd)

	dateButtons := make([]g.Node, 0, len(dates))
	for i, d := range dates {
		isToday := conv.DatesEqual(today, d)
		isInPast := d.Before(today) && !isToday
		isSelectedWeek := conv.DateInRange(d, v.FirstDayOfSelectedWeek, v.LastDayOfSelectedWeek)
		// We can ignore the year because d will always be within one month on
		// either side of the selected day.
		isSelectedMonth := v.SelectedDate.Month() == d.Month()

		classes := gc.Classes{
			"py-1.5 focus:z-10": true,
			"hover:bg-gray-100": !isInPast && !isSelectedWeek,
			"bg-gray-200":       isSelectedWeek,
			"bg-gray-50":        !isSelectedWeek && !isSelectedMonth,
			"bg-white":          isSelectedMonth && !isSelectedWeek,
			"font-semibold":     isSelectedWeek || isToday,
			"text-indigo-600":   isToday,
			"text-gray-900":     isSelectedWeek && isSelectedMonth && !isToday,
			"text-gray-400":     !isSelectedMonth && !isToday,
			"rounded-tl-lg":     i == 0,
			"rounded-tr-lg":     i == 6,
			"rounded-bl-lg":     i == len(dates)-7,
			"rounded-br-lg":     i == len(dates)-1,
		}

		buttonContent := h.Time(
			g.Attr("dateTime", d.Format(time.DateOnly)),
			gc.Classes{
				"mx-auto flex h-7 w-7 items-center justify-center": true,
			},
			g.Text(strconv.Itoa(d.Day())))

		if isInPast || isSelectedWeek {
			dateButtons = append(dateButtons, h.Span(classes, buttonContent))
		} else {
			dateButtons = append(dateButtons, h.A(
				h.Href("?selectedDate="+d.Format(time.DateOnly)),
				classes, buttonContent))
		}
	}

	firstOfNextMonth := conv.FirstOfNextMonth(v.SelectedDate).Format(time.DateOnly)

	return h.Div(
		h.Class("mt-10 text-center lg:col-start-8 lg:col-end-13 lg:row-start-1 lg:mt-9 xl:col-start-9"),
		h.Div(
			h.Class("flex items-center text-gray-900"),
			c.LazyIf(selectedMonth != today.Month() || selectedYear != today.Year(),
				func() g.Node {
					// Do not allow navigating to before the current month or to days before today.
					firstOfPreviousMonth := conv.FirstOfPreviousMonth(v.SelectedDate)
					if firstOfPreviousMonth.Before(today) {
						firstOfPreviousMonth = today
					}

					return h.A(
						h.Href("?selectedDate="+firstOfPreviousMonth.Format(time.DateOnly)),
						h.Class("-m-1.5 flex flex-none items-center justify-center p-1.5 text-gray-400 hover:text-gray-500"),
						c.TextSROnly("Previous month"),
						c.IconChevronLeft(
							h.Class("h-5 w-5")))
				}),
			h.Div(
				h.Class("flex-auto text-sm font-semibold"),
				h.A(
					h.Href("?selectedDate="+today.Format(time.DateOnly)),
					h.Class("hover:text-gray-500"),
					g.Text(fmt.Sprintf("%s %d", selectedMonth.String(), selectedYear)))),
			h.A(
				h.Href("?selectedDate="+firstOfNextMonth),
				h.Class("-m-1.5 flex flex-none items-center justify-center p-1.5 text-gray-400 hover:text-gray-500"),
				c.TextSROnly("Next month"),
				c.IconChevronRight(
					h.Class("h-5 w-5")))),
		h.Div(
			h.Class("mt-6 grid grid-cols-7 text-xs leading-6 text-gray-500"),
			h.Div(g.Text("S")),
			h.Div(g.Text("M")),
			h.Div(g.Text("T")),
			h.Div(g.Text("W")),
			h.Div(g.Text("T")),
			h.Div(g.Text("F")),
			h.Div(g.Text("S"))),
		h.Div(
			h.Class("isolate mt-2 grid grid-cols-7 gap-px rounded-lg bg-gray-200 text-sm shadow ring-1 ring-gray-200"),
			g.Group(dateButtons),
		),
		h.A(
			h.Href(v.Page.ToURL(pages.Names.Event.New)),
			h.Class("block mt-8 w-full rounded-md bg-indigo-600 px-3 py-2 text-sm font-semibold text-white shadow hover:bg-indigo-500 focus-visible:outline focus-visible:outline-2 focus-visible:outline-offset-2 focus-visible:outline-indigo-600"),
			g.Text("Add event")))
}
