package views

import (
	"database/sql"
	"io"
	"net/http"
	"time"

	"go.uber.org/zap"
	g "maragu.dev/gomponents"
	h "maragu.dev/gomponents/html"

	"git.grosinger.net/tgrosinger/tonyslist/pkg/controller"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/conv"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/services"
	c "git.grosinger.net/tgrosinger/tonyslist/pkg/views/components"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/views/layouts"
)

type EventCreateForm struct {
	Title       string `form:"title" valid:"required"`
	Description string `form:"description" valid:"required"`
	Location    string `form:"location" valid:"required"`
	Date        string `form:"date" valid:"required"`
	StartTime   string `form:"start_time" valid:"optional"`
	EndTime     string `form:"end_time" valid:"optional"`

	ProcessedStartTime time.Time
	ProcessedEndTime   sql.NullTime

	Submission controller.FormSubmission
}

func (f *EventCreateForm) Process(logger *zap.Logger, loc *time.Location) {
	startTime, err := conv.ParseTimeParts(f.Date, f.StartTime, loc)
	if err != nil {
		logger.Warn("unable to parse event start date and time",
			zap.String("date", f.Date),
			zap.String("start_time", f.StartTime))
		f.Submission.SetFieldError("Date", "Unable to parse date or start time")
		return
	}

	f.ProcessedStartTime = startTime

	if startTime.Before(time.Now()) {
		f.Submission.SetFieldError("Date", "Events must not be in the past")
	}

	if f.EndTime != "" {
		if f.StartTime == "" {
			f.Submission.SetFieldError("StartTime", "May only specify an end time if start time is also provided")
			return
		}

		endTime, err := conv.ParseTimeParts(f.Date, f.EndTime, loc)
		if err != nil {
			logger.Warn("unable to parse event end date and time",
				zap.String("date", f.Date),
				zap.String("end_time", f.EndTime))
			f.Submission.SetFieldError("EndTime", "Unable to parse end time")
			return
		}

		f.ProcessedEndTime = sql.NullTime{Valid: true, Time: endTime}

		if endTime.Before(startTime) {
			f.Submission.SetFieldError("EndTime",
				"End time must be after start time.")
		}
	}
}

// Sanitize should call the sanitize function on any field which is not
// validated to be of a specific format.
func (f *EventCreateForm) Sanitize(s *services.Sanitizer) {
	f.Title = s.Sanitize(f.Title)
	f.Description = s.SanitizeHTML(f.Description)
}

type EventCreate struct {
	// Location is used to properly determine the current time.
	Location   *time.Location
	ActionHref string
	CancelHref string
	SubmitText string
	Form       EventCreateForm
	Page       controller.Page
}

func (v *EventCreate) Render(w io.Writer) error {
	content := c.WideContainer(
		c.CenteredX(
			c.FullScreenForm(
				http.MethodPost,
				v.ActionHref,
				"",
				"",
				c.FormFooter(nil, v.CancelHref, v.SubmitText, false),
				c.FormSection(
					c.FormHeading("Event Information",
						"Events may only be created by individuals and non-profits."),
					c.CSRF(v.Page),
					c.InputText(c.FormInputParams{
						Name:     "title",
						Display:  "Title",
						Required: true,
						Value:    v.Form.Title,
						Errors:   v.Form.Submission.GetFieldErrors("Title"),
					}),
					c.InputRichTextEditor(c.FormInputParams{
						Name:    "description",
						Display: "Description",
						HintNode: g.Group([]g.Node{
							h.P(
								h.Class("text-sm text-gray-500"),
								g.Text("Write a few sentences about what is happening.")),
						}),
						Required: true,
						Value:    v.Form.Description,
						Errors:   v.Form.Submission.GetFieldErrors("Description"),
					}),
					c.InputText(c.FormInputParams{
						Name:     "location",
						Display:  "Location",
						Required: true,
						Value:    v.Form.Location,
						Errors:   v.Form.Submission.GetFieldErrors("Location"),
					}),
					c.InputDate(c.FormInputParams{
						Name:     "date",
						Display:  "Date",
						Required: true,
						Value:    v.Form.Date,
						Errors:   v.Form.Submission.GetFieldErrors("Date"),
					},
						// Disallow creating events in the past
						h.Min(time.Now().In(v.Location).Format("2006-01-02"))),
					c.InputTime(c.FormInputParams{
						Name:    "start_time",
						Display: "Start Time",
						Hint:    "Optional",
						Value:   v.Form.StartTime,
						Errors:  v.Form.Submission.GetFieldErrors("StartTime"),
					}),
					c.InputTime(c.FormInputParams{
						Name:    "end_time",
						Display: "End Time",
						Hint:    "Optional",
						Value:   v.Form.EndTime,
						Errors:  v.Form.Submission.GetFieldErrors("EndTime"),
					})),
			)),
	)

	l := layouts.Primary{
		Header:  c.EventHeader(v.Page),
		Content: content,
		Page:    v.Page,
	}
	return l.Render(w)
}
