package views

import (
	"fmt"
	"io"
	"time"

	"github.com/yuin/goldmark"
	g "maragu.dev/gomponents"
	comp "maragu.dev/gomponents/components"
	h "maragu.dev/gomponents/html"

	"git.grosinger.net/tgrosinger/tonyslist/constants/pages"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/controller"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
	c "git.grosinger.net/tgrosinger/tonyslist/pkg/views/components"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/views/layouts"
)

type EventDetails struct {
	Event sqlc.Event
	Photo *sqlc.EventPhoto

	// CreatorInfo will contain information about the post creator iff the
	// viewer of this page has moderator permissions and will be able to see the
	// moderator toolbar.
	CreatorInfo *sqlc.GetUserInfoRow

	MarkdownRenderer goldmark.Markdown

	Page controller.Page
}

func (v *EventDetails) Render(w io.Writer) error {
	content := c.WideContainer(
		c.LazyIf(v.Event.Creator == v.Page.AuthUser.ID,
			v.ownerFloatingToolbar),
		c.LazyIf(v.Page.Enforce("event", "moderate"),
			v.moderatorFloatingToolbar),
		v.details(),
	)

	l := layouts.Primary{
		Header:  c.EventHeader(v.Page),
		Content: content,
		Page:    v.Page,
	}
	return l.Render(w)
}

func (v *EventDetails) ownerFloatingToolbar() g.Node {
	colors := "border-gray-300 bg-white text-indigo-600 hover:bg-indigo-50 focus:border-indigo-500 focus:ring-indigo-500"

	photoButtonText := "Add Photo"
	if v.Photo != nil {
		photoButtonText = "Edit Photo"
	}

	buttons := []g.Node{
		c.FloatingToolbarJoinedButtons(
			c.ButtonParams{
				Href:    v.Page.ToURL(pages.Names.Event.Edit, v.Event.ID),
				Display: "Edit",
			},
			c.ButtonParams{
				Href:    v.Page.ToURL(pages.Names.Event.NewPhotos, v.Event.ID),
				Display: photoButtonText,
			},
			colors),
		c.FloatingToolbarButton(c.ButtonParams{
			Href:    v.Page.ToURL(pages.Names.Event.Delete, v.Event.ID),
			Display: "Delete",
		}, colors),
	}

	leftText := h.P(
		h.Class("ml-3 font-medium text-white"),
		g.Text("Your event is live"))

	return c.FloatingToolbar("owner-toolbar", "bg-indigo-600", leftText, g.Group(buttons))
}

func (v *EventDetails) moderatorFloatingToolbar() g.Node {
	editButton := c.FloatingToolbarButton(
		c.ButtonParams{
			Href:    v.Page.ToURL(pages.Names.Event.Edit, v.Event.ID),
			Display: "Edit",
		},
		"border-transparent bg-white text-gray-600 hover:bg-gray-50",
	)
	deleteButton := c.FloatingToolbarButton(
		c.ButtonParams{
			Href:    v.Page.ToURL(pages.Names.Event.Delete, v.Event.ID),
			Display: "Delete",
		},
		"border-transparent bg-white text-gray-600 hover:bg-gray-50",
	)

	name := v.CreatorInfo.Email
	if v.CreatorInfo.Name.Valid {
		name = v.CreatorInfo.Name.String
	}
	left := h.Div(
		h.Class("flex flex-col ml-3 font-medium text-white"),
		h.P(
			g.Text("Moderator Actions")),
		h.Div(
			h.Class("space-x-2 flex text-sm"),
			h.P(
				g.Text("Posted by:")),
			c.TextLink(v.Page.ToURL(pages.Names.Admin.UserDetails, v.Event.Creator), g.Text(name))),
	)

	return c.FloatingToolbar("moderator-toolbar", "bg-gray-600", left, editButton, deleteButton)
}

func (v *EventDetails) details() g.Node {
	loc := v.Page.Cfg.App.Loc

	return h.Div(
		h.Class("mx-auto max-w-2xl py-8 px-4 sm:py-8 sm:px-6 lg:max-w-7xl lg:px-8"),
		h.Div(
			h.Class("mt-4"),
			c.PrimaryHeader(v.Event.Title)),
		h.Div(
			h.Class("lg:grid lg:grid-cols-2 lg:gap-x-8"),
			h.Section(
				h.Aria("labelledby", "information-heading"),
				h.Class("mt-4"),
				h.H2(
					h.ID("information-heading"),
					h.Class("sr-only"),
					g.Text("Event information")),
				h.Div(
					h.Class("mt-4 text-base text-gray-500 markdown-container"),
					g.Raw(v.Event.Description)),
				h.Div(
					h.Class("mt-4"),
					c.ButtonSecondary(v.Page.ToURL(pages.Names.Event.AddCalendar, v.Event.ID),
						c.IconCalendarPlus(),
						g.Text("Add to Calendar")))),
			c.LazyIf(v.Photo != nil, func() g.Node {
				return h.Div(
					h.Class("lg:row-span-2 mt-4 aspect-w-1 aspect-h-1 w-full"),
					h.Img(
						h.Class("h-full w-full object-contain object-center cursor-pointer z-0 absolute inset-0 sm:rounded-lg"),
						h.Src(v.Page.ToUserContentPath(
							fmt.Sprintf("event-%d-%d.jpg", v.Event.ID, v.Photo.ID)))))
			}),
			h.Div(
				comp.Classes{
					"mt-10 flex flex-col divide-y lg:mt-0 lg:self-center": true,
					"lg:col-start-1 lg:row-start-2":                       v.Photo != nil,
					"lg:col-start-2 lg:row-span-2 ":                       v.Photo == nil,
				},
				h.Div(
					h.Class("flex space-x-4 py-4"),
					h.Div(
						h.Class("text-gray-500"),
						c.TextSROnly("Date"),
						c.IconCalendar()),
					h.Div(
						c.TextNormal(v.Event.StartTime.In(loc).Format("Mon Jan _2")))),
				c.LazyIf(v.Event.HasTime,
					func() g.Node {
						return h.Div(
							h.Class("flex space-x-4 py-4"),
							h.Div(
								h.Class("text-gray-500"),
								c.TextSROnly("Time"),
								c.IconClock()),
							h.Div(
								c.LazyIfElse(v.Event.EndTime.Valid,
									func() g.Node {
										return c.TextNormal(fmt.Sprintf("%s to %s",
											v.Event.StartTime.In(loc).Format(time.Kitchen),
											v.Event.EndTime.Time.In(loc).Format(time.Kitchen)))
									},
									func() g.Node {
										return c.TextNormal(v.Event.StartTime.In(loc).Format(time.Kitchen))
									})))
					}),
				h.Div(
					h.Class("flex space-x-4 py-4"),
					h.Div(
						h.Class("text-gray-500"),
						c.TextSROnly("Location"),
						c.IconMapPin()),
					h.Div(
						c.TextNormal(v.Event.Location))))))
}
