package helpers

import (
	"fmt"
	"time"

	"github.com/dustin/go-humanize"
	"github.com/labstack/gommon/random"

	"git.grosinger.net/tgrosinger/tonyslist/config"
)

// CacheBuster stores a random string used as a cache buster for static files.
var CacheBuster = random.String(10)

// humanizeTimes are the times used by RelativeDateTime. It modifies the
// behavior of humanize.Time to cap it at "days ago" instead of going all the
// way up to "weeks ago" or more.
var humanizeTimes = []humanize.RelTimeMagnitude{
	{D: time.Second, Format: "now", DivBy: time.Second},
	{D: 2 * time.Second, Format: "1 second %s", DivBy: 1},
	{D: time.Minute, Format: "%d seconds %s", DivBy: time.Second},
	{D: 2 * time.Minute, Format: "1 minute %s", DivBy: 1},
	{D: time.Hour, Format: "%d minutes %s", DivBy: time.Minute},
	{D: 2 * time.Hour, Format: "1 hour %s", DivBy: 1},
	{D: humanize.Day, Format: "%d hours %s", DivBy: time.Hour},
	{D: 2 * humanize.Day, Format: "1 day %s", DivBy: 1},
	{D: humanize.Week, Format: "%d days %s", DivBy: humanize.Day},
}

func FileNoBuster(filepath string) string {
	return fmt.Sprintf("/%s/%s", config.StaticPrefix, filepath)
}

// File appends a cache buster to a given filepath so it can remain cached until the app is restarted
func File(filepath string) string {
	return fmt.Sprintf("/%s/%s?v=%s", config.StaticPrefix, filepath, CacheBuster)
}

func FormatBool(b bool) string {
	if b {
		return "Yes"
	}
	return "No"
}

// FormatDuration converts a duration into human friendly text description in
// either minutes, hours, or days.
func FormatDuration(d time.Duration) string {
	if d < 2*time.Hour {
		return fmt.Sprintf("%d minutes", int(d.Minutes()))
	}
	if d < 24*time.Hour {
		return fmt.Sprintf("%d hours", int(d.Hours()))
	}
	days := int(d.Hours() / 24)
	if days <= 1 {
		return "1 day"
	}
	return fmt.Sprintf("%d days", days)
}

// RelativeDateTime formats a time into a relative string.
// Capped at "days ago"
//
// RelativeDateTime(someT) -> "3 days ago"
func RelativeDateTime(dt time.Time) string {
	return humanize.CustomRelTime(dt, time.Now(), "ago", "from now", humanizeTimes)
}
