package views

import (
	"fmt"
	"io"

	g "maragu.dev/gomponents"
	gc "maragu.dev/gomponents/components"
	h "maragu.dev/gomponents/html"

	"git.grosinger.net/tgrosinger/tonyslist/config"
	"git.grosinger.net/tgrosinger/tonyslist/constants/pages"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/controller"
	c "git.grosinger.net/tgrosinger/tonyslist/pkg/views/components"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/views/layouts"
)

// NOTE: This page makes limited use of the Typography components because the
// darker backgrounds makes most of the typography components not work here.

type section struct {
	name string
	body []g.Node
}

type topLink struct {
	boldText    string
	description string
	icon        g.Node
	pageName    string
	external    bool
}

var topLinks = []topLink{
	{
		boldText:    "Classifieds",
		description: "Buy, borrow, and sell with your neighbors",
		icon:        c.IconNewspaper(),
		pageName:    pages.Names.Classified.ForSale.Browse,
	},
	// {
	// 	boldText:    "Housing",
	// 	description: "Provide or search for housing",
	// 	icon:        c.IconHouse(),
	// 	pageName:    pages.Names.Housing.ForRent.Browse,
	// },
	{
		boldText:    "Events Calendar",
		description: "What's happening on this island",
		icon:        c.IconCalendarDays(),
		pageName:    pages.Names.Event.Browse,
	},
	{
		boldText:    "Groups",
		description: "Stay in contact with your community",
		icon:        c.IconChatBubbles(),
		pageName:    "https://groups.orcashub.org",
		external:    true,
	},
	{
		boldText:    "Community Resources",
		description: "Information about Orcas and its residents",
		icon:        c.IconCommunityResources(),
		pageName:    "https://wiki.orcashub.org",
		external:    true,
	},
}

type Home struct {
	LeftSectionContent  string
	RightSectionContent string

	Page controller.Page
}

func (v *Home) Render(w io.Writer) error {
	content := h.Div(
		v.TopSection(),
		v.MiddleSection(),
		v.ContactForm(),
	)

	l := layouts.Primary{
		Content: content,
		Page:    v.Page,
	}
	return l.Render(w)
}

func (v *Home) TopSection() g.Node {
	return h.Div(
		h.Class("relative overflow-hidden bg-gray-800"),
		h.Img(
			h.Src(fmt.Sprintf("%s/%s", config.StaticPrefix, "hero.jpg")),
			h.Alt(""),
			h.Class("absolute -z-5 inset-0 h-full w-full object-cover blur-sm opacity-70")),
		h.Div(
			h.Class("relative pt-6 mx-auto max-w-7xl mt-16 sm:mt-24 sm:pb-24"),
			h.Div(
				h.Class("lg:grid lg:grid-cols-12 lg:gap-8"),
				v.TopSectionLinks(),
				v.TopSectionLogin())))
}

func (v *Home) TopSectionLinks() g.Node {
	return h.Div(
		h.Class("px-4 sm:px-6 sm:text-center md:mx-auto md:max-w-2xl lg:col-span-7 lg:flex lg:items-center lg:text-left"),
		h.Section(
			h.H1(
				h.Class("mt-4 text-4xl font-bold tracking-tight text-white sm:text-5xl md:text-6xl"),
				g.Text("Welcome to OrcasHub")),
			h.P(
				h.Class("mt-3 text-base text-gray-100 sm:mt-5 sm:text-xl lg:text-lg xl:text-xl"),
				g.Text("Made on Orcas, for the Orcas Community.")),
			h.Div(
				h.Class("mt-10 max-w-xl text-base leading-7 text-gray-100 lg:max-w-none sm:text-xl lg-text-lg xl:text-xl"),
				g.Group(
					g.Map(topLinks, func(tl topLink) g.Node {
						href := tl.pageName
						if !tl.external {
							href = v.Page.ToURL(tl.pageName)
						}

						return h.A(
							g.If(tl.pageName != "",
								h.Href(href)),
							gc.Classes{
								"block p-4 grid grid-cols-8 rounded-md":                                               true,
								"cursor-pointer hover:bg-gray-700/70 transition duration-100 ease-out hover:ease-in ": tl.pageName != "",
							},
							h.Div(
								h.Class("self-center"),
								c.LazyIf(tl.icon != nil, func() g.Node {
									return tl.icon
								})),
							h.Div(
								h.Class("col-span-6"),
								c.LazyIf(tl.boldText != "", func() g.Node {
									return h.Span(
										h.Class("block font-semibold text-white"),
										g.Text(tl.boldText))
								}),
								h.P(
									g.Text(tl.description))),
							h.Div(
								h.Class("self-center"),
								g.If(tl.pageName != "", c.IconArrowRight(h.Class("h-8 w-8")))))
					})))))
}

func (v *Home) TopSectionLogin() g.Node {
	var authForm g.Node

	if !v.Page.IsAuth {
		r := Register{Page: v.Page}

		authForm = h.Div(
			h.Class("bg-white p-4 sm:mx-auto sm:w-full sm:max-w-lg sm:overflow-hidden sm:rounded-lg"),
			c.SecondaryHeader("Join the community"),
			h.P(
				h.Class("text-2xl font-bold tracking-tight text-indigo-600 sm:text-2xl"),
				g.Text("Receive a digest of new Classified posts.")),
			h.Form(
				h.Class("mt-8"),
				h.Action(v.Page.ToURL(pages.Names.User.Login)),
				h.Method("POST"),
				c.InputEmailFullWidth(c.FormInputParams{
					Name:     "email",
					Display:  "Email Address",
					Required: true,
				}),
				c.MothersMaidenName(),
				c.CSRF(v.Page),
				r.Terms("Sign Up"),
				r.Button("Sign Up", "", "")))
	}

	return h.Section(
		h.Class("mt-16 sm:mt-24 sm:px-4 lg:col-span-5 lg:mt-0"),
		authForm)
}

func (v *Home) MiddleSection() g.Node {
	return h.Div(
		h.Class("my-8 mx-auto flex max-w-7xl px-4 sm:py-6 sm:px-6 lg:px-8"),
		h.Div(
			h.Class("lg:grid lg:grid-cols-12 lg:gap-8"),
			h.Section(
				h.Class("homepage-markdown-container lg:col-span-6"),
				g.Raw(v.LeftSectionContent)),
			h.Section(
				h.Class("homepage-markdown-container lg:col-span-6"),
				g.Raw(v.RightSectionContent))))
}

func (v *Home) ContactForm() g.Node {
	contactP := func(text string) g.Node {
		return h.P(
			h.Class("mt-6 max-w-3xl text-base text-indigo-50"),
			g.Text(text))
	}

	formLabel := func(name, display string) g.Node {
		return h.Label(
			h.For(name),
			h.Class("block text-sm font-medium text-gray-900"),
			g.Text(display))
	}

	textInput := func(name, inputType string) g.Node {
		return h.Div(
			h.Class("mt-1"),
			h.Input(
				h.Name(name),
				h.Type(inputType),
				h.AutoComplete(name),
				h.Class("block w-full rounded-md border-gray-300 py-3 px-4 text-gray-900 shadow-sm focus:border-indigo-500 focus:ring-indigo-500")))
	}

	return h.Div(
		h.Class("bg-gray-100"),
		h.Div(
			h.Class("mx-auto max-w-7xl py-16 px-4 sm:py-24 sm:px-6 lg:px-8"),
			h.Div(
				h.Class("relative bg-white shadow-xl"),
				h.H2(
					h.Class("sr-only"),
					g.Text("Contact us")),

				h.Div(
					h.Class("grid grid-cols-1 lg:grid-cols-3"),
					h.Div(
						h.Class("relative overflow-hidden bg-indigo-700 py-10 px-6 sm:px-10 xl:p-12"),
						h.H3(
							h.Class("text-lg font-medium text-white"),
							g.Text("Contact")),
						contactP("Have feedback or a question? Please send me a message!"),
						contactP("As an individual, I can not respond to every "+
							"feature request message, but all messages will be "+
							"read. Thank you!")),
					h.Div(
						h.Class("py-10 px-6 sm:px-10 lg:col-span-2 xl:p-12"),
						c.TextLarge("Send us a message"),
						h.Form(
							h.Class("mt-6 grid grid-cols-1 gap-y-6 sm:grid-cols-2 sm:gap-x-8"),
							h.Method("POST"),
							h.Action(v.Page.ToURL(pages.Names.Contact)),
							h.Div(
								formLabel("name", "Name"),
								textInput("name", "text")),
							h.Div(
								formLabel("email", "Email"),
								textInput("email", "email")),
							c.MothersMaidenName(),
							h.Div(
								h.Class("sm:col-span-2"),
								formLabel("message", "Message"),
								h.Div(
									h.Class("mt-1"),
									h.Textarea(
										h.Name("message"),
										h.Rows("4"),
										h.Class("block w-full rounded-md border border-gray-300 py-3 px-4 text-gray-900 shadow-sm focus:border-indigo-500 focus:ring-indigo-500")))),
							c.CSRF(v.Page),
							h.Div(
								h.Class("sm:col-span-2 sm:flex sm:justify-end"),
								c.FormSubmit("Submit"))))))))
}
