package views

import (
	"fmt"
	"io"
	"net/http"
	"strings"

	"go.uber.org/zap"
	g "maragu.dev/gomponents"
	h "maragu.dev/gomponents/html"

	"git.grosinger.net/tgrosinger/tonyslist/constants"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/controller"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/conv"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/housing"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/services"
	c "git.grosinger.net/tgrosinger/tonyslist/pkg/views/components"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/views/layouts"
)

type HousingUpdateForm struct {
	Title       string `form:"title" valid:"required"`
	Description string `form:"description" valid:"required"`
	Pets        bool   `form:"pets_allowed"`
	Shared      bool   `form:"shared"`

	// These fields are strings even though they are nillable because otherwise
	// an empty form value is turned into a 0
	Bedrooms  string `form:"bedrooms" valid:"numeric,optional"`
	Bathrooms string `form:"bathrooms" valid:"bathroomCount,optional"`
	Parking   string `form:"parking" valid:"numeric,optional"`

	// This field is only used on for-rent housing posts
	Rent string `form:"rent" valid:"money,optional"`

	Submission controller.FormSubmission
}

type HousingCreateForm struct {
	// Embed the HousingUpdateForm as these fields are needed when
	// creating a post too. The Submission controller from the HousingUpdateForm
	// will be used for this form too.
	HousingUpdateForm

	HousingType string `form:"housing_type" valid:"required"`
	Street      string `form:"street" valid:"optional"`
	Zipcode     string `form:"zipcode" valid:"numeric,stringlength(5|5),optional"`

	PhoneNumber  string `form:"phone_number" valid:"phone,optional"`
	SMSNumber    string `form:"sms_number" valid:"phone,optional"`
	IncludeEmail bool   `form:"include_email"`
}

func (f *HousingCreateForm) ProcessEditableFields() (housing.EditableFields, bool) {
	bedrooms, err := conv.StringToNullInt("bedrooms", f.Bedrooms)
	if err != nil {
		f.Submission.SetFieldError("bedrooms", "Invalid value")
		return housing.EditableFields{}, false
	}

	bathrooms, err := conv.StringToNullFloat64("bathrooms", f.Bathrooms)
	if err != nil {
		f.Submission.SetFieldError("bathrooms", "Invalid value")
		return housing.EditableFields{}, false
	}

	parking, err := conv.StringToNullInt("parking", f.Parking)
	if err != nil {
		f.Submission.SetFieldError("parking", "Invalid value")
		return housing.EditableFields{}, false
	}

	// Strip any commas from the rent field
	f.Rent = strings.ReplaceAll(f.Rent, ",", "")

	rent, err := conv.StringToNullFloat64("rent", f.Rent)
	if err != nil {
		f.Submission.SetFieldError("rent", "Invalid value")
		return housing.EditableFields{}, false
	}

	return housing.EditableFields{
		Bedrooms:     bedrooms,
		Bathrooms:    bathrooms,
		ParkingSpots: parking,
		Pets:         f.Pets,
		Shared:       f.Shared,
		Rent:         rent,
		Street:       conv.StringToNullString(f.Street),
		Zipcode:      conv.StringToNullString(f.Zipcode),
	}, true
}

// ProcessCreate verifies fields which are settable when creating a housing
// post. This process function should not be called when updating a post.
func (f *HousingCreateForm) ProcessCreate(logger *zap.Logger) {
	if f.HousingType != string(constants.ForRent) &&
		f.HousingType != string(constants.Request) {
		f.Submission.SetFieldError("HousingType", "Invalid housing type in request")
	}

	if !f.IncludeEmail && f.PhoneNumber == "" && f.SMSNumber == "" {
		f.Submission.SetFieldError("IncludeEmail", "Must include at least one contact method")
	}
}

// sanitize should call the sanitize function on any field which is not
// validated to be of a specific format.
func (f *HousingUpdateForm) Sanitize(s *services.Sanitizer) {
	f.Rent = strings.ReplaceAll(f.Rent, ",", "")
	f.Title = s.Sanitize(f.Title)
	f.Description = s.SanitizeHTML(f.Description)
}

// sanitize should call the sanitize function on any field which is not
// validated to be of a specific format.
func (f *HousingCreateForm) Sanitize(s *services.Sanitizer) {
	f.Street = s.Sanitize(f.Street)
	f.HousingUpdateForm.Sanitize(s)
}

type HousingCreate struct {
	IsCreate   bool
	ActionHref string
	CancelHref string
	SubmitText string
	Form       HousingCreateForm
	Page       controller.Page
}

func (v *HousingCreate) Render(w io.Writer) error {
	content := c.WideContainer(
		c.CenteredX(
			c.FullScreenForm(
				http.MethodPost,
				v.ActionHref,
				"",
				fmt.Sprintf("{housing_type: '%s'}", v.Form.HousingType),
				c.FormFooter(
					h.Div(
						h.Class("flex flex-col grow"),
						c.LazyIf(v.IsCreate, func() g.Node {
							return h.Span(
								c.XShow("housing_type === 'for-rent'"),
								h.Class("grow text-gray-500"),
								g.Text("Add photos on the following page after clicking next."))
						}),
						h.Span(
							h.Class("grow text-gray-500"),
							g.Text("Posts are active for 14 days, after which they may be renewed."))),
					v.CancelHref, v.SubmitText, false),
				c.FormSection(
					c.FormHeading("Post Information", ""),
					c.CSRF(v.Page),
					c.LazyIf(v.IsCreate, func() g.Node {
						return c.InputSelect(c.FormSelectParams{
							FormInputParams: c.FormInputParams{
								Name:     "housing_type",
								Display:  "Housing Type",
								Required: true,
								Value:    v.Form.HousingType,
								Errors:   v.Form.Submission.GetFieldErrors("HousingType"),
							},
							XModel: "housing_type",
							Options: []c.FormSelectOption{
								{Display: "Housing Available", Value: string(constants.ForRent)},
								{Display: "Housing Request", Value: string(constants.Request)},
							},
						})
					}),
					c.InputText(c.FormInputParams{
						Name:     "title",
						Display:  "Title",
						Required: true,
						Value:    v.Form.Title,
						Errors:   v.Form.Submission.GetFieldErrors("Title"),
					}),
					h.Div(
						c.XShow("housing_type === 'for-rent'"),
						c.LazyIf(v.IsCreate,
							func() g.Node {
								return c.InputStreetAndZip(
									"Street Address",
									v.Form.Street,
									v.Form.Zipcode,
									"A map showing the approximate location will be displayed in the post",
									v.Form.Submission.GetFieldErrors("Street"),
									v.Form.Submission.GetFieldErrors("Zipcode"))
							}),
						c.InputDollar(c.FormInputParams{
							Name:    "rent",
							Display: "Rent",
							Value:   v.Form.Rent,
							Errors:  v.Form.Submission.GetFieldErrors("Rent"),
						})),
					c.InputRichTextEditor(c.FormInputParams{
						Name:    "description",
						Display: "Description",
						HintNode: g.Group([]g.Node{
							c.XIfTemplate(`housing_type === 'for-rent'`,
								h.P(
									h.Class("text-sm text-gray-500"),
									g.Text("Please provide detailed information about the housing available."))),
							c.XIfTemplate(`housing_type === 'request'`,
								h.P(
									h.Class("text-sm text-gray-500"),
									g.Text("Write a few sentences about your housing request."))),
						}),
						Required: true,
						Value:    v.Form.Description,
						Errors:   v.Form.Submission.GetFieldErrors("Description"),
					}),
					c.InputSteppedNumber("0", "1", c.FormInputParams{
						Name:    "bedrooms",
						Display: "Bedrooms",
						Hint:    "Optional",
						Value:   v.Form.Bedrooms,
						Errors:  v.Form.Submission.GetFieldErrors("Bedrooms"),
					}),
					c.InputSteppedNumber("0", "0.25", c.FormInputParams{
						Name:    "bathrooms",
						Display: "Bathrooms",
						Hint:    "Optional",
						Value:   v.Form.Bathrooms,
						Errors:  v.Form.Submission.GetFieldErrors("Bathrooms"),
					}),
					c.InputSteppedNumber("0", "1", c.FormInputParams{
						Name:    "parking",
						Display: "Parking Spaces",
						Hint:    "Optional",
						Value:   v.Form.Parking,
						Errors:  v.Form.Submission.GetFieldErrors("Parking"),
					}),
					c.InputCheckbox(v.Form.Pets, c.FormInputParams{
						Name:    "pets_allowed",
						Display: "Pets Allowed",
					}),
					c.InputCheckbox(v.Form.Shared, c.FormInputParams{
						Name:    "shared",
						Display: "Shared",
						Hint:    "For example, a single room in a house.",
					})),
				c.LazyIf(v.IsCreate, func() g.Node {
					return c.FormSection(
						c.FormHeading("Contact Information",
							"How would you like those interested to contact you? Leave a field blank if you do not want it to be used. Contact information entered will be displayed to all site visitors."),
						c.InputCheckbox(v.Form.IncludeEmail, c.FormInputParams{
							Name:    "include_email",
							Display: "Include Anonymous Email",
							Hint:    "Display an anonymous email address. Messages will forward to the email associated with your account.",
							Errors:  v.Form.Submission.GetFieldErrors("IncludeEmail"),
						}),
						c.InputPhone(c.FormInputParams{
							Name:    "phone_number",
							Display: "Phone Call",
							Value:   v.Form.PhoneNumber,
							Errors:  v.Form.Submission.GetFieldErrors("PhoneNumber"),
						}),
						c.InputPhone(c.FormInputParams{
							Name:    "sms_number",
							Display: "Text/SMS Message",
							Value:   v.Form.SMSNumber,
							Errors:  v.Form.Submission.GetFieldErrors("SMSNumber"),
						}))
				}))),
	)

	l := layouts.Primary{
		Header:  c.HousingHeader(v.Page),
		Content: content,
		Page:    v.Page,
	}
	return l.Render(w)
}
