package views

import (
	"fmt"
	"io"
	"strings"

	g "maragu.dev/gomponents"
	gc "maragu.dev/gomponents/components"
	h "maragu.dev/gomponents/html"

	"git.grosinger.net/tgrosinger/tonyslist/constants"
	"git.grosinger.net/tgrosinger/tonyslist/constants/pages"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/controller"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/postbase"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
	c "git.grosinger.net/tgrosinger/tonyslist/pkg/views/components"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/views/layouts"
)

const reloadEnabled = true

type KioskClassified struct {
	PostBase   postbase.PostBase
	Classified sqlc.Classified
	Photos     []sqlc.Photo
	BodyOnly   bool

	// Should be a valid img src. Can be a base64 encoded image.
	QRCode g.Node

	Page controller.Page
}

func (v *KioskClassified) Render(w io.Writer) error {
	if v.BodyOnly {
		return v.body().Render(w)
	}

	content := h.Div(
		h.Class("mx-auto max-w-7xl h-full"),
		h.Div(
			h.Class("bg-white shadow my-6 mx-8 px-8 rounded-lg h-[calc(100%-3rem)]"),
			v.body()))

	l := layouts.Kiosk{
		Content: content,
		Page:    v.Page,
	}
	return l.Render(w)
}

func (v *KioskClassified) body() g.Node {
	return h.Div(
		c.LazyIf(reloadEnabled, func() g.Node {
			return g.Group([]g.Node{
				c.HxGet(v.Page.ToURL(pages.Names.Classified.Kiosk) + fmt.Sprintf("?current_id=%d", v.PostBase.ID)),
				c.HxTrigger("every 15s"),
				c.HxSwap("outerHTML"),
			})
		}),
		h.Class("mx-auto flex justify-start max-w-7xl gap-x-8 px-8 py-8 h-full"),
		v.postDetails(),
		v.photoGallery())
}

func (v *KioskClassified) typeBadge(t constants.ClassifiedType) g.Node {
	var text string
	switch t {
	case constants.ForSale:
		text = "For Sale"
	case constants.ForFree:
		text = "For Free"
	case constants.WantToBorrow:
		text = "Want to Borrow"
	case constants.WantToBuy:
		text = "Looking For"
	default:
		return nil
	}

	return h.Span(
		gc.Classes{
			"inline-flex items-center rounded-full px-2 py-1 text-lg font-medium ring-1 ring-inset": true,
			"bg-red-100 text-red-900 ring-red-600/10":                                               t == constants.ForSale,
			"bg-yellow-100 text-yellow-900 ring-yellow-600/20":                                      t == constants.ForFree,
			"bg-green-100 text-green-900 ring-green-600/20":                                         t == constants.WantToBorrow,
			"bg-blue-100 text-blue-900 ring-blue-700/10":                                            t == constants.WantToBuy,
		},
		g.Text(text))
}

func (v *KioskClassified) postDetails() g.Node {
	descRaw := v.PostBase.Description
	if strings.HasPrefix(descRaw, "<div>") {
		descRaw = strings.TrimPrefix(descRaw, "<div>")
		descRaw = `<div class="line-clamp-[15]">` + descRaw
	}

	return h.Div(
		h.Class("basis-1/2 flex flex-col overflow-hidden"),
		h.Div(
			h.Class("mt-4 flex-none"),
			c.PrimaryHeader(v.PostBase.Title)),
		h.Section(
			h.Aria("labelledby", "information-heading"),
			h.Class("mt-4 flex-1 flex flex-col overflow-hidden"),
			h.Div(
				h.Class("flex items-center gap-x-4 flex-none"),
				v.typeBadge(v.Classified.ClassifiedType),
				c.LazyIf(v.Classified.Price.Valid, func() g.Node {
					return c.TextLarge("$" + c.TextCurrency(&v.Classified.Price.Float64))
				})),
			h.Div(
				h.Class("mt-4 overflow-hidden text-2xl text-gray-900 markdown-container thellipsis"),
				g.Raw(descRaw))),
		h.Div(
			h.Class("h-40 flex gap-x-8 items-center"),
			h.Div(
				h.Class("h-40 w-40"),
				v.QRCode),
			h.P(
				h.Class("font-semibold text-gray-900 text-2xl"),
				g.Text("Scan to see more details!"))))
}

func (v *KioskClassified) photoGallery() g.Node {
	if len(v.Photos) == 0 {
		return nil
	}

	// Rotate between images.
	imageRotateSec := 3
	initJS := fmt.Sprintf(`
		setInterval(() => {
			selected_idx = (selected_idx + 1) %% photos.length;
		}, %d)
	`, imageRotateSec*1000)

	return h.Div(
		c.XData("{ selected_idx: 0, photos: %s}", v.photoList()),
		h.Class("basis-1/2 flex flex-col"),
		h.Div(
			c.XInit(initJS),
			h.Class("aspect-w-1 aspect-h-1 w-full"),
			h.Img(
				c.XBind("src", "photos[selected_idx]"),
				h.Class("h-full w-full object-contain object-center cursor-pointer z-0 absolute inset-0 rounded-lg"),
				h.Src(v.Page.ToUserContentPath(v.Photos[0].GetRegularFilename())))))
}

// photoList converts the list of photos on the post into a list of urls used
// by alpine.js to switch between the different photo options.
func (v *KioskClassified) photoList() string {
	b := strings.Builder{}
	b.WriteString("[")
	for _, p := range v.Photos {
		b.WriteString("'")
		b.WriteString(v.Page.ToUserContentPath(p.GetRegularFilename()))
		b.WriteString("',")
	}

	b.WriteString("]")
	return b.String()
}
