package layouts

import (
	"io"
	"strconv"
	"strings"

	g "maragu.dev/gomponents"
	gc "maragu.dev/gomponents/components"
	h "maragu.dev/gomponents/html"

	"git.grosinger.net/tgrosinger/tonyslist/constants/pages"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/controller"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/msg"
	c "git.grosinger.net/tgrosinger/tonyslist/pkg/views/components"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/views/helpers"
)

type Primary struct {
	// Header is the nav below the global nav, specific to the page being
	// viewed. If nil there will be no secondary nav.
	Header g.Node

	// Content is the main body of the page below the nav.
	Content g.Node

	Page controller.Page
}

func (l *Primary) Render(w io.Writer) error {
	return h.Doctype(
		h.HTML(
			h.Class("h-full bg-gray-100"),
			l.head(),
			h.Body(
				h.Class("h-full"),
				h.Div(
					h.Class("min-h-full"),
					l.globalNav(),
					l.Header,
					h.Main(
						l.Content),
					l.footer(),
				),
				l.messages())),
	).Render(w)
}

func (l *Primary) head() g.Node {
	title := l.Page.Cfg.App.Name
	if l.Page.Title != "" {
		title += " | " + l.Page.Title
	}

	return h.Head(
		h.TitleEl(
			g.Text(title)),

		// Favicons
		// TODO: Serve these from the root instead
		c.Favicon("image/svg+xml", "", "favicon.svg"),
		c.Favicon("image/png", "180x180", "apple-touch-icon.png"),
		c.Favicon("image/png", "16x16", "favicon-16x16.png"),
		c.Favicon("image/png", "32x32", "favicon-32x32.png"),
		c.Favicon("image/png", "192x192", "android-chrome-192x192.png"),
		c.StaticFileLink("favicon.ico",
			h.Rel("shortcut icon")),

		c.StaticFileLink("site.webmanifest",
			h.Rel("manifest")),
		c.StaticFileLink("safari-pinned-tab.svg",
			h.Rel("mask-icon"),
			g.Attr("color", "#086983")),

		h.Meta(
			h.Name("apple-mobile-web-app-title"),
			h.Content(l.Page.Cfg.App.Name)),
		h.Meta(
			h.Name("application-name"),
			h.Content(l.Page.Cfg.App.Name)),
		h.Meta(
			h.Name("msapplication-TileColor"),
			h.Content("#ffffff")),
		h.Meta(
			h.Name("msapplication-config"),
			h.Content(helpers.FileNoBuster("browserconfig.xml"))),
		h.Meta(
			h.Name("theme-color"),
			h.Content("#ffffff")),
		h.Meta(
			h.Charset("utf-8")),
		h.Meta(
			h.Name("viewport"),
			h.Content("width=device-width, initial-scale=1")),
		h.Meta(
			g.Attr("http-equiv", "X-UA-Compatible"),
			h.Content("IE=edge")),
		c.LazyIf(l.Page.Metatags.Description != "",
			func() g.Node {
				return h.Meta(
					h.Name("description"),
					h.Content(l.Page.Metatags.Description))
			}),
		c.LazyIf(len(l.Page.Metatags.Keywords) > 0,
			func() g.Node {
				return h.Meta(
					h.Name("keywords"),
					h.Content(strings.Join(l.Page.Metatags.Keywords, ", ")))
			}),

		c.LazyIf(l.Page.Metatags.HasFeed,
			func() g.Node {
				return h.Link(
					h.Rel("alternate"),
					h.Type("application/rss+xml"),
					h.Href(l.Page.RSSLink()))
			}),

		h.Link(
			h.Rel("stylesheet"),
			h.Href(helpers.File("css/styles.css"))),

		c.LazyIf(
			strings.HasSuffix(l.Page.Path, "/new") || strings.HasSuffix(l.Page.Path, "/edit"),
			func() g.Node {
				return g.Group([]g.Node{
					h.Script(
						h.Defer(),
						h.Src(helpers.FileNoBuster("js/alpine.js-mask-3.14.7.js"))),
					h.Script(
						h.Defer(),
						h.Src(helpers.FileNoBuster("js/trix-2.1.8.js"))),
					h.Link(
						h.Rel("stylesheet"),
						h.Type("text/css"),
						h.Href(helpers.FileNoBuster("css/trix-2.1.8.css"))),
				})
			}),
		c.LazyIf(strings.HasPrefix(l.Page.Path, "/classified") || strings.HasPrefix(l.Page.Path, "/housing"),
			func() g.Node {
				return h.Script(
					h.Defer(),
					h.Src(helpers.FileNoBuster("js/alpine.js-persist-3.14.7.js")))
			}),
		c.LazyIf(l.Page.Path == "/donate",
			func() g.Node {
				// Must load the Stripe library from Stripe to be PCI compliant.
				return h.Script(
					h.Src("https://js.stripe.com/v3/"))
			}),
		l.htmx(),
		h.Script(
			h.Defer(),
			h.Src(helpers.FileNoBuster("js/alpine.js-3.14.7.js"))))
}

func (l *Primary) htmx() g.Node {
	if strings.HasPrefix(l.Page.Path, l.Page.ToURL(pages.Names.Admin.Home)) ||
		l.Page.Path == l.Page.ToURL(pages.Names.Classified.ForFree.Browse) ||
		l.Page.Path == l.Page.ToURL(pages.Names.Classified.ForSale.Browse) ||
		l.Page.Path == l.Page.ToURL(pages.Names.Classified.WantToBorrow.Browse) ||
		l.Page.Path == l.Page.ToURL(pages.Names.Classified.WantToBuy.Browse) ||
		l.Page.Path == l.Page.ToURL(pages.Names.Housing.Browse) {
		return h.Script(
			h.Defer(),
			h.Src(helpers.FileNoBuster("js/htmx-2.0.4.js")))
	}
	return nil
}

func (l *Primary) globalNav() g.Node {
	return h.Nav(
		h.Class("bg-gray-700"),
		c.XData(`{ mobile_menu_open: false, profile_dropdown_open: false }`),
		h.Div(
			h.Div(
				h.Class("py-6"),
				h.Div(
					h.Class("flex justify-center"),
					h.A(
						h.Href(l.Page.ToURL(pages.Names.Home)),
						c.OrcasHubTopLogo()))),
			h.Div(
				// Mobile menu button
				h.Class("absolute right-2 top-2 flex md:hidden"),
				h.Button(
					h.Type("button"),
					h.Class("inline-flex items-center justify-center rounded-md bg-gray-800 p-2 text-gray-400 hover:bg-gray-700 hover:text-white focus:outline-none focus:ring-2 focus:ring-white focus:ring-offset-2 focus:ring-offset-gray-800"),
					c.XOn("click", "mobile_menu_open=!mobile_menu_open"),
					h.Aria("controls", "mobile-menu"),
					c.XBind("aria-expanded", "mobile_menu_open"),
					c.TextSROnly("Open main menu"),
					h.Div(
						c.XShow("!mobile_menu_open"),
						c.IconMenu()),
					h.Div(
						c.XCloak(),
						c.XShow("mobile_menu_open"),
						c.IconX())))),
		h.Div(
			h.ID("navbar"),
			h.Class("mx-auto max-w-7xl px-4 sm:px-6 lg:px-8"),
			h.Div(
				h.Class("hidden h-16 items-center justify-between md:flex"),
				h.Div(
					h.Class("flex items-center"),
					h.Div(
						h.Class("flex items-baseline space-x-4"),
						l.renderPrimaryNavButtons(l.navButton))),
				h.Div(
					h.Class("hidden md:block"),
					h.Div(
						h.Class("ml-4 flex items-center md:ml-6"),
						c.LazyIfElse(l.Page.IsAuth,
							func() g.Node {
								return g.Group([]g.Node{
									h.Div(
										h.Class("relative ml-3"),
										h.A(
											h.Href(l.Page.ToURL(pages.Names.Donate)),
											h.Target("_blank"),
											h.Class("inline-flex items-center rounded-md border border-transparent bg-indigo-600 px-3 py-2 text-sm font-medium leading-4 text-white shadow-sm hover:bg-indigo-700 focus:outline-none focus:ring-2 focus:ring-indigo-500 focus:ring-offset-2"),
											g.Text("Donate"),
											c.IconGift())),
									h.Div(
										// Profile dropdown
										h.Class("relative ml-3"),
										h.Span(
											c.XOn("click", "profile_dropdown_open=!profile_dropdown_open"),
											c.XBind("aria-expanded", "profile_dropdown_open"),
											c.XBind("class", `{ 'bg-gray-700 text-white': profile_dropdown_open, 'text-gray-300 hover:bg-gray-700 hover:text-white': !profile_dropdown_open }`),
											h.Aria("haspopup", "true"),
											h.Class("cursor-pointer rounded-md px-3 py-2 text-sm font-medium text-gray-300 hover:bg-gray-700 hover:text-white"),
											g.Text("Account")),
										h.Div(
											c.XShow("profile_dropdown_open"),
											c.XCloak(),
											h.Role("menu"),
											h.Aria("orientation", "vertical"),
											h.Aria("labelledby", "user-menu-button"),
											c.XOn("click.outside", "profile_dropdown_open=false"),
											h.TabIndex("-1"),
											h.Class("absolute right-0 z-50 mt-2 w-48 origin-top-right rounded-md bg-white py-1 shadow-lg ring-1 ring-black ring-opacity-5 focus:outline-none"),
											c.XTransition(c.TransitionOptions{
												DurationEnter:  "duration-100",
												DurationLeave:  "duration-75",
												OperationEnter: "ease-out",
												OperationLeave: "ease-in",
												OpacityStart:   "opacity-0",
												OpacityEnd:     "opacity-100",
												ScaleStart:     "scale-95",
												ScaleEnd:       "scale-100",
											}),
											l.dropdownMenuItem(0, "My Posts", l.Page.ToURL(pages.Names.Posts.MyPosts)),
											l.dropdownMenuItem(1, "Account Settings", l.Page.ToURL(pages.Names.User.Account)),
											l.dropdownMenuItem(2, "Sign out", l.Page.ToURL(pages.Names.User.Logout)),
										)),
								})
							},
							func() g.Node {
								isLoginPage := l.Page.Path == "/user/login"
								return h.A(
									h.Href(l.Page.ToURL(pages.Names.User.Login)),
									gc.Classes{
										"px-3 py-2 rounded-md text-sm font-medium":         true,
										"text-gray-300 hover:bg-gray-700 hover:text-white": isLoginPage,
										"bg-gray-900 text-white":                           !isLoginPage,
									},
									g.Text("Login or Register"))
							}))))),
		l.mobileMenu())
}

func (l *Primary) navButton(display, pageName string, current bool, external bool) g.Node {
	href := pageName
	if !external {
		href = l.Page.ToURL(pageName)
	}

	return h.Div(
		h.Class("relative"),
		h.A(
			h.Href(href),
			gc.Classes{
				"px-3 py-2 rounded-md text-sm font-medium":         true,
				"bg-gray-900 text-white":                           current,
				"text-gray-300 hover:bg-gray-700 hover:text-white": !current,
			},
			g.Text(display)))
}

func (l *Primary) dropdownMenuItem(index int, display, href string) g.Node {
	return h.A(
		h.Class("block px-4 py-2 text-sm text-gray-700 hover:bg-gray-100"),
		h.Href(href),
		h.Role("menuitem"),
		h.TabIndex("-1"),
		h.ID("user-menu-item-"+strconv.Itoa(index)),
		g.Text(display))
}

func (l *Primary) mobileMenu() g.Node {
	return h.Div(
		h.Class("md:hidden"),
		h.ID("mobile-menu"),
		c.XCloak(),
		c.XShow("mobile_menu_open"),
		h.Div(
			h.Class("space-y-1 px-2 pb-2 pt-2 sm:px-3"),
			l.renderPrimaryNavButtons(l.mobileMenuItem),
		),
		h.Div(
			h.Class("border-t border-gray-500 px-2 pb-3 pt-2 sm:px-3"),
			c.LazyIfElse(l.Page.IsAuth,
				func() g.Node {
					return g.Group([]g.Node{
						l.mobileMenuItem("My Posts", pages.Names.Posts.MyPosts, false, false),
						l.mobileMenuItem("Account Settings", pages.Names.User.Account, false, false),
						l.mobileMenuItem("Sign out", pages.Names.User.Logout, false, false),
					})
				},
				func() g.Node {
					return l.mobileMenuItem("Login", pages.Names.User.Login,
						l.Page.Path == l.Page.ToURL(pages.Names.User.Login), false)
				})))
}

func (l *Primary) renderPrimaryNavButtons(fn func(display, pageName string, current, external bool) g.Node) g.Node {
	return g.Group([]g.Node{
		fn("Home", pages.Names.Home, l.Page.Path == "" || l.Page.Path == "/", false),
		fn("Classifieds", pages.Names.Classified.ForSale.Browse, strings.HasPrefix(l.Page.Path, "/classified"), false),
		// fn("Housing", pages.Names.Housing.ForRent.Browse, strings.HasPrefix(l.Page.Path, "/housing"), false),
		fn("Events", pages.Names.Event.Browse, strings.HasPrefix(l.Page.Path, "/event"), false),
		fn("Groups", "https://groups.orcashub.org", false, true),
		fn("Resources", "https://wiki.orcashub.org", false, true),
	})
}

func (l *Primary) mobileMenuItem(display, pageName string, current, external bool) g.Node {
	href := pageName
	if !external {
		href = l.Page.ToURL(pageName)
	}

	return h.A(
		h.Href(href),
		gc.Classes{
			"block rounded-md px-3 py-2 text-base font-medium": true,
			"bg-gray-900 text-white":                           current,
			"text-gray-300 hover:bg-gray-700 hover:text-white": !current,
		},
		g.Text(display))
}

func (l *Primary) footer() g.Node {
	return h.Footer(
		h.Class("bg-gray-800"),
		h.Aria("labelledby", "footer-heading"),
		h.H2(
			h.ID("footer-heading"),
			h.Class("sr-only"),
			g.Text("Footer")),
		h.Div(
			h.Class("mx-auto max-w-7xl px-4 py-12 sm:px-6 lg:px-8 lg:py-16"),
			h.Div(
				h.Class("pb-8 xl:grid xl:grid-cols-4 xl:gap-8"),
				h.Div(
					h.Class("grid grid-cols-2 gap-8 xl:col-span-4"),
					h.Div(
						h.Class("md:grid md:grid-cols-2 md:gap-8"),
						l.footerLinkSection("Classifieds", false,
							l.footerLink("For Sale", pages.Names.Classified.ForSale.Browse, false),
							l.footerLink("For Free", pages.Names.Classified.ForFree.Browse, false),
							l.footerLink("Looking For", pages.Names.Classified.WantToBuy.Browse, false),
							l.footerLink("Want to Borrow", pages.Names.Classified.WantToBorrow.Browse, false),
							c.LazyIf(l.Page.IsAuth,
								func() g.Node {
									return l.footerLink("New Post", pages.Names.Classified.New, false)
								})),
						l.footerLinkSection("Housing", true,
							// Deep link to filtered pages
							l.footerLink("For Rent", pages.Names.Housing.Browse, false),
							l.footerLink("Requests", pages.Names.Housing.Browse, false),
							c.LazyIf(l.Page.IsAuth,
								func() g.Node {
									return l.footerLink("New Post", pages.Names.Housing.New, false)
								}))),
					h.Div(
						h.Class("md:grid md:grid-cols-2 md:gap-8"),
						l.footerLinkSection("Other Services", false,
							l.footerLink("Groups", "https://groups.orcashub.org", true),
							l.footerLink("Events", pages.Names.Event.Browse, false),
							l.footerLink("Resources", "https://wiki.orcashub.org", true)),
						l.footerLinkSection("About", true,
							l.footerLink("FAQ", pages.Names.FAQ, false),
							l.footerLink("Terms of Service", pages.Names.TermsOfService, false),
							l.footerLink("Privacy Policy", pages.Names.PrivacyPolicy, false),
							l.footerLink("Donate", pages.Names.Donate, false),
						))))))
}

func (l *Primary) footerLinkSection(heading string, topMargin bool, children ...g.Node) g.Node {
	return h.Div(
		c.LazyIf(topMargin,
			func() g.Node {
				return h.Class("mt-12 md:mt-0")
			}),
		h.H3(
			h.Class("text-base font-medium text-white"),
			g.Text(heading)),
		h.Ul(
			g.Group(children)))
}

// footerLink creates a link for displaying in the footer. If external is true
// the pageName is instead treated as a URL directly instead of reversing it as
// a page name.
func (l *Primary) footerLink(display, pageName string, external bool) g.Node {
	href := pageName
	if !external {
		href = l.Page.ToURL(pageName)
	}
	return h.Li(
		h.A(
			h.Href(href),
			h.Class("text-base text-gray-300 hover:text-white"),
			c.LazyIf(external,
				func() g.Node {
					return h.Target("_blank")
				}),
			g.Text(display)))
}

func (l *Primary) messages() g.Node {
	ctx := l.Page.Context
	return h.Div(
		h.Class("pointer-events-none fixed inset-0 flex items-end px-4 py-6 sm:items-start sm:p-6"),
		h.Aria("live", "assertive"),
		h.Div(
			h.Class("flex w-full flex-col items-center space-y-4 sm:items-end"),
			l.message(msg.Get(ctx, msg.TypeSuccess), c.IconGreenCheckCircle()),
			l.message(msg.Get(ctx, msg.TypeInfo), c.IconBlueInformationCircle()),
			l.message(msg.Get(ctx, msg.TypeWarning), c.IconYellowExclamationTriangle()),
			l.message(msg.Get(ctx, msg.TypeDanger), c.IconRedExclamationTriangle()),
		))
}

func (l *Primary) message(messages []string, icon g.Node) g.Node {
	return g.Group(
		g.Map(messages, func(message string) g.Node {
			return h.Div(
				h.Class("pointer-events-auto w-full max-w-sm overflow-hidden rounded-lg bg-white shadow-lg ring-1 ring-black ring-opacity-5"),
				c.XCloak(),
				c.XShow("show"),
				c.XData(`{show: true}`),
				c.XTransition(c.TransitionOptions{
					DurationEnter:  "duration-300",
					DurationLeave:  "duration-100",
					OperationEnter: "ease-out",
					OperationLeave: "ease-in",
					OpacityStart:   "opacity-0",
					OpacityEnd:     "opacity-100",
					// TODO: Translate
					// x-transition:enter-start="translate-y-2 opacity-0 sm:translate-y-0 sm:translate-x-2"
					// x-transition:enter-end="translate-y-0 opacity-100 sm:translate-x-0"
				}),
				h.Div(
					h.Class("p-4"),
					h.Div(
						h.Class("flex items-center"),
						h.Div(
							h.Class("flex w-0 flex-1 justify-between"),
							h.Div(
								h.Class("flex-shrink-0 pr-4"),
								icon),
							h.P(
								h.Class("w-0 flex-1 text-sm font-medium text-gray-900"),
								g.Raw(message))),
						h.Div(
							h.Class("ml-4 flex flex-shrink-0"),
							h.Button(
								h.Class("inline-flex rounded-md bg-white text-gray-400 hover:text-gray-500 focus:outline-none focus:ring-2 focus:ring-indigo-500 focus:ring-offset-2"),
								c.XOn("click", "show = false"),
								h.Type("button"),
								c.TextSROnly("Close"),
								c.IconMiniX(
									h.Class("h-5 w-5")),
							)))))
		}))
}
