package views

import (
	"fmt"
	"io"
	"strconv"

	g "maragu.dev/gomponents"
	gc "maragu.dev/gomponents/components"
	h "maragu.dev/gomponents/html"

	"git.grosinger.net/tgrosinger/tonyslist/constants"
	"git.grosinger.net/tgrosinger/tonyslist/constants/pages"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/controller"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/generic"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/postbase"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
	c "git.grosinger.net/tgrosinger/tonyslist/pkg/views/components"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/views/helpers"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/views/layouts"
)

type BrowseDisplayType string

const (
	GridDisplayType BrowseDisplayType = "grid"
	ListDisplayType BrowseDisplayType = "list"
)

type PostBrowse struct {
	Posts              []postbase.PostBase
	FeaturePhotos      []sqlc.Photo
	MyPosts            []postbase.PostBase
	DisplayType        BrowseDisplayType
	SortOptions        []constants.SortOption
	CategorySelections []string
	Filters            []constants.Filter
	NewPostLink        string
	PostsListOnly      bool

	// Headers must be explicitly because this view is used for Classified,
	// Housing, and Events.
	Headers g.Node

	Page controller.Page
}

const (
	dataJS = `{
        off_canvas_menu_open: false,
        filters: ((new URLSearchParams(location.search).get('filters') || '').split(',')),
        sort: ((new URLSearchParams(location.search).get('sort') || '')),
	}`

	// initJS contains the Alpine.js watch functions needed to update the URL
	// based on the current state of the filters, search, and sort fields.
	// https://alpinejs.dev/magics/watch
	initJS = `
		$watch('filters', (value) => {
			const url = new URL(window.location.href);
			url.searchParams.set('filters', filters.join(','));
			history.pushState(null, document.title, url.toString());
			location.reload();
		});
		$watch('sort', (value) => {
			const url = new URL(window.location.href);
			url.searchParams.set('sort', sort);
			history.pushState(null, document.title, url.toString());
			location.reload();
		});`
)

func (v *PostBrowse) Render(w io.Writer) error {
	if v.PostsListOnly {
		return v.body().Render(w)
	}

	content := c.FullWidthContainer(
		h.Div(
			c.XData(dataJS),
			c.XInit(initJS),
			// Focus the search box on the forward slash keyboard shortcut.
			c.XOn("keyup.slash.window", "document.getElementById('search').focus()"),
			v.mobileFilterSidebar(),
			h.Div(
				// TODO: Combine this with c.WideContainer
				h.Class("mx-auto max-w-7xl px-4 sm:px-6 lg:px-8"),
				h.Div(
					h.Class("relative z-10 flex items-baseline justify-between border-b border-gray-200 pb-6 pt-8"),
					v.search(),
					h.Div(
						h.Class("flex items-center"),
						v.sortDropdown(),
						v.mobileFilterButton(),
						v.rssButton())),
				h.Section(
					h.Class("pb-24 pt-6"),
					h.Div(
						h.Class("grid grid-cols-1 gap-x-8 gap-y-10 lg:grid-cols-4"),
						// Filters
						h.Form(
							h.Class("hidden lg:block"),
							v.categorySelector(false),
							v.filterSelection(false)),
						v.body()))),
			v.myPostsTable()),
	)

	l := layouts.Primary{
		Header:  v.Headers,
		Content: content,
		Page:    v.Page,
	}
	return l.Render(w)
}

func (v *PostBrowse) mobileFilterButton() g.Node {
	return h.Button(
		c.XClick("off_canvas_menu_open=true"),
		h.Type("button"),
		h.Class("-m-2 ml-4 p-2 text-gray-400 hover:text-gray-500 sm:ml-6 lg:hidden"),
		c.TextSROnly("Filters"),
		c.IconSolidFilter())
}

func (v *PostBrowse) rssButton() g.Node {
	return h.Div(
		h.Class("-m-2 ml-4 p-2 text-gray-400 hover:text-gray-500 sm:ml-6"),
		h.A(
			h.Href(v.Page.RSSLink()),
			h.Target("_blank"),
			c.IconRSS()))
}

func (v *PostBrowse) mobileFilterSidebar() g.Node {
	return h.Div(
		c.XShow("off_canvas_menu_open"),
		h.Class("relative z-40 lg:hidden"),
		h.Role("dialog"),
		h.Aria("modal", "true"),
		h.Div(
			c.XCloak(),
			c.XShow("off_canvas_menu_open"),
			c.XTransition(c.TransitionOptions{
				DurationEnter:  "duration-300",
				DurationLeave:  "duration-300",
				OperationEnter: "ease-linear",
				OperationLeave: "ease-linear",
				OpacityStart:   "opacity-0",
				OpacityEnd:     "opacity-100",
				ScaleStart:     "scale-100",
				ScaleEnd:       "scale-100",
			}),
			h.Class("fixed inset-0 bg-black bg-opacity-25")),
		h.Div(
			h.Class("fixed inset-0 z-40 flex"),
			h.Div(
				c.XCloak(),
				c.XShow("off_canvas_menu_open"),
				g.Attr("x-transition:enter", "transition ease-in-out duration-300 transform"),
				g.Attr("x-transition:enter-start", "translate-x-full"),
				g.Attr("x-transition:enter-end", "translate-x-0"),
				g.Attr("x-transition:leave", "transition ease-in-out duration-300 transform"),
				g.Attr("x-transition:leave-start", "translate-x-0"),
				g.Attr("x-transition:leave-end", "translate-x-full"),
				h.Class("relative ml-auto flex h-full w-full max-w-xs flex-col overflow-y-auto bg-white py-4 pb-12 shadow-xl"),
				h.Div(
					h.Class("flex items-center justify-between px-4"),
					h.Button(
						c.XOn("click", "off_canvas_menu_open=false"),
						h.Type("button"),
						h.Class("-mr-2 flex h-10 w-10 items-center justify-center rounded-md bg-white p-2 text-gray-400"),
						c.TextSROnly("Close menu"),
						c.IconX())),
				h.Form(
					// Filters
					h.Class("mt-4 border-t border-gray-200"),
					v.categorySelector(true),
					v.filterSelection(true)))))
}

func (v *PostBrowse) sortDropdown() g.Node {
	return h.Div(
		c.XData("{sort_dropdown_open: false}"),
		h.Class("relative inline-block text-left"),
		h.Div(
			h.Button(
				c.XOn("click", "sort_dropdown_open=!sort_dropdown_open"),
				h.Type("button"),
				h.Class("group inline-flex justify-center text-sm font-medium text-gray-700 hover:text-gray-900"),
				h.ID("sort-dropdown-button"),
				h.Aria("haspopup", "true"),
				g.Text("Sort"),
				c.IconSolidChevronDown())),
		h.Div(
			c.XCloak(),
			c.XShow("sort_dropdown_open"),
			c.XTransition(c.TransitionOptions{
				DurationEnter:  "duration-100",
				DurationLeave:  "duration-75",
				OperationEnter: "ease-out",
				OperationLeave: "ease-in",
				OpacityStart:   "opacity-0",
				OpacityEnd:     "opacity-100",
				ScaleStart:     "scale-95",
				ScaleEnd:       "scale-100",
			}),
			c.XOn("click.outside", "sort_dropdown_open=false"),
			h.Class("absolute right-0 mt-2 w-40 origin-top-right rounded-md bg-white shadow-2xl ring-1 ring-black ring-opacity-5 focus:outline-none"),
			h.Role("menu"),
			h.Aria("labelledby", "sort-dropdown-button"),
			h.Aria("orientation", "vertical"),
			h.TabIndex("-1"),
			v.sortOptions()))
}

func (v *PostBrowse) sortOptions() g.Node {
	return h.Div(
		h.Class("py-1"),
		h.Role("none"),
		g.Group(
			g.Map(v.SortOptions, func(opt constants.SortOption) g.Node {
				return h.Span(
					c.XBind("class",
						fmt.Sprintf(
							`{
							'text-gray-500': sort !== '%s',
							'font-medium text-gray-900': sort === '%[1]s',
						}`, opt.Name)),
					c.XOn("click", fmt.Sprintf("sort='%s'", opt.Name)),
					h.Class("block px-4 py-2 text-sm font-medium text-gray-900 hover:bg-gray-100"),
					h.Role("menuitem"),
					h.TabIndex("-1"),
					g.Text(opt.Display))
			})))
}

func (v *PostBrowse) filterSelection(mobile bool) g.Node {
	return g.Group(g.Map(v.Filters,
		func(filter constants.Filter) g.Node {
			return h.Div(
				c.XData("{'expanded': true}"),
				gc.Classes{
					"border-gray-200 py-6": true,
					"border-t px-4":        mobile,
					"border-b":             !mobile,
				},
				h.H3(
					gc.Classes{
						"-my-3 flow-root": true,
						"-mx-2":           mobile,
					},
					h.Button(
						c.XOn("click", "expanded=!expanded"),
						h.Type("button"),
						gc.Classes{
							"flex w-full items-center justify-between bg-white py-3 text-gray-400 hover:text-gray-500": true,
							"px-2":    mobile,
							"text-sm": !mobile,
						},
						h.Span(
							h.Class("font-medium text-gray-900"),
							g.Text(filter.SectionDisplay)),
						h.Span(
							h.Class("ml-6 flex items-center"),
							c.XIfTemplate("!expanded",
								c.IconSolidPlusSm()),
							c.XIfTemplate("expanded",
								c.IconSolidMinusSm())))),
				h.Div(
					c.XCloak(),
					c.XShow("expanded"),
					h.Class("pt-6"),
					h.Div(
						h.Class("space-y-4"),
						g.Group(
							g.Map(filter.Options, func(option constants.FilterOption) g.Node {
								id := fmt.Sprintf("filter-%s-%s", filter.SectionName, option.Value)
								if mobile {
									id = "mobile-" + id
								}
								return h.Div(
									h.Class("flex items-center"),
									h.Input(
										h.ID(id),
										h.Value(fmt.Sprintf("%s_%s", filter.SectionName, option.Value)),
										h.Type("checkbox"),
										h.Class("h-4 w-4 rounded border-gray-300 text-indigo-600 focus:ring-indigo-500"),
										c.XModel("filters")),
									h.Label(
										h.For(id),
										gc.Classes{
											"ml-3 text-gray-600": true,
											"min-w-0 flex-1":     mobile,
											"text-sm":            !mobile,
										},
										g.Text(option.Display)))
							})))))
		}))
}

func (v *PostBrowse) categorySelector(mobile bool) g.Node {
	if len(v.CategorySelections) == 0 {
		return nil
	}

	return g.Group([]g.Node{
		h.H3(
			h.Class("sr-only"),
			g.Text("Categories")),
		h.Ul(
			gc.Classes{
				"space-y-4 border-b border-gray-200 pb-6 text-sm font-medium text-gray-900": !mobile,
				"px-2 py-3 font-medium text-gray-900":                                       mobile,
			},
			g.Group(
				g.Map(v.CategorySelections, func(category string) g.Node {
					return h.Li(
						h.A(
							h.Href("#"),
							c.LazyIf(mobile, func() g.Node {
								return h.Class("block px-2 py-3")
							}),
							g.Text(category)))
				}))),
	})
}

func (v *PostBrowse) search() g.Node {
	return h.Div(
		h.Class("mr-2 w-full max-w-lg lg:max-w-xs"),
		c.SRLabel("search", "Search"),
		h.Div(
			h.Class("relative text-gray-400 focus-within:text-gray-600"),
			h.Div(
				h.Class("pointer-events-none absolute inset-y-0 left-0 flex items-center pl-3"),
				c.IconSolidSearch()),
			h.Input(
				h.ID("search"),
				c.HxGet(v.Page.Path),
				c.HxTarget("#posts-grid"),
				c.HxSwap("outerHTML"),
				c.HxTrigger("search, keyup changed delay:300ms"),
				c.HxSync("this:replace"),
				h.Class("peer block w-full rounded-md border border-transparent bg-white py-2 pl-10 pr-3 leading-5 text-gray-900 placeholder-gray-500 focus:border-white focus:outline-none focus:ring-2 focus:ring-white focus:ring-offset-2 focus:ring-offset-indigo-600 sm:text-sm"),
				h.Placeholder("Search"),
				h.Type("search"),
				h.Name("search")),
			h.A(
				h.Href(v.Page.ToURL(pages.Names.Help.Classified.Search)),
				h.Target("_blank"),
				h.Class("invisible absolute inset-y-0 right-0 flex cursor-pointer items-center pr-3 active:visible peer-focus:visible"),
				c.IconCircleQuestion())))
}

func (v *PostBrowse) pager() g.Node {
	if v.Page.Pager.Pages <= 1 {
		return nil
	}

	hasPrevious := v.Page.Pager.Page > 1
	hasNext := v.Page.Pager.Page < v.Page.Pager.Pages

	return h.Nav(
		h.Class("flex items-center justify-between border-t border-gray-200 px-4 sm:px-0"),
		h.Div(
			h.Class("-mt-px flex w-0 flex-1"),
			h.A(
				h.Href("?page="+strconv.Itoa(generic.Max(1, v.Page.Pager.Page-1))),
				gc.Classes{
					"border-t-2 border-transparent pt-4 pr-1 inline-flex items-center text-sm font-medium": true,
					"text-gray-500 hover:text-gray-700":                                                    hasPrevious,
					"text-gray-300":                                                                        !hasPrevious,
				},
				c.LazyIf(!hasPrevious, h.Disabled),
				c.IconSolidArrowLeft(hasPrevious),
				g.Text("Previous"))),
		h.Div(
			h.Class("hidden md:-mt-px md:flex"),
			g.Group(
				g.Map(helpers.PageNums(v.Page.Pager.Pages, v.Page.Pager.Page),
					func(page int) g.Node {
						return v.pagerPageNum(page, page == v.Page.Pager.Page)
					}))),
		h.Div(
			h.Class("-mt-px flex w-0 flex-1 justify-end"),
			h.A(
				h.Href("?page="+strconv.Itoa(generic.Min(v.Page.Pager.Pages, v.Page.Pager.Page+1))),
				gc.Classes{
					"border-t-2 border-transparent pt-4 pl-1 inline-flex items-center text-sm font-medium": true,
					"text-gray-500 hover:text-gray-700":                                                    hasNext,
					"text-gray-300":                                                                        !hasNext,
				},
				c.LazyIf(!hasNext, h.Disabled),
				g.Text("Next"),
				c.IconSolidArrowRight(hasNext))))
}

func (v *PostBrowse) pagerPageNum(i int, current bool) g.Node {
	return h.A(
		h.Href("?page="+strconv.Itoa(i)),
		gc.Classes{
			"inline-flex items-center border-t-2 px-4 pt-4 text-sm font-medium ":         true,
			"border-indigo-500 text-indigo-600":                                          current,
			"border-transparent text-gray-500 hover:border-gray-300 hover:text-gray-700": !current,
		},
		g.Text(strconv.Itoa(i)))
}

func (v *PostBrowse) myPostsTable() g.Node {
	t := MyPostsTable{Page: v.Page}

	if !v.Page.IsAuth {
		return nil
	}

	return h.Div(
		h.Class("px-4 pb-24 sm:px-6 lg:px-8"),
		h.Div(
			h.Class("sm:flex sm:items-center"),
			h.H1(
				h.Class("text-xl font-semibold text-gray-900"),
				g.Text("My Posts"))),
		t.Table(v.MyPosts, v.NewPostLink))
}

func (v *PostBrowse) body() g.Node {
	return h.Div(
		h.Class("space-y-8 lg:col-span-3"),
		h.ID("posts-grid"),
		c.LazyIfElse(v.DisplayType == ListDisplayType,
			v.postsList,
			v.postsGrid))
}

func (v *PostBrowse) postsList() g.Node {
	if len(v.Posts) == 0 {
		return c.EmptyBrowserPlaceholder(
			"New Post",
			v.NewPostLink)
	}

	return g.Group([]g.Node{
		c.BrowserList(v.Page.Context, v.Posts),
		v.pager(),
	})
}

func (v *PostBrowse) postsGrid() g.Node {
	if len(v.Posts) == 0 {
		return c.EmptyBrowserPlaceholder(
			"New Post",
			v.NewPostLink)
	}

	return g.Group([]g.Node{
		c.BrowserGrid(v.Page.Context, v.Page, v.Posts, v.FeaturePhotos),
		v.pager(),
	})
}
