package views

import (
	"database/sql"
	"fmt"
	"io"
	"strconv"
	"strings"

	"github.com/yuin/goldmark"
	g "maragu.dev/gomponents"
	comp "maragu.dev/gomponents/components"
	h "maragu.dev/gomponents/html"

	"git.grosinger.net/tgrosinger/tonyslist/constants"
	"git.grosinger.net/tgrosinger/tonyslist/constants/pages"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/controller"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/conv"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/postbase"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
	c "git.grosinger.net/tgrosinger/tonyslist/pkg/views/components"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/views/helpers"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/views/layouts"
)

type PostDetails struct {
	postbase.PostBase
	Price  sql.NullFloat64
	Photos []sqlc.Photo

	// Housing-only fields
	Bedrooms     conv.NullInt
	Bathrooms    sql.NullFloat64
	ParkingSpots conv.NullInt
	Pets         bool
	Shared       bool

	// CreatorInfo will contain information about the post creator iff the
	// viewer of this page has moderator permissions and will be able to see the
	// moderator toolbar.
	CreatorInfo *sqlc.GetUserInfoRow

	MarkdownRenderer goldmark.Markdown

	// Headers must be explicitly because this view is used for Classified,
	// Housing, and Events.
	Headers g.Node

	Page controller.Page
}

func (v *PostDetails) Render(w io.Writer) error {
	content := c.WideContainer(
		c.LazyIf(v.Page.IsAuth && v.Creator == v.Page.AuthUser.ID,
			v.ownerFloatingToolbar),
		c.LazyIf(v.Page.Enforce("post", "moderate"),
			v.moderatorFloatingToolbar),
		h.Div(
			h.Class("mx-auto max-w-2xl py-8 px-4 sm:py-8 sm:px-6 lg:grid lg:max-w-7xl lg:grid-cols-2 lg:gap-x-8 lg:px-8"),
			v.postDetails(),
			v.photoGallery(),
			v.contactDetails()),
	)

	l := layouts.Primary{
		Header:  v.Headers,
		Content: content,
		Page:    v.Page,
	}
	return l.Render(w)
}

func (v *PostDetails) ownerFloatingToolbar() g.Node {
	colors := "border-gray-300 bg-white text-indigo-600 hover:bg-indigo-50 focus:border-indigo-500 focus:ring-indigo-500"

	photosButtonParams := c.ButtonParams{
		Href:    v.Page.ToURL(pages.Names.Posts.NewPhotos, v.ID),
		Display: "Edit Photos",
	}
	editButtonParams := c.ButtonParams{
		Href:    v.GetEditHref(v.Page.Context.Echo()),
		Display: "Edit",
	}
	deleteButtonParams := c.ButtonParams{
		Href:    v.GetDeleteHref(v.Page.Context.Echo()),
		Display: "Delete",
	}

	var postText string
	var buttons []g.Node
	switch v.GetStatus() {
	case constants.Active:
		postText = "Your post is active."
		soldText := "Delist"
		if v.PostType == constants.Classified && (v.ClassifiedType == constants.ForSale || v.ClassifiedType == constants.ForFree) {
			soldText = "Mark Sold"
		}

		if v.PostType == constants.Classified || v.HousingType != constants.Request {
			buttons = append(buttons, c.FloatingToolbarJoinedButtons(
				editButtonParams,
				photosButtonParams,
				colors,
			))
		} else {
			buttons = append(buttons, c.FloatingToolbarButton(editButtonParams, colors))
		}

		buttons = append(buttons, []g.Node{
			c.FloatingToolbarJoinedButtons(
				c.ButtonParams{
					Href:    v.GetDelistHref(v.Page.Context.Echo()),
					Display: soldText,
				},
				deleteButtonParams,
				colors,
			),
		}...)
	case constants.Unlisted:
		postText = "Your post is not listed."
		buttons = []g.Node{
			c.FloatingToolbarButton(editButtonParams, colors),
			c.FloatingToolbarJoinedButtons(
				c.ButtonParams{
					Href:    v.GetRelistHref(v.Page.Context.Echo()),
					Display: "Relist",
				},
				deleteButtonParams,
				colors,
			),
		}
	case constants.Draft:
		postText = "Your post has not yet been published."
		buttons = []g.Node{
			c.FloatingToolbarButton(editButtonParams, colors),
			c.FloatingToolbarJoinedButtons(
				c.ButtonParams{
					Href:    v.GetAddPhotosHref(v.Page.Context.Echo()),
					Display: "Publish",
				},
				deleteButtonParams,
				colors,
			),
		}
	case constants.Expired:
		postText = "Your post has expired"
		buttons = []g.Node{
			c.FloatingToolbarButton(editButtonParams, colors),
			c.FloatingToolbarJoinedButtons(
				c.ButtonParams{
					Href:    v.GetRelistHref(v.Page.Context.Echo()),
					Display: "Relist",
				},
				deleteButtonParams,
				colors,
			),
		}
	}

	leftText := h.P(
		h.Class("ml-3 font-medium text-white"),
		g.Text(postText))

	return c.FloatingToolbar("owner-toolbar", "bg-indigo-600", leftText, g.Group(buttons))
}

func (v *PostDetails) moderatorFloatingToolbar() g.Node {
	colors := "border-transparent bg-white text-gray-600 hover:bg-gray-50"

	name := v.CreatorInfo.Email
	if v.CreatorInfo.Name.Valid {
		name = v.CreatorInfo.Name.String
	}
	left := h.Div(
		h.Class("flex flex-col ml-3 font-medium text-white"),
		h.P(
			g.Text("Moderator Actions")),
		h.Div(
			h.Class("space-x-2 flex text-sm"),
			h.P(
				g.Text("Posted by:")),
			c.TextLink(v.Page.ToURL(pages.Names.Admin.UserDetails, v.Creator), g.Text(name))),
	)

	editButton := c.FloatingToolbarButton(c.ButtonParams{
		Href:    v.GetEditHref(v.Page.Context.Echo()),
		Display: "Edit",
	}, colors)
	deleteButton := c.FloatingToolbarButton(
		c.ButtonParams{
			Href:    v.GetModeratorDeleteHref(v.Page.Context.Echo()),
			Display: "Delete",
		},
		colors)

	return c.FloatingToolbar("moderator-toolbar", "bg-gray-600", left, editButton, deleteButton)
}

func (v *PostDetails) postDetails() g.Node {
	return h.Div(
		comp.Classes{
			"lg:max-w-lg": true,
			"lg:self-end": len(v.Photos) > 0,
		},
		h.Div(
			h.Class("mt-4"),
			c.PrimaryHeader(v.Title)),
		h.Section(
			h.Aria("labelledby", "information-heading"),
			h.Class("mt-4"),
			h.H2(
				h.ID("information-heading"),
				h.Class("sr-only"),
				g.Text("Posting information")),
			h.Div(
				h.Class("flex items-center"),
				c.LazyIf(v.Price.Valid, func() g.Node {
					return c.TextLarge("$" + c.TextCurrency(&v.Price.Float64))
				}),
				c.LazyIf(v.PublishedAt.Valid, func() g.Node {
					return h.Div(
						h.Class("ml-4 border-l border-gray-300 pl-4"),
						c.TextDescription("Posted "+helpers.RelativeDateTime(v.PublishedAt.Time)))
				})),
			h.Div(
				h.Class("mt-4 text-base text-gray-500 markdown-container"),
				g.Raw(v.Description)),
			c.LazyIf(v.PostType == constants.Housing, func() g.Node {
				return h.Div(
					h.Class("my-5 border-t border-gray-200 px-4 py-5 sm:p-0"),
					h.Dl(
						h.Class("sm:divide-y sm:divide-gray-200"),
						c.LazyIf(v.Bedrooms.Valid, func() g.Node {
							return v.textRow("Bedrooms", strconv.Itoa(v.Bedrooms.Int))
						}),
						c.LazyIf(v.Bathrooms.Valid, func() g.Node {
							return v.textRow("Bathrooms", strconv.FormatFloat(v.Bathrooms.Float64, 'f', 2, 64))
						}),
						c.LazyIf(v.ParkingSpots.Valid, func() g.Node {
							return v.textRow("Parking Spaces", strconv.Itoa(v.ParkingSpots.Int))
						}),
						v.textRow("Pets Allowed", helpers.FormatBool(v.Pets)),
						v.textRow("Shared", helpers.FormatBool(v.Shared)),
					))
			})))
}

func (v *PostDetails) textRow(display, value string) g.Node {
	return v.row(display, c.TextNormal(value))
}

func (v *PostDetails) row(display string, value g.Node) g.Node {
	return h.Div(
		h.Class("py-4 sm:grid sm:grid-cols-3 sm:gap-4 sm:py-5 sm:px-6"),
		h.Dt(c.TextDescription(display)),
		h.Dd(
			h.Class("mt-1 sm:col-span-2 sm:mt-0"),
			value))
}

func (v *PostDetails) photoGallery() g.Node {
	if len(v.Photos) == 0 {
		return nil
	}

	// TODO: Use the Alpine.js teleport directive for the fullscreen images.
	// https://devdojo.com/pines/docs/image-gallery
	// https://alpinejs.dev/directives/teleport

	// TODO: Make photos know how to convert their own filenames?

	// featured contains the featured photo, or the first photo on the post.
	featured := v.GetFeaturedPhoto(v.Photos).UnwrapOr(&v.Photos[0])

	return h.Div(
		c.XData("{ selected_photo: '%d', photos: %s, expanded: false}",
			featured.ID,
			v.photoMap()),
		h.Class("mt-10 flex flex-col-reverse lg:col-start-2 lg:row-span-2 lg:mt-0 lg:self-center"),
		v.photoSelector(),
		h.Div(
			c.XCloak(),
			c.XShow("expanded"),
			h.Class("fixed flex justify-center items-center inset-0 min-w-full z-50 bg-gray-400/50 cursor-pointer"),
			h.Div(
				g.Attr("x-on:click.window", "expanded = false"),
				g.Attr("x-on:keyup.escape.window", "expanded = false")),
			h.Div(
				h.Class("relative mx-5 sm:mx-20"),
				h.Img(
					h.Class("max-w-full max-h-screen"),
					c.XBind("src", "photos[selected_photo]['full']")),
				h.Div(
					c.XShow("Object.keys(photos).length > 1"),
					h.Class("absolute flex items-center justify-between inset-y-0 inset-x-0 text-white"),
					h.Div(
						h.Class("min-h-full flex items-center hover:bg-gray-200/50"),
						// Add keys.length before subtracting so the mod operation is always greater than 0.
						c.XClickStop("const keys = Object.keys(photos); selected_photo = keys[(keys.indexOf(selected_photo) + keys.length - 1) % keys.length]"),
						c.IconChevronLeft(
							h.Class("h-24 w-24"))),
					h.Div(
						h.Class("min-h-full flex items-center hover:bg-gray-200/50"),
						c.XClickStop("const keys = Object.keys(photos); selected_photo = keys[(keys.indexOf(selected_photo) + 1) % keys.length]"),
						c.IconChevronRight(
							h.Class("h-24 w-24"))))),
		),
		h.Div(
			h.Class("aspect-w-1 aspect-h-1 w-full"),
			// Without nextTick, the closing hook will conflict.
			c.XClick("$nextTick(() => { expanded = true })"),
			h.Div(
				h.Class("opacity-0 hover:opacity-100 duration-100 text-white absolute inset-0 z-10 flex items-center justify-center"),
				c.IconExpand()),
			h.Img(
				c.XBind("src", "photos[selected_photo]['preview']"),
				h.Class("h-full w-full object-contain object-center cursor-pointer z-0 absolute inset-0 sm:rounded-lg"),
				h.Src(v.Page.ToUserContentPath(featured.GetRegularFilename())))))
}

// photoMap converts the list of photos on the post into a map which can be used
// by alpine.js to switch between the different photo options.
func (v *PostDetails) photoMap() string {
	b := strings.Builder{}
	b.WriteString("{")
	for _, p := range v.Photos {
		b.WriteString(strconv.Itoa(p.ID))
		b.WriteString(":{")
		b.WriteString("preview:'")
		b.WriteString(v.Page.ToUserContentPath(p.GetPreviewFilename()))
		b.WriteString("', full:'")
		b.WriteString(v.Page.ToUserContentPath(p.GetRegularFilename()))
		b.WriteString("'},")
	}

	b.WriteString("}")
	return b.String()
}

func (v *PostDetails) photoSelector() g.Node {
	if len(v.Photos) <= 1 {
		// Do not display carousel if there is only one photo
		return nil
	}

	return h.Div(
		h.Class("mx-auto mt-6 block w-full max-w-2xl lg:max-w-none"),
		h.Div(
			h.Class("grid grid-cols-4 gap-6"),
			h.Aria("orientation", "horizontal"),
			h.Role("tablist"),
			g.Group(g.Map(v.Photos, func(p sqlc.Photo) g.Node {
				return h.Button(
					c.XData("{ selected: false, photo: '%d' }", p.ID),
					c.XClick("selected = true; selected_photo = photo"),
					h.Class("aspect-w-1 aspect-h-1 relative flex h-24 cursor-pointer items-center justify-center rounded-md bg-white text-sm font-medium uppercase text-gray-900 hover:bg-gray-50 focus:outline-none"),
					h.Aria("controls", "tabs-1-panel-1"),
					h.Role("tab"),
					h.Type("button"),
					h.Span(
						h.Class("absolute inset-0 overflow-hidden rounded-md"),
						h.Img(
							h.Class("h-full w-full object-cover object-center"),
							h.Src(v.Page.ToUserContentPath(p.GetPreviewFilename())))),
					h.Span(
						c.XBind("class", "{ 'ring-indigo-500': selected_photo === photo, 'ring-transparent': selected_photo !== photo }"),
						h.Class("pointer-events-none absolute inset-0 rounded-md ring-2 ring-transparent ring-offset-2"),
						h.Aria("hidden", "true")))
			}))))
}

func (v *PostDetails) contactDetails() g.Node {
	var email g.Node = nil
	if v.AnonymousEmail.Valid {
		addr := fmt.Sprintf("%s@%s", v.AnonymousEmail.String, v.Page.Cfg.Mail.Domain)
		email = v.row("Email", c.TextLink("mailto:"+addr, c.TextNormal(addr)))
	}

	return h.Div(
		comp.Classes{
			"mt-10 lg:max-w-lg lg:self-start": true,
			"lg:col-start-1 lg:row-start-2":   len(v.Photos) > 0,
		},
		h.Section(
			h.Aria("labelledby", "options-heading"),
			h.Div(
				h.Class("border-b border-gray-200 pb-5"),
				c.ListHeader("Contact Information")),
			c.LazyIfElse(
				v.Page.IsAuth,
				func() g.Node {
					return h.Div(
						h.Class("px-4 py-5 sm:p-0"),
						h.Dl(
							h.Class("sm:divide-y sm:divide-gray-200"),
							c.LazyIf(v.PhoneCall.Valid, func() g.Node { return v.textRow("Phone", v.PhoneCall.String) }),
							c.LazyIf(v.PhoneText.Valid, func() g.Node { return v.textRow("SMS/Text", v.PhoneText.String) }),
							email,
						))
				},
				func() g.Node {
					return h.Div(
						h.Class("py-5"),
						h.Span(
							h.Class("text-md font-medium text-gray-500"),
							c.TextLink(
								v.Page.ToURL(pages.Names.User.Login)+"?return_to="+v.Page.Path,
								g.Text("Please login or sign up")),
							g.Text(" to view contact details")))
				})))
}
