package views

import (
	"fmt"
	"io"
	"net/http"
	"strings"

	g "maragu.dev/gomponents"
	h "maragu.dev/gomponents/html"

	"git.grosinger.net/tgrosinger/tonyslist/constants"
	"git.grosinger.net/tgrosinger/tonyslist/constants/pages"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/controller"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/models/sqlc"
	c "git.grosinger.net/tgrosinger/tonyslist/pkg/views/components"
	"git.grosinger.net/tgrosinger/tonyslist/pkg/views/layouts"
)

type PostPhotoUpload struct {
	PublishDisplay           string
	PublishAction            string
	PreviouslyUploadedPhotos []sqlc.Photo
	PostType                 constants.PostType
	PostID                   int
	Token                    string

	Page controller.Page
}

func (v *PostPhotoUpload) Render(w io.Writer) error {
	remainingPhotos := 10 - len(v.PreviouslyUploadedPhotos)
	maxPhotos := remainingPhotos == 0

	photosLabel := "photos"
	if remainingPhotos == 1 {
		photosLabel = "photo"
	}

	content := c.NarrowContainer(
		h.Div(
			h.Class("flex"),
			h.Div(
				h.Class("grow"),
				c.PrimaryHeader("Add Photos")),
			h.Form(
				h.Method(http.MethodPost),
				h.Action(v.PublishAction),
				c.CSRF(v.Page),
				c.FormSubmit(v.PublishDisplay))),
		h.Div(
			h.Class("text-md my-2 max-w-2xl space-y-2 font-medium text-gray-900"),
			c.TextNormal("Add photos one at a time here. The first photo added will be the primary photo."),
			c.TextNormal(
				fmt.Sprintf("Click %s when you are finished.", strings.ToLower(v.PublishDisplay))),
			h.P(
				h.Class("text-base text-gray-700"),
				g.Text(fmt.Sprintf("You may add %d more %s.", remainingPhotos, photosLabel)))),
		c.LazyIf(!maxPhotos, func() g.Node {
			return h.Div(
				h.Input(
					h.Class("mt-4"),
					h.Type("file"),
					h.Accept("image/jpeg, image/png")),
				c.TextDescription("PNG or JPG files, 32MB or smaller."),
				h.Div(
					h.Class("w-full max-w-full border-b border-gray-300"),
					h.Div(
						h.ID("progress-bar"),
						h.Class("h-1 bg-indigo-700"),
						h.Style("width: 0%;"))))
		}),
		c.LazyIf(len(v.PreviouslyUploadedPhotos) > 0, func() g.Node {
			return h.Div(
				h.Div(
					h.Class("py-4"),
					c.TertiaryHeader("Added Photos:")),
				h.Ul(
					h.ID("upload-list"),
					h.Role("list"),
					h.Class("pt-6 grid grid-cols-2 gap-x-4 gap-y-8 sm:grid-cols-3 sm:gap-x-6 lg:grid-cols-4 xl:gap-x-8"),
					g.Group(g.Map(v.PreviouslyUploadedPhotos,
						func(p sqlc.Photo) g.Node {
							return v.uploadedImageThumbnail(p)
						}))))
		}),
		c.LazyIf(!maxPhotos, func() g.Node {
			return g.Group([]g.Node{
				h.Script(
					h.Src("/files/js/tus-4.2.3.js")),
				v.script(),
			})
		}))

	l := layouts.Primary{
		Content: content,
		Page:    v.Page,
	}
	return l.Render(w)
}

func (v *PostPhotoUpload) uploadedImageThumbnail(p sqlc.Photo) g.Node {
	if p.IsMap {
		return nil
	}

	return h.Form(
		h.Method(http.MethodPost),
		h.Action(v.Page.ToURL(pages.Names.Posts.DeletePhoto, v.PostID, p.ID)),
		c.CSRF(v.Page),
		h.Li(
			h.Class("relative"),
			h.Span(
				h.Class("relative inline-block"),
				h.Img(
					h.Class("pointer-events-none object-cover rounded-md"),
					h.Src(v.Page.ToUserContentPath(p.GetThumbnailFilename()))),
				h.Button(
					h.Type("submit"),
					h.Class("absolute right-0 top-0 block h-4 w-4 -translate-y-1/2 translate-x-1/2 transform rounded-full bg-red-400 text-white ring-3 ring-white"),
					c.IconMiniX(
						h.Class("h-4 w-4"))))))
}

func (v *PostPhotoUpload) script() g.Node {
	// Upload resumption is disabled, but check demo if you want to re-enable.
	return h.Script(
		g.Rawf(`
const post_type = %q;
const post_id = %d;
const token = %q;
const support_email = %q;
let upload = null;
let uploadIsRunning = false;
const input = document.querySelector("input[type=file]");
const progressBar = document.querySelector("#progress-bar");

function startUpload() {
	const file = input.files[0];
	if (!file) {
		return;
	}

	const options = {
		endpoint: "/uploads",
		retryDelays: [0, 1000, 3000, 5000],
		uploadDataDuringCreation: true,
		storeFingerprintForResuming: false,
		metadata: {
			ext: file.name.slice((file.name.lastIndexOf(".") - 1 >>> 0) + 2),
			post_type: post_type,
			post_id: post_id,
			token: token
		},
		onError(error) {
			console.error(error);

			if (error.originalRequest) {
				if (window.confirm("Upload failed.\nDo you want to retry?")) {
					upload.start();
					uploadIsRunning = true;
					return;
				}
			} else {
				const message = "Upload failed. Please contact " + support_email + " for assistance"
				window.alert(message);
			}

			reset();
		},
		onProgress(bytesUploaded, bytesTotal) {
			const percentage = ((bytesUploaded / bytesTotal) * 100).toFixed(2);
			progressBar.style.width = percentage + '%%';
			console.log(bytesUploaded, bytesTotal, percentage + '%%');
		},
		onSuccess() {
			input.value = "";
			window.location.reload();
		},
	};

	upload = new tus.Upload(file, options);
	input.disabled = true;
	upload.start();
	uploadIsRunning = true;
}

input.addEventListener("change", startUpload);`,
			v.PostType,
			v.PostID,
			v.Token,
			v.Page.Cfg.Mail.SupportAddress))
}
